/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.services.authentication;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.apache.axis2.context.MessageContext;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.model.SystemPermissions;
import org.wso2.carbon.utils.ServerConstants;
import org.wso2.carbon.registry.core.session.UserRegistry;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.sso.consumer.SSOConstants;
import org.wso2.carbon.sso.consumer.SSOConsumer;
import org.wso2.carbon.core.services.authentication.stats.LoginAttempt;
import org.wso2.carbon.core.services.authentication.stats.LoginStatDatabase;
import org.wso2.carbon.core.services.internal.CarbonServicesServiceComponent;


/**
 * Admin service to manage operations on the <code>AxisConfiguration</code>
 */
public class AuthenticationAdmin implements AuthenticationInterface {

    private static final Log log = LogFactory.getLog(AuthenticationAdmin.class);

    public boolean login(String username, String password, String remoteAddress) throws AuthenticationException {

        MessageContext msgCtx = MessageContext.getCurrentMessageContext();
        HttpServletRequest request =
                (HttpServletRequest) msgCtx.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        HttpSession httpSess = request.getSession();

        Date currentTime = Calendar.getInstance().getTime();
        SimpleDateFormat date = new SimpleDateFormat("'['yyyy-MM-dd HH:mm:ss,SSSS']'");

        if ((username == null) || (password == null) || (remoteAddress == null) ||
                username.trim().equals("") || password.trim().equals("") ||
                remoteAddress.trim().equals("")) {
            log.warn("Failed Administrator login attempt \'"
                    + username + "\' at " + date.format(currentTime) + " from IP address "
                    + remoteAddress);
            httpSess.invalidate();
            //LoginAttempt loginAttempt = new LoginAttempt(username, remoteAddress, currentTime, 
            //      false, "Null credentials");
            //LoginStatDatabase.recordLoginAttempt(loginAttempt);
            return false;
        }

        try {

			UserRealm realm = CarbonServicesServiceComponent.getUserRealm();
			boolean isAuthenticated = false;

			if (MessageContext.getCurrentMessageContext().getAxisService().getAxisConfiguration()
					.getParameter(SSOConstants.SSO_SERVICE_EPR) != null) {
				String epr = (String) MessageContext.getCurrentMessageContext().getAxisService()
						.getAxisConfiguration().getParameter(SSOConstants.SSO_SERVICE_EPR)
						.getValue();
				// If this is the case - what we get for password is considered as the session id
				// for SSO.
				isAuthenticated = new SSOConsumer(epr).isAuthenticated(username, password);
			}
			
			if (!isAuthenticated) {
                isAuthenticated = realm.getAuthenticator().authenticate(username, password);
            } else {
                log.info("SSO enabled login success by " + username);
            }
			
			boolean isAuthorized = realm.getAuthorizer().isUserAuthorized(username, "System",
					"login");
			if (isAuthenticated && isAuthorized) {
				httpSess.setAttribute(ServerConstants.ADMIN_LOGGED_IN, "true");
				httpSess.setAttribute(ServerConstants.ADMIN_USERNAME, username);
				log.info("\'" + username + "\' logged in at " + date.format(currentTime) + " from IP address "
                        + remoteAddress);

                // create a UserRegistry instance for the logged in user. This instance is used to perform Registry
                // actions performed by *this* user. If a Registry action is performed by the system, it is
                // done using the system registry instance, although user's registry is available.

                // in most cases, user registry instance is used only by the Registry UI bundles as those actions
                // are performed directly by the currently logged in user.

                UserRegistry userRegistry = CarbonServicesServiceComponent
                        .getRegistryService().getUserRegistry(username);

                request.getSession().setAttribute(RegistryConstants.USER_REGISTRY, userRegistry);

                if (log.isDebugEnabled()) {
                    log.debug("User Registry instance is set in the session for user " + username);
                }
                LoginAttempt loginAttempt = new LoginAttempt(username, remoteAddress, currentTime, 
                        true, null);
                LoginStatDatabase.recordLoginAttempt(loginAttempt);
                return true;
            } else {
                httpSess.invalidate();
                log.warn("Failed Administrator login attempt \'"
                        + username + "\' at " + date.format(currentTime) + " from IP address "
                        + remoteAddress);
                LoginAttempt loginAttempt = new LoginAttempt(username, remoteAddress, currentTime, 
                        false, "Invalid credentials");
                LoginStatDatabase.recordLoginAttempt(loginAttempt);
                return false;
            }
        } catch (Exception e) {
            String msg = "System error while Authenticating/Authorizing User";
            log.error(msg, e);
            return false;
        }
    }
    
    /**
	 * Supports delegated identity logins. A third party system authenticates the authenticatedUser
	 * - had to accept it's assertion the system needs to authenticates it self to carbon. Also the
	 * only the users who belong the roles having delegate-identity permission could perform this
	 * action.
	 * 
	 * @param username User who could delegate logins for other users.
	 * @param password Password of the user who could delegate logins for other users.
	 * @param authenticatedUser The user who's being authenticated by a third party system.
	 * @param remoteAddress
	 * @return
	 * @throws AuthenticationException
	 */
	public boolean loginWithDelegation(String username, String password, String authenticatedUser,
			String remoteAddress) throws AuthenticationException {

		MessageContext msgCtx = MessageContext.getCurrentMessageContext();
		HttpServletRequest request = (HttpServletRequest) msgCtx
				.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
		HttpSession httpSess = request.getSession();

		Date currentTime = Calendar.getInstance().getTime();
		SimpleDateFormat date = new SimpleDateFormat("'['yyyy-MM-dd HH:mm:ss,SSSS']'");

		if ((username == null) || (password == null) || (remoteAddress == null)
				|| (authenticatedUser == null) || username.trim().equals("")
				|| password.trim().equals("") || authenticatedUser.trim().equals("")
				|| remoteAddress.trim().equals("")) {
			log.warn("Failed login with identity delegation \'" + authenticatedUser + "\' at "
					+ date.format(currentTime) + " from IP address " + remoteAddress);
			httpSess.invalidate();
			return false;
		}

		try {

			UserRealm realm = CarbonServicesServiceComponent.getUserRealm();
			boolean isAuthenticated;

			isAuthenticated = realm.getAuthenticator().authenticate(username, password);

			if (!isAuthenticated) {
				httpSess.invalidate();
				log.warn("Failed Administrator login attempt iwth identity delegation \'"
						+ username + "\' at " + date.format(currentTime) + " from IP address "
						+ remoteAddress);
				LoginAttempt loginAttempt = new LoginAttempt(username, remoteAddress, currentTime, 
                        false, "Invalid credentials");
                LoginStatDatabase.recordLoginAttempt(loginAttempt);
				return false;
			}

			// If we are to trust the user who delegates identity - he should be in a role having
			// delegate-identity permission.
			boolean isDelegateToAuthorized = realm.getAuthorizer().isUserAuthorized(username,
					"System", "delegate-identity");

			// authenticatedUser user should have the permission to login to the system.
			boolean isLoginToAuthorized = realm.getAuthorizer().isUserAuthorized(authenticatedUser,
					"System", "login");

			if (isDelegateToAuthorized && isLoginToAuthorized) {
				httpSess.setAttribute(ServerConstants.ADMIN_LOGGED_IN, "true");
				log.info("\'" + authenticatedUser + "\' logged in at "
						+ date.format(currentTime) + " from IP address " + remoteAddress);

				// create a UserRegistry instance for the logged in user. This instance is used to
				// perform Registry actions performed by *this* user. If a Registry action is
				// performed by the system, it is done using the system registry instance, although
				// user's registry
				// is available. In most cases, user registry instance is used only by the Registry
				// UI bundles as those actions are performed directly by the currently logged in
				// user.
                UserRegistry userRegistry = CarbonServicesServiceComponent.getRegistryService()
                        .getUserRegistry(authenticatedUser);
				request.getSession().setAttribute(RegistryConstants.USER_REGISTRY, userRegistry);
				request.getSession().setAttribute(ServerConstants.ADMIN_USERNAME, authenticatedUser);
				request.getSession().setAttribute("DELEGATED_BY", username);

				if (log.isDebugEnabled()) {
					log.debug("User Registry instance is set in the session for user "
							+ authenticatedUser);
				}
				
				log.info("Identity delegation by " + username
						+ " on behalf of " + authenticatedUser + " at "
						+ date.format(currentTime) + " from IP address "
						+ remoteAddress);

				return true;
			} else {
				httpSess.invalidate();
				log.warn("Failed Administrator login attempt with identity delegation \'"
						+ username + "\' at " + date.format(currentTime) + " from IP address "
						+ remoteAddress);
				LoginAttempt loginAttempt = new LoginAttempt(username, remoteAddress, currentTime, 
                        false, "Invalid credentials");
                LoginStatDatabase.recordLoginAttempt(loginAttempt);
				return false;
			}
		} catch (Exception e) {
			String msg = "System error while Authenticating/Authorizing User with identity delegation";
			log.error(msg, e);
			return false;
		}
	}

    public void logout() throws AuthenticationException {
        String loggedInUser = null;
        String delegatedBy = null;
        Date currentTime = Calendar.getInstance().getTime();
        SimpleDateFormat date = new SimpleDateFormat("'['yyyy-MM-dd HH:mm:ss,SSSS']'");
        MessageContext msgCxt = MessageContext.getCurrentMessageContext();
        HttpServletRequest request = (HttpServletRequest) msgCxt
                .getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        HttpSession session = request.getSession();
        loggedInUser = (String) session.getAttribute(ServerConstants.ADMIN_USERNAME);
        delegatedBy = (String) session.getAttribute("DELEGATED_BY");
        if (delegatedBy == null) {
            log.info("'" + loggedInUser + "' logged out at " + date.format(currentTime));
        } else {
            log.info("'" + loggedInUser + "' logged out at " + date.format(currentTime) +
                     " delegated by " + delegatedBy);
        }
        session.invalidate();
    }

    public String[] getSystemPermissionOfUser(String username) throws Exception {
        try {
            MessageContext msgCtx = MessageContext.getCurrentMessageContext();
            HttpServletRequest request =
                    (HttpServletRequest) msgCtx.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
            HttpSession httpSess = request.getSession();
            UserRealm realm = CarbonServicesServiceComponent.getUserRealm();
            List<String> userPermissions = getUserPermissions(username, realm);
            String[] permissions = userPermissions.toArray(new String[userPermissions.size()]);
            return permissions;
        } catch(Exception e) {
            log.error(e);
            throw e;
        }
    }


    public List<String> getUserPermissions(String username, UserRealm realm) throws Exception{
        List<String> permissions = SystemPermissions.getSystemPermission();
        List<String> userPermissions = new ArrayList<String>();
        Authorizer authorizer = realm.getAuthorizer();
        for(String permission : permissions){
            if(authorizer.isUserAuthorized(username, "System", permission)){
                userPermissions.add(permission);
            }
        }
        return userPermissions;
    }
}
