/*
 * Copyright 2005-2008 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.transports;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.core.persistence.PersistenceManager;
import org.wso2.carbon.core.transports.util.TransportDetails;
import org.wso2.carbon.core.transports.util.TransportParameter;

/**
 * AbstractTransportAdmin encapsulates all the common functionalities required by
 * transport admin implementations. This involves fetching parameters related to
 * transport listeners and transport senders as well as updating them. It is
 * recommended that all transport admin implementations, in their respective components
 * extend from this abstract class to prevent duplication of code. By extending this
 * abstract class, transport admin implementation not only inherit the common
 * transport admin behavior but they also inherit some useful references to the
 * Axis configuration, transport loader and the persistence manager.
 */
public abstract class AbstractTransportAdmin {

    private String transportName;
    protected PersistenceManager pm;
    protected CommonTransportLoader loader;
    protected AxisConfiguration axisConfig;

    private static final Log log = LogFactory.getLog(AbstractTransportAdmin.class);

    /**
     * The implementations of this abstract method should try to determine
     * whether all the dependencies required to enable a particular transport
     * is available in the classpath.
     *
     * @return true if the dependencies are available and false otherwise
     */
    protected abstract boolean dependenciesAvailable();

    public AbstractTransportAdmin(String transportName) {
        if (log.isDebugEnabled()) {
            log.debug("Initializing " + transportName + " transport admin");
        }
        this.transportName = transportName;
        this.axisConfig = CommonTransportUtils.getAxisConfiguration();
        this.loader = new CommonTransportLoader(axisConfig);
    }

    /**
     * Get the set of service specific transport listener parameters
     *
     * @param serviceName name of the service
     * @return an array of TransportParameter objects
     * @throws Exception if the service name is invalid or loading of transport details fails
     */
    public TransportParameter[] getServiceSpecificInParameters(String serviceName) throws Exception {
        TransportDetails details;

        if (serviceName == null) {
            if (log.isDebugEnabled()) {
                log.debug("Invalid service name");
            }
            throw new Exception("Invalid service name");
        }

        details = loader.getTransportDetails(serviceName, this.transportName);

        if (details == null || details.getInParameters() == null ||
                details.getInParameters().length == 0) {
            if (log.isDebugEnabled()) {
                log.debug("No " + this.transportName +" specific transport listener parameters defined" +
                        " for the service " + serviceName);
            }
            return null;
        }

        return details.getInParameters();
    }

    /**
     * Get the set of service specific transport sender parameters
     *
     * @param serviceName name of the service
     * @return an array of TransportParameter objects
     * @throws Exception if the service name is invalid or loading of transport details fails
     */
    public TransportParameter[] getServiceSpecificOutParameters(String serviceName) throws Exception {
        TransportDetails details;

        if (serviceName == null) {
            if (log.isDebugEnabled()) {
                log.debug("Invalid service name");
            }
            throw new Exception("Invalid service name");
        }

        details = loader.getTransportDetails(serviceName, this.transportName);

        if (details == null || details.getOutParameters() == null ||
                details.getOutParameters().length == 0) {
            if (log.isDebugEnabled()) {
                log.debug("No " + this.transportName + " specific transport sender parameters " +
                        "defined for the service " + serviceName);
            }
            return null;
        }

        return details.getOutParameters();
    }

    /**
     * Get the set of global transport listener parameters
     *
     * @return an array of TransportParameter objects
     * @throws Exception if the loading of transport details fails
     */
    public TransportParameter[] getGloballyDefinedInParameters() throws Exception {
        TransportDetails details = loader.getTransportDetails(this.transportName);

        if (details == null) {
            if (log.isDebugEnabled()) {
                log.debug("No " + this.transportName + " specific transport listener parameters defined.");
            }
            return null;
        }

        return details.getInParameters();
    }

    /**
     * Get the set of global transport sender parameters
     *
     * @return an array of TransportParameter objects
     * @throws Exception if the loading of transport details fails
     */
    public TransportParameter[] getGloballyDefinedOutParameters() throws Exception {
        TransportDetails details = loader.getTransportDetails(this.transportName);

        if (details == null) {
            if (log.isDebugEnabled()) {
                log.debug("No " + this.transportName + " specific transport sender parameters defined.");
            }
            return null;
        }

        return details.getOutParameters();
    }

    /**
     * Update the set of global transport listener parameters
     *
     * @param inParams set of transport parameters to be updated
     * @throws Exception if loading the transport details fails
     */
    public void updateGloballyDefinedInParameters(TransportParameter[] inParams) throws Exception {

        boolean callDestroy = false;
		TransportInDescription inTransport = this.axisConfig.getTransportIn(this.transportName);

        if (!dependenciesAvailable()) {
           return;
        }

		// This will update in-memory transport details.
		loader.updateTransportInDetails(this.transportName, inParams);

        if (inTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		} else {
			// This is not active yet - we need to enable it first.
			loader.enableTransport();
		}

		// Let's get the updated configuration.
		inTransport = this.axisConfig.getTransportIn(this.transportName);

        PersistenceManager pm = getPersistenceManager();

		try {
			// This will make sure parameter updates are persistent.
			pm.updateInTransport(this.transportName, inTransport.getParameters());
        } catch (Exception e) {
            String msg = "Error while updating transport listener parameters for " +
                    this.transportName +" transport";
            log.error(msg, e);
			throw new Exception(msg, e);
		}

		// Now we need to re-initialize the transport listener.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportInUpdateThread(inTransport, callDestroy).start();
    }

    /**
     * Update the set of global transport sender parameters
     *
     * @param outParams set of transport parameters to be updated
     * @throws Exception if loading the transport details fails
     */
    public void updateGloballyDefinedOutParameters(TransportParameter[] outParams) throws Exception {

        boolean callDestroy = false;
		TransportOutDescription outTransport = this.axisConfig.getTransportOut(this.transportName);

       if (!dependenciesAvailable()) {
           return;
       }

		// This will update in-memory transport details.
		loader.updateTransportOutDetails(this.transportName, outParams);

        if (outTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		}

		// Let's get the updated configuration.
		outTransport = this.axisConfig.getTransportOut(this.transportName);
        if (outTransport == null) {
            return;
        }

        PersistenceManager pm = getPersistenceManager();

		try {
			// This will make sure parameter updates are persistent.
			pm.updateOutTransport(this.transportName, outTransport.getParameters());
        } catch (Exception e) {
            String msg = "Error while updating transport listener parameters for " +
                    this.transportName +" transport";
            log.error(msg, e);
			throw new Exception(msg, e);
		}

		// Now we need to re-initialize the transport listener.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportOutUpdateThread(outTransport, callDestroy).start();
    }

    /**
     * Update the set of transport listener parameters for a specified service
     *
     * @param serviceName name of the service
     * @param inParams set of transport parameters to be updated
     * @throws Exception if loading of the transport details fails
     */
    public void updateServiceSpecificInParameters(String serviceName,
                                                  TransportParameter[] inParams) throws Exception {

        boolean callDestroy = false;
		TransportInDescription inTransport = this.axisConfig.getTransportIn(this.transportName);

        if (!dependenciesAvailable()) {
            return;
        }

		// This will update in-memory transport details.
		loader.updateTransportInDetails(serviceName, this.transportName, inParams);

        if (inTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		} else {
			// This not active yet - we need to enable it first.
			loader.enableTransport();
		}

		// Let's get the updated configuration.
		inTransport = this.axisConfig.getTransportIn(this.transportName);

        pm = getPersistenceManager();

		try {
			// This will make sure parameter updates are persistent.
			pm.updateInTransport(this.transportName, inTransport.getParameters());
        } catch (Exception e) {
            String msg = "Error while updating transport parameters for " + transportName +" transport";
            log.error(msg, e);
			throw new Exception(msg, e);
		}

		// Now we need to re-initialize transport listener and sender.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportInUpdateThread(inTransport, callDestroy).start();
	}

    /**
     * Update the set of transport sender parameters for a specified service
     *
     * @param serviceName name of the service
     * @param outParams set of transport parameters to be updated
     * @throws Exception if loading of the transport details fails
     */
    public void updateServiceSpecificOutParameters(String serviceName,
                                                   TransportParameter[] outParams) throws Exception {

        boolean callDestroy = false;
        TransportOutDescription outTransport;

        if (!dependenciesAvailable()) {
            return;
        }

		// This will update in-memory transport details.
        loader.updateTransportOutDetails(serviceName, this.transportName, outParams);

        outTransport = this.axisConfig.getTransportOut(this.transportName);

        if (outTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		}

		// Let's get the updated configuration.
        outTransport = this.axisConfig.getTransportOut(this.transportName);
        if (outTransport == null) {
            return;
        }

        pm = getPersistenceManager();

		try {
			// This will make sure parameter updates are persistent.
            pm.updateOutTransport(this.transportName, outTransport.getParameters());
        } catch (Exception e) {
            String msg = "Error while updating transport parameters for " + transportName +" transport";
            log.error(msg, e);
			throw new Exception(msg, e);
		}

		// Now we need to re-initialize transport listener and sender.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportOutUpdateThread(outTransport, callDestroy).start();
	}

    /**
     * Enables the transport listener using the specified set of listener and sender
     * parameters
     *
     * @param inParams transport listener parameters
     * @throws Exception if updating the parameters fails
     */
    public void enableTransportListener(TransportParameter[] inParams) throws Exception {

        this.updateGloballyDefinedInParameters(inParams);
    }

    /**
     * Enables the transport sender using the specified set of listener and sender
     * parameters
     *
     * @param outParams transport sender parameters
     * @throws Exception if updating the parameters fails
     */
    public void enableTransportSender(TransportParameter[] outParams) throws Exception {

        this.updateGloballyDefinedOutParameters(outParams);
    }

    /**
     * Disables the transport listener and removes the related descriptions from the
     * Axis configuration
     *
     * @throws Exception if updating the transport details fails
     */
    public void disableTransportListener() throws Exception {

        // To disable this is an already enabled transport - so let's get the reference to it first.
		TransportInDescription inTransport = this.axisConfig.getTransportIn(this.transportName);

        // Now we are removing the transport from AxisConfiguration.
		this.axisConfig.getTransportsIn().remove(this.transportName);

        pm = getPersistenceManager();

		try {
			// This will update the registry to persist the removal.
			pm.disableTransportListener(this.transportName);
		} catch (Exception e) {
            String msg = "Error while disabling " + this.transportName +" transport";
            log.error(msg, e);
			throw new Exception(msg, e);
		}

		// Now we need to stop transport listener.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportInRemoveThread(inTransport).start();
	}

    /**
     * Disables the transport sender and removes the related descriptions from the
     * Axis configuration
     *
     * @throws Exception if updating the transport details fails
     */
    public void disableTransportSender() throws Exception {

        // To disable this is an already enabled transport - so let's get the reference to it first.
        TransportOutDescription outTransport = this.axisConfig.getTransportOut(this.transportName);

        // Now we are removing the transport from AxisConfiguration.
        this.axisConfig.getTransportsOut().remove(this.transportName);

        pm = getPersistenceManager();

		try {
			// This will update the registry to persist the removal.
			pm.disableTransportSender(this.transportName);
		} catch (Exception e) {
            String msg = "Error while disabling " + this.transportName +" transport";
            log.error(msg, e);
			throw new Exception(msg, e);
		}

		// Now we need to stop the transport sender.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportOutRemoveThread(outTransport).start();
	}

    /**
	 * Re-initializes transport listener in a different thread.
	 */
	private static class TransportInUpdateThread extends Thread {

		TransportInDescription inTransport = null;
        ConfigurationContext config = null;
		boolean callDestroy = false;

		public TransportInUpdateThread(TransportInDescription inTransport, boolean callDestroy) {
			this.config = CommonTransportUtils.getConfigurationContext();
			this.inTransport = inTransport;
            this.callDestroy = callDestroy;
		}

		public void run() {
			try {
				if (callDestroy) {
					// This is the case where already enabled transport being updated.
					// So, we need to call destroy() first.
                    inTransport.getReceiver().stop();
                    inTransport.getReceiver().destroy();
                }

                inTransport.getReceiver().init(config, inTransport);
                inTransport.getReceiver().start();

            } catch (Exception e) {
				log.error("Error while updating globally defined " +
                        "parameters for " + inTransport.getName(), e);
			}
		}
	}

    private static class TransportOutUpdateThread extends Thread {

		TransportOutDescription outTransport = null;
        ConfigurationContext config = null;
		boolean callDestroy = false;

		public TransportOutUpdateThread(TransportOutDescription inTransport, boolean callDestroy) {
			this.config = CommonTransportUtils.getConfigurationContext();
			this.outTransport = inTransport;
            this.callDestroy = callDestroy;
		}

		public void run() {
			try {
				if (callDestroy) {
					// This is the case where already enabled transport being updated.
					// So, we need to call destroy() first.
					outTransport.getSender().stop();
                }

                outTransport.getSender().init(config, outTransport);

            } catch (Exception e) {
				log.error("Error while updating globally defined " +
                        "parameters for " + outTransport.getName(), e);
			}
		}
	}

    private static class TransportInRemoveThread extends Thread {
		TransportInDescription inTransport = null;

        public TransportInRemoveThread(TransportInDescription inTransport) {
			this.inTransport = inTransport;
        }

		public void run() {
			try {
				inTransport.getReceiver().destroy();
				inTransport.getReceiver().stop();
            } catch (Exception e) {
				log.error("Error while stopping " + inTransport.getName() + "  transport listener", e);
			}
		}
	}

    private static class TransportOutRemoveThread extends Thread {
		TransportOutDescription outTransport = null;

        public TransportOutRemoveThread(TransportOutDescription outTransport) {
			this.outTransport = outTransport;
        }

		public void run() {
			try {
				outTransport.getSender().stop();
            } catch (Exception e) {
				log.error("Error while stopping " + outTransport.getName() + "  transport listener", e);
			}
		}
	}

    /**
     * Get the PersistenceManager associated with this transport admin instance.
     *
     * @return a PersistenceManager object
     * @throws AxisFault if creation of the PersistenceManager fails
     */
    protected PersistenceManager getPersistenceManager() throws AxisFault {
		if (pm == null) {
			pm = new PersistenceManager(this.axisConfig);
		}
		return pm;
	}
  
}
