/*
 * Copyright 2005-2008 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.transports;

import org.apache.axis2.AxisFault;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.core.transports.util.TransportParameter;
import org.wso2.carbon.core.transports.util.TransportDetails;

/**
 * This abstract class implements the TransportService interface and encapsulates
 * the functionalities common to all transport service implementations. It is
 * recommended that all transport service implementations extends this class
 * instead of directly implementing the TransportService interface, to prevent
 * code duplication.
 */
public abstract class AbstractTransportService implements TransportService {

    protected CommonTransportLoader loader;
    protected String transportName;

    private static final Log log = LogFactory.getLog(AbstractTransportService.class);

    public AbstractTransportService(String transportName, AxisConfiguration axisConfig) {
        if (log.isDebugEnabled()) {
            log.debug("Initializing " + transportName + " transport service");
        }
        this.transportName = transportName;
        this.loader = new CommonTransportLoader(axisConfig);
    }

    public String getName() {
        return this.transportName;
    }

    public TransportParameter[] getTransportInParameters() {
        try {
			TransportDetails trpDetails = loader.getTransportDetails(this.transportName);
            if (trpDetails != null) {
                return trpDetails.getInParameters();
            }
		} catch (AxisFault e) {
			log.error("Error while loading transport listener parameters " +
                                            "for " + this.transportName + " transport", e);
		}
        return null;
    }

    public TransportParameter[] getTransportOutParameters() {
        try {
            TransportDetails trpDetails = loader.getTransportDetails(this.transportName);
            if (trpDetails != null) {
                return trpDetails.getOutParameters();
            }
		} catch (AxisFault e) {
			log.error("Error while loading transport sender parameters " +
                                            "for " + this.transportName + " transport", e);
		}
        return null;
    }

    public boolean isEnableAtStartup() {
        try {
			// This basically checks whether required configuration details available for the
			// corresponding transport.If so this transport will be available in the management
			// console for enabling/disabling.
			if (loader.getTransportDetails(getName()) != null) {
				return true;
			}
		} catch (AxisFault e) {
			log.error("Error while checking enabled at startup for " + this.transportName + " transport", e);
			return false;
		}
		return false;
    }

    public boolean isListenerActive() {
        try {
			return loader.isListenerActive(this.transportName);
		} catch (AxisFault e) {
			log.error("Error while checking active for " + this.transportName + "  transport", e);
			return false;
		}
    }

    public boolean isSenderActive() {
        try {
			return loader.isSenderActive(this.transportName);
		} catch (AxisFault e) {
			log.error("Error while checking active for " + this.transportName + "  transport", e);
			return false;
		}
    }

    public abstract boolean dependenciesAvailable(TransportParameter[] params);
}
