/*
 * Copyright 2005-2008 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.transports;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.llom.util.AXIOMUtil;
import org.wso2.carbon.core.transports.util.TransportParameter;
import org.wso2.carbon.core.transports.util.TransportDetails;
import org.wso2.carbon.core.persistence.PersistenceManager;


public class CommonTransportAdmin {

    private static final Log log = LogFactory.getLog(CommonTransportAdmin.class);

    private AxisConfiguration axisConfig;
    private CommonTransportLoader loader;

    public CommonTransportAdmin(AxisConfiguration axisConfig) {
        this.axisConfig = axisConfig;
        this.loader = new CommonTransportLoader(axisConfig);
    }

    /**
     * Get the set of global transport listener parameters
     *
     * @param transportName name of the transport
     * @return an array of TransportParameter objects
     * @throws Exception if the loading of transport details fails
     */
    public TransportParameter[] getGloballyDefinedInParameters(String transportName) throws Exception {
        TransportDetails details = loader.getTransportDetails(transportName);

        if (details == null || details.getInParameters() == null) {
            if (log.isDebugEnabled()) {
                log.debug("No " + transportName + " specific transport listener parameters defined.");
            }
            return null;
        }

        return details.getInParameters();
    }

    /**
     * Get the set of global transport sender parameters
     *
     * @param transportName name of the transport
     * @return an array of TransportParameter objects
     * @throws Exception if the loading of transport details fails
     */
    public TransportParameter[] getGloballyDefinedOutParameters(String transportName) throws Exception {
        TransportDetails details = loader.getTransportDetails(transportName);

        if (details == null || details.getOutParameters() == null) {
            if (log.isDebugEnabled()) {
                log.debug("No " + transportName + " specific transport sender parameters defined.");
            }
            return null;
        }

        return details.getOutParameters();
    }

    /**
     * Get the set of service specific transport listener parameters
     *
     * @param transportName name of the transport
     * @param serviceName name of the service
     * @return an array of TransportParameter objects
     * @throws Exception if the service name is invalid or loading of transport details fails
     */
    public TransportParameter[] getServiceSpecificInParameters(String transportName,
                                                               String serviceName) throws Exception {
        TransportDetails details;

        if (serviceName == null) {
            if (log.isDebugEnabled()) {
                log.debug("Invalid service name");
            }
            throw new Exception("Invalid service name");
        }

        details = loader.getTransportDetails(serviceName, transportName);

        if (details == null || details.getInParameters() == null ||
                details.getInParameters().length == 0) {
            if (log.isDebugEnabled()) {
                log.debug("No " + transportName +" specific transport listener parameters defined" +
                        " for the service " + serviceName);
            }
            return null;
        }

        return details.getInParameters();
    }

    /**
     * Get the set of service specific transport sender parameters
     *
     * @param transportName name of the transport
     * @param serviceName name of the service
     * @return an array of TransportParameter objects
     * @throws Exception if the service name is invalid or loading of transport details fails
     */
    public TransportParameter[] getServiceSpecificOutParameters(String transportName, String serviceName) throws Exception {
        TransportDetails details;

        if (serviceName == null) {
            if (log.isDebugEnabled()) {
                log.debug("Invalid service name");
            }
            throw new Exception("Invalid service name");
        }

        details = loader.getTransportDetails(serviceName, transportName);

        if (details == null || details.getOutParameters() == null ||
                details.getOutParameters().length == 0) {
            if (log.isDebugEnabled()) {
                log.debug("No " + transportName + " specific transport sender parameters " +
                        "defined for the service " + serviceName);
            }
            return null;
        }

        return details.getOutParameters();
    }

    /**
     * Update the set of global transport listener parameters
     *
     * @param transportName name of the transport
     * @param inParams set of transport parameters to be updated
     * @throws Exception if loading the transport details fails
     */
    public void updateGloballyDefinedInParameters(String transportName,
                                                  TransportParameter[] inParams) throws Exception {

        if (log.isDebugEnabled()) {
            log.debug("Updating global transport receiver parameters of " + transportName);
        }

        boolean callDestroy = false;
		TransportInDescription inTransport = this.axisConfig.getTransportIn(transportName);

        // This will update in-memory transport details.
		loader.updateTransportInDetails(transportName, inParams);

        if (inTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		} else {
			// This is not active yet - we need to enable it first.
			loader.enableTransport();
		}

		// Let's get the updated configuration.
		inTransport = axisConfig.getTransportIn(transportName);

        // This will make sure parameter updates are persistent.
        PersistenceManager pm = new PersistenceManager(axisConfig);
        pm.updateInTransport(transportName, inTransport.getParameters());

		// Now we need to re-initialize the transport listener.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportInUpdateThread(inTransport, callDestroy).start();
    }

    /**
     * Update the set of global transport sender parameters
     *
     * @param transportName name of the transport
     * @param outParams set of transport parameters to be updated
     * @throws Exception if loading the transport details fails
     */
    public void updateGloballyDefinedOutParameters(String transportName,
                                                   TransportParameter[] outParams) throws Exception {

        if (log.isDebugEnabled()) {
            log.debug("Updating global transport sender parameters of " + transportName);
        }

        boolean callDestroy = false;
		TransportOutDescription outTransport = this.axisConfig.getTransportOut(transportName);

		// This will update in-memory transport details.
		loader.updateTransportOutDetails(transportName, outParams);

        if (outTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		}

		// Let's get the updated configuration.
		outTransport = this.axisConfig.getTransportOut(transportName);
        if (outTransport == null) {
            return;
        }

        PersistenceManager pm = new PersistenceManager(axisConfig);
        pm.updateOutTransport(transportName, outTransport.getParameters());

		// Now we need to re-initialize the transport listener.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportOutUpdateThread(outTransport, callDestroy).start();
    }

    /**
     * Update the set of transport listener parameters for a specified service
     *
     * @param transportName name of the transport
     * @param serviceName name of the service
     * @param inParams set of transport parameters to be updated
     * @throws Exception if loading of the transport details fails
     */
    public void updateServiceSpecificInParameters(String transportName, String serviceName,
                                                  TransportParameter[] inParams) throws Exception {

        boolean callDestroy = false;
		TransportInDescription inTransport = this.axisConfig.getTransportIn(transportName);

		// This will update in-memory transport details.
		loader.updateTransportInDetails(serviceName, transportName, inParams);

        if (inTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		} else {
			// This not active yet - we need to enable it first.
			loader.enableTransport();
		}

		// Let's get the updated configuration.
		inTransport = this.axisConfig.getTransportIn(transportName);

        PersistenceManager pm = new PersistenceManager(axisConfig);
        pm.updateInTransport(transportName, inTransport.getParameters());

		// Now we need to re-initialize transport listener and sender.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportInUpdateThread(inTransport, callDestroy).start();
	}

    /**
     * Update the set of transport sender parameters for a specified service
     *
     * @param transportName name of the transport
     * @param serviceName name of the service
     * @param outParams set of transport parameters to be updated
     * @throws Exception if loading of the transport details fails
     */
    public void updateServiceSpecificOutParameters(String transportName, String serviceName,
                                                   TransportParameter[] outParams) throws Exception {

        boolean callDestroy = false;
        TransportOutDescription outTransport;

		// This will update in-memory transport details.
        loader.updateTransportOutDetails(serviceName, transportName, outParams);

        outTransport = this.axisConfig.getTransportOut(transportName);

        if (outTransport != null) {
			// Transport already active - All transports already loaded in to axis2configuration are
			// considered as active.Here we need to call the destroy() of TransportListener.
			callDestroy = true;
		}

		// Let's get the updated configuration.
        outTransport = this.axisConfig.getTransportOut(transportName);
        if (outTransport == null) {
            return;
        }

        PersistenceManager pm = new PersistenceManager(axisConfig);
        pm.updateOutTransport(transportName, outTransport.getParameters());

        // Now we need to re-initialize transport listener and sender.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportOutUpdateThread(outTransport, callDestroy).start();
	}

    /**
     * Disables the transport listener and removes the related descriptions from the
     * Axis configuration
     *
     * @param transportName name of the transport
     * @throws Exception if updating the transport details fails
     */
    public void disableTransportListener(String transportName) throws Exception {

        // To disable this is an already enabled transport - so let's get the reference to it first.
		TransportInDescription inTransport = this.axisConfig.getTransportIn(transportName);

        // Now we are removing the transport from AxisConfiguration.
		this.axisConfig.getTransportsIn().remove(transportName);

        PersistenceManager pm = new PersistenceManager(axisConfig);
        pm.disableTransportListener(transportName);

		// Now we need to stop transport listener.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportInRemoveThread(inTransport).start();
	}

    /**
     * Disables the transport sender and removes the related descriptions from the
     * Axis configuration
     *
     * @param transportName name of the transport
     * @throws Exception if updating the transport details fails
     */
    public void disableTransportSender(String transportName) throws Exception {

        // To disable this is an already enabled transport - so let's get the reference to it first.
        TransportOutDescription outTransport = this.axisConfig.getTransportOut(transportName);

        // Now we are removing the transport from AxisConfiguration.
        this.axisConfig.getTransportsOut().remove(transportName);

        PersistenceManager pm = new PersistenceManager(axisConfig);
        pm.disableTransportSender(transportName);

		// Now we need to stop the transport sender.
		// This runs on a different thread - we need not to wait till this finishes to return back.
		new TransportOutRemoveThread(outTransport).start();
	}

    /**
     * Add a new transport parameter to a specified transport
     *
     * @param transportName name of the transport listener
     * @param serviceName name of the service or null if the parameter is global
     * @param parameter new transport parameter
     * @param listener true for listener and false for sender
     * @throws Exception on error
     */
    public void addTransportParameter(String transportName, String serviceName,
                                      TransportParameter parameter,
                                      boolean listener) throws Exception {

        boolean updateInMemory;

        // Check whether the transport is enabled and available from the axis configuration
        if (listener) {
            updateInMemory = loader.isListenerActive(transportName);
        } else {
            updateInMemory = loader.isSenderActive(transportName);
        }


        if (!updateInMemory) {
            // If the transport is not enabled we simply need to
            // add the new parameter to the registry
            Parameter trpParameter = new Parameter();
            trpParameter.setName(parameter.getName());
            trpParameter.setValue(parameter.getValue());
            if (parameter.getParamElement() != null) {
                OMElement elem = AXIOMUtil.stringToOM(parameter.getParamElement());
                trpParameter.setParameterElement(elem);
            }

            PersistenceManager pm = new PersistenceManager(axisConfig);
            pm.addTransportParameter(transportName, trpParameter, listener);
            log.info("Saved " + transportName + " transport parameter named " + parameter.getName());

        } else {
            TransportParameter[] trpParams = getTransportParameters(transportName, serviceName, listener);
            boolean update = false;

            // Check whether the parameter is already available
            for (int i = 0; i < trpParams.length; i++) {
                if (trpParams[i].getName().equals(parameter.getName())) {
                    // Overwrite the existing parameter
                    trpParams[i] = parameter;
                    update = true;
                    break;
                }
            }

            if (!update) {
                // Create a new array with the new parameter and all the existing ones
                TransportParameter[] newTrpParams = new TransportParameter[trpParams.length + 1];
                System.arraycopy(trpParams, 0, newTrpParams, 0, trpParams.length);
                newTrpParams[newTrpParams.length - 1] = parameter;
                updateTransportParameters(transportName, serviceName, listener, newTrpParams);
                log.info("Saved " + transportName + " transport parameter named " + parameter.getName());
            } else {
                updateTransportParameters(transportName, serviceName, listener, trpParams);
                log.info("Updated " + transportName + " transport parameter named " + parameter.getName());
            }
        }
    }

    public void removeTransportParameter(String transportName, String service, String parameterName,
                                         boolean listener) throws Exception {
        boolean updateInMemory;

        // Check whether the transport is enabled and available from the axis configuration
        if (listener) {
            updateInMemory = loader.isListenerActive(transportName);
        } else {
            updateInMemory = loader.isSenderActive(transportName);
        }

        if (!updateInMemory) {
            // If the transport is not available we simply need to remove the
            // parameter from the registry
            PersistenceManager pm = new PersistenceManager(axisConfig);
            pm.removeTransportParameter(transportName, parameterName, listener);
        } else {
            TransportParameter[] trpParams = getTransportParameters(transportName, service, listener);
            boolean update = false;

            if (trpParams.length == 0) {
                return;
            }

            // Make sure the parameter to be deleted exists
            for (TransportParameter p : trpParams) {
                if (p.getName().equals(parameterName)) {
                    update = true;
                    break;
                }
            }

            if (update) {
                TransportParameter[] newTrpParams = new TransportParameter[trpParams.length - 1];
                int j = 0;
                for (TransportParameter p : trpParams) {
                    if (!p.getName().equals(parameterName)) {
                        newTrpParams[j++] = p;
                    }
                }

                updateTransportParameters(transportName, service, listener, newTrpParams);
            }
        }
    }

    private TransportParameter[] getTransportParameters(String protocol, String service,
                                                        boolean listener) throws Exception {

        if (service == null) {
            if (listener) {
                return getGloballyDefinedInParameters(protocol);
            } else {
                return getGloballyDefinedOutParameters(protocol);
            }
        } else {
            if (listener) {
                return getServiceSpecificInParameters(protocol, service);
            } else {
                return getServiceSpecificOutParameters(protocol, service);
            }
        }
    }

    private void updateTransportParameters(String protocol, String service, boolean listener,
                                           TransportParameter[] params) throws Exception {

        if (service == null) {
            if (listener) {
                updateGloballyDefinedInParameters(protocol, params);
            } else {
                updateGloballyDefinedOutParameters(protocol, params);
            }
        } else {
            if (listener) {
                updateServiceSpecificInParameters(protocol, service, params);
            } else {
                updateServiceSpecificOutParameters(protocol, service, params);
            }
        }
    }

    /**
	 * Re-initializes transport listener in a different thread.
	 */
	private static class TransportInUpdateThread extends Thread {

		TransportInDescription inTransport = null;
        ConfigurationContext config = null;
		boolean callDestroy = false;

		public TransportInUpdateThread(TransportInDescription inTransport, boolean callDestroy) {
			this.config = CommonTransportUtils.getConfigurationContext();
			this.inTransport = inTransport;
            this.callDestroy = callDestroy;
		}

		public void run() {
			try {
				if (callDestroy) {
					// This is the case where already enabled transport being updated.
					// So, we need to call destroy() first.
                    inTransport.getReceiver().stop();
                    inTransport.getReceiver().destroy();
                }

                inTransport.getReceiver().init(config, inTransport);
                inTransport.getReceiver().start();

            } catch (Exception e) {
				log.error("Error while updating globally defined " +
                        "parameters for " + inTransport.getName(), e);
			}
		}
	}

    private static class TransportOutUpdateThread extends Thread {

		TransportOutDescription outTransport = null;
        ConfigurationContext config = null;
		boolean callDestroy = false;

		public TransportOutUpdateThread(TransportOutDescription inTransport, boolean callDestroy) {
			this.config = CommonTransportUtils.getConfigurationContext();
			this.outTransport = inTransport;
            this.callDestroy = callDestroy;
		}

		public void run() {
			try {
				if (callDestroy) {
					// This is the case where already enabled transport being updated.
					// So, we need to call destroy() first.
					outTransport.getSender().stop();
                }

                outTransport.getSender().init(config, outTransport);

            } catch (Exception e) {
				log.error("Error while updating globally defined " +
                        "parameters for " + outTransport.getName(), e);
			}
		}
	}

    private static class TransportInRemoveThread extends Thread {
		TransportInDescription inTransport = null;

        public TransportInRemoveThread(TransportInDescription inTransport) {
			this.inTransport = inTransport;
        }

		public void run() {
			try {
				inTransport.getReceiver().destroy();
				inTransport.getReceiver().stop();
            } catch (Exception e) {
				log.error("Error while stopping " + inTransport.getName() + "  transport listener", e);
			}
		}
	}

    private static class TransportOutRemoveThread extends Thread {
		TransportOutDescription outTransport = null;

        public TransportOutRemoveThread(TransportOutDescription outTransport) {
			this.outTransport = outTransport;
        }

		public void run() {
			try {
				outTransport.getSender().stop();
            } catch (Exception e) {
				log.error("Error while stopping " + outTransport.getName() + "  transport listener", e);
			}
		}
	}

}
