/*
 * Copyright 2005-2008 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.transports;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.llom.util.AXIOMUtil;
import org.apache.axis2.AxisFault;
import org.apache.axis2.deployment.DeploymentConstants;
import org.apache.axis2.deployment.util.Utils;
import org.apache.axis2.description.*;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.util.XMLUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.core.persistence.PersistenceManager;
import org.wso2.carbon.core.transports.util.TransportDetails;
import org.wso2.carbon.core.transports.util.TransportParameter;
import org.wso2.carbon.core.transports.util.TransportSummary;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.utils.ServerConstants;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import java.io.*;
import java.net.URL;
import java.util.*;

/**
 * Contains common functionality that can be shared across all transport management bundles -
 * related to load transport relates settings from either the AxisConfiguration or the transport
 * specific configuration files.
 */
public class CommonTransportLoader implements DeploymentConstants {

	private AxisConfiguration axisConfiguration = null;

    private final String TRANSPORT_RECEIVER = "transportReceiver";
    private final String TRANSPORT_SENDER = "transportSender";
    private static Log log = LogFactory.getLog(CommonTransportLoader.class);

    /**
     * Used to construct a general CommonTransportLoader instance which is not bound
     * to any particular transport management bundle
     *
     * @param config Axis2 configuration
     */
    public CommonTransportLoader(AxisConfiguration config) {
        this.axisConfiguration = config;
    }

    /**
     * This method checks whether the configuration of a particular transport is available in
     * the registry. If it does then the method will simply return. Otherwise it will attempt
     * to load the transport configurations from the specified transport configuration file.
     * Once loaded the transport configurations will be permanently stored in the registry so
     * the subsequent calls to this method will always find the transport configurations in
     * the registry.
     *
     * @param transport Name of the transport
     * @param configFileName name of the transport configuration file
     * @param configFileURL URL to the transport configuration file available in the transport bundle
     */
    public void saveTransportConfiguration(String transport, String configFileName, URL configFileURL) {

        PersistenceManager pm;
        Resource trpResource;
        OMElement configXML;

        try {
            pm = new PersistenceManager(axisConfiguration);
            trpResource = pm.getTransport(transport, true);

            // Check whether the transport details are in the registry
            if (trpResource != null) {
                // Transport details found in the registry - Return immediately
                if (log.isDebugEnabled()) {
                    log.debug("Transport configuration for " + transport + " already available in " +
                            "the registry.");
                }
                trpResource.discard();
                return;
            }

            // We need to build the transport configurations from the specified transport config files
            configXML = parseTransportConfiguration(transport, configFileName, configFileURL);
            if (configXML == null) {
                log.error("Error while parsing the transport configuration file");
                return;
            }

            saveTransportReceiverConfiguration(transport, pm, configXML);
            saveTransportSenderConfiguration(transport, pm, configXML);

        } catch (Exception e) {
            log.error("Error while saving transport configuration in the " +
                         "registry for " + transport, e);
        }
    }

    private void saveTransportReceiverConfiguration(String transport, PersistenceManager pm,
                                                    OMElement configXML) {

        TransportInDescription transportIn;
        TransportBuilder transportBuilder = new TransportBuilder();
        OMElement trpReceiver = configXML.getFirstChildWithName(new QName(TRANSPORT_RECEIVER));

        try {
            if (trpReceiver != null) {
                transportIn = transportBuilder.processTransportReceiver(trpReceiver);
                Parameter receiverClassParam = transportIn.getParameter("implClassName");
                String receiverClass = receiverClassParam.getValue().toString();
                // implClassName property no longer needed
                transportIn.removeParameter(receiverClassParam);

                pm.addInTransportOnStartup(transport, false, transportIn.getParameters());
                pm.setTransportProperty(transport, true, "class", receiverClass);

            } else {
                log.warn("No transport receiver configuration found for " + transport);
            }

        } catch (Exception e) {
            log.error("A problem occured while saving the transport receiver configuration " +
                    "for " + transport + " in the registry", e);
        }
    }

    private void saveTransportSenderConfiguration(String transport, PersistenceManager pm,
                                                  OMElement configXML) {

        TransportOutDescription transportOut;
        TransportBuilder transportBuilder = new TransportBuilder();
        OMElement trpSender = configXML.getFirstChildWithName(new QName(TRANSPORT_SENDER));
        try {
            if (trpSender != null) {
                transportOut = transportBuilder.processTransportSender(trpSender);
                Parameter senderClassParam = transportOut.getParameter("implClassName");
                String senderClass = senderClassParam.getValue().toString();
                // implClassName Property no longer needed.
                transportOut.removeParameter(senderClassParam);
                
                pm.addOutTransportOnStartup(transport, false, transportOut.getParameters());
                pm.setTransportProperty(transport, false, "class", senderClass);

            } else {
                log.warn("No transport sender configuration found for " + transport);
            }

        } catch (Exception e) {
            log.error("A problem occured while saving the transport sender configuration " +
                    "for " + transport + " in the registry", e);
        }

    }

    private OMElement parseTransportConfiguration(String transport, String configFileName, URL configFileURL) {

        boolean configLoaded = false;
        InputStream configFileStream = null;

        File trpConfigFile = new File(System.getProperty(ServerConstants.CARBON_HOME) + File.separator +
                "conf" + File.separator + configFileName);
        if (trpConfigFile.exists()) {
            log.info("Transport configuration file provided for transport: " + transport +
                    ", File: " + trpConfigFile.getAbsolutePath());
            try {
                configFileStream = new FileInputStream(trpConfigFile);
                configLoaded = true;
            } catch (FileNotFoundException e) {
                log.error("Error while opening the transport configuration " +
                        "file: " + trpConfigFile.getAbsolutePath(), e);
            }
        }

        if (!configLoaded) {
	    if (log.isDebugEnabled()) {
	            log.debug("Reading the transport configuration from " + configFileURL);
            }

            try {
                configFileStream = configFileURL.openStream();
            } catch (IOException e) {
                log.error("Error while opening the URL: " + configFileURL, e);
            }
        }

        if (configFileStream != null) {
            try {
                return (OMElement) XMLUtils.toOM(configFileStream);
            } catch (XMLStreamException e) {
                log.error("Error while building the object model from the transport configuration", e);
            }
        }
        return null;
    }

    /**
	 * Loads transport attributes corresponding to the given in transport name and populates a HashMap
	 * with those.
	 * 
	 * @param transport Name of the transport
	 * @return transport parameters as a map
	 * @throws AxisFault on error
	 */
	public Map<String, String> loadTransportInSettings(String transport) throws AxisFault {
		TransportDetails details;
		TransportParameter[] transportParams;
		Map<String, String> params;

		details = getTransportDetails(transport);
		params = new HashMap<String, String>();

		if (details != null) {
			transportParams = details.getInParameters();
			for (TransportParameter transportPara : transportParams) {
				params.put(transportPara.getName(), transportPara.getValue());
			}
		}

		return params;
	}

    /**
	 * Loads transport attributes corresponding to the given in transport name and populates a HashMap
	 * with those.
	 *
	 * @param transport Name of the transport
	 * @return transport parameters as a map
	 * @throws AxisFault on error
	 */
	public Map<String, String> loadTransportOutSettings(String transport) throws AxisFault {
		TransportDetails details;
		TransportParameter[] transportParams;
		Map<String, String> params;

		details = getTransportDetails(transport);
		params = new HashMap<String, String>();

		if (details != null) {
			transportParams = details.getOutParameters();
            if (transportParams == null) {
                return null;
            }
            for (TransportParameter transportPara : transportParams) {
				params.put(transportPara.getName(), transportPara.getValue());
			}
		}

		return params;
	}

    /**
	 * If a given transport listener is present in AxisConfiguration - then it's considered to be active.
	 * 
	 * @param transport Name of the transport protocol.
	 * @return true if the specified transport is in the active state and false otherwise
	 * @throws AxisFault on error
	 */
	public boolean isListenerActive(String transport) throws AxisFault {
		return (axisConfiguration.getAxisConfiguration().getTransportIn(transport) != null);
	}

    /**
	 * If a given transport sender is present in AxisConfiguration - then it's considered to be active.
	 *
	 * @param transport Name of the transport protocol.
	 * @return true if the specified transport is in the active state and false otherwise
	 * @throws AxisFault on error
	 */
	public boolean isSenderActive(String transport) throws AxisFault {
		return (axisConfiguration.getAxisConfiguration().getTransportOut(transport) != null);
	}

    /**
	 * This will make sure - when we enable a transport globally - appropriate end-points of all the
	 * services being updated.
	 * 
	 * @throws AxisFault on error
	 */
	public void enableTransport() throws AxisFault {
		Map services = axisConfiguration.getServices();
		Iterator serviceNames = services.keySet().iterator();

		while (serviceNames.hasNext()) {
			String serviceName = (String) serviceNames.next();
			// Need a way to get the service even if service is not active.
			AxisService axisService = axisConfiguration.getServiceForActivation(serviceName);

			// We do not need to worry about adminServices.
			if (axisService.getParameter("adminService") == null
					|| !"true".equals(axisService.getParameter("adminService").getValue()
							.toString())) {
				Utils.addEndpointsToService(axisService, axisConfiguration);
			}
		}
	}

	/**
	 * Lists all active transports - that is which are available in AxisConfiguration.
	 * 
	 * @return an array of TransportSummary objects
	 */
	public TransportSummary[] listTransports() {
		Map transports = getAxisConfig().getTransportsIn();
        Map trpOuts = getAxisConfig().getTransportsOut();
        List transCollection = new ArrayList();

		for (Iterator iter = transports.values().iterator(); iter.hasNext();) {
			TransportInDescription inTransport = (TransportInDescription) iter.next();
			TransportSummary summary = new TransportSummary();
			summary.setProtocol(inTransport.getName());
			// This is active, since we found this in AxisConfiguration.
			summary.setListenerActive(true);
            summary.setSenderActive(trpOuts.containsKey(inTransport.getName()));
            transCollection.add(summary);
		}

		return (TransportSummary[]) transCollection.toArray(new TransportSummary[transCollection
				.size()]);
	}

    /**
	 * Reads transport in details for the given transport protocol.
	 * 
	 * @param transportProtocol Name of the transport.
	 * @return a TransportDetails object
	 * @throws AxisFault on error
	 */
	public TransportDetails getTransportDetails(String transportProtocol) throws AxisFault {
		TransportInDescription inTransport;
        TransportOutDescription outTransport;

        // See whether the transport is available in the AxisConfiguration first.
		// If so - this transport is already being enabled - only enables transports are present in
		// the AxisConfiguration.
		inTransport = getAxisConfig().getTransportIn(transportProtocol);
        outTransport = getAxisConfig().getTransportOut(transportProtocol);

        if (inTransport == null) {
			// This transport is not yet enabled.
			// So we can't read it's attributes from AxisConfiguration.
            TransportBuilder trpBuilder = new TransportBuilder();
            inTransport = trpBuilder.getTransportInDescription(transportProtocol, axisConfiguration, false);
		}

        if (outTransport == null) {
            TransportBuilder trpBuilder = new TransportBuilder();
            outTransport = trpBuilder.getTransportOutDescription(transportProtocol, axisConfiguration, false);
        }

        return getTransportDetails(inTransport, outTransport, transportProtocol);
	}

    /**
	 * Reads transport details for the given transport protocol and given service.
	 * 
	 * @param serviceName Name of the service.
	 * @param transportProtocol Name of the transport.
	 * @return a TransportDetails object
	 * @throws AxisFault on error
	 */
	public TransportDetails getTransportDetails(String serviceName, String transportProtocol)
			throws AxisFault {

		AxisService service = getAxisConfig().getServiceForActivation(serviceName);

		// See whether the transport is available in the AxisConfiguration first.
		// If so - this transport is already being enabled - only enables transports are present in
		// the AxisConfiguration.
		TransportInDescription inTransport = service.getAxisConfiguration().getTransportIn(transportProtocol);
        TransportOutDescription outTransport = service.getAxisConfiguration().getTransportOut(transportProtocol);

        if (inTransport == null) {
			// This transport is not yet enabled.
			// So we can't read it's attributes from AxisConfiguration.
            TransportBuilder trpBuilder = new TransportBuilder();
            inTransport = trpBuilder.getTransportInDescription(transportProtocol, axisConfiguration, false);
		}

        if (outTransport == null) {
            TransportBuilder trpBuilder = new TransportBuilder();
            outTransport = trpBuilder.getTransportOutDescription(transportProtocol, axisConfiguration, false);
        }

        return getTransportDetails(inTransport, outTransport, transportProtocol);
    }




    /**
	 * Builds a TransportDetails instance out of TransportInDescription attributes.
	 * 
	 * @param transportIn a valid TransportInDescription
     * @param transportOut a valid TransportOutDescription
     * @param name Name of the transport
	 * @return a TransportDetails object
	 * @throws AxisFault on error
	 */
	private TransportDetails getTransportDetails(TransportInDescription transportIn,
                                                 TransportOutDescription transportOut,
                                                 String name) throws AxisFault {
		TransportDetails details;
		List<TransportParameter> inParams, outParams;
		int intValue = -1;

		if (transportIn == null) {
			// In valid inputs provided - let's ignore further proceeding.
			return null;
		}

		details = new TransportDetails();
		inParams = new ArrayList<TransportParameter>();
		details.setListenerActive(isListenerActive(name));
        details.setSenderActive(isSenderActive(name));

        // Transport in parameters
		for (Iterator paramIter = transportIn.getParameters().iterator(); paramIter.hasNext();) {
			Parameter param = (Parameter) paramIter.next();
			TransportParameter paramDO = new TransportParameter();
			paramDO.setName(param.getName());
			OMElement paramEle = param.getParameterElement();

			try {
				if (param.getValue() != null) {
					// At this point the param.getValue() can be anything.
					intValue = Integer.parseInt(param.getValue().toString());
					// At this point the param.getValue() must be an integer.
				}
			} catch (NumberFormatException e) {
				// At this point the param.getValue()is not an integer.
				intValue = 0;
			}

			// We need to avoid invalid values.
			// That all the non-integer values can be considered as valid.
			// If it is integer - then it should be greater than or equal to 0.
			if (intValue >= 0) {
				if (paramEle != null) {
					paramDO.setParamElement(paramEle.toString());
				}
				paramDO.setValue(param.getValue().toString());
				inParams.add(paramDO);
			}
		}

        if (inParams.size() > 0) {
            details.setInParameters(inParams.toArray(new TransportParameter[inParams.size()]));
        }

        if (transportOut != null) {
            outParams = new ArrayList<TransportParameter>();

            // Transport out parameters
            for (Iterator paramIter = transportOut.getParameters().iterator(); paramIter.hasNext();) {
                Parameter param = (Parameter) paramIter.next();
                TransportParameter paramDO = new TransportParameter();
                paramDO.setName(param.getName());
                OMElement paramEle = param.getParameterElement();

                try {
                    if (param.getValue() != null) {
                        // At this point the param.getValue() can be anything.
                        intValue = Integer.parseInt(param.getValue().toString());
                        // At this point the param.getValue() must be an integer.
                    }
                } catch (NumberFormatException e) {
                    // At this point the param.getValue()is not an integer.
                    intValue = 0;
                }

                // We need to avoid invalid values.
                // That all the non-integer values can be considered as valid.
                // If it is integer - then it should be greater than or equal to 0.
                if (intValue >= 0) {
                    if (paramEle != null) {
                        paramDO.setParamElement(paramEle.toString());
                    }
                    paramDO.setValue(param.getValue().toString());
                    outParams.add(paramDO);
                }
            }

            if (outParams.size() > 0) {
                details.setOutParameters(outParams.toArray(new TransportParameter[outParams.size()]));
            }
        }

        return details;
	}

    /**
	 * Updates transport details in AxisConfiguration.
	 * 
	 * @param serviceName Name of the service.
	 * @param transportProtocol Name of the transport protocol.
	 * @param params Updated parameter set.
	 * @throws AxisFault on error
	 */
	public void updateTransportInDetails(String serviceName, String transportProtocol,
			TransportParameter[] params) throws AxisFault {

		AxisService service = axisConfiguration.getServiceForActivation(serviceName);
		TransportInDescription inTransport = service.getAxisConfiguration().getTransportIn(transportProtocol);

		if (inTransport == null) {
			// We have no transports defined for this.
			// This is the case - where a transport not found in AxisConfiguration being enabled.

			// Let's build the TransportInDescription from the configuration file.
			TransportInDescription transport = new TransportBuilder().getTransportInDescription(transportProtocol,
                    axisConfiguration, true);
			if (transport != null) {
				// We have a valid TransportInDescription.
				// Add to AxisConfiguration - in memory.
				getAxisConfig().addTransportIn(transport);
				updateTransportDetails(transportProtocol, transport, params);
			}
		} else {
			// Updating transport details of an already enabled transport.
			// In other words - this already presents in the AxisConfiguration.
			updateTransportDetails(transportProtocol, inTransport, params);
		}
	}

     /**
	 * Updates transport details in AxisConfiguration.
	 *
	 * @param serviceName Name of the service.
	 * @param transportProtocol Name of the transport protocol.
	 * @param params Updated parameter set.
	 * @throws AxisFault on error
	 */
	public void updateTransportOutDetails(String serviceName, String transportProtocol,
			TransportParameter[] params) throws AxisFault {

		AxisService service = getAxisConfig().getServiceForActivation(serviceName);
		TransportOutDescription outTransport = service.getAxisConfiguration().
                getTransportOut(transportProtocol);

		if (outTransport == null) {
			// We have no transports defined for this.
			// This is the case - where a transport not found in AxisConfiguration being enabled.

			// Let's build the TransportOutDescription from the configuration file.
			TransportOutDescription transport = new TransportBuilder().getTransportOutDescription(transportProtocol,
                    axisConfiguration, true);
			if (transport != null) {
				// We have a valid TransportOutDescription.
				// Add to AxisConfiguration - in memory.
				getAxisConfig().addTransportOut(transport);
				updateTransportDetails(transportProtocol, transport, params);
			}
		} else {
			// Updating transport details of an already enabled transport.
			// In other words - this already presents in the AxisConfiguration.
			updateTransportDetails(transportProtocol, outTransport, params);
		}
	}

    /**
	 * Updates transport details in AxisConfiguration.
	 * 
	 * @param transportProtocol Name of the transport protocol.
	 * @param params Updated parameter set.
	 * @throws AxisFault on error
	 */
	public void updateTransportInDetails(String transportProtocol, TransportParameter[] params)
			throws AxisFault {
		TransportInDescription inTransport;

		inTransport = getAxisConfig().getTransportIn(transportProtocol);

		if (inTransport == null) {
			// We have no transports defined for this.
			// This is the case - where a transport not found in AxisConfiguration being enabled.

			// Let's build the TransportInDescription from the configuration file.
			TransportInDescription transport = new TransportBuilder().getTransportInDescription(transportProtocol,
                    axisConfiguration, true);
			if (transport != null) {
				// We have a valid TransportInDescription.
				// Add to AxisConfiguration - in memory.
				getAxisConfig().addTransportIn(transport);
				inTransport = getAxisConfig().getTransportIn(transportProtocol);
				updateTransportDetails(transportProtocol, inTransport, params);
			}
		} else {
			// Updating transport details of an already enabled transport.
			// In other words - this already presents in the AxisConfiguration.
			updateTransportDetails(transportProtocol, inTransport, params);
		}
	}

    /**
     * Updates transport details in AxisConfiguration.
	 *
	 * @param transportProtocol Name of the transport protocol.
	 * @param params Updated parameter set.
	 * @throws AxisFault on error
	 */
	public void updateTransportOutDetails(String transportProtocol, TransportParameter[] params)
			throws AxisFault {

        TransportOutDescription outTransport = getAxisConfig().getTransportOut(transportProtocol);

		if (outTransport == null) {
			// We have no transports defined for this.
			// This is the case - where a transport not found in AxisConfiguration being enabled.

			// Let's build the TransportOutDescription from the configuration file.
			TransportOutDescription transport = new TransportBuilder().getTransportOutDescription(transportProtocol,
                    axisConfiguration, true);
			if (transport != null) {
				// We have a valid TransportOutDescription.
				// Add to AxisConfiguration - in memory.
				getAxisConfig().addTransportOut(transport);
				outTransport = getAxisConfig().getTransportOut(transportProtocol);
				updateTransportDetails(transportProtocol, outTransport, params);
			}
		} else {
			// Updating transport details of an already enabled transport.
			// In other words - this already presents in the AxisConfiguration.
			updateTransportDetails(transportProtocol, outTransport, params);
		}
	}

    /**
	 * Updates transport details in AxisConfiguration. This method does not perform
     * partial updates. Whenever an exception is encountered it rollsback the parameters
     * to the previous consistent state and simply throws an AxisFault.
	 *
     * @param transportName Name of the transport
	 * @param transport TransportInDescription or TransportOutDescription
	 * @param transParams Updated parameter set.
	 * @throws AxisFault on error
	 */
	protected void updateTransportDetails(String transportName, ParameterInclude transport,
			TransportParameter[] transParams) throws AxisFault {

        if (transParams == null) {
            return;
        }

        Map<String, Parameter> updatedParameters = new HashMap<String, Parameter>();
        List<Parameter> deletedParams = new ArrayList<Parameter>();
        Parameter param, newParameter;

        for (TransportParameter tp : transParams) {
            if (tp == null || tp.getName() == null) {
                continue;
            }

            // Load the corresponding parameter from memory, first.
            param = transport.getParameter(tp.getName());
            if (param != null && !param.isLocked()) {
                // We have the parameter from memory and it's not locked - so we can update.
                try {
                    newParameter = getParameter(tp);
                    updatedParameters.put(newParameter.getName(), newParameter);
                } catch (XMLStreamException e) {
                    throw new AxisFault("Error parsing XML parameter", e);
                }

            } else if (param == null) {
                //adding new parameter
                try {
                    newParameter = getParameter(tp);
                    updatedParameters.put(newParameter.getName(), newParameter);

                } catch (XMLStreamException e) {
                    throw new AxisFault("Error parsing XML parameter", e);
                }

            } else {
                log.warn("Attempt made to update the locked parameter: " + tp.getName());
            }
        }

        // Now add all new parameters (newly added ones and existing updated ones)
        for (Parameter p : updatedParameters.values()) {
            transport.addParameter(p);
        }

        // Find the deleted parameters
        for (Parameter p : transport.getParameters()) {
            if (!updatedParameters.containsKey(p.getName()) && !p.isLocked()) {
                deletedParams.add(p);
            }
        }

        // Delete the unnecessary parameters
        for (Parameter deletedParam : deletedParams) {
            log.info("Removing transport parameter " + deletedParam.getName() + " from " + transportName);
            transport.removeParameter(deletedParam);
        }
    }

    private Parameter getParameter(TransportParameter tp) throws XMLStreamException {
        Parameter param = new Parameter();
        param.setName(tp.getName());

        if (tp.getParamElement() != null) {
            OMElement paramElement = AXIOMUtil.stringToOM(tp.getParamElement());
            param.setParameterElement(paramElement);
            param.setValue(paramElement);
            param.setParameterType(Parameter.OM_PARAMETER);
        } else {
            try {
                OMElement paramElement = AXIOMUtil.stringToOM("<parameter name=\"" +
                                    tp.getName() + "\">" + tp.getValue() + "</parameter>");
                param.setParameterElement(paramElement);
            } catch (XMLStreamException e) {
                // ignore
            }
            param.setValue(tp.getValue());
            param.setParameterType(Parameter.TEXT_PARAMETER);

        }

        return param;
    }


    private AxisConfiguration getAxisConfig() {
		return axisConfiguration;
	}
}
