/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.transports;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.axis2.deployment.DeploymentConstants;
import org.apache.axis2.deployment.DeploymentErrorMsgs;
import org.apache.axis2.deployment.DeploymentException;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.ParameterInclude;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.axis2.i18n.Messages;
import org.apache.axis2.transport.TransportListener;
import org.apache.axis2.transport.TransportSender;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.core.persistence.PersistenceManager;
import org.wso2.carbon.registry.core.Collection;

import javax.xml.namespace.QName;
import java.util.Iterator;
import java.util.List;

public class TransportBuilder implements DeploymentConstants {

	private static Log log = LogFactory.getLog(TransportBuilder.class);

	/**
	 * Builds a TransportInDescription object from an XML configuration. The expected
     * XML configuration should be as follows.
     *
     * <transportReceiver name="myTransport" class="my.transport.listener.Class">
     *     <parameter name="param1">value 1</parameter>
     *     <parameter name="param1">value 1</parameter>
     * </transportReceiver>
	 * 
	 * @param transport An OMElement which contains transport configurations.
	 * @return a transport in description
	 * @throws DeploymentException on error
	 */
	protected TransportInDescription processTransportReceiver(OMElement transport) 
            throws DeploymentException {

        TransportInDescription transportIn;

        // getting transport Name
        OMAttribute trsName = transport.getAttribute(new QName(ATTRIBUTE_NAME));

        if (trsName == null) {
            throw new DeploymentException("Name of the transport is not specified in the " +
                    "receiver configuration");
        }

        String name = trsName.getAttributeValue();
        transportIn = new TransportInDescription(name);

        // transport implementation class
        OMAttribute trsClass = transport.getAttribute(new QName(TAG_CLASS_NAME));
        if (trsClass != null) {
            String className = trsClass.getAttributeValue().trim();
            addClassNameAsProperty(transportIn, className);

        } else {
            throw new DeploymentException("No transport receiver class is specified for " + name);
        }

        try {
            Iterator itr = transport.getChildrenWithName(new QName(TAG_PARAMETER));
            processParameters(itr, transportIn);
            return transportIn;
        } catch (AxisFault axisFault) {
            throw new DeploymentException(axisFault);
        }
    }

    /**
	 * Builds a TransportOutDescription object based on the specified XML configuration.
     * The XML should be as follows.
     *
     * <transportSender name="myTransport" class="my.transport.sender.Class">
     *     <parameter name="param1">value 1</parameter>
     *     <parameter name="param1">value 1</parameter>
     * </transportSender>
	 *
	 * @param transport An Iterator of OMElement which contains transport configurations.
	 * @return an array of transport out descriptions
	 * @throws DeploymentException on error
	 */
    protected TransportOutDescription processTransportSender(OMElement transport) 
            throws DeploymentException {

        TransportOutDescription transportOut;

        // getting transport Name
        OMAttribute transportName = transport.getAttribute(new QName(ATTRIBUTE_NAME));

        if (transportName == null) {
            throw new DeploymentException("Name of the transport is not specified in " +
                    "the sender configuration");
        }

        String name = transportName.getAttributeValue();
        transportOut = new TransportOutDescription(name);

        // transport sender implementation class
        OMAttribute transportClass = transport.getAttribute(new QName(TAG_CLASS_NAME));

        if (transportClass != null) {
            String className = transportClass.getAttributeValue().trim();
            addClassNameAsProperty(transportOut, className);
            
        } else {
            throw new DeploymentException("No transport sender class is specified for " + name);
        }

        try {
            Iterator itr = transport.getChildrenWithName(new QName(TAG_PARAMETER));
            processParameters(itr, transportOut);
            return transportOut;
        } catch (AxisFault e) {
            throw new DeploymentException(e);
        }

    }

    private void addClassNameAsProperty(ParameterInclude trpDesc, String className) {
        try {
            trpDesc.addParameter(new Parameter("implClassName", className));
        } catch (AxisFault axisFault) {
            log.error("Error while setting the class name property for transport description", axisFault);
        }
    }

    /**
     * Constructs a transport in description for the specified transport by using the
     * details available in the registry. Actual transport listener object is created only
     * if the createListener flag is set. Therefore when we are only interested in transport
     * listener parameters we can get around without creating the transport listener by passing
     * false to createListener parameter.
     *
     * @param transportName name of the transport
     * @param axisConfiguration the Axis2 configuration
     * @param createListener true if the builder should try to create the actual transport listener
     * @return a TransportInDescription object
     */
    public TransportInDescription getTransportInDescription(String transportName,
                                                               AxisConfiguration axisConfiguration,
                                                               boolean createListener) {

        TransportInDescription trpIn = new TransportInDescription(transportName);
        try {
            PersistenceManager pm = new PersistenceManager(axisConfiguration);
            Collection trpResource = (Collection) pm.getTransport(transportName, true);
            if (trpResource == null) {
                // Nothing in the registry related to this transport
                // Return null
                return null;
            }

            String clazz = trpResource.getProperty("class");
            if (clazz == null) {
                // That's odd! Transport resource is in the registry but the class has not been defined
                // Probably the user removed the entries from the axis2.xml or the database itself
                log.info("Transport listener configuration for " + transportName + " has been removed" +
                        " from the system. Cleaning up the related entries in the registry.");
                pm.removeTransport(transportName, true);
                // We can't proceed so return null
                return null;
            }

            // If asked for create the listener object
            if (createListener) {
                TransportListener listener = (TransportListener) Class.forName(clazz).newInstance();
                trpIn.setReceiver(listener);
            }

            List<Parameter> params = pm.getTransportParameters(transportName, true);
            for (Parameter p : params) {
                trpIn.addParameter(p);
            }

        } catch (Exception e) {
            // We cannot initialize the transport in description
            // So we cannot enable it - Return null
            log.error("Error while creating transport in description for " + transportName, e);
            return null;
        }
        // All done - Return the constructed transport in description
        return trpIn;
    }

    /**
     * Constructs a transport out description for the specified transport by using the
     * details available in the registry. Actual transport sender object is created only
     * if the createSender flag is set. Therefore when we are only interested in transport
     * sender parameters we can get around without creating the transport sender by passing
     * false to createSender parameter.
     *
     * @param transportName name of the transport
     * @param axisConfiguration the Axis2 configuration
     * @param createSender true if the builder should try to create the actual transport sender
     * @return a TransportOutDescription object
     */
    public TransportOutDescription getTransportOutDescription(String transportName,
                                                               AxisConfiguration axisConfiguration,
                                                               boolean createSender) {

        TransportOutDescription trpOut = new TransportOutDescription(transportName);
        try {
            PersistenceManager pm = new PersistenceManager(axisConfiguration);
            Collection trpResource = (Collection) pm.getTransport(transportName, false);
            if (trpResource == null) {
                // Nothing in the registry related to this transport
                // Return null
                return null;
            }

            String clazz = trpResource.getProperty("class");
            if (clazz == null) {
                // That's odd! Transport resource is in the registry but the class has not been defined
                // Probably the user removed the entries from the axis2.xml or the database itself
                log.info("Transport sender configuration for " + transportName + " has been removed" +
                        " from the system. Cleaning up the related entries in the registry.");
                pm.removeTransport(transportName, false);
                // We can't proceed so return null
                return null;
            }

            if (createSender) {
                TransportSender sender = (TransportSender) Class.forName(clazz).newInstance();
                trpOut.setSender(sender);
            }

            List<Parameter> params = pm.getTransportParameters(transportName, false);
            for (Parameter p : params) {
                trpOut.addParameter(p);
            }

        } catch (Exception e) {
            // We cannot initialize the transport out description
            // So we cannot enable it - Return null
            log.error("Error while creating transport out description for " + transportName, e);
            return null;
        }
        return trpOut;
    }

    /**
	 * Processes transport parameters from the given configuration.
	 * 
	 * @param parameters An Iterator of OMElement which contains transport parameters.
	 * @param parameterInclude ParameterInclude
	 * @throws DeploymentException on error
	 */
	private void processParameters(Iterator<OMElement> parameters, ParameterInclude parameterInclude)
			throws DeploymentException {
		while (parameters.hasNext()) {
			// this is to check whether some one has locked the parameter at the
			// top level
			OMElement parameterElement = parameters.next();
			Parameter parameter = new Parameter();
			// setting parameterElement
			parameter.setParameterElement(parameterElement);
			// setting parameter Name
			OMAttribute paramName = parameterElement.getAttribute(new QName(ATTRIBUTE_NAME));
			if (paramName == null) {
				throw new DeploymentException(Messages.getMessage(
						DeploymentErrorMsgs.BAD_PARAMETER_ARGUMENT, parameterElement.toString()));
			}
			parameter.setName(paramName.getAttributeValue());
			// setting parameter Value (the child element of the parameter)
			OMElement paramValue = parameterElement.getFirstElement();
			if (paramValue != null) {
				parameter.setValue(parameterElement);
				parameter.setParameterType(Parameter.OM_PARAMETER);
			} else {
				String paratextValue = parameterElement.getText();

				parameter.setValue(paratextValue);
				parameter.setParameterType(Parameter.TEXT_PARAMETER);
			}
			// setting locking attribute
			OMAttribute paramLocked = parameterElement.getAttribute(new QName(ATTRIBUTE_LOCKED));

			if (paramLocked != null) {
				String lockedValue = paramLocked.getAttributeValue();
				if (BOOLEAN_TRUE.equals(lockedValue)) {
					parameter.setLocked(true);
				} else {
					parameter.setLocked(false);
				}
			}
			try {
				parameterInclude.addParameter(parameter);
			} catch (AxisFault axisFault) {
				throw new DeploymentException(axisFault);
			}
		}
	}
}