/*                                                                             
 * Copyright 2004,2005 The Apache Software Foundation.                         
 *                                                                             
 * Licensed under the Apache License, Version 2.0 (the "License");             
 * you may not use this file except in compliance with the License.            
 * You may obtain a copy of the License at                                     
 *                                                                             
 *      http://www.apache.org/licenses/LICENSE-2.0                             
 *                                                                             
 * Unless required by applicable law or agreed to in writing, software         
 * distributed under the License is distributed on an "AS IS" BASIS,           
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    
 * See the License for the specific language governing permissions and         
 * limitations under the License.                                              
 */
package org.wso2.carbon.registry.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.session.UserRegistry;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.utils.CarbonUtils;

import javax.sql.DataSource;
import java.io.Reader;

/**
 * The Registry implementation used by Carbon. The main purpose of this class is to enable
 * Carbon to run in Registry READ-ONLY mode
 */
public class CarbonRegistry extends UserRegistry {

    private static final Log log = LogFactory.getLog(CarbonRegistry.class);

    public CarbonRegistry(String userName, String password,
                          Registry coreRegistry,
                          UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {
        super(userName, password, coreRegistry, defaultRealm, dataSource);
    }

    public CarbonRegistry(String userName, Registry coreRegistry,
                          UserRealm defaultRealm)
            throws RegistryException {
        super(userName, coreRegistry, defaultRealm);
    }

    public CarbonRegistry(String userName, Registry coreRegistry,
                          UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {
        super(userName, coreRegistry, defaultRealm, dataSource);
    }

    @Override
    public String put(String suggestedPath, Resource resource) throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to put a resource at " + suggestedPath);
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot put resource at " + suggestedPath + " since the registry is read-only");
            }
            return suggestedPath;
        }
        return super.put(suggestedPath, resource);
    }

    @Override
    public void delete(String path) throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to delete a resource at " + path);
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot delete resource at " + path + " since the registry is read-only");
            }
            return;
        }
        super.delete(path);
    }

    @Override
     public String move(String currentPath, String newPath)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to move resource " + currentPath + " to " + newPath);
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot move resource since the registry is read-only");
            }
            return currentPath;
        }
        return super.move(currentPath, newPath);
    }

    @Override
    public void editComment(String commentPath, String text)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to edit comment");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot edit comment since the registry is read-only");
            }
            return;
        }
        super.editComment(commentPath, text);
    }

    @Override
    public String addComment(String resourcePath, Comment comment)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to add comment");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot add comment since the registry is read-only");
            }
            return resourcePath; //TODO: Check this
        }
        return super.addComment(resourcePath, comment);
    }

    @Override
    public void removeTag(String path, String tag) throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to remove tag of resource at " + path);
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot remove tag since the registry is read-only");
            }
            return;
        }
        super.removeTag(path, tag);
    }

    @Override
    public void rateResource(String resourcePath, int rating)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to rate resource at " + resourcePath);
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot rate resource since the registry is read-only");
            }
            return;
        }
        super.rateResource(resourcePath, rating);
    }

    @Override
    public void removeAssociation(String sourcePath, String associationPaths,
                                  String associationType)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to remove association");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot remove association since the registry is read-only");
            }
            return;
        }
        super.removeAssociation(sourcePath, associationPaths, associationType);
    }

    @Override
    public void addAssociation(String sourcePath, String targetPath,
                               String associationType)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying add association");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot add association since the registry is read-only");
            }
            return;
        }
        super.addAssociation(sourcePath, targetPath, associationType);
    }

    @Override
    public void restoreVersion(String versionPath) throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to restore version");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot restore version since the registry is read-only");
            }
            return;
        }
        super.restoreVersion(versionPath);
    }

    @Override
    public void applyTag(String resourcePath, String tag)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to apply tag");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot apply tag since the registry is read-only");
            }
            return;
        }
        super.applyTag(resourcePath, tag);
    }

    @Override
    public String importResource(String suggestedPath, String sourceURL,
                                 Resource metadata) throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to import resource");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot import resource since the registry is read-only");
            }
            return sourceURL;
        }
        return super.importResource(suggestedPath, sourceURL, metadata);
    }

    @Override
    public String copy(String sourcePath, String targetPath)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to copy resource");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot copy resource since the registry is read-only");
            }
            return sourcePath;
        }
        return super.copy(sourcePath, targetPath);
    }

    @Override
    public String rename(String currentPath, String newPath)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to rename resource");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot rename resource since the registry is read-only");
            }
            return currentPath;
        }
        return super.rename(currentPath, newPath);
    }

    @Override
    public void createVersion(String path) throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to create version");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot create version since the registry is read-only");
            }
            return;
        }
        super.createVersion(path);
    }

    @Override
    public void associateAspect(String resourcePath, String aspect)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to associate aspect");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot associate aspect since the registry is read-only");
            }
            return;
        }
        super.associateAspect(resourcePath, aspect);
    }

    @Override
    public void createLink(String path, String target)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to creat link");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot create link since the registry is read-only");
            }
            return;
        }
        super.createLink(path, target);
    }

    @Override
    public void createLink(String path, String target, String subTargetPath)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to creat link");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot create link since the registry is read-only");
            }
            return;
        }
        super.createLink(path, target, subTargetPath);
    }

    @Override
    public void removeLink(String path)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to remove link");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot remove link since the registry is read-only");
            }
            return;
        }
        super.removeLink(path);
    }

    @Override
    public void restore(String path, Reader reader)
            throws RegistryException {
        boolean registryReadOnly = CarbonUtils.isRegistryReadOnly();
        if(log.isDebugEnabled()){
            log.debug("Trying to restore");
            log.debug("Registry ReadOnly mode: " + registryReadOnly);
        }
        // Check whether put is allowed. If this node is operating in read-only mode
        // do not put  the  resource
        if(registryReadOnly){
            if(log.isDebugEnabled()){
                log.debug("Cannot restore the dump, since the registry is read-only");
            }
            return;
        }
        super.restore(path, reader);
    }
}
