/*
 * Copyright (c) 2007, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.registry.core;

import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.apache.axiom.om.OMElement;

import java.util.Map;
import java.util.Date;
import java.util.ArrayList;
import java.util.List;
import java.io.Writer;
import java.io.Reader;

/**
 * This is a Registry facade which allows the user to "root" their Registry someplace other
 * than the real root.
 */
public class ChrootRegistry implements Registry {
    // The URL prefix to use for all accesses - must end with a slash
    protected String basePrefix = "";
    protected Registry backend;

    public ChrootRegistry(Registry backend, String basePrefix) {
        if (backend == null) {
            throw new IllegalArgumentException("Can't have null backend");
        }
        this.backend = backend;
        if (basePrefix != null) {
            if (!basePrefix.endsWith("/")) {
                basePrefix += "/";
            }
            this.basePrefix = basePrefix;
        } else {
            this.basePrefix = "";
        }
    }

    /**
     * Get an absolute path for the given path argument, taking into
     * account both initial double-slashes (indicating an absolute path)
     * and any basePrefix that has been established.
     *
     * This is the converse of getRelativePath().
     *
     * @param path a relative path
     * @return an absolute path into the "real" registry.
     */
    public String getRealPath(String path) {
        if (path == null || path.length() == 0) return path;

        if (path.startsWith("//")) {
            // This is an absolute path, so just strip the doubled slash
            return path.substring(1);
        }

        // Relative path, so prepend basePrefix appropriately
        if (path.startsWith("/")) {
            return basePrefix + path.substring(1);
        }

        return basePrefix + path;
    }
    
    /**
     * Take an absolute path in the "real" registry and convert it to a relative
     * path suitable for this particular RemoteRegistry (which may be rooted at a
     * particular place).
     *
     * This is the converse of getRealPath().
     *
     * @param absolutePath a full path from the root of the registry, starting with "/"
     * @return a relative path which generates the correct absolute path
     */
    public String getRelativePath(String absolutePath) {
        // No worries if there's no base prefix
        if (basePrefix == null || basePrefix.length() == 0 || absolutePath == null)
            return absolutePath;

        // this is to avoid the problem when we need to filter /x when the basePrefix is /x/
        String truncatedBasePrefix = basePrefix.substring(0, basePrefix.length() - 1);

        if (absolutePath.startsWith(truncatedBasePrefix)) {
            String returnStr = absolutePath.substring(basePrefix.length() - 1);
            return (!returnStr.startsWith("/")) ? "/" + returnStr: returnStr;
        }

        // Somewhere else, so make sure there are dual slashes at the beginning
        return "/" + absolutePath;
    }

    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {
        return getRelativePath(backend.importResource(getRealPath(suggestedPath),
                                                      sourceURL,
                                                      metadata));
    }

    public String rename(String currentPath, String newName) throws RegistryException {
        if (newName.startsWith(RegistryConstants.ROOT_PATH)) {
            // this is not an absolute path, so needded to be transformed
            newName = getRealPath(newName);
        }
        return getRelativePath(backend.rename(getRealPath(currentPath), newName));
    }

    public String move(String currentPath, String newPath) throws RegistryException {
        return getRelativePath(backend.move(getRealPath(currentPath), getRealPath(newPath)));
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {
        return getRelativePath(backend.copy(getRealPath(sourcePath), getRealPath(targetPath)));
    }

    public void createVersion(String path) throws RegistryException {
        backend.createVersion(getRealPath(path));
    }

    public String[] getVersions(String path) throws RegistryException {
        String[] versions = backend.getVersions(getRealPath(path));
        for (int i = 0; i < versions.length; i++) {
            String versionPath = versions[i];
            versions[i] = getRelativePath(versionPath);
        }
        return versions;
    }

    public void restoreVersion(String versionPath) throws RegistryException {
        backend.restoreVersion(getRealPath(versionPath));
    }

    public void addAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {
        backend.addAssociation(getRealPath(sourcePath), getRealPath(targetPath), associationType);
    }

    public void removeAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {
        backend.removeAssociation(getRealPath(sourcePath),
                                  getRealPath(targetPath),
                                  associationType);
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {
        Association [] associations = backend.getAllAssociations(getRealPath(resourcePath));
        for ( Association association : associations) {
            if (association != null) {
                association.setSourcePath(getRelativePath(association.getSourcePath()));
                association.setDestinationPath(getRelativePath(association.getDestinationPath()));
            }
        }
        return associations;
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {
        Association [] associations = backend.getAssociations(getRealPath(resourcePath), associationType);
        for ( Association association : associations) {
            if (association != null) {
                association.setSourcePath(getRelativePath(association.getSourcePath()));
                association.setDestinationPath(getRelativePath(association.getDestinationPath()));
            }
        }
        return associations;
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {
        backend.applyTag(getRealPath(resourcePath), tag);
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {
        TaggedResourcePath[] result = backend.getResourcePathsWithTag(tag);
        for (TaggedResourcePath trp : result) {
            String path = trp.getResourcePath();
            trp.setResourcePath(getRelativePath(path));
        }
        return result;
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {
        return backend.getTags(getRealPath(resourcePath));
    }

    public void removeTag(String path, String tag) throws RegistryException {
        backend.removeTag(getRealPath(path), tag);
    }

    public String addComment(String resourcePath, Comment comment) throws RegistryException {
        return backend.addComment(getRealPath(resourcePath), comment);
    }

    public void editComment(String commentPath, String text) throws RegistryException {
        backend.editComment(getRealPath(commentPath), text);
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {
        Comment[] comments = backend.getComments(getRealPath(resourcePath));
        for (Comment comment : comments) {
            fixComment(comment);
        }
        return comments;
    }

    public void fixComment(Comment comment) {
        comment.setPath(getRelativePath(comment.getPath()));
        comment.setResourcePath(getRelativePath(comment.getResourcePath()));
        comment.setCommentPath(getRelativePath(comment.getCommentPath()));
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {
        backend.rateResource(getRealPath(resourcePath), rating);
    }

    public float getAverageRating(String resourcePath) throws RegistryException {
        return backend.getAverageRating(getRealPath(resourcePath));
    }

    public int getRating(String path, String userName) throws RegistryException {
        return backend.getRating(getRealPath(path), userName);
    }

    public Collection executeQuery(String path, Map parameters) throws RegistryException {
        Collection collection = backend.executeQuery(getRealPath(path), parameters);
        String[] results = collection.getChildren();
        List<String> filteredResult = new ArrayList<String>();

        // this is to avoid the problem when we need to filter /x when the basePrefix is /x/
        String truncatedBasePrefix;
        if (basePrefix == null || basePrefix.length() == 0) {
            truncatedBasePrefix = "";
        } else {
            truncatedBasePrefix = basePrefix.substring(0, basePrefix.length() - 1);
        }
        for (int i = 0; i < results.length; i ++) {

            if (results[i].startsWith(truncatedBasePrefix)) {
                if (results[i] instanceof String) {
                    filteredResult.add(results[i]);
                }
            }
        }
        String[] filteredResultArr = filteredResult.toArray(new String[filteredResult.size()]);
        collection.setContent(filteredResultArr);

        fixCollection((CollectionImpl)collection);
        return collection;
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from,
                              Date to,
                              boolean recentFirst) throws RegistryException {
        LogEntry[] logEntries = backend.getLogs(getRealPath(resourcePath), action, userName, from, to, recentFirst);
        return fixLogEntries(logEntries);
    }

    private LogEntry[] fixLogEntries(LogEntry[] logEntries) {
        List<LogEntry> fixedLogEntries = new ArrayList<LogEntry>();
        for(LogEntry logEntry: logEntries) {
            String logPath = logEntry.getResourcePath();
            // this is to avoid the problem when we need to filter /x when the basePrefix is /x/
            String truncatedBasePrefix;
            if (basePrefix == null || basePrefix.length() == 0) {
                truncatedBasePrefix = "";
            } else {
                truncatedBasePrefix = basePrefix.substring(0, basePrefix.length() - 1);
            }
            if (!logPath.startsWith(truncatedBasePrefix)) {
                continue;
            }
            logEntry.setResourcePath(getRelativePath(logPath));
            fixedLogEntries.add(logEntry);
        }
        return fixedLogEntries.toArray(new LogEntry[fixedLogEntries.size()]);
    }

    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {
        throw new UnsupportedOperationException("Sorry we need to implement this method");
    }

    public String[] getAvailableAspects() {
        return backend.getAvailableAspects();
    }

    public void associateAspect(String resourcePath, String aspect) throws RegistryException {
        backend.associateAspect(getRealPath(resourcePath), aspect);
    }

    public void invokeAspect(String resourcePath, String aspectName, String action)
           throws RegistryException {
        backend.invokeAspect(getRealPath(resourcePath), aspectName, action);
    }

    public String[] getAspectActions(String resourcePath, String aspectName)
           throws RegistryException {
        return backend.getAspectActions(getRealPath(resourcePath), aspectName);
    }

    public void beginTransaction() throws RegistryException {
        backend.beginTransaction();
    }

    public void commitTransaction() throws RegistryException {
        backend.commitTransaction();
    }

    public void rollbackTransaction() throws RegistryException {
        backend.rollbackTransaction();
    }

    public RegistryContext getRegistryContext() {
        return backend.getRegistryContext();
    }

    public Collection searchContent(String keywords) throws RegistryException {
        // TODO - need to localize a search with a prefix....
        return backend.searchContent(keywords);
    }

    public void createLink(String path, String target) throws
                                                       RegistryException {
        backend.createLink(getRealPath(path), getRealPath(target));
    }

    public void createLink(String path, String target, String subTargetPath)
            throws RegistryException {
        backend.createLink(getRealPath(path), target, subTargetPath);
    }

    public void removeLink(String path) throws RegistryException {
        backend.removeLink(getRealPath(path));
    }

    public Resource newResource() throws RegistryException {
        return backend.newResource();
    }

    public Collection newCollection() throws RegistryException {
        return backend.newCollection();
    }

    public void fixMountPoints(Resource resource) {
        String mountPoint = resource.getProperty("registry.mountpoint");
        if (mountPoint != null) {
            resource.setProperty("registry.mountpoint", getRelativePath(mountPoint));
        }
        String targetPoint = resource.getProperty("registry.targetpoint");
        if (targetPoint != null) {
            resource.setProperty("registry.targetpoint", getRelativePath(targetPoint));
        }
    }

    public Resource get(String path) throws RegistryException {
        Resource resource = backend.get(getRealPath(path));
        if (resource instanceof Collection) {
            fixCollection((CollectionImpl)resource);
        }
        // need to set the calling path to resource as well
        ((ResourceImpl)resource).setPath(path);
        fixMountPoints(resource);
        return resource;
    }

    public Resource getMetaData(String path) throws RegistryException {
        Resource resource = backend.getMetaData(getRealPath(path));
        if (resource instanceof Collection) {
            fixCollection((CollectionImpl)resource);
        }
        // need to set the calling path to resource as well
        ((ResourceImpl)resource).setPath(path);
        fixMountPoints(resource);
        return resource;
    }

    public void fixCollection(CollectionImpl collection) throws RegistryException {
        Object content = collection.getContent();
        collection.setPath(getRelativePath(collection.getPath()));
        if (content instanceof String[]) {
            String [] paths = (String[])content;
            for (int i = 0; i < paths.length; i++) {
                paths[i] = getRelativePath(paths[i]);
            }
        } else if (content instanceof Resource[]) {
            Resource [] resources = (Resource[])content;
            for (Resource resource : resources) {
                ((ResourceImpl)resource).setPath(getRelativePath(resource.getPath()));
                if (resource instanceof Comment) {
                    Comment comment = (Comment)resource;
                    comment.setResourcePath(getRelativePath(comment.getResourcePath()));
                }
            }
        }
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {
        CollectionImpl result = (CollectionImpl) backend.get(getRealPath(path), start, pageSize);
        // Convert paths to relative ones on the way out
        fixCollection(result);
        // need to set the calling path to resource as well
        result.setPath(path);
        fixMountPoints(result);

        return result;
    }

    public Collection getRelativeCollection(CollectionImpl source) throws RegistryException {
        source.setContent(null);
        return source;
    }

    public boolean resourceExists(String path) throws RegistryException {
        return backend.resourceExists(getRealPath(path));
    }

    public String put(String suggestedPath, Resource resource) throws RegistryException {
        return getRelativePath(backend.put(getRealPath(suggestedPath), resource));
    }

    public void delete(String path) throws RegistryException {
        backend.delete(getRealPath(path));
    }

    // check in, check out functionality
    public void restore(String path, Reader reader) throws RegistryException {
        backend.restore(getRealPath(path), reader);
    }

    public void dump(String path, Writer writer) throws RegistryException {
        backend.dump(getRealPath(path), writer);
    }

    public String getEventingServiceURL(String path) throws RegistryException {
        return backend.getEventingServiceURL(getRealPath(path));
    }

    public void setEventingServiceURL(String path, String eventingServiceURL) throws RegistryException {
        backend.setEventingServiceURL(getRealPath(path), eventingServiceURL);
    }
}
