/*
 * Copyright (c) 2007, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.registry.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

public class CollectionImpl extends ResourceImpl implements Collection {

    private static Log log = LogFactory.getLog(CollectionImpl.class);

    protected int childCount;

    public CollectionImpl() {
        childCount = -1;
    }

    public CollectionImpl(String[] paths) {
        content = paths;
        childCount = paths.length;
    }
    
    public CollectionImpl(String path, ResourceDO resourceDO) {
        super(path, resourceDO);
        childCount = -1;
    }


    public void setContent(Object content) throws RegistryException {
        if (content == null) return;
        // note that string contents are allowed in collection to support custom generated UIs.
        if (content instanceof String[] ||
                content instanceof Resource[] ||
                content instanceof String) {
            super.setContent(content);
            return;
        }
        throw new IllegalArgumentException("Invalid content for collection. " +
                "Content of type " + content.getClass().toString() +
                " is not allowed for collections.");
    }



    public void setContentWithNoUpdate(Object content) throws RegistryException {
        if (content == null) return;
        // note that string contents are allowed in collection to support custom generated UIs.
        if (content instanceof String[] ||
                content instanceof Resource[] ||
                content instanceof String) {
            super.setContentWithNoUpdate(content);
            return;
        }
        throw new IllegalArgumentException("Invalid content for collection. " +
                "Content of type " + content.getClass().toString() +
                " is not allowed for collections.");
    }

    public String[] getChildren() throws RegistryException {
        return (String[]) getContent();
    }

    public String[] getChildren(int start, int pageLen) throws RegistryException {

        try {
            if (content == null) {

                String[] childPaths = new String[0];

                if (Transaction.isStarted()) {

                    childPaths = resourceDAO.
                            getChildren(this, start, pageLen, Transaction.getConnection());

                } else {

                    Connection conn = null;

                    try {
                        conn = dataSource.getConnection();
                        childPaths = resourceDAO.
                            getChildren(this, start, pageLen, conn);

                    } catch (SQLException e) {

                        String msg = "Failed to get the child paths " + pageLen + " child paths from " +
                                start + " of resource " + getPath() + ". " + e.getMessage();
                        log.error(msg, e);
                        throw new RegistryException(msg, e);

                    } finally {

                        if (conn != null) {
                            try {
                                conn.close();
                            } catch (SQLException e) {
                                log.error("Failed to close the database connection used to " +
                                        "get child paths of collection " + getPath());
                            }
                        }
                    }
                }

                return childPaths;

            } else {

                if (content instanceof String[]){

                    List<String> childPathList = new ArrayList <String> ();
                    String childNodes [] = (String[])content;
                    int limit = start + pageLen;
                    if (start > childNodes.length) {
                        return new String[0];
                    }
                    if (limit > childNodes.length) {
                        limit = childNodes.length;
                    }
                    for (int i = start ; i < limit; i++) {
                        String childNode = childNodes[i];
                        childPathList.add(childNode);
                    }
                    return childPathList.toArray(new String[childPathList.size()]);
                }
            }
            return new String[0];
        } finally {
            clearSessionInformation();
        }
    }

    public List<Resource> getChildResources() {
        return null;
    }

    public int getChildCount() throws RegistryException {

        try {
            setSessionInformation();

            if (childCount != -1) {
                return childCount;

            } else if (content != null && content instanceof String[]) {

                String[] childPaths = (String[]) content;
                return childPaths.length;

            } else {

                if (Transaction.isStarted()) {

                    childCount = resourceDAO.getChildCount(this, Transaction.getConnection());

                } else {

                    Connection conn = null;

                    try {
                        conn = dataSource.getConnection();
                        childCount = resourceDAO.getChildCount(this, conn);

                    } catch (SQLException e) {

                        String msg = "Failed to get the child count of resource " +
                                getPath() + ". " + e.getMessage();
                        log.error(msg, e);
                        throw new RegistryException(msg, e);

                    } finally {

                        if (conn != null) {
                            try {
                                conn.close();
                            } catch (SQLException e) {
                                log.error("Failed to close the database connection used to " +
                                        "get child count of collection " + getPath());
                            }
                        }
                    }
                }
            }

            return childCount;

        } finally {
            clearSessionInformation();
        }
    }


    public void setChildCount(int count) {
        childCount = count;
    }


    /**
     * Collection's content is a string array, which contains paths of its children. These paths
     * are loaded on demand to increase performance. It is recommended to use getChildren(...)
     * method to get child paths of a collection, which provides pagination. Calling this method
     * will load all child paths.
     *
     * @return String array of child paths.
     * @throws RegistryException On any error.
     */
    public Object getContent() throws RegistryException {
        try {
            setSessionInformation();
            
            if (content == null) {
                if (Transaction.isStarted()) {                                                        
                    resourceDAO.fillChildren(this, 0, -1, Transaction.getConnection());
                } else {
                    
                    Connection conn = null;
                    
                    try {
                        conn = dataSource.getConnection();
                        resourceDAO.fillChildren(this, 0, -1, conn);

                    } catch (SQLException e) {

                        String msg = "Failed to get child paths of " +
                                getPath() + ". " + e.getMessage();
                        log.error(msg, e);
                        throw new RegistryException(msg, e);

                    } finally {

                        if (conn != null) {
                            try {
                                conn.close();
                            } catch (SQLException e) {
                                log.error("Failed to close the database connection opened for " +
                                        "accessing the child paths of " + getPath(), e);
                            }
                        }
                    }
                }
            }

            return content;

        } finally {
            clearSessionInformation();
        }
    }

    public ResourceImpl getShallowCopy() throws RegistryException {
        CollectionImpl newCollection = new CollectionImpl();
        fillCollectionCopy(newCollection);
        return newCollection;
    }

    public void fillCollectionCopy(CollectionImpl collection) throws RegistryException {
        super.fillResourceCopy(collection);
        collection.setChildCount(this.childCount);
    }
}
