/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dao.ResourceDAO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.user.core.UserRealm;

import javax.sql.DataSource;
import java.io.*;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Properties;

/**
 * Represents any file or collection stored in the registry. It encapsulates both the content
 * of the entity and its metadata. In addition to files and collections, instances of Resource
 * are used to represent results of runtime queries. In such secenarios, comments, tags, ratings as
 * well as collections of comments, etc. are also represented by Resource objects.
 *
 * Each resource instance contains a unique path within a Registry instance. Registry.get(...)
 * method invocation using this path gives an instance of that resource. This path can be
 * combined with the base URL of the registry server to generate a URI for the resource.
 *
 * <strong>Handling content</strong>
 *
 * Contents of resource can be set either as an input stream or an object. If an input stream is
 * set, an unique file based content will be created for the given input stream. If an object of
 * type byte[] or String is set, an inmemeory input stream will be created from it and then a
 * unique file based input stream is created. If an object of any other type is set, there should
 * be a handler to convert it to an input stream before reaching the repository.
 *
 * When a resource is retrieved from the database layer, its content is not retrieved from the
 * database. Instead, UUID referring to the database content is stored in dbBasedContentID. When
 * the content is first accessed, a file based input stream is created from the database content
 * and fileBasedContentID is set (fileBasedContentID = dbBasedContentID).
 *
 */
public class ResourceImpl implements Resource {

    private static final Log log = LogFactory.getLog(ResourceImpl.class);

    /**
     * UUID to identify the resource. currently this is same as the path
     */
    protected String id;

    /**
     * Snapshot for which this resource instance is created. Any resource and all its versioned
     * data can be accessed only using the resource ID and the version number. But snapshot ID
     * is required to track currently accessing snapshot. Note that it is not possible to
     * derive snapshot ID from the resource ID and the version number, as there can be multiple
     * snapshots using the same version.
     */
    protected long snapshotID;

    /**
     * Version of this instance of the resource. Versioned attributes (e.g. content,
     * last modified date) can be accessed by using the combination of id and versionNumber.
     */
    protected long versionNumber;

    /**
     * Username of the user who added the resource to the registry.
     */
    protected String authorUserName;

    /**
     * Time at which the resource is first added to the registry.
     */
    protected Timestamp createdTime;

    /**
     * Username of the user who modified the resource most recently.
     */
    protected String lastUpdaterUserName;

    /**
     * Time at which the resource modified most recently.
     */
    protected Timestamp lastModified;

    /**
     * Description about the resource. This may contain any text including HTML fragments.
     */
    protected String description;

    /**
     * Unique path of the resource within the registry. This is generated by appending all
     * ascendant paths to the resource name separated by "/". For example assume that the resource
     * name is "users.xml". config.xml if inside the collection named "config". config collection
     * is inside the root level collection named "servers". Then the path of the resource is
     * /servers/config/users.xml.
     */
    protected String path;

    protected long matchingSnapshotID;

    protected String permanentPath;

    /**
     * Media type of the resource. Each resource can have a media type associated with it. This can
     * be either a standart MIME media type or a custom media type defined by the users of the
     * registry. Media type is used to activate media type handlers defined in the registry.
     * Thus, by defining a media type for a resource and by registering a media type handler to
     * handle that media type, it is possible to apply special processing for resources.
     */
    protected String mediaType;

    /**
     * Path of the parent collection of the resource. If the resource path is
     * /servers/config/users.xml, parent path is /servers/config.
     */
    protected String parentPath;

    /**
     * Used to detect whether the resource content is modified after it is retrieved from the
     * Registry. If this is set to true at the time of adding the resource back to the Registry, new
     * version will be created.
     */
    protected boolean contentModified;

    /**
     * Used to detect whether properties of this resource is modified since retrieval from the
     * repository. If true, properties will be updated and new version will be created upon putting
     * this resource to the repository.
     */
    protected boolean propertiesModified;

    /**
     * Determines whether the resource is subjected to changes, which causes it to create a new
     * version. If true, a new version will be created upon adding the resource to the repository.
     */
    protected boolean versionableChange;

    /**
     * Normal resources have the state RegistryConstants.ACTIVE_STATE (100) Deleted resources have
     * the state RegistryConstants.DELETED_STATE (101)
     */
    protected int state;

    /**
     * Properties associated with the resource. A resource can contain zero or more properties,
     * where each property is a name->value pair. Both name and the value should be strings.
     */
    protected Properties properties = new Properties();

    /**
     * UUID of the content stored in the database. This ID is set when a resource is retrieved from
     * the database. Upon the first getContentStream() call, a file based content will be created
     * from this database based content and file based content ID will be set.
     */
    protected int dbBasedContentID = -1;

    /**
     * Content of the resource. Object and the type stored in this field depends on the resource
     * type. If the resource is a file with no special media type handling, this contains an array
     * of bytes (byte[]) conatining the raw bytes of the file. If the resource is a collection, this
     * contains a String[] containing the paths of child resources. If the resource is processed
     * by a media type handler, it is upto the media type handler to set a content. In that case
     * content can be anything ranging from String to custom type. Therefore, clients of the API
     * should be aware of the media type and the content for that media type.
     */
    protected Object content;

    /**
     * Specified whether the resource is a collection (directory) or a file. By default all
     * resources are considered as files. If you want to make a collection, you have to explicitly
     * set the directory field to true.
     */
    protected boolean directory = false;

    /**
     * Content of the resource, represented as a input stream.
     */
    //protected InputStream contentStream;


    /**
     * Registry datasource. This is to be used only by the resource implemetation and users of the
     * resource are not needed to use this. Some attributes of the resource (e.g. content) is
     * fetched upon the first request to optimize the response time. Getters of such attributes use
     * this datasource to fetch the values from the database.
     */
    protected DataSource dataSource;

    /**
     * Resources and collections can be accessed outside the user registry. There are some operations in these entities
     * that require fetching fresh data, that are not authorized previously. For authorizing those data, it is
     * necessary to store the user name and the realm of the user who retrieved the resource/collection instance.
     */
    protected String userName;

    /**
     * Resources and collections can be accessed outside the user registry. There are some operations in these entities
     * that require fetching fresh data, that are not authorized previously. For authorizing those data, it is
     * necessary to store the user name and the realm of the user who retrieved the resource/collection instance.
     */
    protected UserRealm userRealm;

    /**
     * Indicates whether this resource instance is accessed inside or outside a active session.
     */
    private boolean insideSession;

    /**
     * Resource DAO instance to access the database directly. Note that the VersionedResourceDAO is
     * thread safe, so it is possible to keep a single instance for the resource.
     */
    protected ResourceDAO resourceDAO;

    /**
     * path id,
     * if collection - id of the path of the collection
     * if non-collection - id of the path of the non-collection
     */
    int pathID;
    /**
     * the name of the resource
     * if collection - NULL
     * non-collection - the name of the resource..
     */
    String name;

    public ResourceImpl() {
        resourceDAO = new ResourceDAO();
        id = null;
        snapshotID = -1;
    }
    
    public ResourceImpl(String path, ResourceDO resourceDO)  {

        this.id = path;
        this.snapshotID = -1;
        this.versionNumber = resourceDO.getVersion();
        this.authorUserName = resourceDO.getAuthor();
        this.createdTime = resourceDO.getCreatedOn();
        this.lastUpdaterUserName = resourceDO.getLastUpdator();
        this.lastModified = resourceDO.getLastUpdatedOn();
        this.description = resourceDO.getDescription();
        this.path = path;
        this.matchingSnapshotID = -1;
        this.mediaType = resourceDO.getMediaType();
        this.parentPath = RegistryUtils.getParentPath(path);
        this.contentModified = true;
        this.propertiesModified = true;
        this.versionableChange = true;
        this.state = -1;
        this.dbBasedContentID = resourceDO.getContentID();
        this.content = null;
        this.dataSource = null;
        this.userName = null;
        this.userRealm = null;
        this.pathID = resourceDO.getPathID();
        this.name = resourceDO.getName();

        resourceDAO = new ResourceDAO();
    }

    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }

    public void setUserName(String userName) {
        this.userName = userName;
    }

    public void setUserRealm(UserRealm userRealm) {
        this.userRealm = userRealm;
    }

    public String getId() {
        return path;
    }

    public void setId(String path) {
        this.path = path;
    }

    public long getSnapshotID() {
        return snapshotID;
    }

    public void setSnapshotID(long snapshotID) {
        this.snapshotID = snapshotID;
    }

    public long getVersionNumber() {
        return versionNumber;
    }

    public void setVersionNumber(long versionNumber) {
        this.versionNumber = versionNumber;
    }

    public String getAuthorUserName() {
        return authorUserName;
    }

    public void setAuthorUserName(String authorUserName) {
        this.authorUserName = authorUserName;
    }

    public Date getCreatedTime() {
        return createdTime;
    }

    public void setCreatedTime(Date createdTime) {
        this.createdTime = new Timestamp(createdTime.getTime());
    }

    public Date getLastModified() {
        return lastModified;
    }

    public void setLastModified(Date lastModified) {
        this.lastModified = new Timestamp(lastModified.getTime());
    }

    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
        versionableChange = true;
    }

    public String getPath() {
        return path;
    }

    public void setPath(String path) {
        this.path = path == null ? null : preparePath(path);
    }

    public long getMatchingSnapshotID() {
        return matchingSnapshotID;
    }

    public void setMatchingSnapshotID(long matchingSnapshotID) {
        this.matchingSnapshotID = matchingSnapshotID;

        if (matchingSnapshotID == -1) {
            permanentPath = null;
        } else {
            permanentPath =
                    path + RegistryConstants.URL_SEPARATOR + "version:" + matchingSnapshotID;
        }
    }

    public String getPermanentPath() {
        return permanentPath;
    }

    public String getMediaType() {
        return mediaType;
    }

    public void setMediaType(String mediaType) {
        this.mediaType = mediaType;
        versionableChange = true;
    }

    public String getParentPath() {
        if (parentPath != null) return parentPath;
        if ((path == null) || path.length() == 1) return null;
        int i = path.lastIndexOf('/');
        return path.substring(0, i);
    }

    public void setParentPath(String parentPath) {
        this.parentPath = parentPath;
    }

    public int getState() {
        return state;
    }

    public void setState(int state) {
        this.state = state;
    }

    public String getProperty(String key) {

        List propValues = (List) properties.get(key);
        if (propValues == null) {
            return null;
        } else {
            return (String) propValues.get(0);
        }
    }

    public List getPropertyValues(String key) {

        return (List) properties.get(key);
    }

    public Properties getProperties() {

        return properties;
    }

    public void removeProperty(String key) {
        if(key != null) {
            properties.remove(key);
        }

        setPropertiesModified(true);
    }

    public void removePropertyValue(String key, String value) {

        List propValues = (List) properties.get(key);

        if (propValues != null) {
            propValues.remove(value);
            setPropertiesModified(true);
        }
    }

    public void editPropertyValue(String key, String oldValue, String newValue) {

        List propValues = (List) properties.get(key);

        if (propValues != null) {
            propValues.remove(oldValue);
            propValues.add(newValue);
            setPropertiesModified(true);
        }
    }

    public void setProperty(String key, String value) {
        List propValues = new ArrayList();
        propValues.add(value);
        properties.put(key, propValues);

        setPropertiesModified(true);
    }

    public void setProperty(String key, List<String> value) {
        properties.put(key, value);

        setPropertiesModified(true);
    }

    public void addProperty(String key, String value) {
        addPropertyWithNoUpdate(key, value);

        setPropertiesModified(true);
    }


    public void addPropertyWithNoUpdate(String key, String value) {
        List propValues = (List) properties.get(key);
        if (propValues != null) {
            // no need to check if the value is already there. we should permit adding repeating
            // values
            propValues.add(value);
        } else {
            propValues = new ArrayList();
            propValues.add(value);
            properties.put(key, propValues);
        }
    }

    public void setProperties(Properties properties) {
        if(properties != null) {
            this.properties = properties;
        }

        setPropertiesModified(true);
    }

    public int getDbBasedContentID() {
        return dbBasedContentID;
    }

    public void setDbBasedContentID(int dbBasedContentID) {
        this.dbBasedContentID = dbBasedContentID;
    }

    public InputStream getContentStream() throws RegistryException {

        if (content == null) {
            throw new RegistryException("Resource content is empty.");
        }

        if (content instanceof byte[]) {
            return new ByteArrayInputStream((byte[]) content);

        } else if (content instanceof String) {
            byte[] contentBytes = ((String) content).getBytes();
            return new ByteArrayInputStream(contentBytes);

        } else {
            throw new RegistryException(
                    "Cannot return input stream for content of type: " + content.getClass().getName());
        }
    }

    /**
     * Invalidates the current file based content and creates a new file based content for the
     * new content stream. Given content stream will be closed after completing this method.
     *
     * @param contentStream input stream containing the new content
     */
    public void setContentStream(InputStream contentStream) throws RegistryException {

        setContentStreamWithNoUpdate(contentStream);

        setContentModified(true);
    }

    public void setContentStreamWithNoUpdate(InputStream contentStream) throws RegistryException {
        
        content = RegistryUtils.getByteArray(contentStream);
    }

    public Object getContent() throws RegistryException {

        return content;
    }

    public void setContent(Object content) throws RegistryException {

        setContentWithNoUpdate(content);

        setContentModified(true);
    }


    public void setContentWithNoUpdate(Object content) throws RegistryException {

        this.content = content;
    }

    public void prepareContentForPut() throws RegistryException {

        if (content instanceof String) {
            content = ((String) content).getBytes();
        } else if (content instanceof InputStream) {
            content = RegistryUtils.getByteArray((InputStream) content);
        }
    }

    private String preparePath(String path) {

        String preparedPath = path;

        // make sure that the path does not end with a "/"
        if (!path.equals(RegistryConstants.ROOT_PATH) &&
                path.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            preparedPath =
                    path.substring(0, path.length() - RegistryConstants.PATH_SEPARATOR.length());
        }

        return preparedPath;
    }


    public String getLastUpdaterUserName() {
        return lastUpdaterUserName;
    }

    public void setLastUpdaterUserName(String lastUpdaterUserName) {
        this.lastUpdaterUserName = lastUpdaterUserName;
    }

    public boolean isContentModified() {
        return contentModified;
    }

    /**
     * This method is used to explicitly set the content modified state of the resource. Normal
     * users of the Registry API should not call this method.
     *
     * @param contentModified true if we want to add a new version upon putting this to the
     * registry. false otherwise.
     */
    public void setContentModified(boolean contentModified) {
        this.contentModified = contentModified;
        setLastModified(new Date());
        versionableChange = true;
    }

    public boolean isPropertiesModified() {
        return propertiesModified;
    }

    public void setPropertiesModified(boolean propertiesModified) {
        this.propertiesModified = propertiesModified;
        setLastModified(new Date());
        versionableChange = true;
    }


    public void setPropertiesModifiedWithNoUpdate(boolean propertiesModified) {
        this.propertiesModified = propertiesModified;
    }

    public boolean isVersionableChange() {
        return versionableChange;
    }

    public void setVersionableChange(boolean versionableChange) {
        setLastModified(new Date());
        this.versionableChange = versionableChange;
    }

    public List<String> getAspects() {
        return getPropertyValues(Aspect.AVAILABLE_ASPECTS);
    }

    public void addAspect(String name) {
        List aspects = getPropertyValues(Aspect.AVAILABLE_ASPECTS);
        if (aspects == null)
            aspects = new ArrayList<String>();
        aspects.add(name);
        setProperty(Aspect.AVAILABLE_ASPECTS, aspects);
    }

    public void removeAspect(String name) {
        List aspects = getPropertyValues(Aspect.AVAILABLE_ASPECTS);
        if (aspects != null)
            aspects.remove(name);
    }

    public void discard() {

    }

    protected void setSessionInformation() {

        if (CurrentSession.getRealm() != null) {
            // some one else has the session information, which indicates that we are inside an active session.
            // so don't mess with those session information.
            insideSession = true;
        } else {
            insideSession = false;
            CurrentSession.setUser(userName);
            CurrentSession.setRealm(userRealm);
        }
    }

    protected void clearSessionInformation() {

        if (!insideSession) {
            CurrentSession.removeUser();
            CurrentSession.removeRealm();
        }
    }

    public int getPathID() {
        return pathID;
    }

    public void setPathID(int pathID) {
        this.pathID = pathID;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public ResourceIDImpl getResourceIDImpl() {
        ResourceIDImpl resourceIDImpl = new ResourceIDImpl();
        resourceIDImpl.setCollection(this instanceof CollectionImpl);
        resourceIDImpl.setPathID(pathID);
        if (name == null && !(this instanceof CollectionImpl)) {
            name = RegistryUtils.getResourceName(path);
        }
        resourceIDImpl.setName(name);
        resourceIDImpl.setPath(path);

        return resourceIDImpl;
    }

    public ResourceDO getResourceDO() {
        ResourceDO resourceDO = new ResourceDO();
        resourceDO.setPathID(this.getPathID());
        resourceDO.setName(this.name);
        resourceDO.setVersion(this.versionNumber);
        resourceDO.setMediaType(this.mediaType);
        resourceDO.setAuthor(this.authorUserName);
        resourceDO.setCreatedOn(this.createdTime);
        resourceDO.setLastUpdator(this.lastUpdaterUserName);
        resourceDO.setLastUpdatedOn(this.lastModified);
        resourceDO.setDescription(this.description);
        resourceDO.setContentID(this.dbBasedContentID);

        return resourceDO;
    }


    public ResourceImpl getShallowCopy() throws RegistryException {
        ResourceImpl newResource = new ResourceImpl();
        fillResourceCopy(newResource);

        return newResource;
    }

    protected ResourceImpl fillResourceCopy(ResourceImpl resource) throws RegistryException {
        resource.setId(this.id);
        resource.setSnapshotID(this.snapshotID);
        resource.setVersionNumber(this.versionNumber);
        resource.setAuthorUserName(this.authorUserName);
        resource.setCreatedTime(this.createdTime);
        resource.setDescription(this.description);
        resource.setPath(this.path);
        resource.setMatchingSnapshotID(this.matchingSnapshotID);
        resource.setMediaType(this.mediaType);
        resource.setParentPath(this.parentPath);
        resource.setContentModified(this.contentModified);
        resource.setPropertiesModified(this.propertiesModified);
        resource.setVersionableChange(this.versionableChange);
        resource.setState(this.state);
        resource.setProperties(this.properties);
        resource.setDbBasedContentID(dbBasedContentID);
        if (this.content != null) {
            resource.setContent(this.content);
        }
        resource.setDataSource(this.dataSource);
        resource.setUserName(this.userName);
        resource.setUserRealm(this.userRealm);
        resource.setPathID(this.pathID);
        resource.setName(this.name);

        // having these two fields at the end is a must, as above field have an affect of them
        resource.setLastUpdaterUserName(this.lastUpdaterUserName);
        resource.setLastModified(this.lastModified);

        return resource;
    }
}
