/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core;

import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

/**
 * This class represents resource paths used inside the embedded registry. i.e. paths in the form
 * /c1/c2/r1;version:2;view:1
 */
public class ResourcePath {

    private static final String VERSION_PARAMETER_NAME = "version";

    private String completePath;

    private String path;

    private boolean currentVersion = true;

    private Map <String, String> parameters = new HashMap <String, String> ();

    public ResourcePath(String rawPath) {

        completePath = rawPath;

        if (!completePath.startsWith(RegistryConstants.ROOT_PATH)) {
            completePath = RegistryConstants.ROOT_PATH + completePath;
        }

        String[] parts = completePath.split(RegistryConstants.URL_SEPARATOR);

        path = parts[0];
        if (!path.equals(RegistryConstants.ROOT_PATH) && path.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            // trip the last path Seperator
            path = path.substring(0, path.length() - RegistryConstants.PATH_SEPARATOR.length());
        }

        for (int i = 1; i < parts.length; i++) {
            String[] paramParts = parts[i].split(RegistryConstants.URL_PARAMETER_SEPARATOR);
            String key = paramParts[0];
            String value = null;
            if (paramParts.length > 1) {
                value = paramParts[1];
            }

            // if a parameter has more than one value, they are just appended to the value.
            // e.g. raw path: /c1/r1;p1:v1:v2
            // parameter name: p1
            // parameter value: v1:v2
            for (int j = 2; j < paramParts.length; j++) {
                value = value + RegistryConstants.URL_PARAMETER_SEPARATOR + paramParts[j];
            }

            parameters.put(key, value);

            if (VERSION_PARAMETER_NAME.equals(key)) {
                currentVersion = false;
            }
        }
    }


    public String getCompletePath() {
        return completePath;
    }

    public String getPath() {
        return path;
    }

    public boolean parameterExists(String key) {
        return parameters.containsKey(key);
    }

    public String getParameterValue(String key) {
        return parameters.get(key);
    }

    public void setParameter(String key, String value) {

        parameters.put(key, value);
        completePath = path + getParameterPath();
    }

    public void appendPath(String pathToAppend) {

        if (pathToAppend == null || pathToAppend.length() == 0) {
            return;
        }

        if (pathToAppend.startsWith(RegistryConstants.PATH_SEPARATOR)) {
            pathToAppend = pathToAppend.substring(RegistryConstants.PATH_SEPARATOR.length());
        }

        if (!path.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            path = path + RegistryConstants.PATH_SEPARATOR;
        }

        path = path + pathToAppend;
        completePath = path + getParameterPath();
    }

    public boolean isCurrentVersion() {
        return currentVersion;
    }

    public String getPathWithVersion() {

        if (isCurrentVersion()) {
            return getPath();
        }

        String versionString = getParameterValue(RegistryConstants.VERSION_PARAMETER_NAME);
        if (versionString != null) {
            return getPath() +
                    RegistryConstants.URL_SEPARATOR +
                    RegistryConstants.VERSION_PARAMETER_NAME +
                    RegistryConstants.URL_PARAMETER_SEPARATOR +
                    versionString;
        }

        return getPath();
    }

    public String toString() {
        return completePath;
    }

    private String getParameterPath() {
        String paramPath = "";
        Iterator <String> iKeys = parameters.keySet().iterator();
        while (iKeys.hasNext()) {
            String paramKey = iKeys.next();
            String paramValue = parameters.get(paramKey);
            paramPath = paramPath + RegistryConstants.URL_SEPARATOR +
                    paramKey + RegistryConstants.URL_PARAMETER_SEPARATOR + paramValue;

            if (VERSION_PARAMETER_NAME.equals(paramKey)) {
                currentVersion = false;
            }
        }

        return paramPath;
    }
}
