 /*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.config;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.Aspect;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.config.StaticConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.i18n.Messages;
import org.wso2.carbon.registry.core.jdbc.handlers.CustomEditManager;
import org.wso2.carbon.registry.core.jdbc.handlers.EditProcessor;
import org.wso2.carbon.registry.core.jdbc.handlers.Handler;
import org.wso2.carbon.registry.core.jdbc.handlers.filters.Filter;
import org.wso2.carbon.registry.core.jdbc.dao.PathCache;
import org.wso2.carbon.utils.CarbonUtils;
import org.wso2.carbon.CarbonConstants;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;

import java.io.File;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Builds the registry configuration from xml document. Configuration has to be given as an
 * input stream. Registry configuration consists of details of datasources, handlers and
 * aspects. These information is exracted from the configuration populates the necessary
 * components.
 */
public class RegistryConfigurationProcessor {

    private static final Log log = LogFactory.getLog(RegistryConfigurationProcessor.class);

    /**
     * Read XML configuration from the passed InputStream, or from the classpath.
     *
     * @param in an InputStream containing XML data, or null.
     * @param registryContext the RegistryContext to populate
     * @throws RegistryException if there's a problem
     */
    public static void populateRegistryConfig(InputStream in, RegistryContext registryContext)
            throws RegistryException {
        if (in == null) {
            in = Thread.currentThread().getContextClassLoader().getResourceAsStream(
                    "org/wso2/carbon/registry/core/servlet/registry.xml");
        }
        if (in == null) {
            throw new RegistryException(Messages.getMessage("inputstream.null",
                    "org/wso2/servlet/registry.xml"));
        }
        try {
            StAXOMBuilder buillder = new StAXOMBuilder(in);
            OMElement configElement = buillder.getDocumentElement();

            if (configElement != null) {
                OMElement registryRootEle =
                        configElement.getFirstChildWithName(new QName("registryRoot"));
                if (registryRootEle != null) {
                    String registryRoot = registryRootEle.getText();
                    if (registryRoot != null && !registryRoot.equals(RegistryConstants.ROOT_PATH)) {
                        if (registryRoot.endsWith(RegistryConstants.PATH_SEPARATOR)) {
                            registryRoot = registryRoot.substring(0, registryRoot.length() - 1);
                        } else if (!registryRoot.startsWith(RegistryConstants.PATH_SEPARATOR)) {
                            registryRoot = RegistryConstants.ROOT_PATH + registryRoot;
                        }
                    } else {
                        registryRoot = null;
                    }
                    registryContext.setRegistryRoot(registryRoot);
                }

                Iterator dbconfgs = configElement.getChildrenWithName(new QName("dbConfig"));
                while (dbconfgs.hasNext()) {
                    OMElement dbconfig = (OMElement)dbconfgs.next();
                    DataBaseConfiguration dataBaseConfiguration = new DataBaseConfiguration();

                    String dbName = dbconfig.getAttributeValue(new QName("name"));
                    if (dbName == null) {
                        throw new RegistryException(Messages.getMessage("dbname.null"));
                    }
                    dataBaseConfiguration.setConfigName(dbName);
                    OMElement dataSource = dbconfig.getFirstChildWithName(new QName("dataSource"));
                    if (dataSource != null) {
                        dataBaseConfiguration.setDataSourceName(dataSource.getText());
                    } else {
                        OMElement userName = dbconfig.getFirstChildWithName(new QName("userName"));
                        if (userName != null) {
                            dataBaseConfiguration.setUserName(userName.getText());
                        }

                        OMElement password = dbconfig.getFirstChildWithName(new QName("password"));
                        if (password != null) {
                            dataBaseConfiguration.setPassWord(password.getText());
                        }

                        OMElement url = dbconfig.getFirstChildWithName(new QName("url"));
                        String dbUrl = url.getText();
                        if (dbUrl != null) {
                        	if(dbUrl.contains(CarbonConstants.CARBON_HOME_PARAMETER)){
                        		File carbonHomeDir;
                        		carbonHomeDir = new File(CarbonUtils.getCarbonHome());
                        		String path = carbonHomeDir.getPath();
                        		path = path.replaceAll(Pattern.quote("\\"), "/");
                        		if(carbonHomeDir.exists() && carbonHomeDir.isDirectory()){
                        			dbUrl = dbUrl.replaceAll(Pattern.quote(CarbonConstants.CARBON_HOME_PARAMETER),path);
                        		}
                        		else{
                        			 log.warn("carbon home invalid");
                        			 String[] tempStrings1 = dbUrl.split(Pattern.quote(CarbonConstants.CARBON_HOME_PARAMETER));
                         			 String tempUrl = tempStrings1[1];
                         			 String[] tempStrings2 = tempUrl.split("/");
                         			 for(int i=0; i< tempStrings2.length -1; i++){
                         				dbUrl = tempStrings1[0] + tempStrings2[i]+ "/";
                         			 }
                         			dbUrl = dbUrl + tempStrings2[tempStrings2.length -1];
                         		}
                        		
                        		url.setText(dbUrl);
                        	}
                        }
                        dataBaseConfiguration.setDbUrl(url.getText());
                        

                        OMElement driverName =
                                dbconfig.getFirstChildWithName(new QName("driverName"));
                        if (driverName != null) {
                            dataBaseConfiguration.setDriverName(driverName.getText());
                        }

                        OMElement maxWait = dbconfig.getFirstChildWithName(new QName("maxWait"));
                        if (maxWait != null) {
                            dataBaseConfiguration.setMaxWait(maxWait.getText());
                        }

                        OMElement maxActive = dbconfig.getFirstChildWithName(new QName("maxActive"));
                        if (maxActive != null) {
                            dataBaseConfiguration.setMaxActive(maxActive.getText());
                        }

                        OMElement maxIdle = dbconfig.getFirstChildWithName(new QName("maxIdle"));
                        if (maxIdle != null) {
                            dataBaseConfiguration.setMaxIdle(maxIdle.getText());
                        }

                        OMElement minIdle = dbconfig.getFirstChildWithName(new QName("minIdle"));
                        if (minIdle != null) {
                            dataBaseConfiguration.setMinIdle(minIdle.getText());
                        }                        

                        OMElement validationQuery = dbconfig.getFirstChildWithName(new QName("validationQuery"));
                        if (validationQuery != null) {
                            dataBaseConfiguration.setValidationQuery(validationQuery.getText());
                        }                   
                    }
                    registryContext.addDBConfig(dbName, dataBaseConfiguration);
                }

                // loading one-time startup configurations
                OMElement staticConfigElement = configElement.getFirstChildWithName(new QName("staticConfiguration"));
                if (staticConfigElement != null) {
                    Iterator staticConfigs = staticConfigElement.getChildElements();
                    while (staticConfigs.hasNext()) {
                        OMElement staticConfig = (OMElement)staticConfigs.next();

                        if (staticConfig.getLocalName().equals("versioningProperties")) {
                            String versioningProperties = staticConfig.getText();
                            StaticConfiguration.setVersioningProperties(versioningProperties.equals("true"));
                        } else if (staticConfig.getLocalName().equals("versioningComments")) {
                            String versioningComments = staticConfig.getText();
                            StaticConfiguration.setVersioningComments(versioningComments.equals("true"));
                        } else if (staticConfig.getLocalName().equals("versioningTags")) {
                            String versioningTags = staticConfig.getText();
                            StaticConfiguration.setVersioningTags(versioningTags.equals("true"));
                        } else if (staticConfig.getLocalName().equals("versioningRatings")) {
                            String versioningRatings = staticConfig.getText();
                            StaticConfiguration.setVersioningRatings(versioningRatings.equals("true"));
                        } else if (staticConfig.getLocalName().equals("versioningAssociations")) {
                            String versioningAssociations = staticConfig.getText();
                            StaticConfiguration.setVersioningAssociations(versioningAssociations.equals("true"));
                        }else if (staticConfig.getLocalName().equals("profilesPath")) {
                            String profilespath = staticConfig.getText();
                            if(profilespath != null){
                                registryContext.setProfilesPath(profilespath);
                            }
                        } else if (staticConfig.getLocalName().equals("servicePath")) {
                            String servicepath = staticConfig.getText();
                            if (!servicepath.startsWith(RegistryConstants.PATH_SEPARATOR)) {   //if user give the path like test or test/
                                servicepath = RegistryConstants.PATH_SEPARATOR + servicepath;
                            }
                            if (servicepath.endsWith(RegistryConstants.PATH_SEPARATOR)) {
                                servicepath = servicepath.substring(0,(servicepath.length() -1)); //if user give the path like this /test/
                            }
                            if(servicepath != null){
                                registryContext.setServicepath(servicepath);
                            }
                        }
                    }
                }


                // loading cache configurations
                OMElement cacheConfigElement = configElement.getFirstChildWithName(new QName("cacheConfiguration"));
                if (cacheConfigElement != null) {
                    OMElement pathCacheElement = cacheConfigElement.getFirstChildWithName(new QName("pathCache"));
                    Iterator pathCacheConfigChilds = pathCacheElement.getChildElements();
                    int maxElementsInMemory = RegistryConstants.DEFAULT_PATH_CACHE_SIZE;
                    boolean eternal = RegistryConstants.DEFAULT_PATH_CACHE_IS_ETERNAL;
                    long timeToLiveSeconds = RegistryConstants.DEFAULT_PATH_CACHE_TIMETO_LIVE;
                    while (pathCacheConfigChilds.hasNext()) {
                        OMElement pathCacheSubConfig = (OMElement)pathCacheConfigChilds.next();

                        if (pathCacheSubConfig.getLocalName().equals("maxElementsInMemory")) {
                            String maxElementsInMemoryStr = pathCacheSubConfig.getText();
                            if (maxElementsInMemoryStr != null) {
                                maxElementsInMemory = Integer.parseInt(maxElementsInMemoryStr.trim());
                            }
                        } else if (pathCacheSubConfig.getLocalName().equals("eternal")) {
                            String eternalStr = pathCacheSubConfig.getText();
                            eternal =eternalStr.equals("true");
                        } else if (pathCacheSubConfig.getLocalName().equals("timeToLiveSeconds")) {
                            String timeToLiveStr = pathCacheSubConfig.getText();
                            if (timeToLiveStr != null) {
                                timeToLiveSeconds = Long.parseLong(timeToLiveStr.trim());
                            }
                        }
                    }

                    PathCache.createPathCache(maxElementsInMemory, eternal, timeToLiveSeconds);
                }

                OMElement currentConfigElement =
                        configElement.getFirstChildWithName(new QName("currentDBConfig"));
                if (currentConfigElement == null) {
                    throw new RegistryException(Messages.getMessage("db.not.defined"));
                }

                String currentConfigName = currentConfigElement.getText();
                registryContext.selectDBConfig(currentConfigName);

                initializeHandlers(configElement, registryContext);
                readRemoteInstances(configElement, registryContext);

                // process query processor config
                Iterator queryProcessors = configElement.
                        getChildrenWithName(new QName("queryProcessor"));
                while (queryProcessors.hasNext()) {

                    QueryProcessorConfiguration queryProcessorConfiguration =
                            new QueryProcessorConfiguration();

                    OMElement queryProcessorElement = (OMElement)queryProcessors.next();
                    OMElement queryType = queryProcessorElement.
                            getFirstChildWithName(new QName("queryType"));
                    if (queryType != null) {
                        queryProcessorConfiguration.setQueryType(queryType.getText());
                    }

                    OMElement processorName = queryProcessorElement.
                            getFirstChildWithName(new QName("processor"));
                    if (processorName != null) {
                        queryProcessorConfiguration.
                                setProcessorClassName(processorName.getText());
                    }

                    registryContext.addQueryProcessor(queryProcessorConfiguration);
                }

                initializeAspects(configElement, registryContext);

                OMElement versionConfig = configElement.getFirstChildWithName(new QName("versionResourcesOnChange"));
                if (versionConfig != null && "true".equals(versionConfig.getText())) {
                    registryContext.setVersionOnChange(true);
                } else {
                    registryContext.setVersionOnChange(false);
                }
            }

        } catch (XMLStreamException e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public static OMElement getRegistryConfigAsXML(RegistryContext registryContext) {

        OMFactory factory = OMAbstractFactory.getOMFactory();
        OMElement root = factory.createOMElement("wso2registry", null);

        if (registryContext.getDefaultDataBaseConfiguration() != null) {
            OMElement currrentConfigElement = factory.createOMElement("currentConfig", null);
            currrentConfigElement
                    .setText(registryContext.getDefaultDataBaseConfiguration().getConfigName());
            root.addChild(currrentConfigElement);
        }

        Iterator values = registryContext.getDBConfigNames();
        while (values.hasNext()) {
            DataBaseConfiguration dataBaseConfiguration = (DataBaseConfiguration)values.next();
            OMElement config = factory.createOMElement("dbconfig", null);
            OMElement url = factory.createOMElement("url", null);
            url.setText(dataBaseConfiguration.getDbUrl());
            config.addChild(url);
            OMElement userName = factory.createOMElement("userName", null);
            userName.setText(dataBaseConfiguration.getUserName());
            config.addChild(userName);
            OMElement password = factory.createOMElement("password", null);
            password.setText(dataBaseConfiguration.getPassWord());
            config.addChild(password);
            OMElement driverName = factory.createOMElement("driverName", null);
            driverName.setText(dataBaseConfiguration.getDriverName());
            config.addChild(driverName);
            config.addAttribute("name", dataBaseConfiguration.getConfigName(), null);
            root.addChild(config);
        }
        return root;
    }

    public static void initializeAspects(OMElement configElement, RegistryContext registryContext) throws RegistryException {
        Iterator aspectElement = configElement.
                getChildrenWithName(new QName("aspect"));
        if (aspectElement != null) {
            while (aspectElement.hasNext()) {
                OMElement aspect = (OMElement) aspectElement.next();
                String name = aspect.getAttributeValue(new QName("name"));
                String clazz = aspect.getAttributeValue(new QName("class"));
                Aspect aspectInstance = null;
                try {
                    Class handlerClass = Class.forName(clazz);

                    if (aspect.getChildElements().hasNext()) {
                        try {
                            Constructor constructor =
                                    handlerClass.getConstructor(OMElement.class);
                            try {
                                aspectInstance = (Aspect)constructor.newInstance(aspect);
                            } catch (Exception e) {
                                throw new RegistryException("Couldn't instanitate", e);
                            }
                        } catch (NoSuchMethodException e) {
                            // Throw error because the specified config won't be used?
                        }
                    }

                    if (aspectInstance == null) {
                        aspectInstance = (Aspect) handlerClass.newInstance();
                    }
                    registryContext.addAspect(name, aspectInstance);
                } catch (Exception e) {
                    throw new RegistryException("Unable to load the class : " + clazz) ;
                }
                if (name == null || clazz == null) {
                    throw new RegistryException("Invalid aspect elemet , required " +
                            "values are missing " + aspect.toString());
                }
            }
        }
    }

    private static void initializeHandlers(OMElement configElement, RegistryContext registryContext)
            throws RegistryException {

        // process handler configurations

        CustomEditManager customEditManager = registryContext.getCustomEditManager();

        try {
            Iterator<OMElement> handlerConfigs =
                    configElement.getChildrenWithName(new QName("handler"));
            while (handlerConfigs.hasNext()) {

                OMElement handlerConfigElement =  handlerConfigs.next();

                String handlerClassName = handlerConfigElement.getAttributeValue(new QName("class"));
                String methodsValue = handlerConfigElement.getAttributeValue(new QName("methods"));

                String[] methods = null;
                if (methodsValue != null) {

                    methods = methodsValue.split(",");
                    for (int i = 0; i < methods.length; i++) {
                        methods[i] = methods[i].trim();
                    }
                }

                Class handlerClass;
                try {
                    handlerClass = Class.forName(handlerClassName);
                } catch (ClassNotFoundException e) {
                    String msg = "Could not find the handler class " + handlerClassName +
                            ". This handler will not be registered. All handler and " +
                            "filter classes should be in the class path of the Registry.";
                    log.error(msg, e);
                    continue;
                }
                Handler handler = (Handler) handlerClass.newInstance();

                // set configured properties of the handler object
                Iterator<OMElement> handlerProps =
                        handlerConfigElement.getChildrenWithName(new QName("property"));
                while (handlerProps.hasNext()) {
                    OMElement propElement = handlerProps.next();

                    String propName = propElement.getAttributeValue(new QName("name"));
                    String propType = propElement.getAttributeValue(new QName("type"));

                    if (propType != null && "xml".equals(propType)) {

                        String setterName = getSetterName(propName);
                        Method setter = handlerClass.getMethod(setterName, OMElement.class);
                        setter.invoke(handler, propElement);

                    } else {

                        String setterName = getSetterName(propName);
                        Method setter = handlerClass.getMethod(setterName, String.class);
                        String propValue = propElement.getText();
                        setter.invoke(handler, propValue);
                    }
                }

                // initialize and configure the filter for this handler
                OMElement filterElement =
                        handlerConfigElement.getFirstChildWithName(new QName("filter"));
                String filterClassName = filterElement.getAttributeValue(new QName("class"));

                Class filterClass;
                try {
                    filterClass = Class.forName(filterClassName);
                } catch (ClassNotFoundException e) {
                    String msg = "Could not find the filter class " +
                            filterClassName + ". " + handlerClassName +
                            " will not be registered. All configured handler, filter and " +
                            "edit processor classes should be in the class " +
                            "path of the Registry.";
                    log.error(msg, e);
                    continue;
                }
                Filter filter = (Filter) filterClass.newInstance();

                // set configured properties of the filter object
                Iterator<OMElement> filterProps =
                        filterElement.getChildrenWithName(new QName("property"));
                while (filterProps.hasNext()) {
                    OMElement propElement = filterProps.next();

                    String propName = propElement.getAttributeValue(new QName("name"));
                    String propValue = propElement.getText();

                    String setterName = getSetterName(propName);
                    Method setter = filterClass.getMethod(setterName, String.class);
                    setter.invoke(filter, propValue);
                }

                OMElement editElement =
                        handlerConfigElement.getFirstChildWithName(new QName("edit"));
                if (editElement != null) {
                    String processorKey = editElement.getAttributeValue(new QName("processor"));
                    String processorClassName = editElement.getText();

                    Class editProcessotClass;
                    try {
                        editProcessotClass = Class.forName(processorClassName);
                    } catch (ClassNotFoundException e) {
                        String msg = "Could not find the edit processor class " +
                                processorClassName + ". " + handlerClassName +
                                " will not be registered. All configured handler, filter and " +
                                "edit processor classes should be in the class " +
                                "path of the Registry.";
                        log.error(msg, e);
                        continue;
                    }
                    EditProcessor editProcessor = (EditProcessor) editProcessotClass.newInstance();

                    customEditManager.addProcessor(processorKey, editProcessor);
                }

                registryContext.getHandlerManager().addHandler(methods, filter, handler);
            }

        } catch (Exception e) {

            String msg = "Could not initialize custom handlers. Caused by: " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private static void readRemoteInstances(OMElement configElement,
                                            RegistryContext registryContext) throws RegistryException {
        try {
            Iterator<OMElement> remoteConfigs =
                    configElement.getChildrenWithName(new QName("remoteInstance"));
            List<String> idList = new ArrayList();

            while (remoteConfigs.hasNext()) {
                OMElement remoteConfigElement = remoteConfigs.next();

                String url = remoteConfigElement.getAttributeValue(new QName("url"));
                String id = remoteConfigElement.getFirstChildWithName(new QName("id")).getText();

                if (idList.contains(id)) {
                    String msg = "Two remote instances can't have the same id.";
                    log.error(msg);
                    throw new RegistryException(msg);
                }
                idList.add(id);

                String trustedUser = remoteConfigElement.getFirstChildWithName(new QName("username")).getText();
                String trustedPwd = remoteConfigElement.getFirstChildWithName(new QName("password")).getText();

                RemoteConfiguration remoteConfiguration = new RemoteConfiguration();
                remoteConfiguration.setId(id);
                remoteConfiguration.setUrl(url);
                remoteConfiguration.setTrustedUser(trustedUser);
                remoteConfiguration.setTrustedPwd(trustedPwd);

                registryContext.getRemoteInstances().add(remoteConfiguration);

            }
        } catch (Exception e) {
            String msg = "Could not read remote instance configuration. Caused by: " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }

    private static String getSetterName(String varName) {

        String setterName;

        if (varName.length() == 1) {
            setterName = "set" + varName.substring(0, 1).toUpperCase();
        } else {
            setterName = "set" +
                    varName.substring(0, 1).toUpperCase() + varName.substring(1, varName.length());
        }

        return setterName;
    }
}
