/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.internal;

import org.osgi.framework.BundleContext;
import org.osgi.service.component.ComponentContext;
import org.wso2.carbon.registry.core.jdbc.EmbeddedRegistry;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.jdbc.realm.RegistryRealm;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.config.RegistryConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.service.RegistryService;
import org.wso2.carbon.registry.core.session.UserRegistry;
import org.wso2.carbon.registry.core.ResourcePath;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.app.RemoteRegistryService;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.utils.CarbonUtils;
import org.wso2.carbon.utils.ServerConstants;
import org.wso2.carbon.registry.app.Utils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.File;

/**
 * @scr.component name="registry.core.dscomponent"" immediate="true"
 * @scr.reference name="user.realm.default" interface="org.wso2.carbon.user.core.UserRealm"
 * cardinality="1..1" policy="dynamic" target="(RealmGenre=Delegating)" bind="setUserRealmDelegating"
 * unbind="unsetUserRealmDelegating"
 */
public class RegistryCoreServiceComponent {

    private static final String CARBON_COLLECTION_NAME = "carbon";
    private static final String CARBON_COLLECTION_PARENT_PROPERTY = "carbon.registry.root";
    private static final String CARBON_REGISTRY_CLEAN = "carbon.registry.clean";
    private static RegistryConfiguration registryConfig;
  
    private static UserRealm userRealmService;

    private static final Log log = LogFactory.getLog(RegistryCoreServiceComponent.class);

    protected void activate(ComponentContext ctxt) {
        try {
            start(ctxt.getBundleContext());
            log.debug("Registry Core bundle is activated ");
        } catch (Throwable e) {
            log.error("Failed to activate Registry Core bundle ", e);
        }
    }

    protected void deactivate(ComponentContext ctxt) {
        log.debug("Registry Core bundle is deactivated ");
    }

    public void start(BundleContext bundleContext) throws Exception {

        RegistryCoreServiceComponent.registryConfig = getRegistryConfiguration();

        UserRealm registryRealm =
                new RegistryRealm(RegistryCoreServiceComponent.getUserRealmService());

        //Setting carbon.home system property. Need to fix this
        String carbonHome = System.getProperty(ServerConstants.CARBON_HOME);
        if (carbonHome != null) {
            System.setProperty(ServerConstants.CARBON_HOME, carbonHome);
        }

        ServerConfiguration serverConfig = getServerConfiguration();
        setSystemTruststore(serverConfig);
                
        RegistryService registryService;

        if (RegistryConfiguration.REMOTE_REGISTRY
                .equals(RegistryCoreServiceComponent.registryConfig.getRegistryType())) {
            registryService = getRemoteRegistryService(registryRealm,
                    RegistryCoreServiceComponent.registryConfig);
        } else {
            registryService = getEmbbedRegistryService(registryRealm);
            Utils.setEmbeddedRegistry((EmbeddedRegistry) registryService);
        }

        String registryMode = "READ-WRITE";
        if (CarbonUtils.isRegistryReadOnly()) {
            registryMode = "READ-ONLY";
        }
        String registryRoot = System.getProperty("carbon.registry.root");
        if (registryRoot != null) {
            log.info("Registry Root    : " + registryRoot);
        }
        log.info("Registry Mode    : " + registryMode);
        if (!RegistryConfiguration.EMBEDDED_REGISTRY.equals(
                RegistryCoreServiceComponent.registryConfig.getRegistryType())) {
            log.info("Registry Type    : "
                     + RegistryCoreServiceComponent.registryConfig.getRegistryType());
        }

        bundleContext.registerService(RegistryService.class.getName(), registryService, null);
        
    }

    private File getConfigFile() throws RegistryException {
        String carbonHome = System.getProperty(ServerConstants.CARBON_HOME);
        if (carbonHome != null) {
            String configPath = carbonHome + File.separator +
                    "conf" + File.separator + "registry.xml";
            File registryXML = new File(configPath);
            if (!registryXML.exists()) {
                String msg = "Registry configuration file (registry.xml) file does " +
                        "not exist in the path " + configPath;
                log.error(msg);
                throw new RegistryException(msg);
            }
            return registryXML;
        } else {
            String msg = "carbon.home system property is not set. It is required to to derive " +
                    "the path of the Registry configuration file (registry.xml).";
            log.error(msg);
            throw new RegistryException(msg);
        }
    }

    private void addCarbonRootCollection(UserRegistry userRegistry) throws RegistryException {

        String carbonPathParent = System.getProperty(CARBON_COLLECTION_PARENT_PROPERTY);

        String carbonPath;
        if (carbonPathParent != null) {
            ResourcePath rootPath = new ResourcePath(carbonPathParent);
            rootPath.appendPath(CARBON_COLLECTION_NAME);
            carbonPath = rootPath.getPath();
        } else {
            carbonPath = RegistryConstants.ROOT_PATH + CARBON_COLLECTION_NAME;
        }


        String carbonRoot;
        int index = carbonPath.indexOf(RegistryConstants.PATH_SEPARATOR, 1);
        if (index == -1) {
            carbonRoot = carbonPath;
        } else {
            carbonRoot = carbonPath.substring(0, index);
        }

        boolean inTransaction = Transaction.isStarted();
        try {
            if (!inTransaction) {
                userRegistry.beginTransaction();
            }

            if (System.getProperty(CARBON_REGISTRY_CLEAN) != null) {
                // clean the system property to prevent the registry cleanup at the next
                // restart, if commanded through the admin console, which will carry all the
                // System properties
                System.clearProperty(CARBON_REGISTRY_CLEAN);
                if (userRegistry.resourceExists(carbonPath)) {
                    userRegistry.delete(carbonPath);
                    log.info("Cleaned the registry space of carbon at path : " + carbonPath);
                }
            }

            Collection carbonRootCollection = userRegistry.newCollection();
            String carbonRootDescription = "Root of the system collections used by the Carbon server.";
            carbonRootCollection.setDescription(carbonRootDescription);
            userRegistry.put(carbonRoot, carbonRootCollection);
            carbonRootCollection.discard();

            UserRealm userRealm = userRegistry.getUserRealm();
            AuthorizationUtils.setAdminAuthorizations(carbonRoot, userRealm);

            Collection carbonCollection = userRegistry.newCollection();
            userRegistry.put(carbonPath, carbonCollection);
            carbonCollection.discard();

            if (!inTransaction) {
                userRegistry.commitTransaction();
            }

        } catch (Exception e) {
            String msg = "Unable to setup collections used by the Carbon server.";
            log.error(msg, e);
            if (!inTransaction) {
                userRegistry.rollbackTransaction();
            }
            throw new RegistryException(e.getMessage(), e);
        }

    }

    private RegistryService getEmbbedRegistryService(UserRealm registryRealm) throws Exception {

        InputStream configInputStream = new FileInputStream(getConfigFile());
        RegistryContext registryContext = new RegistryContext(configInputStream, registryRealm);
        registryContext.setSetup(System.getProperty(RegistryConstants.SETUP_PROPERTY) != null);
        EmbeddedRegistry embeddedRegistry = new EmbeddedRegistry(registryContext);

        UserRegistry systemRegistry = embeddedRegistry.getSystemRegistry();
        addCarbonRootCollection(systemRegistry);

        return embeddedRegistry;
    }

    private RegistryService getRemoteRegistryService(UserRealm registryRealm,
                                                     RegistryConfiguration regConfig) throws Exception {

        String url = regConfig.getValue(RegistryConfiguration.URL);
        String username = regConfig.getValue(RegistryConfiguration.USERNAME);
        String password = regConfig.getValue(RegistryConfiguration.PASSWORD);
        String chroot = regConfig.getValue(RegistryConfiguration.REGISTRY_ROOT);

        RemoteRegistryService registryService =
                new RemoteRegistryService(url, username, password, registryRealm, chroot);

        UserRegistry systemRegistry = registryService.getSystemRegistry();
        addCarbonRootCollection(systemRegistry);

        return registryService;
    }

    private RegistryConfiguration getRegistryConfiguration() throws RegistryException {
        String path = CarbonUtils.getServerXml();
        return new RegistryConfiguration(path);
    }

    private void setSystemTruststore(ServerConfiguration serverConfiguration) throws RegistryException {
        if (serverConfiguration != null) {
            String location = serverConfiguration.getFirstProperty("Security.KeyStore.Location");
            String password = serverConfiguration.getFirstProperty("Security.KeyStore.Password");
            String type = serverConfiguration.getFirstProperty("Security.KeyStore.Type");

            if (location != null && password != null) {
                System.setProperty("javax.net.ssl.trustStore", location);
                System.setProperty("javax.net.ssl.trustStoreType", type);
                System.setProperty("javax.net.ssl.trustStorePassword", password);
            }
        }
    }

    private ServerConfiguration getServerConfiguration() throws RegistryException {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        try {
            File carbonXML = new File(CarbonUtils.getServerXml());
            InputStream inSXml = new FileInputStream(carbonXML);
            serverConfig.init(inSXml);
        } catch (Exception e) {
            log.error("Error initializing the Server Configuration", e);
            throw new RegistryException("Error initializing the Server Configuration", e);
        }
        return serverConfig;
    }

    protected void setUserRealmDelegating(UserRealm userRealmDelegating) {
        userRealmService = userRealmDelegating;
    }

    protected void unsetUserRealmDelegating(UserRealm userRealmDelegating) {
        userRealmService = null;
    }

    public static UserRealm getUserRealmService() {
        return userRealmService;
    }
    
    public static RegistryConfiguration getRegistryConfig() {
        return registryConfig;
    }

    public static void setRegistryConfig(RegistryConfiguration registryConfig) {
        RegistryCoreServiceComponent.registryConfig = registryConfig;
    }
}
