/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.handlers.Handler;
import org.wso2.carbon.registry.core.jdbc.handlers.filters.Filter;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.*;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Date;
import java.util.Map;
import java.io.OutputStream;
import java.io.Writer;
import java.io.Reader;

public class AtomicRegistry implements Registry {

    public static final Log log = LogFactory.getLog(AtomicRegistry.class);

    private BasicRegistry basicRegistry;

    private DataSource dataSource;

    public AtomicRegistry(
            BasicRegistry basicRegistry, DataSource dataSource) {

        this.basicRegistry = basicRegistry;
        this.dataSource = dataSource;
    }

    public void beginTransaction() throws RegistryException {

        if (Transaction.isStarted()) {
            String msg = "Failed to start new transaction. Transaction is already " +
                    "started by this thread. Please close started transactions before " +
                    "starting new transaction.";

            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn;
        try {
            conn = dataSource.getConnection();
            if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
                conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
            }
            conn.setAutoCommit(false);
        } catch (SQLException e) {
            String msg = "Failed to start new registry transaction. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        Transaction.setConnection(conn);
    }

    public void rollbackTransaction() throws RegistryException {

        if (!Transaction.isStarted()) {
            String msg = "Failed to rollback transaction. A transaction is not started.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn = Transaction.getConnection();
        try {
            conn.rollback();

        } catch (SQLException e) {
            String msg = "Failed to rollback transaction. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } finally {
            endTransaction();
        }
    }

    public void commitTransaction() throws RegistryException {

        if (!Transaction.isStarted()) {
            String msg = "Failed to commit transaction. A transaction is not started.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn = Transaction.getConnection();
        try {
            conn.commit();

        } catch (SQLException e) {
            String msg = "Failed to commit transaction. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } finally {
            endTransaction();
        }
    }

    private void endTransaction() throws RegistryException {

        if (Transaction.isStarted()) {
            Connection conn = Transaction.getConnection();
            try {
                conn.close();

            } catch (SQLException e) {
                String msg = "Failed to close transaction. " + e.getMessage();
                log.error(msg);
                throw new RegistryException(msg);

            } finally {
                Transaction.setStarted(false);
                Transaction.setConnection(null);
            }
        }

    }

    public Resource newResource() throws RegistryException {
        String msg = "Atomic registry cannot create resource instances. Please use " +
                "JDBCRegistry for instantiating new resource and collection instances.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public Collection newCollection() throws RegistryException {
        String msg = "Atomic registry cannot create collection instances. Please use " +
                "JDBCRegistry for instantiating new resource and collection instances.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public RegistryContext getRegistryContext() {
        return null;
    }

    public void setRegistryContext(RegistryContext registryContext) {
    }

    public void addHandler(Filter filter, Handler handler) throws RegistryException {
        String msg = "Atomic registry does not support adding handlers.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Resource get(String path) throws RegistryException {
        boolean success = false;
        try {
            beginTransaction();

            Resource resource = basicRegistry.get(path);

            success = true;

            return resource;

        } finally {
            if (success) {
                commitTransaction();
            } else {
                rollbackTransaction();
            }
        }
    }
    /**
     * Returns meta data of the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Resource getMetaData(String path) throws RegistryException {
        boolean success = false;
        try {
            beginTransaction();

            Resource resource = basicRegistry.getMetaData(path);

            success = true;

            return resource;

        } finally {
            if (success) {
                commitTransaction();
            } else {
                rollbackTransaction();
            }
        }
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {
        boolean success = false;
        try {
            beginTransaction();

            Collection collection = basicRegistry.get(path, start, pageSize);

            success = true;

            return collection;

        } finally {
            if (success) {
                commitTransaction();
            } else {
                rollbackTransaction();
            }
        }
    }

    public boolean resourceExists(String path) throws RegistryException {

        boolean exists;

        try {
            beginTransaction();

            exists = basicRegistry.resourceExists(path);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to the existence of the resource " + path + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg);

        }

        return exists;
    }

    /**
     * Adds or updates resources in the registry. If there is no resource in the given path,
     * resource is added. If a resource already exist in the given path, it will be replaced with
     * the new resource. Implementations may use a concurrency control method to prevent concurrency
     * issues.
     *
     * @param suggestedPath the path which we'd like to use for the new resource.
     * @param resource      Resource instance for the new resource
     * @return the actual path that the server chose to use for our Resource
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException
     *
     */
    public synchronized String put(String suggestedPath, Resource resource)
            throws RegistryException {

        try {
            beginTransaction();

            String actualPath = basicRegistry.put(suggestedPath, resource);

            commitTransaction();

            return actualPath;

        } catch (Exception e) {

            String msg = "Failed to put resource " + suggestedPath +
                    ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);

        }
    }

    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {

        try {
            beginTransaction();

            String actualPath = basicRegistry.
                    importResource(suggestedPath, sourceURL, metadata);

            commitTransaction();

            return actualPath;

        } catch (Exception e) {

            String msg = "Failed to import resource " + suggestedPath +
                    ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    /**
     * Deletes the resource in the given path. If the path refers to a directory, all child
     * resources of the directory will also be deleted.
     *
     * @param path Path of the resource to be deleted.
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException
     *          : if the user is not authorized
     */
    public synchronized void delete(String path) throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.delete(path);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to delete resource " + path +
                    ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);

        }
    }

    public String rename(String currentPath, String newName) throws RegistryException {

        try {

            beginTransaction();

            String newPath  = basicRegistry.rename(currentPath, newName);

            commitTransaction();

            return newPath;

        } catch (Exception e) {

            String msg = "Failed to rename resource " + currentPath + " to the new name " +
                    newName + ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        try {

            beginTransaction();

            String movedPath  = basicRegistry.move(currentPath, newPath);

            commitTransaction();

            return movedPath;

        } catch (Exception e) {

            String msg = "Failed to move resource " + currentPath + " to the new path " +
                    newPath + ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);

        }
    }

    public synchronized String copy(String sourcePath, String targetPath) throws RegistryException {

        try {

            beginTransaction();

            String copiedPath  = basicRegistry.copy(sourcePath, targetPath);

            commitTransaction();

            return copiedPath;

        } catch (Exception e) {

            String msg = "Failed to copy resource " + sourcePath + " to the path " +
                    targetPath + ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void createVersion(String path) throws RegistryException {

        boolean success = false;
        try {

            beginTransaction();

            basicRegistry.createVersion(path);

            success = true;

        } finally {

            if (success) {
                commitTransaction();
            } else {
                rollbackTransaction();
            }
        }
    }

    public String[] getVersions(String path) throws RegistryException {

        boolean success = false;
        try {

            beginTransaction();

            String[] versionPaths = basicRegistry.getVersions(path);

            success = true;

            return versionPaths;

        } finally {

            if (success) {
                commitTransaction();
            } else {
                rollbackTransaction();
            }
        }
    }

    public synchronized void restoreVersion(String versionPath) throws RegistryException {

        boolean success = false;
        try {

            beginTransaction();

            basicRegistry.restoreVersion(versionPath);

            success = true;

        } finally {

            if (success) {
                commitTransaction();
            } else {
                rollbackTransaction();
            }
        }
    }

    public void addAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.addAssociation(sourcePath, targetPath, associationType);

            commitTransaction();

        } catch(Exception e) {

            String msg = "Failed to add association of type " +
                    associationType + " between resources " + sourcePath +
                    " and " + targetPath + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);

        }
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        Association[] associations;
        try {

            beginTransaction();

            associations = basicRegistry.getAllAssociations(resourcePath);

            commitTransaction();

        } catch(Exception e) {

            String msg = "Failed to get all associations of the resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);

        }

        return associations;
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

        Association[] associations;
        try {

            beginTransaction();

            associations = basicRegistry.getAssociations(resourcePath, associationType);

            commitTransaction();

        } catch(Exception e) {

            String msg = "Failed to get associations of type " + associationType +
                    " for the resource " + resourcePath + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);

        }

        return associations;

    }

    public void removeAssociation(String sourcePath, String associationPath, String associationType)
            throws RegistryException {
        try {
            beginTransaction();
            basicRegistry.removeAssociation(sourcePath, associationPath, associationType);
            commitTransaction();
        } catch(Exception e) {
            String msg = "Failed to remove association of type " + associationType +
                    " between resources " + sourcePath + " and " + associationPath +
                    ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }

    }

////////////////////////////////////////////////////////
// Tagging
////////////////////////////////////////////////////////

    /**
     * Applies the given tag to the resource in the given path. If the given tag is not defined in
     * the registry, it will be defined.
     *
     * @param resourcePath Path of the resource to be tagged.
     * @param tag          Tag. Any string can be used for the tag.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void applyTag(String resourcePath, String tag) throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.applyTag(resourcePath, tag);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to tag resource " + resourcePath + " with the tag " +
                    tag + ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    /**
     * Returns the paths of the resources, which are tagged using the given tag.
     *
     * @param tag Tag
     * @return Resource paths
     */
    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        TaggedResourcePath[] taggedPaths;

        try {

            beginTransaction();

            taggedPaths = basicRegistry.getResourcePathsWithTag(tag);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to get resource paths tagged with tag " + tag +
                    ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }

        return taggedPaths;
    }

    /**
     * Returns all tags used for tagging the given resource.
     *
     * @param resourcePath Path of the resource
     * @return Tags
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public Tag[] getTags(String resourcePath) throws RegistryException {

        try {

            beginTransaction();

            Tag[] tags = basicRegistry.getTags(resourcePath);

            commitTransaction();

            return tags;

        } catch (Exception e) {

            String msg = "Failed to get tags applied for the resource " + resourcePath +
                    ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void removeTag(String path, String tag) throws RegistryException {

        try {
            beginTransaction();

            basicRegistry.removeTag(path, tag);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to remove the tag " + tag + " from the resource " + path +
                    ". All database operations will be rolled back. " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

////////////////////////////////////////////////////////
// Comments
////////////////////////////////////////////////////////

    /**
     * Adds a comment to a resource.
     *
     * @param resourcePath Path of the resource to add the comment.
     * @param comment      Comment instance for the new comment.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized String addComment(String resourcePath, Comment comment)
            throws RegistryException {

        try {

            beginTransaction();

            String commentPath = basicRegistry.addComment(resourcePath, comment);

            commitTransaction();

            return commentPath;

        } catch (Exception e) {

            String msg = "Failed to add comment " + comment.getText() + " to the resource " +
                    resourcePath + ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }


    public void editComment(String commentPath, String text) throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.editComment(commentPath, text);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to edit comment " + commentPath +
                    ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    /**
     * Returns all comments for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Array of Comment objects.
     */
    public Comment[] getComments(String resourcePath) throws RegistryException {

        try {

            beginTransaction();

            Comment[] comments =  basicRegistry.getComments(resourcePath);

            commitTransaction();

            return comments;

        } catch (Exception e) {

            String msg = "Failed get comments on resource " + resourcePath +
                    ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

////////////////////////////////////////////////////////
// Ratings
////////////////////////////////////////////////////////

    /**
     * Rates the given resource.
     *
     * @param resourcePath Path of the resource.
     * @param rating       rating value.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void rateResource(String resourcePath, int rating)
            throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.rateResource(resourcePath, rating);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed rate the resource " + resourcePath + " with rating " + rating +
                    ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    /**
     * Returns the average rating (from all users) for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return average rating.
     */
    public float getAverageRating(String resourcePath) throws RegistryException {

        try {
            beginTransaction();

            float rating = basicRegistry.getAverageRating(resourcePath);

            commitTransaction();

            return rating;

        } catch (Exception e) {

            String msg = "Failed get average rating of the resource " + resourcePath +
                    ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public int getRating(String path, String userName) throws RegistryException {

        try {
            beginTransaction();

            int rating = basicRegistry.getRating(path, userName);

            commitTransaction();

            return rating;

        } catch (Exception e) {

            String msg = "Failed to get the rating of the resource " + path + " done by user " +
                    userName + ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

//////////////////////////////////////////////////////////////////////////////
// Extensible searching API
//////////////////////////////////////////////////////////////////////////////

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        try {

            beginTransaction();

            Collection resultCollection = basicRegistry.executeQuery(path, parameters);

            commitTransaction();

            return resultCollection;

        } catch (Exception e) {

            String msg = "Failed to execute query " + path +
                    ". All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from, Date to, boolean recentFirst) throws RegistryException {

        try {

            beginTransaction();

            LogEntry[] logEntryList = basicRegistry.
                    getLogs(resourcePath, action, userName, from, to, recentFirst);

            commitTransaction();

            return logEntryList;

        } catch (Exception e) {

            String msg = "Failed to get log. . All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }


    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {
        try {
            beginTransaction();
            LogEntryCollection logEntryCollection = basicRegistry.
                    getLogCollection(resourcePath, action, userName, from, to, recentFirst);
            commitTransaction();
            return logEntryCollection;
        } catch (Exception e) {
            String msg = "Failed to get log. . All database operations will be rolled back. " +
                    e.getMessage();
            log.error(msg, e);
            rollbackTransaction();
            throw new RegistryException(msg, e);
        }
    }

    public String[] getAvailableAspects() {
        return basicRegistry.getAvailableAspects();
    }

    public void addAspect(String name, Aspect aspect) throws RegistryException {
        String msg = "Atomic registry does not support adding aspects.";
        log.error(msg);
        throw new RegistryException(msg);
    }

    public void associateAspect(String resourcePath, String name) throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.associateAspect(resourcePath, name);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to associate aspect " + name + " with the resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public Aspect getResourceAspect(Resource resource, String aspectName)
            throws RegistryException {

        try {

            beginTransaction();

            Aspect aspect = basicRegistry.getResourceAspect(resource, aspectName);

            commitTransaction();

            return aspect;

        } catch (Exception e) {

            String msg = "Failed to get aspect " + aspectName + " of the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void invokeAspect(String resourcePath, String aspectName, String action)
            throws RegistryException {

        try {

            beginTransaction();

            basicRegistry.invokeAspect(resourcePath, aspectName, action);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to invoke action '" + action + "' of aspect " + aspectName +
                    " for the resource " + resourcePath + ": " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public String[] getAspectActions(String resourcePath, String aspectName)
            throws RegistryException {

        try {

            beginTransaction();

            String[] aspectActions =
                    basicRegistry.getAspectActions(resourcePath, aspectName);

            commitTransaction();

            return aspectActions;

        } catch (Exception e) {

            String msg = "Failed to get aspect actions of aspect " + aspectName +
                    " for the resource " + resourcePath + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public Collection searchContent(String keywords) throws RegistryException {
        try {

            beginTransaction();

            Collection result = basicRegistry.searchContent(keywords);

            commitTransaction();

            return result;

        } catch (Exception e) {

            String msg = "Failed to search content for the keywords " + keywords + ". " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void createLink(String path, String target) throws RegistryException {
        try {
            beginTransaction();

            basicRegistry.createLink(path, target);

            commitTransaction();
            
        } catch (Exception e) {

            String msg = "Failed to create link " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void createLink(String path, String target, String targetSubPath) throws RegistryException {
        try {
            beginTransaction();

            basicRegistry.createLink(path, target, targetSubPath);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to create link " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void removeLink(String path) throws RegistryException {
        try {
            beginTransaction();

            basicRegistry.removeLink(path);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to remove link " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    // check in, check out functionality
    public void restore(String path, Reader reader) throws RegistryException {
        try {
            beginTransaction();

            basicRegistry.restore(path, reader);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to restore: " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void dump(String path, Writer writer) throws RegistryException {

        try {
            beginTransaction();

            basicRegistry.dump(path, writer);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to dump: " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public String getEventingServiceURL(String path) throws RegistryException {
        try {
            beginTransaction();

            String eventingServiceURL = basicRegistry.getEventingServiceURL(path);

            commitTransaction();

            return eventingServiceURL;

        } catch (Exception e) {

            String msg = "Failed to get eventing service URL: " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }

    public void setEventingServiceURL(String path, String eventingServiceURL) throws RegistryException {
        try {
            beginTransaction();

            basicRegistry.setEventingServiceURL(path, eventingServiceURL);

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to set eventing service URL: " + e.getMessage();
            log.error(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }
    }
}
