/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Searcher;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.config.RemoteConfiguration;
import org.wso2.carbon.registry.core.config.StaticConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.exceptions.ResourceNotFoundException;
import org.wso2.carbon.registry.core.jdbc.dao.*;
import org.wso2.carbon.registry.core.jdbc.handlers.HandlerManager;
import org.wso2.carbon.registry.core.jdbc.handlers.RequestContext;
import org.wso2.carbon.registry.core.jdbc.handlers.builtin.MountHandler;
import org.wso2.carbon.registry.core.jdbc.handlers.builtin.SymLinkHandler;
import org.wso2.carbon.registry.core.jdbc.handlers.filters.Filter;
import org.wso2.carbon.registry.core.jdbc.handlers.filters.URLMatcher;
import org.wso2.carbon.registry.core.jdbc.queries.QueryProcessorManager;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.utils.VersionedPath;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;

import java.io.IOException;
import java.io.Writer;
import java.io.Reader;
import java.util.*;

/**
 * This class contains JDBC based implementations of all Registry API methods. This invokes handlers
 * through the HandlerManager for all the methods that allow handlers. Once the handler invocation
 * is complete and if the processing is not completed, it handles the request using the generic
 * procedure. For the methods that do not allow handlers, it simply processes the request using the
 * database layer.
 *
 * A valid database connection has to be available in the Transaction's thread local varaible,
 * when invoking methods of this class. Mehtods of this class neither start nor commit
 * the transactions when performing the operations. The caller of this class should manage database
 * transactions outside this class.
 *
 * There is only one instance of this class per registry instance. This instance can be invoked by
 * many threads. So it has to be thread safe.
 */
public class BasicRegistry implements Registry {

    public static final Log log = LogFactory.getLog(BasicRegistry.class);

    /**
     * Reference to the RegistryContext instance.
     */
    private RegistryContext registryContext;

    /**
     * Repository instance. This is used to handle basic resource storage operations of current
     * versions.
     */
    private Repository repository;

    /**
     * VersionRepository instance. This is used to handle resource storage operations of
     * archived (old) versions.
     */
    private VersionRepository versionRepository;

    /**
     * This is used to execute custom queries.
     */
    private QueryProcessorManager queryProcessorManager;

    /**
     * This is used to interact with the database layer to perform resource related operatons.
     */
    private CommentsDAO commentsDAO;
    private RatingsDAO ratingsDAO;
    private TagsDAO tagsDAO;

    /**
     * This is used to interact with the database layer to perform associations related operations.
     */
    private AssociationDAO associationDAO = new AssociationDAO();

    /**
     * This is used to interact with the database layer to perform activity logging related
     * operations.
     */
    private LogsDAO logsDAO = new LogsDAO();

    /**
     * The URL of the WS-Eventing Service
     */
    private String defaultEventingServiceURL;

    /**
     * Dictionary of URLs of the WS-Eventing Services
     */
    //TODO: Write new comparator for this map
    private Map eventingServiceURLs = new TreeMap<String, String>();

    /**
     * Instantiates the BasicRegistry instance. Objects passed to the constructor should be shared
     * between JDBCRegisty and BasicRegistry instances.
     *
     * @param registryContext RegistryContext instance containing the registry configuration.
     * @param repository Repository instance.
     * @param versionRepository VersionRepository instance.
     * @param queryProcessorManager QueryProcessorManager instance.
     */
    public BasicRegistry(RegistryContext registryContext,
                         Repository repository,
                         VersionRepository versionRepository,
                         QueryProcessorManager queryProcessorManager) {
        this.registryContext = registryContext;
        this.repository = repository;
        this.versionRepository = versionRepository;
        this.queryProcessorManager = queryProcessorManager;

        commentsDAO = StaticConfiguration.getCommentsDAO();
        ratingsDAO = StaticConfiguration.getRatingsDAO();
        tagsDAO = StaticConfiguration.getTagsDAO();
    }

    /**
     * BasicRegistry does not implement transaction related methods as transactions has to be
     * handled outside of the BasicRegistry.
     */
    public void beginTransaction() throws RegistryException {
    }

    /**
     * BasicRegistry does not implement transaction related methods as transactions has to be
     * handled outside of the BasicRegistry.
     */
    public void commitTransaction() throws RegistryException {
    }

    /**
     * BasicRegistry does not implement transaction related methods as transactions has to be
     * handled outside of the BasicRegistry.
     */
    public void rollbackTransaction() throws RegistryException {
    }

    public Resource newResource() throws RegistryException {
        ResourceImpl resource = new ResourceImpl();
        resource.setAuthorUserName(CurrentSession.getUser());
        return resource;
    }

    public Collection newCollection() throws RegistryException {
        CollectionImpl coll = new CollectionImpl();
        coll.setAuthorUserName(CurrentSession.getUser());
        return coll;
    }

    public RegistryContext getRegistryContext() {
        return registryContext;
    }

    public void setRegistryContext(RegistryContext registryContext) {
        this.registryContext = registryContext;
    }

    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Resource get(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        // check if this path refers to a resource referred by a URL query (e.g. comment)

        RequestContext requestContext = new RequestContext(this, repository, versionRepository);
        requestContext.setResourcePath(resourcePath);

        Resource resource = registryContext.getHandlerManager().get(requestContext);

        // resource may have been fetched from the repository, to be used by handlers. if it is
        // done, it has to be stored in the request context. we can just use that resource
        // without fetching it again from the repository.
        if (resource == null) {
            resource = requestContext.getResource();
        }

        if (resource == null) {
            VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
            if (versionedPath.getVersion() == -1) {
                resource = repository.get(resourcePath.getPath());
            } else {
                resource = versionRepository.get(versionedPath);
            }
        }

        if (resource == null) {
            throw new ResourceNotFoundException(path);
        }

        return resource;
    }


    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Resource getMetaData(String path) throws RegistryException {
        ResourcePath resourcePath = new ResourcePath(path);
        Resource resource;
      
        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        if (versionedPath.getVersion() == -1) {
            resource = repository.getMetaData(resourcePath.getPath());
        } else {
            resource = versionRepository.getMetaData(versionedPath);
        }

        if (resource == null) {
            throw new ResourceNotFoundException(path);
        }

        return resource;
    }

    /**
     * Retrives a collection from the registry. This method supports the pagination of the
     * collection. Starting child entry number and number of entries per page can be specified.
     *
     * @param path Path of the collection
     * @param start Starting child entry number
     * @param pageSize Number of entries per page
     * @return Collection instance filled with specified subset of child paths
     * @throws RegistryException
     */
    public Collection get(String path, int start, int pageSize) throws RegistryException {

        Collection collection;

        ResourcePath resourcePath = new ResourcePath(path);
        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        if (versionedPath.getVersion() == -1) {
            collection = repository.get(resourcePath.getPath(), start, pageSize);
        } else {
            collection = versionRepository.get(versionedPath, start, pageSize);
        }

        return collection;
    }

    /**
     * Checks whether a resource exists in the registry.
     *
     * @param path Path of resource
     * @return true if resource exists, false otherwise.
     * @throws RegistryException
     */
    public boolean resourceExists(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        boolean output = registryContext.getHandlerManager().resourceExists(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);

        return versionedPath.getVersion() == -1 &&
                repository.resourceExists(resourcePath.getPath());
    }

    /**
     * Adds or updates resources in the registry. If there is no resource in the given path,
     * resource is added. If a resource already exist in the given path, it will be replaced with
     * the new resource.
     *
     * @param suggestedPath the path which we'd like to use for the new resource.
     * @param resource Resource instance for the new resource
     * @return the actual path that the server chose to use for our Resource
     * @throws RegistryException This exception is thrown for all database layer and handler
     * exceptions.
     */
    public String put(String suggestedPath, Resource resource)
            throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(suggestedPath);

        RequestContext requestContext = new RequestContext(this, repository, versionRepository);
        requestContext.setResourcePath(resourcePath);
        requestContext.setResource(resource);

        if (!RegistryConstants.ROOT_PATH.equals(resourcePath.getPath())) {
            registryContext.getHandlerManager().putChild(requestContext);
        }

        registryContext.getHandlerManager().put(requestContext);

        String actualPath = requestContext.getActualPath();

        if (!requestContext.isProcessingComplete()) {
            ((ResourceImpl) resource).prepareContentForPut();
            
            actualPath = suggestedPath;
            repository.put(resourcePath.getPath(), resource);
        }

        if (actualPath == null) {
            return suggestedPath;
        } else {
            return actualPath;
        }
    }

    /**
     * Creates a new resource by importing the reosurce content from the given source URL.
     *
     * @param suggestedPath Path to put the resource.
     * @param sourceURL URL to import resource content.
     * @param metadata Metadata for the new resource.
     * @return Path where the resource is stored. Note that this may differ from the suggestedPath
     * as some modifications to the suggestedPath will be done if it does not comply with Registry
     * standards.
     * @throws RegistryException
     */
    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(suggestedPath);

        RequestContext importChildContext = new RequestContext(this, repository, versionRepository);
        importChildContext.setResourcePath(resourcePath);
        registryContext.getHandlerManager().importChild(importChildContext);

        RequestContext requestContext = new RequestContext(this, repository, versionRepository);
        requestContext.setResourcePath(resourcePath);
        requestContext.setSourceURL(sourceURL);
        requestContext.setResource(metadata);

        registryContext.getHandlerManager().importResource(requestContext);
        String savedPath = requestContext.getActualPath();

        if (!requestContext.isProcessingComplete()) {
            savedPath = suggestedPath;

            // if some handlers have updated the metadata *without completing the request* we should
            // capture the updated metadata here.
            if (requestContext.getResource() != null) {
                metadata = requestContext.getResource();
            }

            repository.importResource(resourcePath.getPath(), sourceURL, metadata);
        }

        if (savedPath != null) {
            logsDAO.addLog(savedPath, CurrentSession.getUser(), LogEntry.UPDATE, null);
        }

        return suggestedPath;
    }

    /**
     * Deletes the resource in the given path. If the path refers to a directory, all child
     * resources of the directory will also be deleted.
     *
     * @param path Path of the resource to be deleted.
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException
     *          : if the user is not authorized
     */
    public void delete(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);
        RequestContext requestContext = new RequestContext(this, repository, versionRepository);
        requestContext.setRegistryContext(registryContext);
        requestContext.setResourcePath(resourcePath);

        registryContext.getHandlerManager().delete(requestContext);

        if (!requestContext.isProcessingComplete()) {
            repository.delete(resourcePath.getPath());
        }

        logsDAO.addLog(
                resourcePath.getPath(), CurrentSession.getUser(), LogEntry.DELETE_RESOURCE, null);
    }

    public String rename(String currentPath, String newName) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(currentPath);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setRegistryContext(registryContext);
        context.setSourcePath(currentPath);
        context.setTargetPath(newName);
        String newPath = registryContext.getHandlerManager().rename(context);

        if (!context.isProcessingComplete()) {
            newPath  = repository.rename(resourcePath, newName);
        }
        logsDAO.addLog(currentPath, CurrentSession.getUser(), LogEntry.RENAME, newName);

        return newPath;
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        ResourcePath currentResourcePath = new ResourcePath(currentPath);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setSourcePath(currentPath);
        context.setTargetPath(newPath);
        context.setRegistryContext(registryContext);
        String movedPath = registryContext.getHandlerManager().move(context);

        if (!context.isProcessingComplete()) {
            movedPath  = repository.move(currentResourcePath, newPath);
        }
        logsDAO.addLog(currentPath, CurrentSession.getUser(), LogEntry.MOVE, newPath);

        return movedPath;
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {

        ResourcePath sourceResourcePath = new ResourcePath(sourcePath);
        ResourcePath targetResourcePath = new ResourcePath(targetPath);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setSourcePath(sourcePath);
        context.setTargetPath(targetPath);
        String copiedPath = registryContext.getHandlerManager().copy(context);

        if (!context.isProcessingComplete()) {
            copiedPath  = repository.copy(sourceResourcePath, targetResourcePath);
        }
        logsDAO.addLog(sourcePath, CurrentSession.getUser(), LogEntry.COPY, targetPath);

        return copiedPath;
    }

    public void createVersion(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        registryContext.getHandlerManager().createVersion(context);

        if (context.isProcessingComplete()) {
            return;
        }

        versionRepository.createSnapshot(resourcePath, true, true);
    }

    public void createVersion(Resource resource) throws RegistryException {
        versionRepository.createSnapshot(resource, true, true);
    }

    public String[] getVersions(String path) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(new ResourcePath(path));
        String[] output = registryContext.getHandlerManager().getVersions(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        return versionRepository.getVersions(path);
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setVersionPath(versionPath);
        registryContext.getHandlerManager().restoreVersion(context);

        if (context.isProcessingComplete()) {
            return;
        }

        ResourcePath versionedResourcePath = new ResourcePath(versionPath);
        String path = versionedResourcePath.getPath();
        if (repository.resourceExists(path)) {
            // if the target already have resources, delete them..
            repository.prepareResourceRestore(path);
        }
        versionRepository.restoreVersion(versionedResourcePath);
        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.RESTORE, null);
    }

    ////////////////////////////////////////////////////////
    // Associations
    ////////////////////////////////////////////////////////
    public void addAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath sourceResourcePath = new ResourcePath(sourcePath);
        if (repository.resourceExists(sourceResourcePath.getPath())) {
            sourcePath = sourceResourcePath.getPath();
        }

        ResourcePath targetResourcePath = new ResourcePath(targetPath);
        if (repository.resourceExists(targetResourcePath.getPath())) {
            targetPath = targetResourcePath.getPath();
        }

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setSourcePath(sourcePath);
        context.setTargetPath(targetPath);
        context.setAssociationType(associationType);

        registryContext.getHandlerManager().addAssociation(context);

        if (!context.isProcessingComplete()) {
            associationDAO.addAssociation(sourcePath, targetPath, associationType);
        }
    }

    public void removeAssociation(
            String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath sourceResourcePath = new ResourcePath(sourcePath);
        if (repository.resourceExists(sourceResourcePath.getPath())) {
            sourcePath = sourceResourcePath.getPath();
        }

        ResourcePath targetResourcePath = new ResourcePath(targetPath);
        if (repository.resourceExists(targetResourcePath.getPath())) {
            targetPath = targetResourcePath.getPath();
        }


        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setSourcePath(sourcePath);
        context.setTargetPath(targetPath);
        context.setAssociationType(associationType);

        registryContext.getHandlerManager().removeAssociation(context);

        if (!context.isProcessingComplete()) {
            associationDAO.removeAssociation(sourcePath, targetPath, associationType);
        }
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (repository.resourceExists(processedPath.getPath())) {
            resourcePath = processedPath.getPath();
        }

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(new ResourcePath(resourcePath));

        Association[] associations = registryContext.getHandlerManager().getAllAssociations(context);

        if (!context.isProcessingComplete()) {
            return associationDAO.getAllAssociations(resourcePath);
        } else {
            return associations;
        }


    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (repository.resourceExists(processedPath.getPath())) {
            resourcePath = processedPath.getPath();
        }

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(new ResourcePath(resourcePath));
        context.setAssociationType(associationType);

        Association[] associations = registryContext.getHandlerManager().getAssociations(context);

        if (!context.isProcessingComplete()) {
            return associationDAO.getAllAssociationsForType(resourcePath, associationType);
        } else {
            return associations;
        }
    }


    ////////////////////////////////////////////////////////
    // Tagging
    ////////////////////////////////////////////////////////

    /**
     * Applies the given tag to the resource in the given path. If the given tag is not defined in
     * the registry, it will be defined.
     *
     * @param resourcePath Path of the resource to be tagged.
     * @param tag Tag. Any string can be used for the tag.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public void applyTag(String resourcePath, String tag) throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(processedPath);
        context.setTag(tag);

        registryContext.getHandlerManager().applyTag(context);

        if (context.isProcessingComplete()) {
            return;
        }

        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to apply tag to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        // break the comma separated words into multiple tags
        String[] tags = tag.split(",");

        ResourceImpl resource = tagsDAO.getResourceWithMinimumData(resourcePath);
        if (resource == null) {
            String msg = "Failed to apply tag " + tag + " on resource " + resourcePath +
                    ". Resource " + resourcePath + " does not exist.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String userName = CurrentSession.getUser();

        if (!AuthorizationUtils.authorize(resource.getPath(), ActionConstants.GET)) {
            String msg = "Failed to apply tag " + tag + " on resource " + resourcePath +
                    ". User " + userName + " is not authorized to read the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        for (int i = 0; i < tags.length; i++) {

            tags[i] = tags[i].trim();

            if (tags[i].length() == 0 || tags[i].equals(" ")) {
                continue;
            }

            if (tagsDAO.taggingExists(tags[i], resource, userName)) {
                // Already there, don't bother doing it again.
                continue;
            }

            tagsDAO.addTagging(tags[i], resource, userName);
            logsDAO.addLog(resourcePath, userName, LogEntry.TAG, tags[i]);
        }
    }

    /**
     * Returns the paths of the resources, which are tagged using the given tag.
     *
     * @param tag Tag
     * @return Resource paths
     */
    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setTag(tag);

        TaggedResourcePath[] output = registryContext.getHandlerManager().getResourcePathsWithTag(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        // break the tags from spaces
        String[] tags = tag.trim().split(",");

        for (int i = 0; i < tags.length; i++) {
            tags[i] = tags[i].trim();
        }

        List taggedPaths = new ArrayList();

        List pathList = tagsDAO.getPathsWithAnyTag(tags);
        Iterator iPaths = pathList.iterator();
        while (iPaths.hasNext()) {
            String path = (String)iPaths.next();

            TaggedResourcePath taggedResourcePath = new TaggedResourcePath();
            taggedResourcePath.setResourcePath(path);


            ResourceImpl resourceImpl = tagsDAO.getResourceWithMinimumData(path);

            for (String currentTag : tags) {
                long count = tagsDAO.getTagCount(resourceImpl, currentTag);
                taggedResourcePath.addTagCount(currentTag, count);
            }

            taggedPaths.add(taggedResourcePath);
        }

        return (TaggedResourcePath[])taggedPaths
                .toArray(new TaggedResourcePath[taggedPaths.size()]);
    }

    /**
     * Returns all tags used for tagging the given resource.
     *
     * @param path Path of the resource
     * @return Tags
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public Tag[] getTags(String path) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(new ResourcePath(path));

        Tag[] output = registryContext.getHandlerManager().getTags(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        path = RegistryUtils.prepareGeneralPath(path);

        ResourcePath resourcePath = new ResourcePath(path);
        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        ResourceImpl resourceImpl;
        if (versionedPath.getVersion() == -1) {
            resourceImpl = tagsDAO.getResourceWithMinimumData(resourcePath.getPath());
        } else {
            resourceImpl = (ResourceImpl)versionRepository.getMetaData(versionedPath);
        }
        
        if(resourceImpl == null) {
            String msg = "Failed to get tags of " + path +
                    ". The resource doesn't exists.";
            log.debug(msg);
            return new Tag[0];
        }
        return tagsDAO.getTagsWithCount(resourceImpl);
    }

    public void removeTag(String path, String tag) throws RegistryException {

        ResourcePath processedPath = new ResourcePath(path);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(processedPath);
        context.setTag(tag);

        registryContext.getHandlerManager().removeTag(context);

        if (context.isProcessingComplete()) {
            return;
        }

        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to remove tag from the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        path = processedPath.getPath();

        path = RegistryUtils.prepareGeneralPath(path);

        String user = CurrentSession.getUser();

        boolean adminUser = false;

        // check is the user belongs to the admin role
        try {
            UserRealm userRealm = CurrentSession.getRealm();
            String[] roles = userRealm.getUserStoreReader().getUserRoles(user);
            if (RegistryUtils.containsString(RegistryConstants.ADMIN_ROLE, roles)) {
                adminUser = true;
            }

        } catch (UserStoreException e) {

            String msg = "Failed to get roles of the current user. " +
                    "User will be considered as non-admin user.\n" + e.getMessage();
            log.error(msg, e);
            adminUser = false;
        }

        // check if the user is the admin user
        // TODO - do we really need to do this check?  Won't this user always be in the admin role?
        if (RegistryConstants.ADMIN_USER.equals(user)) {
            adminUser = true;
        }


        if (adminUser) {
            ResourceImpl resource = tagsDAO.getResourceWithMinimumData(path);
            tagsDAO.removeTags(resource, tag);
            return;
        }

        ResourceImpl resource = (ResourceImpl)repository.getMetaData(path);

        String author = resource.getAuthorUserName();
        if (user.equals(author)) {
            tagsDAO.removeTags(resource, tag);
        } else {
            tagsDAO.removeTags(resource, tag, user);
        }
    }

    ////////////////////////////////////////////////////////
    // Comments
    ////////////////////////////////////////////////////////

    /**
     * Adds a comment to a resource.
     *
     * @param resourcePath Path of the resource to add the comment.
     * @param comment      Comment instance for the new comment.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public String addComment(String resourcePath, Comment comment)
            throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(processedPath);
        context.setComment(comment);
        String output = registryContext.getHandlerManager().addComment(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to add comment to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        String userName = CurrentSession.getUser();

        try {
            ResourceImpl resource = commentsDAO.getResourceWithMinimumData(resourcePath);
            if (resource == null) {
                String msg = "Failed to add comment " + comment.getText() + " to resource " +
                        resourcePath + ". Resource " + resourcePath + " does not exist.";
                log.error(msg);
                throw new RegistryException(msg);
            }

            int commentId = commentsDAO.addComment(resource, userName, comment);
            String commentPath = resourcePath + RegistryConstants.URL_SEPARATOR + "comments:" +
                    commentId;
            logsDAO.addLog(
                    resourcePath, userName, LogEntry.COMMENT, comment.getText());

            return commentPath;
        }
        catch(RegistryException e) {
            String msg = "Failed to add comment " + comment.getText() + " on resource " +
                    resourcePath + ". User " + userName + ".";
            log.error(msg);
            throw new RegistryException(msg);
        }
    }


    public void editComment(String commentPath, String text) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);

        Comment comment = new Comment();
        comment.setCommentPath(commentPath);
        comment.setText(text);

        context.setComment(comment);
        registryContext.getHandlerManager().editComment(context);

        if (context.isProcessingComplete()) {
            return;
        }

        String[] parts = commentPath.split(RegistryConstants.URL_SEPARATOR);
        String commentPart = parts[1];

        if (parts.length == 2 && commentPart.startsWith("comments:")) {
            String commentId = parts[1].substring(9);
            commentsDAO.updateComment(Long.parseLong(commentId), text);
        }
    }

    /**
     * Returns all comments for the given resource.
     *
     * @param path Path of the resource.
     * @return Array of Comment objects.
     */
    public Comment[] getComments(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        RequestContext context = new RequestContext(this, repository, versionRepository);

        context.setResourcePath(resourcePath);
        Comment[] output = registryContext.getHandlerManager().getComments(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        ResourceImpl resourceImpl;
        if (versionedPath.getVersion() == -1) {
            resourceImpl = commentsDAO.getResourceWithMinimumData(resourcePath.getPath());
        } else {
            resourceImpl = (ResourceImpl)versionRepository.getMetaData(versionedPath);
        }
        
        if (resourceImpl == null) {
            return new Comment[0];
        }
        return commentsDAO.getComments(resourceImpl);
    }

    ////////////////////////////////////////////////////////
    // Ratings
    ////////////////////////////////////////////////////////

    /**
     * Rates the given resource.
     *
     * @param resourcePath Path of the resource.
     * @param rating       rating value.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public void rateResource(String resourcePath, int rating)
            throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(new ResourcePath(resourcePath));
        context.setRating(rating);
        registryContext.getHandlerManager().rateResource(context);

        if (context.isProcessingComplete()) {
            return;
        }

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to apply rating to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        String userName = CurrentSession.getUser();

        ResourceImpl resourceImpl = ratingsDAO.getResourceWithMinimumData(resourcePath);

        if (resourceImpl != null) {

            if (!AuthorizationUtils.authorize(resourcePath, ActionConstants.GET)) {
                String msg = "Failed to rate resource " + resourcePath + " with rating " + rating +
                        ". User " + userName + " is not authorized to read the resource.";
                log.error(msg);
                throw new RegistryException(msg);
            }

            // check for the existing rating.
            int rateID = ratingsDAO.getRateID(resourceImpl, userName);
            if (rateID > -1) {
                ratingsDAO.updateRating(resourceImpl, rateID, rating);

                if(log.isDebugEnabled()) {
                    log.debug("Updated the rating on the resource " +
                        resourcePath + " by user " + userName);
                }
            } else {
                ratingsDAO.addRating(resourceImpl, userName, rating);
            }

            logsDAO.addLog(resourcePath,
                    userName, LogEntry.RATING, Integer.toString(rating));

        } else {
            //TODO:
            //String msg = Messages.getMessage("rate.on.null.artfact", resourcePath);
            String msg = "Rate on Null resource " + resourcePath;
            log.error(msg);
            throw new RegistryException(msg);
        }
    }

    /**
     * Returns the average rating (from all users) for the given resource.
     *
     * @param path Path of the resource.
     * @return average rating.
     */
    public float getAverageRating(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        float rating = registryContext.getHandlerManager().getAverageRating(context);

        if (context.isProcessingComplete()) {
            return rating;
        }

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        ResourceImpl resourceImpl;
        if (versionedPath.getVersion() == -1) {
            resourceImpl = ratingsDAO.getResourceWithMinimumData(resourcePath.getPath());
        } else {
            resourceImpl = (ResourceImpl)versionRepository.getMetaData(versionedPath);
        }

        if (resourceImpl == null) {
            //TODO:
            //String msg = Messages.getMessage("rate.on.null.artfact", resourcePath);
            String msg = "Rate on Null resource " + path;
            log.debug(msg);
            return 0;
        }
        return ratingsDAO.getAverageRating(resourceImpl);
    }

    public int getRating(String path, String userName) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        context.setUserName(userName);
        int rating = registryContext.getHandlerManager().getRating(context);

        if (context.isProcessingComplete()) {
            return rating;
        }

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        ResourceImpl resourceImpl;
        if (versionedPath.getVersion() == -1) {
            resourceImpl = ratingsDAO.getResourceWithMinimumData(resourcePath.getPath());
        } else {
            resourceImpl = (ResourceImpl)versionRepository.getMetaData(versionedPath);
        }
        if (resourceImpl == null) {
            //TODO:
            //String msg = Messages.getMessage("rate.on.null.artfact", path);
            String msg = "Rate on Null resource " + path;
            log.error(msg);
            throw new RegistryException(msg);
        }
        return ratingsDAO.getRating(resourceImpl, userName);
    }

    //////////////////////////////////////////////////////////////////////////////
    // Extensible searching API
    //////////////////////////////////////////////////////////////////////////////

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);

        context.setResourcePath(new ResourcePath(path));
        context.setQueryParameters(parameters);
        Collection output = registryContext.getHandlerManager().executeQuery(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        // we have to get the stored query without checking the user permissions.
        // all query actions are blocked for all users. they are allowed to read the queries, only when executing them.
        String purePath = RegistryUtils.getPureResourcePath(path);
        Resource query = repository.get(purePath);

        return queryProcessorManager.executeQuery(this, query, parameters);
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from, Date to, boolean recentFirst) throws RegistryException {

        List logEntryList  =
                logsDAO.getLogs(resourcePath, action, userName, from, to, recentFirst);

        // We go on two iterations to avoid null values in the following array. Need better way to
        // in a single iteration 
        for (int i = logEntryList.size() - 1 ; i >= 0 ; i--) {
            LogEntry logEntry = (LogEntry)logEntryList.get(i);
            if (logEntry == null){
                logEntryList.remove(i);
            }
        }
        
        LogEntry[] logEntries = new LogEntry[logEntryList.size()];
        for (int i = 0; i < logEntryList.size(); i++) {
            logEntries[i] = (LogEntry)logEntryList.get(i);
        }

        return logEntries;
    }


    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {
        LogEntryCollection logEntryCollection = new LogEntryCollection();
        logEntryCollection.setLogCount(logsDAO.getLogsCount(resourcePath,action,userName,from,to,recentFirst));
        return logEntryCollection;
    }

    public String[] getAvailableAspects() {
        return registryContext.getAspectNames();
    }

    public void associateAspect(String resourcePath, String name) throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to associate aspect to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        //TODO need to do the security validation here
        Resource resource = get(resourcePath);
        if ((resource.getAspects() == null) || (!resource.getAspects().contains(name))) {
            Aspect aspect = getAspect(name);
            if (aspect == null) {
                throw new RegistryException("Couldn't find aspect '" + name + "'");
            }
            aspect.associate(resource, this);
            resource.addAspect(name);
            put(resource.getPath(), resource);
        }
        
    }

    private Aspect getAspect(String name) throws RegistryException {
        return registryContext.getAspect(name);
    }

    //public void addAspect(String name, Aspect aspect) {
    //    aspects.put(name, aspect);
    //}

    public Aspect getResourceAspect(Resource resource, String aspectName)
            throws RegistryException {

        Aspect aspect = getAspect(aspectName);
        if (aspect == null) {
            throw new RegistryException("Aspect '" + aspectName + "' is not registered!");
        }

        // Confirm this aspect is associated with this Resource
        List<String> resourceAspects = resource.getAspects();
        if (resourceAspects == null || !resourceAspects.contains(aspectName)) {
            throw new RegistryException("Resource at '" + resource.getPath() +
                    "' not associated with aspect '" + aspectName + "'");
        }

        return aspect;
    }

    public void invokeAspect(String resourcePath, String aspectName, String action)
            throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to invoke aspect of the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }
        resourcePath = processedPath.getPath();

        Resource resource = get(resourcePath);

        Aspect aspect = getResourceAspect(resource, aspectName);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResource(resource);

//        List aspectNames = resource.getPropertyValues(Aspect.AVAILABLE_ASPECTS);
//        if (aspectNames == null) {
//            throw new RegistryException("No aspect are associated with the resource");
//        }
        context.setAspect(aspect);
        context.setAction(action);
        registryContext.getHandlerManager().invokeAspect(context);

        if (!context.isProcessingComplete()) {
            aspect.invoke(context, action);
        }
        resource.discard();
    }

    public String[] getAspectActions(String resourcePath, String aspectName)
            throws RegistryException {
        
        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to get aspect actions of the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }
        resourcePath = processedPath.getPath();

        Resource resource = get(resourcePath);
        Aspect aspect = getResourceAspect(resource, aspectName);

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResource(resource);
        return aspect.getAvailableActions(context);
    }

    public Collection searchContent (String keywords) throws RegistryException {
        // TODO: to modified

        RequestContext context = new RequestContext(this, repository, versionRepository);

        context.setKeywords(keywords);
        Collection output = registryContext.getHandlerManager().searchContent(context);

        if (context.isProcessingComplete()) {
            return output;
        }

        try{
            Searcher searcher = new IndexSearcher(RegistryContext.getSingleton().getJdbcDir());
            Query query = new QueryParser("content", new StandardAnalyzer()).parse(keywords);
            Hits hits = searcher.search(query);
            org.wso2.carbon.registry.core.Collection collection = new CollectionImpl();
            String [] paths = new String [hits.length()];
            for (int i = 0; i < hits.length(); i++) {
                String id = hits.doc(i).get("id");
                String path = id;
                paths[i] = path;
            }
            collection.setContent(paths);
            return collection;
        } catch (IOException e) {
            String msg = "Failed to search content";
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } catch (ParseException e) {
            String msg = "Failed to parse the query";
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void createLink (String path, String target) throws RegistryException {
        if (path.equals(target)) {
            String msg = "Path and target are same, path = target = " + path +
                    ". You can't create a symlink to itself.";
            log.error(msg);
            throw new RegistryException(msg);
        }
        // first put the data..
        Resource oldResource = repository.getMetaData(target);
        Resource resource;
        if (oldResource != null) {
            if (oldResource instanceof Collection) {
                resource = new CollectionImpl();
            } else {
                resource = new ResourceImpl();
            }
            repository.put(path, resource);
            resource.discard();
        } else {
            repository.put(path, new CollectionImpl());
        }
        
        HandlerManager hm = registryContext.getHandlerManager();

        ResourcePath resourcePath = new ResourcePath(path);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        context.setTargetPath(target);
        hm.createLink(context);

        if (context.isProcessingComplete()) {
            return;
        }

        boolean isRemote = false;
        // check whether the link is a symbolic link
        if (!(resourceExists(target))) {
            isRemote = true;
            List remoteInstances = registryContext.getRemoteInstances();
            boolean instanceFound = false;
            RemoteConfiguration config = null;
            for (int i=0; i < remoteInstances.size(); i++) {
                config = (RemoteConfiguration)remoteInstances.get(i);
                if (config.getId().equals(target)) {
                    instanceFound = true;
                    break;
                }
            }
            if (!instanceFound) {
                String msg = "Target mount point not found.";
                log.error(msg);
                throw new RegistryException(msg);
            }

            MountHandler handler = new MountHandler();
            handler.setUserName(config.getTrustedUser());
            handler.setPassword(config.getTrustedPwd());
            handler.setId(target);
            handler.setConURL(config.getUrl());
            handler.setMountPoint(path);
            hm.addHandler(getMountingMethods(), getMountingMatcher(path), handler);

        } else {

            SymLinkHandler handler = new SymLinkHandler();
            handler.setMountPoint(path);
            handler.setTargetPoint(target);
            hm.addHandler(getMountingMethods(), getMountingMatcher(path), handler);
        }

        //persist mount details
        resource = new ResourceImpl();
        String relativePath = RegistryUtils.getRelativePath(path);
        resource.addProperty("path", relativePath);
        resource.addProperty("target", RegistryUtils.getRelativePath(target));
        repository.put(RegistryUtils.getAbsolutePath("/system/mount") + "/" + relativePath.replace("/","-"), resource);

        logsDAO.addLog(path, CurrentSession.getUser(), isRemote ? LogEntry.CREATE_REMOTE_LINK : LogEntry.CREATE_SYMBOLIC_LINK, target);
    }

    public void createLink(String path, String target,
                           String targetSubPath) throws RegistryException {

        Collection collection = new CollectionImpl();

        repository.put(path, collection);
        collection.discard();

        HandlerManager hm = registryContext.getHandlerManager();

        ResourcePath resourcePath = new ResourcePath(path);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        context.setTargetPath(target);
        context.setTargetSubPath(targetSubPath);
        hm.createLink(context);

        if (context.isProcessingComplete()) {
            return;
        }

        List remoteInstances = registryContext.getRemoteInstances();
        boolean instanceFound = false;
        RemoteConfiguration config = null;
        for (int i = 0; i < remoteInstances.size(); i++) {
            config = (RemoteConfiguration) remoteInstances.get(i);
            if (config.getId().equals(target)) {
                instanceFound = true;
                break;
            }
        }
        if (!instanceFound) {
            String msg = "Target mount point not found.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        MountHandler handler = new MountHandler();
        handler.setUserName(config.getTrustedUser());
        handler.setPassword(config.getTrustedPwd());
        handler.setId(target);
        handler.setConURL(config.getUrl());
        handler.setMountPoint(path);
        handler.setSubPath(targetSubPath);
        hm.addHandler(getMountingMethods(), getMountingMatcher(path), handler);
        targetSubPath = handler.getSubPath();

        /*if (!CurrentSession.getUser().equals(RegistryConstants.SYSTEM_USER))*/ {
            ResourceImpl resource = new ResourceImpl();
            String relativePath = RegistryUtils.getRelativePath(path);
            resource.addProperty("path", relativePath);
            resource.addProperty("target", target);
            resource.addProperty("subPath", targetSubPath);
            repository.put(RegistryUtils.getAbsolutePath("/system/mount") + "/" + relativePath.replace("/","-"), resource);
        }

        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.CREATE_REMOTE_LINK, target);
    }

    public void removeLink(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(resourcePath);
        registryContext.getHandlerManager().removeLink(context);

        if (context.isProcessingComplete()) {
            return;
        }

        Collection mountCollectiion = (Collection)repository.get(
                RegistryUtils.getAbsolutePath(RegistryConstants.SYSTEM_PATH+"/mount"));
        String [] mountResources = (String [])mountCollectiion.getContent();
        Resource resource = null;
        for (int i=0; i<mountResources.length; i++) {
            String mountResName = mountResources[i].substring(mountResources[i].lastIndexOf("/") + 1);
            String relativePath = RegistryUtils.getRelativePath(path);
            if (mountResName.equals(relativePath.replace("/","-"))) {
                resource = repository.get(mountResources[i]);
                break;
            }
        }

        if (resource == null) {
            String msg = "Couldn't find the mount point to remove. ";
            log.error(msg);
            throw new RegistryException(msg);
        }

        repository.delete(resource.getPath());
        if (repository.resourceExists(path)) {
            repository.delete(path);
        }

        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.REMOVE_LINK, null);
    }

    // check in, check out functionality

    public void restore(String path, Reader reader) throws RegistryException {

        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setDumpingReader(reader);
        context.setResourcePath(new ResourcePath(path));
        registryContext.getHandlerManager().restore(context);
        if (context.isProcessingComplete()) {
            return;
        }

        boolean rootResourceExists = repository.resourceExists(path);
        if (rootResourceExists) {
            // if the target already have resources, delete them..
            repository.prepareResourceRestore(path);
        }
        repository.restore(path, reader, rootResourceExists);
        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.RESTORE, null);
    }

    public void dump(String path, Writer writer) throws RegistryException {
        RequestContext context = new RequestContext(this, repository, versionRepository);
        context.setResourcePath(new ResourcePath(path));
        context.setDumpingWriter(writer);
        registryContext.getHandlerManager().dump(context);
        if (context.isProcessingComplete()) {
            return;
        }
        repository.dump(path, writer);
    }

    public String getEventingServiceURL(String path) throws RegistryException {
        if (path == null || eventingServiceURLs.size() == 0) {
            return defaultEventingServiceURL;
        }
        Set<Map.Entry<String, String>> entries = eventingServiceURLs.entrySet();
        for (Map.Entry<String, String> e : entries) {
            if (e.getValue() == null) {
                // Clean-up step
                eventingServiceURLs.remove(e.getKey());
                continue;
            } else if (path.matches(e.getKey())) {
                return e.getValue();
            }
        }
        return defaultEventingServiceURL;
    }

    public void setEventingServiceURL(String path, String eventingServiceURL) throws RegistryException {
        if (path == null) {
            this.defaultEventingServiceURL = eventingServiceURL;
        } else {
            this.eventingServiceURLs.put(path, eventingServiceURL);
        }
    }

    private URLMatcher getMountingMatcher(String path) {
        URLMatcher matcher = new URLMatcher();
        String matchedWith = path +"($|" + RegistryConstants.PATH_SEPARATOR + ".*)";
        matcher.setResourceExistsPattern(matchedWith);
        matcher.setGetPattern(matchedWith);
        matcher.setPutPattern(matchedWith);
        matcher.setDeletePattern(matchedWith);
        matcher.setRenamePattern(matchedWith);
        matcher.setMovePattern(matchedWith);
        matcher.setCopyPattern(matchedWith);
        matcher.setGetAverageRatingPattern(matchedWith);
        matcher.setGetRatingPattern(matchedWith);
        matcher.setRateResourcePattern(matchedWith);
        matcher.setGetCommentsPattern(matchedWith);
        matcher.setAddCommentPattern(matchedWith);
        matcher.setGetTagsPattern(matchedWith);
        matcher.setApplyTagPattern(matchedWith);
        matcher.setGetAllAssociationsPattern(matchedWith);
        matcher.setGetAssociationsPattern(matchedWith);
        matcher.setAddAssociationPattern(matchedWith);
        matcher.setDumpPattern(matchedWith);
        matcher.setRestorePattern(matchedWith);
        matcher.setRemoveAssociationPattern(matchedWith);
        matcher.setImportPattern(matchedWith);
        matcher.setCreateLinkPattern(matchedWith);
        
        return matcher;
    }

    private String [] getMountingMethods() {
        String[] methods = {Filter.RESOURCE_EXISTS, Filter.GET, Filter.PUT, Filter.DELETE, Filter.RENAME,
                Filter.MOVE, Filter.COPY, Filter.GET_AVERAGE_RATING, Filter.GET_RATING,
                Filter.RATE_RESOURCE, Filter.GET_COMMENTS, Filter.ADD_COMMENT, Filter.GET_TAGS,
                Filter.APPLY_TAG, Filter.GET_ALL_ASSOCIATIONS, Filter.GET_ASSOCIATIONS,
                Filter.ADD_ASSOCIATION, Filter.DUMP, Filter.RESTORE,
                Filter.REMOVE_ASSOCIATION, Filter.IMPORT, Filter.CREATE_LINK};

        return methods;
    }
}
