/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.CarbonRegistry;
import org.wso2.carbon.registry.core.ChrootRegistry;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.clustering.NodeGroupLock;
import org.wso2.carbon.registry.core.config.RegistryConfiguration;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.internal.RegistryCoreServiceComponent;
import org.wso2.carbon.registry.core.service.RegistryService;
import org.wso2.carbon.registry.core.session.UserRegistry;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.utils.dbcreator.DatabaseCreator;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * This is a core class used by application that use registry in the embedded mode. This class is
 * used to create embedded registry instances for user sessions.
 *
 * UserRegistry is the embedded mode implementation of the Registry API. In this mode, all registry
 * accesses has to be done using a UserRegistry instance. And there has to be separate UserRegistry
 * instance for each user to access the registry. These UserRegistry instances has be obtained
 * from the EmbeddeedRegistry. It is recommended to have only one EmbeddedRegistry instance per
 * application. But there can be exceptions, where it is required to maintain two or more registries
 * pointing to different datasources.
 *
 * Applications should initialize an EmbeddedRegistry instance at the startup using following code.
 *
 * InputStream configStream = new FileInputStream("/projects/registry.xml");
 * RegistryContext registryContext = new RegistryContext(configStream);
 * EmbeddedRegistry embeddedRegistry = new EmbeddedRegistry(registryContext);
 *
 * After initializing an EmbeddedRegistry instance it should be stored in some globally accessible
 * location, so that it can be used by necessary modules to create UserRegistry instances. From
 * this, it is possible to create UserRegistry instances using varous parameter combinations
 * documeneted in getUserRegistry(...) methods.
 *
 * UserRegistry adminRegistry = embeddedRegistry.getUserRegistry("admin", "admin");
 */
public class EmbeddedRegistry implements RegistryService {

    private static final Log log = LogFactory.getLog(EmbeddedRegistry.class);

    private Registry registry;
    private UserRealm defaultRealm;
    protected RegistryContext registryContext;

    /**
     * Intantiates the EmbeddedRegistry using the configuration given in the context and the given
     * UserRealm. Datasource given in the context will be used for the resource store. User store
     * is accessed from the given UserRealm, which may point to a different source.
     *
     * @param context RegistryContex containing the configuration.
     * @throws RegistryException If error
     */
    public EmbeddedRegistry(RegistryContext context) throws RegistryException {
        this.registryContext = context;
        configure(context.getUserRealm());
    }

    /**
     * This contructor is used by the inherited InMemoryEmbeddedRegistry class as it has to be
     * instantiated using the default contructor.
     */
    protected EmbeddedRegistry() {
    }

    protected void configure(UserRealm defaultRealm) throws RegistryException {

        this.defaultRealm = defaultRealm;

        if(log.isTraceEnabled()) {
            log.trace("Configuring the embedded registry ...");
        }

        DataSource dataSource = registryContext.getDataSource();
        DatabaseCreator databaseCreator = new DatabaseCreator(dataSource);
        
        NodeGroupLock.init(dataSource, null);

        try {
            if(log.isTraceEnabled()) {
                log.trace("Obtaining a cluster wide database lock ...");
            }

            NodeGroupLock.lock(NodeGroupLock.INITIALIZE_LOCK);

            if(log.isTraceEnabled()) {
                log.trace("Cluster wide database lock obtained successfully.");
            }

            if (registryContext.isSetup()) {

                try {
                    // check whether the tables are already created with a query
                    Connection conn = dataSource.getConnection();
                    String sql = "SELECT REG_PATH_ID FROM REG_PATH WHERE REG_PATH_VALUE='/'";
                    Statement stmt = conn.createStatement();
                    stmt.executeQuery(sql);
                }
                catch (SQLException e)
                {
                    // mean the database tables are not needed
                    if (log.isTraceEnabled()) {
                        log.trace("Creating database tables ...");
                    }
                    try {
                        databaseCreator.createRegistryDatabase();
                    } catch (Exception ex) {
                        String msg = "Error in creatig the database";
                        log.error(msg);
                        throw new RegistryException(msg, ex);
                    }

                    if (log.isTraceEnabled()) {
                        log.trace("Database tables created successfully.");
                    }
                }
            } else {
                if (log.isTraceEnabled()) {
                    log.trace("Registry is not initialized in setup mode. " +
                            "Registry database tables will not be created.");
                }
            }

            if(log.isTraceEnabled()) {
                log.trace("Creating the JDBC Registry instance ..");
            }

            JDBCRegistry jdbcRegistry = new JDBCRegistry(registryContext, defaultRealm);
            String chroot = jdbcRegistry.getRegistryContext().getRegistryRoot();
            if (chroot != null) {
                registry = new ChrootRegistry(jdbcRegistry, chroot);
            } else {
                registry = jdbcRegistry;
            }

            RegistryUtils.addSystemCollection(registry, defaultRealm);
            RegistryUtils.addMountCollection(registry, defaultRealm);
            if (registryContext.getProfilesPath() == null) {
                registryContext.setProfilesPath(RegistryConstants.PROFILES_PATH);
            }
            RegistryUtils.addUserProfileCollection(jdbcRegistry, defaultRealm, registryContext.getProfilesPath());
            if (registryContext.getServicepath() == null) {
                registryContext.setServicepath(RegistryConstants.GOVERNANCE_SERVICE_PATH);
            }
            RegistryUtils.addServiceStoreCollection(jdbcRegistry, defaultRealm, registryContext.getServicepath());
            RegistryUtils.addServiceConfigResources(registry, defaultRealm);
            // since we are storing the mount information as relative paths, the following will use the absolute registry
            RegistryUtils.registerMountPoints(registry, defaultRealm);


            if(log.isTraceEnabled()) {
                log.trace("JDBC Registry instance created successfully.");
            }

        } finally {
            if(log.isTraceEnabled()) {
                log.trace("Releasing a cluster wide database lock ...");
            }

            NodeGroupLock.unlock(NodeGroupLock.INITIALIZE_LOCK);
            if(log.isTraceEnabled()) {
                log.trace("Cluster wide database lock released successfully.");
            }
        }
    }

    /**
     * Creates a UserRegistry instance for anonymous user. Permissions set for anonymous user will
     * be applied for all operations performed using this instance.
     *
     * @return UserRegistry for the anonymous user.
     * @throws RegistryException
     */
    public UserRegistry getUserRegistry() throws RegistryException {
        return new CarbonRegistry(
                RegistryConstants.ANONYMOUS_USER,
                RegistryConstants.ANONYMOUS_PASSWORD,
                registry,
                defaultRealm,
                registryContext.getDataSource());
    }

    /**
     * Returns a registry to be used for system operations. Human users should not be allowed
     * log in using this registry.
     *
     * @return User registry for system user.
     * @throws RegistryException
     */
    public UserRegistry getSystemRegistry() throws RegistryException {
        
        RegistryConfiguration regConfig = RegistryCoreServiceComponent.getRegistryConfig();
        String username = regConfig.getValue(RegistryConfiguration.SYSTEM_USER_NAME);

        return new CarbonRegistry(username,
                                  registry,
                                  defaultRealm,
                                  registryContext.getDataSource());
    }

    /**
     * Creates UserRegistry instances for normal users. Applications should use this method to
     * create UserRegistry instances, unless there is a specific need documented in other methods.
     * User name and the password will be authenticated by the EmbeddedRegistry before creating
     * the requested UserRegistry instance.
     *
     * @param userName User name of the user.
     * @param password Password of the user.
     * @return UserRegistry instance for the given user.
     * @throws RegistryException
     */
    public UserRegistry getUserRegistry(String userName, String password) throws RegistryException {
        return new CarbonRegistry(userName, password, registry, defaultRealm,
                registryContext.getDataSource());
    }

    /**
     * Creates a UserRegistry instance for the given user. This method will NOT authenticate the
     * user before creating the UserRegistry instance. It assumes that the user is authenticated
     * outside the EmbeddedRegistry.
     *
     * @param userName User name of the user.
     * @return UserRegistry instance for the given user.
     * @throws RegistryException
     */
    public UserRegistry getUserRegistry(String userName) throws RegistryException {
        return new CarbonRegistry(userName, registry, defaultRealm,
                registryContext.getDataSource());
    }
}
