/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.*;
import org.apache.axiom.om.util.Base64;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.exceptions.ResourceNotFoundException;
import org.wso2.carbon.registry.core.jdbc.dao.*;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.CommentDO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.TaggingDO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.RatingDO;
import org.wso2.carbon.registry.core.secure.AuthorizationFailedException;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.config.StaticConfiguration;

import javax.sql.DataSource;
import javax.xml.namespace.QName;
import javax.xml.stream.*;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.*;
import java.sql.Timestamp;

/**
 * Encapsulates the retrieving, storing, modifing and deleting of resources. This class only deals
 * with the current versions of resources and it is unware of any versioning or snapshot activity.
 * Only the current version related tables are accessed and updated from the methods of this class.
 */
public class Repository {

    public static final Log log = LogFactory.getLog(Repository.class);

    private ResourceDAO resourceDAO = new ResourceDAO();
    private ResourceVersionDAO resourceVersionDAO = new ResourceVersionDAO();
    private AssociationDAO associationDAO = new AssociationDAO();
    private LogsDAO logsDAO = new LogsDAO();
    private CommentsDAO commentsDAO;
    private RatingsDAO ratingsDAO;
    private TagsDAO tagsDAO;

    /**
     * Determines whether to version resources automatically when a resource is modified. Versionable modifications
     * are adding new resources, changing content and changing properties.
     */
    private boolean versionOnChange = false;

    VersionRepository versionRepository;

    /**
     * Datasource is set on all resources get from the repository.
     */
    private DataSource dataSource;

    public Repository(DataSource dataSource,
                      VersionRepository versionRepository, boolean versionOnChange) {
        this.dataSource = dataSource;
        this.versionRepository = versionRepository;
        this.versionOnChange = versionOnChange;
        commentsDAO = StaticConfiguration.getCommentsDAO();
        ratingsDAO = StaticConfiguration.getRatingsDAO();
        tagsDAO = StaticConfiguration.getTagsDAO();
    }

    /**
     * Checks if a pure resource exists in the given path.
     *
     * @param path Path of a possible pure resource.
     * @return ture if a resource exists in the given path. false otherwise.
     */
    public boolean resourceExists(String path) throws RegistryException {

        String purePath = RegistryUtils.getPureResourcePath(path);

        return resourceDAO.resourceExists(purePath);
    }

    /**
     * Gets the meta data of resource referred by the given path.
     *
     * @param path Path of a "pure" resource. Path should map to an actual resource stored
     *             in the database. Paths refering to virtual resource are not handled (e.g. /c1/r1;comments).
     * @return Resource refered by the given path. Resource can be a file or a collection.
     */
    public Resource getMetaData(String path) throws RegistryException {

        String purePath = RegistryUtils.getPureResourcePath(path);

        Resource resource = resourceDAO.getResourceMetaData(purePath);
        if(resource == null) {
            return null;
        }

        if (!AuthorizationUtils.authorize(purePath, ActionConstants.GET)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "read the resource " + purePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        ((ResourceImpl) resource).setDataSource(dataSource);

        return resource;
    }

    /**
     * Gets the pure resource referred by the given path.
     *
     * @param path Path of a "pure" resource. Path should map to an actual resource stored
     *             in the database. Paths refering to virtual resource are not handled (e.g. /c1/r1;comments).
     * @return Resource refered by the given path. Resource can be a file or a collection.
     */
    public Resource get(String path) throws RegistryException {

        String purePath = RegistryUtils.getPureResourcePath(path);

        Resource resource = resourceDAO.getResourceMetaData(purePath);
        if(resource == null) {
            return null;
        }

        if (!AuthorizationUtils.authorize(purePath, ActionConstants.GET)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "read the resource " + purePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        resourceDAO.fillResource((ResourceImpl)resource);
        ((ResourceImpl) resource).setDataSource(dataSource);

        return resource;
    }

    public Collection get(String path, int start, int pageLen) throws RegistryException {
        String purePath = RegistryUtils.getPureResourcePath(path);

        CollectionImpl resource = (CollectionImpl)resourceDAO.getResourceMetaData(purePath);
        if(resource == null) {
            return null;
        }

        if (!AuthorizationUtils.authorize(purePath, ActionConstants.GET)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "read the resource " + purePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        resourceDAO.fillResource(resource, start, pageLen);
        resource.setDataSource(dataSource);

        return resource;
    }

    /**
     * Adds or updates the resource in the given path with the given resource. Put is executed if
     * the current user has authorization to do so. Below is the method of evaluating
     * authorizations.
     * <p/>
     * <ul>
     * <li> user should have WRITE permissions for the resource, if he wants to update the
     * resource.</li>
     * <li>user should habe WRITE permissions for the parent of the new resource, if he wants to
     * add a new resource.</li>
     * </ul>
     *
     * @param path     Path of the resource to be added or updated. Path should only refer to current
     *                 version a pure resource. i.e. Path should not contain any version information as it is
     *                 impossible to add or update old versions.
     * @param resource Resource to be added or updated.
     */
    public void put(String path, Resource resource) throws RegistryException {

        // Refer org.wso2.carbon.registry.core.test.jdbc.JDBCRegistryTest
        if (path.matches(".*[~!@#$;%^*()+={}\\[\\]\\|\\\\<>].*")) {  //"[~!@#$;%^*()+={}[]|\\<>]"
            throw new RegistryException("The path '" + path + "' contains one or more illegal "
                    + "characters (~!@#$;%^*()+={}[]|\\<>)");
        }

        String purePath = RegistryUtils.getPureResourcePath(path);

        ResourceIDImpl resourceID = resourceDAO.getResourceID(purePath, resource instanceof CollectionImpl);
        boolean resourceExists = false;
        if (resourceID != null) {

            // existance of the resource id doesn't mean the existance of the
            // resource, so need to verify resource existance

            // load the meta data for the existing resource
            ResourceDO oldResourceDO = resourceDAO.getResourceDO(resourceID);

            if (oldResourceDO != null) {
                // the resource does exists
                resourceExists = true;

                // copying the id attribute for the resource
                ((ResourceImpl) resource).setPathID(resourceID.getPathID());
                ((ResourceImpl) resource).setName(resourceID.getName());
                ((ResourceImpl) resource).setPath(resourceID.getPath());
                // get the original details
                ((ResourceImpl) resource).setCreatedTime(oldResourceDO.getCreatedOn());
                ((ResourceImpl) resource).setAuthorUserName(oldResourceDO.getAuthor());

                        
                // we are always creating versions for resources (files), if the resource has changed.
                if (!(resource instanceof Collection) && this.versionOnChange &&
                        ((ResourceImpl) resource).isVersionableChange()) {
                    ResourceImpl oldResourceImpl = new ResourceImpl(resourceID.getPath(), oldResourceDO);
                    versionRepository.createSnapshot(oldResourceImpl, false, false);
                }
                else {
                    // just delete the resource and content
                    // delete the old entry from the resource table
                    removeResource((ResourceImpl)resource, false);
                }
                update(resourceID, (ResourceImpl)resource, oldResourceDO);
            }
        }
        if (!resourceExists){

            // now we are checking whether there is an entry with the inverse type
            ResourceIDImpl inverseResourceID = resourceDAO.getResourceID(purePath,
                        !(resource instanceof CollectionImpl));
            if (inverseResourceID != null) {
                ResourceDO inverseResourceDO = resourceDAO.getResourceDO(inverseResourceID);
                if (inverseResourceDO != null) {
                    // well, infact there is an inverse type=> we are deleting the resource
                    deleteSubtree(inverseResourceID, inverseResourceDO, false);
                }
            }

            // resource does not exists. add the resource.
            add(purePath, (ResourceImpl)resource);

        }
    }

    /**
     * Creates a resource with the content imported from the source URL and metadata extracted from
     * the given metadata resource instance. Then the created resource is put to the registry using
     * the Repository.put() method.
     *
     * @param path      Path to put the resource
     * @param sourceURL URL to import resource content
     * @param metadata  Metadata for the new resource is extracted from this metadata resource
     * @return Actual path where the new resource is stored in the registry
     * @throws RegistryException for all exceptional scenarios
     */
    public String importResource(String path, String sourceURL, Resource metadata)
            throws RegistryException {

        String purePath = RegistryUtils.getPureResourcePath(path);

        URL url;
        try {
            if (sourceURL != null && sourceURL.toLowerCase().startsWith("file:")) {
                String msg = "The source URL must not be file in the server's local file system";
                throw new RegistryException(msg);
            }
            url = new URL(sourceURL);
        } catch (MalformedURLException e) {
            String msg = "Given source URL is not valid.";
            throw new RegistryException(msg, e);
        }

        try {
            URLConnection uc = url.openConnection();
            InputStream in = uc.getInputStream();
            String mediaType = metadata.getMediaType();
            if (mediaType == null)
                mediaType = uc.getContentType();
            metadata.setMediaType(mediaType);
            metadata.setDescription(metadata.getDescription());
            metadata.setContentStream(in);
            put(purePath, metadata);

        } catch (IOException e) {

            String msg = "Could not read from the given URL: " + sourceURL;
            throw new RegistryException(msg, e);
        }

        return purePath;
    }

    /**
     * Deletes the pure resource refered by the path.
     *
     * @param path path Path of the resource deleted. Path should only refer to current
     *             version a pure resource. i.e. Path should not contain any version information as it is
     *             impossible to delete old versions.
     */
    public void delete(String path) throws RegistryException {

        path = RegistryUtils.getPureResourcePath(path);

        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        ResourceDO resourceDO = resourceDAO.getResourceDO(resourceID);
        if (resourceDO == null) {
            boolean isCollection = resourceID.isCollection();
            // then we will check for non-collections as the getResourceID only check the collection exist
            if (isCollection) {
                resourceID = resourceDAO.getResourceID(path, false);
                if (resourceID != null) {
                    resourceDO = resourceDAO.getResourceDO(resourceID);
                }
            }
            if (resourceDO == null) {
                String msg = "Failed to delete resource " + path + ". Resource does not exist.";
                log.error(msg);
                throw new RegistryException(msg);
            }
        }

        if (!AuthorizationUtils.authorize(path, ActionConstants.DELETE)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "delete the resource " + path + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        deleteSubtree(resourceID, resourceDO, false);
    }

    /**
     * This will delete the entire resource, except it keeps the authorization in the root.
     * @param path
     * @throws RegistryException
     */
    public void prepareResourceRestore(String path) throws RegistryException {

        path = RegistryUtils.getPureResourcePath(path);

        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        ResourceDO resourceDO = resourceDAO.getResourceDO(resourceID);
        if (resourceDO == null) {
            boolean isCollection = resourceID.isCollection();
            // then we will check for non-collections as the getResourceID only check the collection exist
            if (isCollection) {
                resourceID = resourceDAO.getResourceID(path, false);
                if (resourceID != null) {
                    resourceDO = resourceDAO.getResourceDO(resourceID);
                }
            }
            if (resourceDO == null) {
                return;
            }
        }

        deleteSubtree(resourceID, resourceDO, true);
    }

    private void deleteSubtree(ResourceIDImpl resourceID,
                               ResourceDO resourceDO,
                               boolean keepRootAuthorization) throws RegistryException {

        if(resourceID.isCollection()) {
            // recursively call for all the resources in the tree..
            ArrayList<ResourceIDImpl> childIDs = resourceDAO.getChildPathIds(resourceID);

            Iterator<ResourceIDImpl> iChildIDs = childIDs.iterator();
            while (iChildIDs.hasNext()) {
                ResourceIDImpl childID = iChildIDs.next();
                ResourceDO childResourceDO = resourceDAO.getResourceDO(childID);
                deleteSubtree(childID, childResourceDO, keepRootAuthorization);
            }
        }

        ResourceImpl resourceImpl;
        if (resourceID.isCollection()) {
            resourceImpl = new CollectionImpl(resourceID.getPath(), resourceDO);
        }
        else {
            resourceImpl = new ResourceImpl(resourceID.getPath(), resourceDO);
        }

        // now do the versioning as delete is considered as a change, where
        // non-collections are versioned automatically unless it is configured otherwise
        if (!(resourceImpl instanceof CollectionImpl) && this.versionOnChange &&
                resourceImpl.isVersionableChange()) {
           
            // we are creating snapshot without renewing old contents..
            versionRepository.createSnapshot(resourceImpl, false, true);

        }
        else {
            // just delete the resource and content
            // delete the old entry from the resource table
            boolean isResourcePathVersioned = resourceVersionDAO.isResourceHistoryExist(
                    resourceImpl.getResourceIDImpl());
            removeResource(resourceImpl, isResourcePathVersioned);

            if (!StaticConfiguration.isVersioningComments() && !isResourcePathVersioned) {
                commentsDAO.removeComments(resourceImpl);
            }
            if (!StaticConfiguration.isVersioningTags() && !isResourcePathVersioned) {
                tagsDAO.removeTags(resourceImpl);
            }
            if (!StaticConfiguration.isVersioningRatings() && !isResourcePathVersioned) {
                ratingsDAO.removeRatings(resourceImpl);
            }
            if (!StaticConfiguration.isVersioningAssociations() && !isResourcePathVersioned) {
                associationDAO.removeAllAssociations(resourceImpl.getPath());
            }


            if (!keepRootAuthorization && !isResourcePathVersioned) {
                // for an empty root, we are not clearing the autherization..
                AuthorizationUtils.clearAuthorizations(resourceID.getPath());
            }
        }
    }

    private void removeResource(ResourceImpl resourceImpl,
                                boolean keepProperties) throws RegistryException {
        resourceDAO.deleteResource(resourceImpl.getResourceDO());
        if (!(resourceImpl instanceof CollectionImpl)) {
            int contentID = resourceImpl.getDbBasedContentID();
            if (contentID > 0) {
                // delete the old content stream from the latest table
                resourceDAO.deleteContentStream(contentID);
            }
        }
        if (!StaticConfiguration.isVersioningProperties() && !keepProperties) {
            resourceDAO.removeProperties(resourceImpl.getResourceDO());
        }
    }

    /**
     * Renames the resource at oldPath with the given newName. If oldPath is
     * "/projects/esb/config.xml" and newName is "synapse.xml" then the new path of the resource
     * will be "/projects/esb/synapse.xml". If the renamed resource is a collection, paths of
     * all its desendants will also be changed.
     *
     * @param oldResourcePath Path of the resource to be renamed.
     * @param newName New name for the resource. Name should not contain "/".
     * @return New path of the resource.
     * @throws RegistryException
     */
    public String rename(ResourcePath oldResourcePath, String newName) throws RegistryException {

        if (!oldResourcePath.isCurrentVersion()) {
            String msg = "Failed to rename the resource " + oldResourcePath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String oldPath = RegistryUtils.getPureResourcePath(oldResourcePath.getPath());

        if (!newName.startsWith(RegistryConstants.PATH_SEPARATOR)) {
            // relative URL
            String oldDir =
                    oldPath.substring(0, oldPath.lastIndexOf(RegistryConstants.PATH_SEPARATOR) + 1);
            newName = oldDir + newName;
        }

        if (!AuthorizationUtils.authorize(oldPath, ActionConstants.PUT)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "rename the resource " + oldResourcePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        move(oldResourcePath, newName);

        return newName;
    }

    /**
     * Moves the resource at oldPath to the newPath. If the resource is a collection, all its
     * descendant resources will also be moved.
     *
     * @param oldResourcePath Path of a existing resource
     * @param newPath New path of the resource
     * @return the actual new path
     * @throws RegistryException
     */
    public String move(ResourcePath oldResourcePath, String newPath) throws RegistryException {

        if (!oldResourcePath.isCurrentVersion()) {
            String msg = "Failed to copy the resource " + oldResourcePath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        // we are implementing move as copying the resource to the new path and deleting the old
        // resource. therefore, the moved resource will be threated as a new resource (i.e without
        // a version hostory. but the resource at old path (which is deleted) has its version
        // hostory. so it is possible to restore the parent collection of old path and get old
        // versions of that resource.

        String oldPath = oldResourcePath.getPath();
        if (newPath.equals(oldPath)) {
            return newPath;
        }


        // prepare the target parent path
        String targetParentPath = RegistryUtils.getParentPath(newPath);
        // first create a parent path for the target if doesnt exist
        ResourceIDImpl targetParentResourceID = resourceDAO.getResourceID(targetParentPath, true);
        if (targetParentResourceID == null || !resourceDAO.resourceExists(targetParentResourceID)) {
            addEmptyCollection(targetParentPath);
            if(targetParentResourceID == null) {
                targetParentResourceID = resourceDAO.getResourceID(targetParentPath, true);
            }
        }
        else if (!AuthorizationUtils.authorize(targetParentPath, ActionConstants.PUT)) {
            String msg = "Resource Move failed. User " +
                    CurrentSession.getUser() + " is not authorized to update " +
                    "the parent collection of target " + targetParentPath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        // get the source resource
        ResourceImpl sourceResource = (ResourceImpl)getMetaData(oldPath);
        if (sourceResource ==  null) {
            throw new ResourceNotFoundException(oldPath);
        }
        ResourceIDImpl sourceID = sourceResource.getResourceIDImpl();

        if (!(sourceResource instanceof CollectionImpl)) {
            prepareMove(oldPath, newPath);
            // get the source resource
            // just rename the resource + all the community features.
            ResourceIDImpl targetID = resourceDAO.getResourceID(newPath, false);
            if (targetID == null) {
                // create resourceID
                targetID = resourceDAO.createResourceID(newPath, targetParentResourceID, false);
            }

            //start moving resource + all the community features
            resourceDAO.moveResources(sourceID, targetID);
            resourceDAO.moveProperties(sourceID, targetID);
            tagsDAO.moveTags(sourceID, targetID);
            ratingsDAO.moveRatings(sourceID, targetID);
            commentsDAO.moveComments(sourceID, targetID);

            if (!StaticConfiguration.isVersioningAssociations()) {
                // replace the right hand side associations
                associationDAO.replaceAssociations(oldPath, newPath);
                // replace the left hand side associations
                associationDAO.copyAssociations(oldPath, newPath);
                associationDAO.removeAllAssociations(oldPath);
            }
            return newPath;
        }
        // handles all the colleciton moves from here
        moveRecursively(sourceID, newPath, targetParentResourceID);

        return newPath;
    }

    private String moveRecursively(ResourceIDImpl sourceID,
                                   String targetPath,
                                   ResourceIDImpl targetParentResourceID) throws RegistryException {
        prepareMove(sourceID.getPath(), targetPath);

        ArrayList<ResourceIDImpl> sourceChildIDs = resourceDAO.getChildPathIds(sourceID);
        for (ResourceIDImpl sourceChildID: sourceChildIDs) {
            if (!sourceChildID.isCollection()) {
                // prepare all the non-collections for move
                String sourceChildPath = sourceChildID.getPath();
                String sourceChildResourceName = RegistryUtils.getResourceName(sourceChildPath);
                // above will will be used to form the targetChildPath
                String targetChildPath = targetPath + (targetPath.endsWith(RegistryConstants.PATH_SEPARATOR)?"":
                                            RegistryConstants.PATH_SEPARATOR) + sourceChildResourceName;
                // if it is a resource, we only need to prepare the move as all
                // the non-collection moves happens in below operation
                prepareMove(sourceChildPath, targetChildPath);

                // just we have to copy associations manually
                if (!StaticConfiguration.isVersioningAssociations()) {
                    // replace the right hand side associations
                    associationDAO.replaceAssociations(sourceChildPath, targetChildPath);
                    // replace the left hand side associations
                    associationDAO.copyAssociations(sourceChildPath, targetChildPath);
                    associationDAO.removeAllAssociations(sourceChildPath);
                }
            }
        }

        ResourceIDImpl targetID = resourceDAO.getResourceID(targetPath, sourceID.isCollection());
        if (targetID == null) {
            targetID = resourceDAO.createResourceID(targetPath, targetParentResourceID, sourceID.isCollection());
        }

        resourceDAO.moveResourcePaths(sourceID, targetID);
        resourceDAO.movePropertyPaths(sourceID, targetID);
        tagsDAO.moveTagPaths(sourceID, targetID);
        ratingsDAO.moveRatingPaths(sourceID, targetID);
        commentsDAO.moveCommentPaths(sourceID, targetID);

        if (!StaticConfiguration.isVersioningAssociations()) {
            // replace the right hand side associations
            associationDAO.replaceAssociations(sourceID.getPath(), targetPath);
            // replace the left hand side associations
            associationDAO.copyAssociations(sourceID.getPath(), targetPath);
            associationDAO.removeAllAssociations(sourceID.getPath());
        }

        // recursively call this for all the child collections

        for (ResourceIDImpl sourceChildID: sourceChildIDs) {
            if (sourceChildID.isCollection()) {
                String sourceChildPath = sourceChildID.getPath();
                String sourceChildResourceName = RegistryUtils.getResourceName(sourceChildPath);
                // above will will be used to form the targetChildPath
                String targetChildPath = targetPath + (targetPath.endsWith(RegistryConstants.PATH_SEPARATOR)?"":
                                            RegistryConstants.PATH_SEPARATOR) + sourceChildResourceName;
                // if it is a collection we are calling this recursively
                moveRecursively(sourceChildID, targetChildPath, targetID);
            }
        }
        return targetPath;
    }

    /**
     * this will prepare the move by checking the authorizations etc..
     * @param sourcePath
     * @param targetPath
     * @return
     * @throws RegistryException
     */
    private void prepareMove(String sourcePath, String targetPath) throws RegistryException {
        // check the authorization
        // read and delete authorization needed for the source (old path)
        if (!AuthorizationUtils.authorize(sourcePath, ActionConstants.GET)) {
            String msg = "Resource Move failed. User " + CurrentSession.getUser() +
                    " is not authorized to read the resource at " + sourcePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        if (!AuthorizationUtils.authorize(sourcePath, ActionConstants.DELETE)) {
            String msg = "Resource Move failed. User " + CurrentSession.getUser() +
                    " is not authorized to delete the resource at " + sourcePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        // check for existance of target path, if target exist delete them
        ResourceIDImpl targetExistingResourceID = resourceDAO.getResourceID(targetPath);
        ResourceDO targetExistingResourceDO = null;
        if (targetExistingResourceID != null) {
            targetExistingResourceDO = resourceDAO.getResourceDO(targetExistingResourceID);
            if (targetExistingResourceDO == null && targetExistingResourceID.isCollection()) {
                // we have to check the possiblity non collection having this path
                targetExistingResourceID = resourceDAO.getResourceID(targetPath, false);
                if (targetExistingResourceID != null) {
                    targetExistingResourceDO = resourceDAO.getResourceDO(targetExistingResourceID);
                }
            }

            if (targetExistingResourceDO != null) {
                // there has been a resource in this place..
                if (!AuthorizationUtils.authorize(targetPath, ActionConstants.PUT)) {
                    String msg = "Resource Move failed. User " +
                            CurrentSession.getUser() + " is not authorized to update " +
                            "the target path " + targetPath + ".";
                    log.error(msg);
                    throw new AuthorizationFailedException(msg);
                }

                // remove the resource, since we are overwriting that..
                removeIndividual(targetExistingResourceID, targetExistingResourceDO);
            }
        }

        // and clear authorization for source if it no resource history exist
        if (!resourceVersionDAO.isResourceHistoryExist(sourcePath)) {
            AuthorizationUtils.clearAuthorizations(sourcePath);
        }
    }

    public void removeIndividual(ResourceIDImpl resourceID, ResourceDO resourceDO) throws RegistryException{
        ResourceImpl resource;
        if (resourceID.isCollection()) {
            resource = new CollectionImpl(resourceID.getPath(), resourceDO);
        }
        else {
            resource = new ResourceImpl(resourceID.getPath(), resourceDO);
        }
        // we are always creating versions for resources (files), if the resource has changed.
        if (!(resource instanceof Collection) && this.versionOnChange &&
                resource.isVersionableChange()) {
            versionRepository.createSnapshot(resource, false, false);
        }
        else {
            // just delete the resource and content
            // delete the old entry from the resource table
            removeResource(resource, false);
        }
    }

    public String copy(ResourcePath sourceResourcePath,
                       ResourcePath targetResourcePath)
            throws RegistryException {


        if (!sourceResourcePath.isCurrentVersion() || !targetResourcePath.isCurrentVersion()) {
            String msg = "Failed to copy the resource " + sourceResourcePath + " to path " +
                    targetResourcePath + ". Both paths should refer to current versions.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String sourcePath = sourceResourcePath.getPath();
        String targetPath = targetResourcePath.getPath();
        
        sourcePath = RegistryUtils.getPureResourcePath(sourcePath);
        targetPath = RegistryUtils.getPureResourcePath(targetPath);

        if(sourcePath.equals(targetPath)) {
            // nothing special to do..
            return targetPath;
        }
        else if(sourcePath.indexOf(targetPath) == 0) {
            // if the target path is a part of the source path, we can't continue
            String msg = "Failed to copy the resource " + sourceResourcePath + " to path " +
                    targetResourcePath + ". The target path is a part of the source path.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        ResourceImpl sourceResource = (ResourceImpl) get(sourcePath);
        if (sourceResource instanceof CollectionImpl) {
            resourceDAO.fillChildren((CollectionImpl)sourceResource, 0, -1);
        }

        if(resourceDAO.resourceExists(targetPath)) {
            // We need to delete the old one if it is already exists..  
            delete(targetPath);
        }

        ResourceImpl targetResource = sourceResource.getShallowCopy();
        put(targetPath, targetResource);


        if (sourceResource instanceof CollectionImpl) {
            // copy all the children recursively
            CollectionImpl collection = (CollectionImpl)sourceResource;
            String[] childSourcePaths = (String[])collection.getContent();

            for(int i = 0; i < childSourcePaths.length; i ++) {
                String childSourcePath = childSourcePaths[i];
                String childResourceName = RegistryUtils.getResourceName(childSourcePath);
                String childTargetPath = targetPath + RegistryConstants.PATH_SEPARATOR + childResourceName;


                ResourcePath childSourceResourcePath = new ResourcePath(childSourcePath);
                ResourcePath childTargetResourcePath = new ResourcePath(childTargetPath);

                copy(childSourceResourcePath, childTargetResourcePath);
            }
        }

        commentsDAO.copyComments(sourceResource, targetResource);
        tagsDAO.copyTags(sourceResource, targetResource);
        ratingsDAO.copyRatings(sourceResource, targetResource);
        associationDAO.copyAssociations(sourceResource.getPath() ,targetResource.getPath());
        return targetPath;
    }


    private void add(String path, ResourceImpl resource) throws RegistryException {

        // first add all non-existing parent collections. note that whether the user has
        // permission to add this resource depends on the permissions of the nearest ascendant.

        String parentPath = RegistryUtils.getParentPath(path);

        // first, let's check if there is a parent for this resource. we do this to speed up the
        // common use case where new resources are added to an existing parent.

        ResourceIDImpl parentResourceID = resourceDAO.getResourceID(parentPath, true);
        if (parentResourceID == null || !resourceDAO.resourceExists(parentResourceID)) {
            addEmptyCollection(parentPath);
            if(parentResourceID == null) {
                parentResourceID = resourceDAO.getResourceID(parentPath, true);
            }
        }
        else if (!AuthorizationUtils.authorize(parentPath, ActionConstants.PUT)) {
            // in order to add a new resource we are checkinig the permission of the parent path +
            // current path.
            String msg = "Failed to add new resource " + path + ". User " +
                    CurrentSession.getUser() + " is not authorized to update " +
                    "the parent collection " + parentPath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        if (!AuthorizationUtils.authorize(path, ActionConstants.PUT)) {
            String msg = "Failed to add new resource " + path + ". User " +
                    CurrentSession.getUser() + " is not authorized to update " +
                    "the resource path " + path + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        // then add the new resource/collection
        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.ADD, null);
        resourceDAO.add(path, parentResourceID, resource);
    }

    /**
     * updating a resource
     * @param resourceID
     * @param resource
     * @return true, if the resource already exist, false otherwise
     * @throws RegistryException
     */
    private void update(ResourceIDImpl resourceID, ResourceImpl resource,
                           ResourceDO oldResourceDO) throws RegistryException {

        if (!AuthorizationUtils.authorize(resourceID.getPath(), ActionConstants.PUT)) {
            String msg = "Resource update failed. User " + CurrentSession.getUser() +
                    " is not authorized to update the resource " + resourceID.getPath() + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        // we ignore the resource.isContentModified()), isPropertiesModifed
        //  as we have to store the version correctly

        logsDAO.addLog(resourceID.getPath(), CurrentSession.getUser(), LogEntry.UPDATE, null);
        resourceDAO.update(resource);

        ResourceImpl oldResourceImpl;
        if (oldResourceDO.getName() == null) {
            // this is a collection
            oldResourceImpl = new CollectionImpl(resourceID.getPath(), oldResourceDO);
        }
        else {
            oldResourceImpl = new ResourceImpl(resourceID.getPath(), oldResourceDO);
        }
        // copying old required attributes to the new versioni
        commentsDAO.copyComments(oldResourceImpl, resource);
        tagsDAO.copyTags(oldResourceImpl, resource);
        ratingsDAO.copyRatings(oldResourceImpl, resource);
    }

    /**
     * This will create an empty collection, if there is no such resource exist
     * if the parent of the empty collection doesn't exist, this will create they as well.
     *
     * @param path Path of which all non-existent collections are added.
     * @throws RegistryException If any ancestor of the given path is a resource.
     */
    private void addEmptyCollection(String path) throws RegistryException {
        // first need to check whether there is a resource (non-collection)
        // where it is asking to create a collection
        ResourceIDImpl assumedResourceID = resourceDAO.getResourceID(path, false);
        if (assumedResourceID != null && resourceDAO.resourceExists(assumedResourceID)) {
            String msg = "Failed to add new Collection " + path + "There already exist " +
                        "non collection resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String parentPath = RegistryUtils.getParentPath(path);
        if (parentPath == null) {
            return;
        }
        
        ResourceIDImpl parentResourceID = resourceDAO.getResourceID(parentPath, true);
        if (parentResourceID == null || !resourceDAO.resourceExists(parentResourceID)) {
            addEmptyCollection(parentPath);
            if(parentResourceID == null) {
                // since this is a collection there is no much performance hit here
                parentResourceID = resourceDAO.getResourceID(parentPath, true);
            }
        }
        else if (!AuthorizationUtils.authorize(
                    parentPath, ActionConstants.PUT)) {
            String msg = "Failed to add new resource. User " +
                    CurrentSession.getUser() + " does not have authorization to " +
                    "update the collection " + parentPath + ".";
            log.error(msg);
            throw new RegistryException(msg);
        }
        
        if (!AuthorizationUtils.authorize(
                path, ActionConstants.PUT)) {
            String msg = "Failed to add new resource. User " +
                    CurrentSession.getUser() + " does not have authorization to " +
                    "add the collection at" + path + ".";
            log.error(msg);
            throw new RegistryException(msg);
        }


        CollectionImpl collection = new CollectionImpl();
        logsDAO.addLog(path, CurrentSession.getUser(), LogEntry.ADD, null);
        resourceDAO.add(path, parentResourceID, collection);
    }


    // check in, check out functionality

    public void restore(String path, Reader reader, boolean rootResourceExists) throws RegistryException {
        if (!path.equals("/") && path.endsWith("/")) {
            // remove the / suffix
            path = path.substring(0, path.length() -1);
        }
        if (path.equals(RegistryConstants.ROOT_PATH)) {
            if (!AuthorizationUtils.authorize(path, ActionConstants.PUT)) {
                String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                        "check in to the path " + path + ".";
                log.error(msg);
                throw new AuthorizationFailedException(msg);
            }
        }
        else {
            String parentPath = RegistryUtils.getParentPath(path);
            ResourceImpl parentResource = resourceDAO.getResourceMetaData(parentPath);
            if (parentResource == null) {
                addEmptyCollection(parentPath);
            }
            else {
                if (!(parentResource instanceof CollectionImpl)) {
                    String msg = "Cannont restore into a non-collection at " + parentPath + ".";
                    log.error(msg);
                    throw new RegistryException(msg);
                }
                if (!rootResourceExists) {
                    // hm, we are adding an element to the parent so we need to check the parents permission
                    if (!AuthorizationUtils.authorize(parentPath, ActionConstants.PUT)) {
                        String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                                "check in to the parent path " + parentPath + ".";
                        log.error(msg);
                        throw new AuthorizationFailedException(msg);
                    }
                }
            }
        }
        try {
            XMLStreamReader xmlReader =
               XMLInputFactory.newInstance().createXMLStreamReader(reader);
            restoreRecursively(path, xmlReader, path);
        } catch (XMLStreamException e) {
            String msg = "Failed to serialize the dumped element at " + path + ".";
            log.error(msg);
            throw new RegistryException(msg, e);
        }
    }

    public void dump(String path, Writer writer) throws RegistryException {
        if (!path.equals("/") && path.endsWith("/")) {
            // remove the / suffix
            path = path.substring(0,  path.length() - 1);
        }

        try {
            XMLOutputFactory xof = XMLOutputFactory.newInstance();
            XMLStreamWriter xmlWriter = xof.createXMLStreamWriter(writer);

            xmlWriter.writeStartDocument();
            dumpRecursively(path, xmlWriter, path);
            xmlWriter.writeEndDocument();

        } catch (XMLStreamException e) {
            String msg = "Failed to serialize the dumped element at " + path + ".";
            log.error(msg);
            throw new RegistryException(msg, e);
        }
    }


    private void restoreRecursively(String path, XMLStreamReader xmlReader, String originalPath) throws RegistryException,
                                                                                XMLStreamException {
        // we need to check the authorization with the current resource path at this point
        if (!AuthorizationUtils.authorize(path, ActionConstants.PUT)) {
            String msg = "Failed to checkin resource " + path + ". User " +
                    CurrentSession.getUser() + " is not authorized to update " +
                    "the current collection path " + path + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        while (!xmlReader.isStartElement() && xmlReader.hasNext()) {
             xmlReader.next();
        }

        if(!xmlReader.hasNext()) {
            // nothing to parse
            return;
        }


        if (!xmlReader.getLocalName().equals("resource")) {
            String msg = "Invalid dump to restore at " + path;
            log.error(msg);
            throw new RegistryException(msg);
        }

        String incommingParentPath = xmlReader.getAttributeValue(null, "path");

        String isCollectionString = xmlReader.getAttributeValue(null, "isCollection");
        boolean isCollection = isCollectionString.equals("true");

        if (path.equals(RegistryConstants.ROOT_PATH) && !isCollection) {
            // you can not put non collection to the root path
            String msg = "Illigal to restore a non-collection inplace of root collection.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        ResourceImpl resourceImpl;
        byte[] contentBytes = new byte[0];
        if (isCollection) {
            resourceImpl = new CollectionImpl();
        }
        else {
            resourceImpl = new ResourceImpl();
        }

        List<CommentDO> commentDOList = new ArrayList<CommentDO>();
        List<TaggingDO> taggingDOList = new ArrayList<TaggingDO>();
        List<RatingDO> ratingDOList = new ArrayList<RatingDO>();

        boolean isCreateatorExist = false;
        boolean isCreatedTimeExist = false;
        boolean isUpdatorExist = false;
        boolean isUpdatedTimeExist = false;

        // traversing to the next element
        do {
            xmlReader.next();
        } while (!xmlReader.isStartElement() && xmlReader.hasNext());

        while (xmlReader.hasNext()) {
            String localName = xmlReader.getLocalName();
            
            // setMediaType
            if (localName.equals("mediaType")) {
                String text = xmlReader.getElementText();
                if (text != null) {
                    resourceImpl.setMediaType(text);
                }
                // now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // creator
            else if (localName.equals("creator")) {
                String text = xmlReader.getElementText();
                if (text != null) {
                    resourceImpl.setAuthorUserName(text);
                    isCreateatorExist = true;
                }
                // now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // createdTime
            else if (localName.equals("createdTime")) {
                String text = xmlReader.getElementText();
                if (text != null) {
                    long date = Long.parseLong(text);
                    resourceImpl.setCreatedTime(new Timestamp(date));
                    isCreatedTimeExist = true;
                }
                // now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // setLastUpdater
            else if (localName.equals("lastUpdater")) {
                String text = xmlReader.getElementText();
                if (text != null) {
                    resourceImpl.setLastUpdaterUserName(text);
                    isUpdatorExist = true;
                }
                // now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // LastModified
            else if (localName.equals("lastModified")) {
                String text = xmlReader.getElementText();
                if (text != null) {
                    long date = Long.parseLong(text);
                    resourceImpl.setLastModified(new Timestamp(date));
                    isUpdatedTimeExist = true;
                }
                // now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // get description
            else if (localName.equals("description")) {
                String text = xmlReader.getElementText();
                if (text != null) {
                    resourceImpl.setDescription(text);
                }
                // now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // get properties
            else if (localName.equals("properties")) {
                // iterating trying to find the childs..
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                while (xmlReader.hasNext() &&
                        xmlReader.getLocalName().equals("property")) {
                    String key = xmlReader.getAttributeValue(null, "key");
                    String text = xmlReader.getElementText();
                    if (text.equals("")) {
                        text = null;
                    }
                    if (text != null) {
                        resourceImpl.addPropertyWithNoUpdate(key, text);
                    }
                    do {
                         xmlReader.next();
                    } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                }
            }
            // get content
            else if (localName.equals("content")) {
                String text = xmlReader.getElementText();
                // we keep content as base64 encoded
                if (text != null) {
                    contentBytes = Base64.decode(text);
                }
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
            // getting comment information
            else if (localName.equals("comments")) {
                // iterating trying to find the childs..
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                while (xmlReader.hasNext() &&
                        xmlReader.getLocalName().equals("comment")) {
                    CommentDO commentDO = new CommentDO();

                    do {
                         xmlReader.next();
                    } while (!xmlReader.isStartElement() && xmlReader.hasNext());

                    localName = xmlReader.getLocalName();
                    while (xmlReader.hasNext() &&
                            (localName.equals("user") || localName.equals("text"))) {
                        if (localName.equals("user")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                commentDO.setCommentedUser(text);
                            }
                        }
                        else if (localName.equals("text")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                commentDO.setCommentText(text);
                            }
                        }
                        
                        do {
                             xmlReader.next();
                        } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                        if (xmlReader.hasNext()) {
                            localName = xmlReader.getLocalName();
                        }
                    }
                    commentDOList.add(commentDO);
                }
            }
            // getting tagging information
            else if (localName.equals("taggings")) {
                // iterating trying to find the childs..
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                while (xmlReader.hasNext() &&
                        xmlReader.getLocalName().equals("tagging")) {
                    
                    TaggingDO taggingDO = new TaggingDO();

                    do {
                         xmlReader.next();
                    } while (!xmlReader.isStartElement() && xmlReader.hasNext());

                    localName = xmlReader.getLocalName();
                    while (xmlReader.hasNext() &&
                            (localName.equals("user") || localName.equals("date") || localName.equals("tagName"))) {
                        if (localName.equals("user")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                taggingDO.setTaggedUserName(text);
                            }
                        }
                        else if (localName.equals("date")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                long date = Long.parseLong(text);
                                taggingDO.setTaggedTime(new Date(date));
                            }
                        }
                        else if (localName.equals("tagName")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                taggingDO.setTagName(text);
                            }
                        }
                        do {
                             xmlReader.next();
                        } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                        if (xmlReader.hasNext()) {
                            localName = xmlReader.getLocalName();
                        }
                    }
                    taggingDOList.add(taggingDO);
                }
            }
            // getting rating information
            else if (localName.equals("ratings")) {
                // iterating trying to find the childs..
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                while (xmlReader.hasNext() &&
                        xmlReader.getLocalName().equals("rating")) {
                    RatingDO ratingDO = new RatingDO();

                    do {
                         xmlReader.next();
                    } while (!xmlReader.isStartElement() && xmlReader.hasNext());

                    localName = xmlReader.getLocalName();
                    while (xmlReader.hasNext() &&
                            (localName.equals("user") || localName.equals("date") || localName.equals("rate"))) {
                        if (localName.equals("user")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                ratingDO.setRatedUserName(text);
                            }
                        }
                        else if (localName.equals("date")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                long date = Long.parseLong(text);
                                ratingDO.setRatedTime(new Date(date));
                            }
                        }
                        else if (localName.equals("rate")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                int ratingValue = Integer.parseInt(text);
                                ratingDO.setRating(ratingValue);
                            }
                        }
                        do {
                             xmlReader.next();
                        } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                        if (xmlReader.hasNext()) {
                            localName = xmlReader.getLocalName();
                        }
                    }
                    ratingDOList.add(ratingDO);
                }
            }

            // getting rating information
            else if (localName.equals("associations")) {
                // iterating trying to find the childs..
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                while (xmlReader.hasNext() &&
                        xmlReader.getLocalName().equals("association")) {
                    String source = null;
                    String destination = null;
                    String type = null;

                    do {
                         xmlReader.next();
                    } while (!xmlReader.isStartElement() && xmlReader.hasNext());

                    localName = xmlReader.getLocalName();
                    while (xmlReader.hasNext() &&
                            (localName.equals("source") || localName.equals("destination") || localName.equals("type"))) {
                        if (localName.equals("source")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                source = text;
                            }
                        }
                        else if (localName.equals("destination")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                destination = text;
                            }
                        }
                        else if (localName.equals("type")) {
                            String text = xmlReader.getElementText();
                            if (text != null) {
                                type = text;
                            }
                        }
                        do {
                             xmlReader.next();
                        } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                        if (xmlReader.hasNext()) {
                            localName = xmlReader.getLocalName();
                        }
                    }
                    // get the source and destination as absolute paths
                    source = RegistryUtils.getAbsolutePathToOriginal(source, originalPath);
                    destination = RegistryUtils.getAbsolutePathToOriginal(destination, originalPath);
                    associationDAO.addAssociation(source, destination, type);
                }
            }
            // getting childs, just storing in arraylist now, will used at the end
            else if (localName.equals("childs")) {
                // we keep the stream to call this function recursively
                break;
            }else if (localName.equals("resource")) {
                // we keep the stream to call this function recursively
                break;
            }
            else {
                // we don't mind having unwanted elements, now go to the next element
                do {
                     xmlReader.next();
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
            }
        }

        // completing the empty fields
        if (!isCreateatorExist) {
            String creator = CurrentSession.getUser();
            resourceImpl.setAuthorUserName(creator);
        }
        if (!isCreatedTimeExist) {
            long now = System.currentTimeMillis();
            resourceImpl.setCreatedTime(new Date(now));
        }
        if (!isUpdatorExist) {
            String updator = CurrentSession.getUser();
            resourceImpl.setLastUpdaterUserName(updator);
        }
        if (!isUpdatedTimeExist) {
            long now = System.currentTimeMillis();
            resourceImpl.setLastModified(new Date(now));
        }

        // creating the resourceID
        ResourceIDImpl resourceID = resourceDAO.getResourceID(path, isCollection);
        if (resourceID == null) {
            // need to create a resourceID
            String parentPath = RegistryUtils.getParentPath(path);

            ResourceIDImpl parentResourceID = resourceDAO.getResourceID(parentPath, true);
            if (parentResourceID == null || !resourceDAO.resourceExists(parentResourceID)) {
                addEmptyCollection(parentPath);
                if(parentResourceID == null) {
                    parentResourceID = resourceDAO.getResourceID(parentPath, true);
                }
            }
            resourceDAO.createAndApplyResourceID(path, parentResourceID, resourceImpl);
            resourceID = resourceImpl.getResourceIDImpl();
        }
        else {
            resourceImpl.setPathID(resourceID.getPathID());
            resourceImpl.setPath(path);
            resourceImpl.setName(resourceID.getName());
        }

        // adding resource followed by content (for nonCollection
        if (!isCollection) {
            int contentId = 0;
            if (contentBytes.length > 0) {
                contentId = resourceDAO.addContentBytes(new ByteArrayInputStream(contentBytes));
            }
            resourceImpl.setDbBasedContentID(contentId);
        }

        ResourceDO resourceDO = resourceImpl.getResourceDO();
        resourceDAO.addResourceDO(resourceDO);
        resourceImpl.setVersionNumber(resourceDO.getVersion());

        // adding the properties.
        resourceDAO.addProperties(resourceImpl);

        // adding comments
        commentsDAO.addComments(resourceImpl,
                commentDOList.toArray(new CommentDO[commentDOList.size()]));

        // adding tags
        tagsDAO.addTaggings(resourceImpl,
                taggingDOList.toArray(new TaggingDO[taggingDOList.size()]));

        // adding ratings
        ratingsDAO.addRatings(resourceImpl,
                ratingDOList.toArray(new RatingDO[ratingDOList.size()]));

        // cleaning up the lists
        commentDOList = null;
        taggingDOList = null;
        ratingDOList = null;
        resourceImpl = null;

        if (!xmlReader.hasNext() || !xmlReader.getLocalName().equals("childs")) {
            // finished the recursion
            return;
        }

        do {
            xmlReader.next();
            if (xmlReader.isEndElement() && xmlReader.getLocalName().equals("childs")) {
                // this means empty childs, just quit from here
                // before that we have to set the cursor to the start of the next element
                do {
                    xmlReader.next();    
                } while (!xmlReader.isStartElement() && xmlReader.hasNext());
                return;
            }
        } while (!xmlReader.isStartElement() && xmlReader.hasNext());

        while (xmlReader.hasNext() && xmlReader.getLocalName().equals("resource")) {
            // prepare the childs absolute path
            String incommingChildPath = xmlReader.getAttributeValue(null, "path");
            if (!incommingChildPath.startsWith(incommingParentPath)) {
                break;
            }

            String relativeChildPath;
            if (incommingParentPath.equals(RegistryConstants.ROOT_PATH)) {
                relativeChildPath = incommingChildPath;
            }
            else {
                relativeChildPath = incommingChildPath.substring(incommingParentPath.length());
            }

            String absoluteChildPath;
            if (path.equals(RegistryConstants.ROOT_PATH)) {
                absoluteChildPath = relativeChildPath;
            }
            else {
                absoluteChildPath = path + relativeChildPath;
            }

            // call the check in method recursively
            restoreRecursively(absoluteChildPath, xmlReader, originalPath);
        }
    }

    private void dumpRecursively(String path, XMLStreamWriter xmlWriter, String originalPath) throws RegistryException,
                                                                            XMLStreamException{
        // adding resource meta data
        ResourceImpl resource = resourceDAO.getResourceMetaData(path);
        if (resource == null) {
            return;
        }
        
        if (!AuthorizationUtils.authorize(path, ActionConstants.GET)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "check out the path " + path + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        
        xmlWriter.writeStartElement("resource");

        // adding path as an attribute
        String relativePath = RegistryUtils.getRelativePathToOriginal(path, originalPath);
        xmlWriter.writeAttribute("path", relativePath);

        // adding isCollection as an attribute
        xmlWriter.writeAttribute("isCollection", (resource instanceof CollectionImpl) ? "true": "false");
        OMElement child;

        // set media type
        String mediaType = resource.getMediaType();
        OMFactory factory = OMAbstractFactory.getOMFactory(); 
        child = factory.createOMElement(new QName("mediaType"));
        child.setText(mediaType);
        child.serialize(xmlWriter);

        // set creator
        String creator = resource.getAuthorUserName();
        child = factory.createOMElement(new QName("creator"));
        child.setText(creator);
        child.serialize(xmlWriter);

        // set createdTime
        Date createdTime = resource.getCreatedTime();
        child = factory.createOMElement(new QName("createdTime"));
        child.setText(Long.toString(createdTime.getTime()));
        child.serialize(xmlWriter);

        // set updator
        String updator = resource.getLastUpdaterUserName();
        child = factory.createOMElement(new QName("lastUpdater"));
        child.setText(updator);
        child.serialize(xmlWriter);

        // set LastModified
        Date lastModified = resource.getLastModified();
        child = factory.createOMElement(new QName("lastModified"));
        child.setText(Long.toString(lastModified.getTime()));
        child.serialize(xmlWriter);

        // set Description
        String description = resource.getDescription();
        child = factory.createOMElement(new QName("description"));
        child.setText(description);
        child.serialize(xmlWriter);

        // fill properties
        resourceDAO.fillResourceProperties(resource);
        Properties properties = resource.getProperties();
        if (properties != null && properties.size() > 0) {
            // properties will be kept inside the <properties> element
            OMElement propertiesOM = factory.createOMElement(new QName("properties"));
            Iterator it = properties.keySet().iterator();
            while (it.hasNext()) {
                String key = (String) it.next();
                List<String> propValues = resource.getPropertyValues(key);

                for(int i = 0; i < propValues.size(); i ++) {
                    String value = propValues.get(i);
                    OMElement propertyOM = factory.createOMElement(new QName("property"));

                    // adding the key and value as attributes
                    OMAttribute keyAttribute = factory.createOMAttribute("key", null, key);
                    propertyOM.addAttribute(keyAttribute);

                    if(value != null) {
                        propertyOM.setText(value);
                    }

                    propertiesOM.addChild(propertyOM);
                }
            }
            propertiesOM.serialize(xmlWriter);
        }

        // adding contents..
        if (!(resource instanceof CollectionImpl)) {
            resourceDAO.fillResourceContent(resource);

            byte[] content = (byte[])resource.getContent();
            if(content != null) {
                child = factory.createOMElement(new QName("content"));
                child.setText(Base64.encode(content));
                child.serialize(xmlWriter);
            }
        }
        // getting comment information
        Comment[] comments = commentsDAO.getComments(resource);
        if (comments != null && comments.length > 0) {
            child = factory.createOMElement(new QName("comments"));
            for (Comment comment : comments) {
                OMElement commentElement = factory.createOMElement(new QName("comment"));
                String user = comment.getAuthorUserName();
                String text = comment.getText();

                OMElement userElement = factory.createOMElement(new QName("user"));
                userElement.setText(user);
                commentElement.addChild(userElement);

                OMElement textElement = factory.createOMElement(new QName("text"));
                textElement.setText(text);
                commentElement.addChild(textElement);

                child.addChild(commentElement);
            }
            child.serialize(xmlWriter);
        }
        // getting taggging
        TaggingDO[] taggings = tagsDAO.getTagging(resource);
        if (taggings != null && taggings.length > 0) {
            child = factory.createOMElement(new QName("taggings"));
            for (TaggingDO tagging : taggings) {
                OMElement taggingElement = factory.createOMElement(new QName("tagging"));
                String user = tagging.getTaggedUserName();
                Date date = tagging.getTaggedTime();
                String tagName = tagging.getTagName();

                OMElement userElement = factory.createOMElement(new QName("user"));
                userElement.setText(user);
                taggingElement.addChild(userElement);

                OMElement dateElement = factory.createOMElement(new QName("date"));
                String dateString = Long.toString(date.getTime());
                dateElement.setText(dateString);
                taggingElement.addChild(dateElement);

                OMElement textElement = factory.createOMElement(new QName("tagName"));
                textElement.setText(tagName);
                taggingElement.addChild(textElement);

                child.addChild(taggingElement);
            }
            child.serialize(xmlWriter);
        }
        // getting ratings
        RatingDO[] ratings = ratingsDAO.getResourceRatingDO(resource);
        if (ratings != null && ratings.length > 0) {
            child = factory.createOMElement(new QName("ratings"));
            for (RatingDO rating : ratings) {
                OMElement ratingElement = factory.createOMElement(new QName("rating"));
                String user = rating.getRatedUserName();
                Date date = rating.getRatedTime();
                int rate = rating.getRating();

                OMElement userElement = factory.createOMElement(new QName("user"));
                userElement.setText(user);
                ratingElement.addChild(userElement);

                OMElement dateElement = factory.createOMElement(new QName("date"));
                String dateString = Long.toString(date.getTime());
                dateElement.setText(dateString);
                ratingElement.addChild(dateElement);

                OMElement textElement = factory.createOMElement(new QName("rate"));
                String rateString = String.valueOf(rate);
                textElement.setText(rateString);
                ratingElement.addChild(textElement);

                child.addChild(ratingElement);
            }
            child.serialize(xmlWriter);
        }

        Association[] associations = associationDAO.getAllAssociations(path);
        if (associations != null && associations.length > 0) {
            child = factory.createOMElement(new QName("associations"));
            for (Association association: associations) {
                OMElement associationElement = factory.createOMElement(new QName("association"));
                String source = association.getSourcePath();
                String destination = association.getDestinationPath();
                String type = association.getAssociationType();

                // getting the relative paths
                source = RegistryUtils.getRelativePathToOriginal(source, originalPath);
                destination = RegistryUtils.getRelativePathToOriginal(destination, originalPath);

                OMElement sourceElement = factory.createOMElement(new QName("source"));
                sourceElement.setText(source);
                associationElement.addChild(sourceElement);

                OMElement destinationElement = factory.createOMElement(new QName("destination"));
                destinationElement.setText(destination);
                associationElement.addChild(destinationElement);

                OMElement typeElement = factory.createOMElement(new QName("type"));
                typeElement.setText(type);
                associationElement.addChild(typeElement);

                child.addChild(associationElement);
            }
            child.serialize(xmlWriter);
        }

        // getting childs and applying dump recursively
        if (resource instanceof CollectionImpl) {
            CollectionImpl collection =(CollectionImpl)resource;
            resourceDAO.fillChildren(collection, 0, -1);
            String childPaths[] = collection.getChildren();

            xmlWriter.writeStartElement("childs");
            for (String childPath : childPaths) {
                dumpRecursively(childPath, xmlWriter, originalPath);
            }
            xmlWriter.writeEndElement();
        }
        xmlWriter.writeEndElement();
    }
}
