/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dao.*;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.VersionInputStream;
import org.wso2.carbon.registry.core.jdbc.utils.VersionRetriever;
import org.wso2.carbon.registry.core.secure.AuthorizationFailedException;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.utils.VersionedPath;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.StaticConfiguration;

import javax.sql.DataSource;
import java.util.ArrayList;
import java.util.List;
import java.io.*;

public class VersionRepository {

    private static Log log = LogFactory.getLog(VersionRepository.class);

    private DataSource dataSource;

    private ResourceDAO resourceDAO = new ResourceDAO();
    private ResourceVersionDAO resourceVersionDAO = new ResourceVersionDAO();
    private CommentsDAO commentsDAO = new CommentsDAO();
    private RatingsDAO ratingsDAO = new RatingsDAO();
    private TagsDAO tagsDAO = new TagsDAO();

    public VersionRepository(DataSource dataSource) {
        this.dataSource = dataSource;
        commentsDAO = StaticConfiguration.getCommentsDAO();
        ratingsDAO = StaticConfiguration.getRatingsDAO();
        tagsDAO = StaticConfiguration.getTagsDAO();
    }

    public void createSnapshot(Resource resource,
                               boolean isRenewing,
                               boolean keepProperties) throws RegistryException {
        ResourceImpl resourceImpl = (ResourceImpl)resource;
        // archiving the old root resource
        createVersionForResource(resourceImpl, isRenewing, keepProperties);
        
        long version = resourceImpl.getVersionNumber();
        boolean isCollection =  resourceImpl instanceof CollectionImpl;
        ResourceIDImpl rootResourceID = resourceImpl.getResourceIDImpl();
        ArrayList<Long> versionList = new ArrayList<Long>();
        versionList.add(new Long(version));
        if (isCollection) {
            // for collection we have to iterate through children
            addDescendants(rootResourceID, versionList, isRenewing, keepProperties);
        }
        // wrap the arraylist into stream
        InputStream versionsInputStream = new VersionInputStream(versionList);

        int pathId = rootResourceID.getPathID();
        String resourceName = rootResourceID.getName();
        resourceVersionDAO.createSnapshot(pathId, resourceName, versionsInputStream);
    }

    public void createSnapshot(ResourcePath resourcePath,
                               boolean isRenewing,
                               boolean keepProperties) throws RegistryException {

         if (!resourcePath.isCurrentVersion()) {
            String msg = "Failed to create snapshot of the resource " + resourcePath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String path = resourcePath.getPath();
        path = RegistryUtils.getPureResourcePath(path);

        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        if (resourceID == null) {
            String msg = "Failed to get resource id to create a snaphost to the resouce " +
                   path + ". ";
            throw new RegistryException(msg);
            
        }
        ResourceImpl resource = resourceDAO.getResourceMetaData(resourceID);
        createSnapshot(resource, isRenewing, keepProperties);
    }

    /**
     * Here the versions of the descendeatns of the resource is added to the versionList array.
     * immediate descendents will be added to the list before others
     * @param resourceID
     * @param versionList
     * @throws RegistryException
     */
    public void addDescendants(ResourceIDImpl resourceID,
                               ArrayList<Long> versionList,
                               boolean isRenewing,
                               boolean keepProperties)  throws RegistryException{
        ArrayList<ResourceIDImpl> childIds = resourceDAO.getChildPathIds(resourceID);
        // immediate childs will be added before others
        for(int i = 0; i < childIds.size(); i ++) {
            ResourceIDImpl childId = childIds.get(i);

            ResourceImpl childResourceImpl = resourceDAO.getResourceWithNoUpdate(childId);
            long version = childResourceImpl.getVersionNumber();
            versionList.add(version);

            // we are archiving all the resources
            createVersionForResource(childResourceImpl, isRenewing, keepProperties);
        }
        // then next immediate levels will be added..
        for(int i = 0; i < childIds.size(); i ++) {
            ResourceIDImpl childId = childIds.get(i);
            if (childId.isCollection()) {
                // add descendants recursivly
                addDescendants(childId, versionList, isRenewing, keepProperties);
            }
        }
        // we have to explicitly copy the collections to the history table
        ResourceDO resourceDO =  resourceDAO.getResourceDO(resourceID);
        if (resourceID.isCollection() &&
                !resourceVersionDAO.isResourceHistoryExist(resourceDO.getVersion())) {
            resourceVersionDAO.putResourceToHistory(resourceDO);
        }
    }

    private void createVersionForResource(ResourceImpl resourceImpl,
                                          boolean isRenewing,
                                          boolean keepProperties) throws RegistryException{
        ResourceDO resourceDO = resourceImpl.getResourceDO();
        if(resourceDO.getVersion() <= 0) {
            // we need to fetch the resource from the database
            resourceDO = resourceDAO.getResourceDO(resourceImpl.getResourceIDImpl());
            isRenewing = false;
        }
        if (isRenewing) {
            // retrive the old content and properties before versioning
            if (!(resourceImpl instanceof CollectionImpl)) {
                // we have to load the content
                resourceDAO.fillResourceContentWithNoUpdate(resourceImpl);
            }
            if (StaticConfiguration.isVersioningProperties() || !keepProperties) {
                // we need to load the properties as well
                resourceDAO.fillResourcePropertiesWithNoUpdate(resourceImpl);
            }
        }
        resourceVersionDAO.versionResource(resourceDO, keepProperties);
        if (isRenewing) {
            // we add a new copy to the resource table with a newer version
            ResourceImpl renewedResourceImpl = resourceImpl.getShallowCopy();
            if (!(renewedResourceImpl instanceof CollectionImpl)) {
                resourceDAO.addContent(renewedResourceImpl);
            }
            resourceDAO.addResourceWithNoUpdate(renewedResourceImpl);
            if(StaticConfiguration.isVersioningProperties() || !keepProperties) {
                // if the properties are not versioned, we can here safely assume
                // the properties are already there, so no need to add twice
                resourceDAO.addProperties(renewedResourceImpl);
            }

            commentsDAO.copyComments(resourceImpl, renewedResourceImpl);
            tagsDAO.copyTags(resourceImpl, renewedResourceImpl);
            ratingsDAO.copyRatings(resourceImpl, renewedResourceImpl);
        }
    }

    public String[] getVersions(String resourcePath) throws RegistryException {

        resourcePath = RegistryUtils.getPureResourcePath(resourcePath);

        Long[] snapshotNumbers = resourceVersionDAO.getSnapshotIDs(resourcePath);

        List <String> versionPaths = new ArrayList();
        for (Long snapshotNumber : snapshotNumbers) {
            String versionPath = resourcePath + RegistryConstants.URL_SEPARATOR +
                    "version:" + snapshotNumber;
            versionPaths.add(versionPath);
        }

        return versionPaths.toArray(new String[versionPaths.size()]);
    }


    public Resource getMetaData(VersionedPath versionedPath) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(versionedPath.getPath());

        if (!AuthorizationUtils.authorize(versionedPath.getPath(), ActionConstants.GET)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "read the resource " + versionedPath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        long snapshotID = versionedPath.getVersion();

        ResourceImpl resourceImpl = resourceVersionDAO.get(resourceID, snapshotID);

        if (resourceImpl == null) {
            String msg = "Resource " + versionedPath.getPath() +
                    " does not have a version " + versionedPath.getVersion();
            log.error(msg);
            throw new RegistryException(msg);
        }
        resourceImpl.setDataSource(dataSource);

        resourceImpl.setPath(versionedPath.getPath());
        resourceImpl.setSnapshotID(snapshotID);
        resourceImpl.setMatchingSnapshotID(snapshotID);
        return resourceImpl;
    }

    public Resource get(VersionedPath versionedPath) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(versionedPath.getPath());

        if (!AuthorizationUtils.authorize(versionedPath.getPath(), ActionConstants.GET)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "read the resource " + versionedPath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }
        
        long snapshotID = versionedPath.getVersion();

        ResourceImpl resourceImpl = resourceVersionDAO.get(resourceID, snapshotID);

        int contentId = resourceImpl.getDbBasedContentID();
        if(contentId > 0) {
            resourceVersionDAO.fillResourceContentArchived(resourceImpl);
        }
        resourceDAO.fillResourcePropertiesWithNoUpdate(resourceImpl);
        
        if (resourceImpl == null) {
            String msg = "Resource " + versionedPath.getPath() +
                    " does not have a version " + versionedPath.getVersion();
            log.error(msg);
            throw new RegistryException(msg);
        }
        resourceImpl.setDataSource(dataSource);
        
        resourceImpl.setPath(versionedPath.getPath());
        resourceImpl.setSnapshotID(snapshotID);
        resourceImpl.setMatchingSnapshotID(snapshotID);
        return resourceImpl;
    }

    public Collection get(VersionedPath versionedPath, int start, int pageLen)
            throws RegistryException {

        ResourceIDImpl resourceID = resourceDAO.getResourceID(versionedPath.getPath());

        long snapshotID = versionedPath.getVersion();

        CollectionImpl collectionImpl = resourceVersionDAO.get(resourceID, snapshotID, start, pageLen);

        if (collectionImpl == null) {
            String msg = "Resource " + versionedPath.getPath() +
                    " does not have a version " + versionedPath.getVersion();
            log.error(msg);
            throw new RegistryException(msg);
        }

        collectionImpl.setDataSource(dataSource);
        collectionImpl.setPath(versionedPath.getPath());
        collectionImpl.setSnapshotID(snapshotID);
        collectionImpl.setMatchingSnapshotID(snapshotID);
        return collectionImpl;
    }

    public void restoreVersion(ResourcePath resourcePath) throws RegistryException {

        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        if (versionedPath.getVersion() == -1) {
            String msg = "Failed to restore resource. " +
                    versionedPath + " is not a valid version path.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String versionedResourcePath = versionedPath.getPath();
        long snapshotID = versionedPath.getVersion();


        ResourceIDImpl resourceID = resourceDAO.getResourceID(versionedResourcePath);
        if (!AuthorizationUtils.authorize(versionedResourcePath, ActionConstants.PUT)) {
            String msg = "User " + CurrentSession.getUser() + " is not authorized to " +
                    "restore the resource " + versionedResourcePath + ".";
            log.error(msg);
            throw new AuthorizationFailedException(msg);
        }

        restoreSnapshotNetwork(resourceID, snapshotID);
    }

    private void restoreSnapshotNetwork(ResourceIDImpl resourceID, long snapshotID)
            throws RegistryException {


        // the algorithm to restore snapshot now..
        // check if resource exist in the the snapshot
        // if no return failure
        // delete the current resource.
        // restore the versoned things
        //      get each resource from resource_history and copy to resource
        //      restore contents
        // That's all

        VersionRetriever versionRetriever = resourceVersionDAO.getVersionList(resourceID, snapshotID);

        if (versionRetriever == null) {
            String msg = "The snapshot " + snapshotID + " doesn't contain the " +
                        "resource " + resourceID.getPath();
            log.error(msg);
            throw new AuthorizationFailedException(msg);  
        }
        int versionIndex = 0;
        while (true) {
            long version = versionRetriever.getVersion(versionIndex);
            if (version == -1) {
                // no more stream
                break;
            }
            // restore resource and content
            String path = resourceVersionDAO.restoreResources(version);
            String parentPath = RegistryUtils.getParentPath(path);
            versionIndex ++;
        }
    }
}
