/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.Association;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

public class AssociationDAO {

    private static final Log log = LogFactory.getLog(AssociationDAO.class);

    public void addAssociation(String sourcePath,
                               String targetPath,
                               String associationType) throws RegistryException {
        if (isAssociationExists(sourcePath, targetPath, associationType)) {
            // not to duplicate the associations.
            return;
        }
        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        try {
            String propSQL =
                    "INSERT INTO REG_ASSOCIATION (REG_SOURCEPATH, REG_TARGETPATH, REG_ASSOCIATION_TYPE) VALUES (?,?,?)";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, sourcePath);
            ps.setString(2, targetPath);
            ps.setString(3, associationType);
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to add association between resources " +
                    sourcePath + " and " + targetPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    private boolean isAssociationExists(String sourcePath,
                                       String targetPath,
                                       String associationType) throws RegistryException {
        boolean isExists = false;
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String propSQL =
                    "SELECT REG_SOURCEPATH FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=? AND REG_TARGETPATH=? AND REG_ASSOCIATION_TYPE=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, sourcePath);
            ps.setString(2, targetPath);
            ps.setString(3, associationType);
            result = ps.executeQuery();

            if (result.next()) {
                isExists = true;    
            }

        } catch (SQLException e) {
            String msg = "Failed to check the existance of the association between resources " +
                    sourcePath + " and " + targetPath + " for association type " +
                    associationType + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
                if (result != null) {result.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
             }
        }
        return isExists;
    }

    public void removeAssociation(String sourcePath,
                                  String targetPath,
                                  String associationType) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        try {
            String propSQL =
                    "DELETE FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=? AND REG_TARGETPATH=? AND REG_ASSOCIATION_TYPE=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, sourcePath);
            ps.setString(2, targetPath);
            ps.setString(3, associationType);
            ps.executeUpdate();

        } catch (SQLException e) {
            String msg = "Failed to remove association between resources " +
                    sourcePath + " and " + targetPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }


    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ArrayList arrayList;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String propSQL = "SELECT REG_SOURCEPATH,REG_TARGETPATH,REG_ASSOCIATION_TYPE FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=? OR REG_TARGETPATH=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourcePath);
            ps.setString(2, resourcePath);
            result = ps.executeQuery();
            arrayList = new ArrayList();

            while (result.next()) {
                Association association = new Association();
                association.setSourcePath(result.getString("REG_SOURCEPATH"));
                association.setDestinationPath(result.getString("REG_TARGETPATH"));
                association.setAssociationType(result.getString("REG_ASSOCIATION_TYPE"));
                arrayList.add(association);
            }
        } catch (SQLException e) {
            String msg = "Failed to get all associations of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        Association associations[] = new Association[arrayList.size()];
        for (int i = 0; i < arrayList.size(); i++) {
            associations[i] = (Association) arrayList.get(i);
        }
        return associations;
    }

    public Association[] getAllAssociationsForType(String resourcePath, String associationType)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ArrayList arrayList;
        try {
            String propSQL = "SELECT REG_SOURCEPATH, REG_TARGETPATH FROM REG_ASSOCIATION " +
                             "WHERE (REG_SOURCEPATH=? OR REG_TARGETPATH=?) AND REG_ASSOCIATION_TYPE=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourcePath);
            ps.setString(2, resourcePath);
            ps.setString(3, associationType);
            result = ps.executeQuery();
            arrayList = new ArrayList();

            while (result.next()) {
                Association association = new Association();
                association.setSourcePath(result.getString("REG_SOURCEPATH"));
                association.setDestinationPath(result.getString("REG_TARGETPATH"));
                association.setAssociationType(associationType);
                arrayList.add(association);
            }
        } catch (SQLException e) {
            String msg = "Failed to get associations of type " +
                    associationType + " for resource " + resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        Association associations[] = new Association[arrayList.size()];
        for (int i = 0; i < arrayList.size(); i++) {
            associations[i] = (Association) arrayList.get(i);
        }
        return associations;
    }

    public void replaceAssociations(String oldPath, String newPath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps2 = null;
        try {
            String sql2 = "UPDATE REG_ASSOCIATION SET REG_TARGETPATH=? WHERE REG_TARGETPATH=?";

            ps2 = conn.prepareStatement(sql2);
            ps2.setString(1, newPath);
            ps2.setString(2, oldPath);
            ps2.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to replace the associations of " + oldPath +
                    " by reassociating them to " + newPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps2 != null) {ps2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public static void removeAllAssociations(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=? OR REG_TARGETPATH=?";

        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourcePath);
            ps.setString(2, resourcePath);
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to remove associations of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void copyAssociations(String fromPath ,String toPath ) throws RegistryException {
    	PreparedStatement ps = null;
    	ResultSet result = null;
        try {
            Connection conn = Transaction.getConnection();
            String propSQL = "SELECT REG_TARGETPATH, REG_ASSOCIATION_TYPE FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, fromPath);
            result = ps.executeQuery();
            ArrayList arrayList = new ArrayList();

            while (result.next()) {
                Association association = new Association();
                association.setSourcePath(toPath);
                association.setDestinationPath(result.getString("REG_TARGETPATH"));
                association.setAssociationType(result.getString("REG_ASSOCIATION_TYPE"));
                arrayList.add(association);
            }

            for (int i = 0; i < arrayList.size(); i++) {
                Association association = (Association) arrayList.get(i);
                addAssociation(toPath,association.getDestinationPath() , association.getAssociationType());
            }
        } catch (SQLException e) {
        	String msg = "SQLException occurred during copyAssociations";
            log.error(msg,e);
            throw new RegistryException(msg,e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }
}
