/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.CommentDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;

public class CommentsDAO {

    private static final Log log = LogFactory.getLog(CommentsDAO.class);
    ResourceDAO resourceDAO = new ResourceDAO();

    public int addComment(ResourceImpl resource, String userID, Comment comment)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps1 = null, ps2 = null;
        int commentId = -1;
        try {

            String sql1, sql2;

            sql1 = "INSERT INTO REG_COMMENT (REG_COMMENT_TEXT," +
                    "REG_USER_ID, REG_COMMENTED_TIME) VALUES (?, ?, ?)";
            sql2 = "INSERT INTO REG_RESOURCE_COMMENT (REG_COMMENT_ID, REG_PATH_ID, REG_RESOURCE_NAME) "+
                       "VALUES (?, ?, ?)";


            ps1 = conn.prepareStatement(sql1, new String[] {"REG_ID"});
            ps2 = conn.prepareStatement(sql2);

            // prepare to execute query2 for the comments
            ps1.setString(1, comment.getText());
            ps1.setString(2, userID);

            long now = System.currentTimeMillis();
            ps1.setTimestamp(3, new Timestamp(now));
            ps1.executeUpdate();
            ResultSet resultset1 = ps1.getGeneratedKeys();
            if (resultset1.next()) {
                // setting the RES_COMMENTS_ID
                commentId = resultset1.getInt(1);

                ps2.setInt(1, commentId);
                ps2.setInt(2, resource.getPathID());
                ps2.setString(3, resource.getName());
                ps2.executeUpdate();
            }
            if(resultset1 != null) {
                resultset1.close();
            }

        } catch (SQLException e) {

            String msg = "Failed to add comments to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }  finally {
            try {
                if (ps1 != null) { ps1.close(); }
                if (ps2 != null) { ps2.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return commentId;
    }

    public void copyComments(ResourceImpl sourceResource, ResourceImpl targetResource)
                throws RegistryException {
        if(sourceResource.getPath() == targetResource.getPath()) {
            // no special copying needed
            return;
        }
        Comment[] comments = getComments(sourceResource);
        CommentDO[] commentDOs = new CommentDO[comments.length];
        for (int i = 0; i < comments.length; i++) {
            CommentDO commentDO = new CommentDO();
            commentDO.setCommentedUser(comments[i].getUser());
            commentDO.setCommentText(comments[i].getText());
            commentDOs[i] = commentDO;
        }
        addComments(targetResource, commentDOs);      
    }

    public void addComments(ResourceImpl resource, CommentDO[] commentDOs)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps1 = null, ps2 = null;
        try {

            String sql1, sql2;

            sql1 = "INSERT INTO REG_COMMENT (REG_COMMENT_TEXT," +
                    "REG_USER_ID, REG_COMMENTED_TIME) VALUES (?, ?, ?)";
            sql2 = "INSERT INTO REG_RESOURCE_COMMENT (REG_COMMENT_ID, REG_PATH_ID, REG_RESOURCE_NAME) "+
                       "VALUES (?, ?, ?)";

            ps1 = conn.prepareStatement(sql1,  new String[] {"REG_ID"});
            ps2 = conn.prepareStatement(sql2);

            for (CommentDO comment : commentDOs){
                // prepare to execute query2 for the comments
                ps1.setString(1, comment.getCommentText());
                ps1.setString(2, comment.getCommentedUser());

                long now = System.currentTimeMillis();
                ps1.setTimestamp(3, new Timestamp(now));
                ps1.executeUpdate();
                ResultSet resultset1 = ps1.getGeneratedKeys();
                if (resultset1.next()) {
                    // setting the RES_COMMENTS_ID
                    int commentId = resultset1.getInt(1);

                    ps2.setInt(1, commentId);
                    ps2.setInt(2, resource.getPathID());
                    ps2.setString(3, resource.getName());

                    ps2.executeUpdate();
                    ps2.clearParameters();
                }
                if(resultset1 != null) {
                    resultset1.close();
                }
                ps2.clearParameters();
            }

        } catch (SQLException e) {

            String msg = "Failed to add comments to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) { ps1.close(); }
                if (ps2 != null) { ps2.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void updateComment(long commentId, String text) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement s = null;
        try {
            String sql = "UPDATE REG_COMMENT SET REG_COMMENT_TEXT=?,REG_COMMENTED_TIME=? WHERE REG_ID=?";
            long now = System.currentTimeMillis();
            s = conn.prepareStatement(sql);
            s.setString(1, text);
            s.setTimestamp(2, new Timestamp(now));
            s.setLong(3, commentId);
            s.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to update the comment with ID " + commentId +
                    " with text " + text + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void deleteComment(long commentId) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement s = null;
        PreparedStatement s1 = null;
        try {
            String sql = "DELETE FROM REG_RESOURCE_COMMENT WHERE REG_COMMENT_ID=?";

            s = conn.prepareStatement(sql);
            s.setLong(1, commentId);
            s.executeUpdate();

            sql = "DELETE FROM REG_COMMENT WHERE REG_ID=?";

            s1 = conn.prepareStatement(sql);
            s1.setLong(1, commentId);
            s1.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to delete the comment with ID " +
                    commentId + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (s != null) {s.close();}
                if (s1 != null) {s1.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public Comment[] getComments(ResourceImpl resource) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<Comment> commentList = new ArrayList<Comment>();
        PreparedStatement s = null;
        ResultSet results = null;
        String path = resource.getPath();
        try {
            String sql;
            if(resource instanceof CollectionImpl) {
                sql = "SELECT C.REG_ID, C.REG_COMMENT_TEXT, C.REG_USER_ID, C.REG_COMMENTED_TIME " +
                    "FROM REG_COMMENT C, REG_RESOURCE_COMMENT RC WHERE C.REG_ID=RC.REG_COMMENT_ID AND " +
                     "RC.REG_PATH_ID = ? AND RC.REG_RESOURCE_NAME IS NULL"; 
                s = conn.prepareStatement(sql);
                s.setInt(1, resource.getPathID());
            }
            else {
                sql = "SELECT C.REG_ID, C.REG_COMMENT_TEXT, C.REG_USER_ID, C.REG_COMMENTED_TIME " +
                    "FROM REG_COMMENT C, REG_RESOURCE_COMMENT RC WHERE C.REG_ID=RC.REG_COMMENT_ID AND " +
                     "RC.REG_PATH_ID = ? AND RC.REG_RESOURCE_NAME = ?";
                s = conn.prepareStatement(sql);
                s.setInt(1, resource.getPathID());
                s.setString(2, resource.getName());
            }

            results = s.executeQuery();
            while (results.next()) {
                Comment comment = new Comment();
                comment.setText(results.getString(DatabaseConstants.COMMENT_TEXT_FIELD));
                comment.setUser(results.getString(DatabaseConstants.COMMENTED_USER_ID_FIELD));
                comment.setTime(results.getTimestamp(DatabaseConstants.COMMENTED_TIME_FIELD));
                comment.setResourcePath(path);
                String commentPath = path + RegistryConstants.URL_SEPARATOR + "comments:" +
                            results.getInt("REG_ID");
                comment.setPath(commentPath);
                comment.setCommentPath(commentPath);
                comment.setParentPath(path + RegistryConstants.URL_SEPARATOR + "comments");
                comment.setCommmentID(results.getLong("REG_ID"));
                commentList.add(comment);
            }

        } catch (SQLException e) {

            String msg = "Failed to get comments on resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return commentList.toArray(new Comment[commentList.size()]);
    }



    public void removeComments(ResourceImpl resource) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps1 = null, ps2 = null;
        ResultSet results = null;
        String path = resource.getPath();
        Comment[] comments = getComments(resource);
        try {
            String sql1 = "DELETE FROM REG_RESOURCE_COMMENT WHERE REG_COMMENT_ID = ?";
            ps1 = conn.prepareStatement(sql1);

            String sql2 = "DELETE FROM REG_COMMENT WHERE REG_ID = ?";
            ps2 = conn.prepareStatement(sql2);

            for(Comment comment: comments) {
                ps1.setLong(1, comment.getCommmentID());
                ps2.setLong(1, comment.getCommmentID());
                
                ps1.addBatch();
                ps2.addBatch();
            }

            if(comments.length > 0) {
                try {
                    ps1.executeBatch();
                    ps2.executeBatch();
                } catch (SQLException e) {
                    ps1.clearBatch();
                    ps2.clearBatch();
                    // the exception will be handled in the next catch block
                    throw e;
                }
            }

        } catch (SQLException e) {

            String msg = "Failed to get comments on resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps1 != null) {ps1.close();}
                if (ps2 != null) {ps2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public Comment getComment(long commentID, String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement s =  null;
        ResultSet result = null;
        try {
            String sql = "SELECT C.REG_ID, C.REG_COMMENT_TEXT, C.REG_USER_ID, C.REG_COMMENTED_TIME " +
                    "FROM REG_COMMENT C WHERE C.REG_ID=?";

            s = conn.prepareStatement(sql);
            s.setLong(1, commentID);

            result = s.executeQuery();

            Comment comment = null;
            if (result.next()) {
                String commentPath = resourcePath + RegistryConstants.URL_SEPARATOR + "comments:" +
                        result.getInt("REG_ID");

                comment = new Comment();
                comment.setText(result.getString(DatabaseConstants.COMMENT_TEXT_FIELD));
                comment.setUser(result.getString(DatabaseConstants.COMMENTED_USER_ID_FIELD));
                comment.setTime(result.getTimestamp(DatabaseConstants.COMMENTED_TIME_FIELD));
                comment.setResourcePath(resourcePath);
                comment.setPath(commentPath);
                comment.setCommentPath(commentPath);
                comment.setParentPath(resourcePath + RegistryConstants.URL_SEPARATOR + "comments");
            }
            return comment;

        } catch (SQLException e) {

            String msg = "Failed to get comment with ID " + commentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public String[] getResourcePathsOfComments(Long[] commentIDs, Connection conn) throws RegistryException {

        if(commentIDs.length == 0) return new String[0];
        StringBuffer sqlBuf = new StringBuffer();
        sqlBuf.append("SELECT DISTINCT RC.REG_COMMENT_ID, RC.REG_PATH_ID, RC.REG_RESOURCE_NAME FROM " +
            "REG_RESOURCE_COMMENT RC WHERE (");
        for (int i = 0; i < commentIDs.length; i++) {
            if (i > 0) sqlBuf.append(" OR ");
            sqlBuf.append("RC.REG_COMMENT_ID=?");
        }
        sqlBuf.append(")");

        List<String> commentPathList = new ArrayList<String>();

        ResultSet results = null;
        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sqlBuf.toString());

            for (int i = 0; i < commentIDs.length; i++) {
                s.setLong(i + 1, commentIDs[i]);
            }

            results = s.executeQuery();
            while (results.next()) {
                long commentID = results.getLong("REG_COMMENT_ID");
                
                int pathId = results.getInt("REG_PATH_ID");
                String resourceName = results.getString("REG_RESOURCE_NAME");
                String path = resourceDAO.getPath(pathId, resourceName, true);
                if (path != null) {
                    String commentPath = path + RegistryConstants.URL_SEPARATOR + "comments:" + commentID;
                    commentPathList.add(commentPath);
                }
            }

        } catch (SQLException e) {


            String msg = "Failed to get the resource for the set of comment ids."
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return commentPathList.toArray(new String[commentPathList.size()]);
    }

    public ResourceImpl getResourceWithMinimumData(String path) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        if (resourceID == null) {
            return null;
        }
        ResourceImpl resourceImpl;
        if (resourceID.isCollection()) {
            resourceImpl = new CollectionImpl();
        }
        else {
            resourceImpl = new ResourceImpl();
        }
        resourceImpl.setName(resourceID.getName());
        resourceImpl.setPathID(resourceID.getPathID());
        resourceImpl.setPath(resourceID.getPath());
        return resourceImpl;
    }
    
    public void moveComments(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE_COMMENT SET REG_PATH_ID=? WHERE " +
                             "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.executeUpdate();
            }
            else {
                String sql = "UPDATE REG_RESOURCE_COMMENT SET REG_PATH_ID=?, REG_RESOURCE_NAME=? WHERE " +
                             "REG_PATH_ID=? AND REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move comments from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    /**
     * this function will move the paths from one path id to anthor regardless of the resource name
     * @param source
     * @param target
     * @throws RegistryException
     */
    public void moveCommentPaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE_COMMENT SET REG_PATH_ID=? WHERE REG_PATH_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move comment paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }
}
