/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.CommentDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;

public class CommentsVersionDAO extends CommentsDAO{

    private static final Log log = LogFactory.getLog(CommentsVersionDAO.class);
    ResourceDAO resourceDAO = new ResourceDAO();

    public int addComment(ResourceImpl resource, String userID, Comment comment)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps1 = null, ps2 = null;
        int commentId = -1;
        try {

            String sql1, sql2;

            sql1 = "INSERT INTO REG_COMMENT (REG_COMMENT_TEXT," +
                    "REG_USER_ID, REG_COMMENTED_TIME) VALUES (?, ?, ?)";
            sql2 = "INSERT INTO REG_RESOURCE_COMMENT (REG_COMMENT_ID, REG_VERSION) VALUES (?, ?)";

            ps1 = conn.prepareStatement(sql1, new String[] {"REG_ID"});
            ps2 = conn.prepareStatement(sql2);

            // prepare to execute query2 for the comments
            ps1.setString(1, comment.getText());
            ps1.setString(2, userID);

            long now = System.currentTimeMillis();
            ps1.setTimestamp(3, new Timestamp(now));
            ps1.executeUpdate();
            ResultSet resultset1 = ps1.getGeneratedKeys();
            if (resultset1.next()) {
                // setting the RES_COMMENTS_ID
                commentId = resultset1.getInt(1);

                ps2.setInt(1, commentId);
                ps2.setLong(2, resource.getVersionNumber());
                ps2.executeUpdate();
            }
            if(resultset1 != null) {
                resultset1.close();
            }

        } catch (SQLException e) {

            String msg = "Failed to add comments to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }  finally {
            try {
                if (ps1 != null) { ps1.close(); }
                if (ps2 != null) { ps2.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return commentId;
    }

    public void copyComments(ResourceImpl sourceResource, ResourceImpl targetResource)
                throws RegistryException {

        Comment[] comments = getComments(sourceResource);
        CommentDO[] commentDOs = new CommentDO[comments.length];
        for (int i = 0; i < comments.length; i++) {
            CommentDO commentDO = new CommentDO();
            commentDO.setCommentedUser(comments[i].getUser());
            commentDO.setCommentText(comments[i].getText());
            commentDOs[i] = commentDO;
        }
        addComments(targetResource, commentDOs);
    }

    public void addComments(ResourceImpl resource, CommentDO[] commentDOs)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps1 = null, ps2 = null;
        try {

            String sql1, sql2;

            sql1 = "INSERT INTO REG_COMMENT (REG_COMMENT_TEXT," +
                    "REG_USER_ID, REG_COMMENTED_TIME) VALUES (?, ?, ?)";
            sql2 = "INSERT INTO REG_RESOURCE_COMMENT (REG_COMMENT_ID, REG_VERSION) VALUES (?, ?)";

            ps1 = conn.prepareStatement(sql1,  new String[] {"REG_ID"});
            ps2 = conn.prepareStatement(sql2);

            for (CommentDO comment : commentDOs){
                // prepare to execute query2 for the comments
                ps1.setString(1, comment.getCommentText());
                ps1.setString(2, comment.getCommentedUser());

                long now = System.currentTimeMillis();
                ps1.setTimestamp(3, new Timestamp(now));
                ps1.executeUpdate();
                ResultSet resultset1 = ps1.getGeneratedKeys();
                if (resultset1.next()) {
                    // setting the RES_COMMENTS_ID
                    int commentId = resultset1.getInt(1);

                    ps2.setInt(1, commentId);
                    ps2.setLong(2, resource.getVersionNumber());

                    ps2.executeUpdate();
                    ps2.clearParameters();
                }
                if(resultset1 != null) {
                    resultset1.close();
                }
                ps2.clearParameters();
            }

        } catch (SQLException e) {

            String msg = "Failed to add comments to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) { ps1.close(); }
                if (ps2 != null) { ps2.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }


    public Comment[] getComments(ResourceImpl resource) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<Comment> commentList = new ArrayList<Comment>();
        PreparedStatement s = null;
        ResultSet results = null;
        String path = resource.getPath();
        try {
            String sql= "SELECT C.REG_ID, C.REG_COMMENT_TEXT, C.REG_USER_ID, C.REG_COMMENTED_TIME " +
                    "FROM REG_COMMENT C, REG_RESOURCE_COMMENT RC WHERE C.REG_ID=RC.REG_COMMENT_ID AND " +
                     "RC.REG_VERSION = ?";
            s = conn.prepareStatement(sql);
            s.setLong(1, resource.getVersionNumber());

            results = s.executeQuery();
            while (results.next()) {
                Comment comment = new Comment();
                comment.setText(results.getString(DatabaseConstants.COMMENT_TEXT_FIELD));
                comment.setUser(results.getString(DatabaseConstants.COMMENTED_USER_ID_FIELD));
                comment.setTime(results.getTimestamp(DatabaseConstants.COMMENTED_TIME_FIELD));
                comment.setResourcePath(path);
                String commentPath = path + RegistryConstants.URL_SEPARATOR + "comments:" +
                            results.getInt("REG_ID");
                comment.setPath(commentPath);
                comment.setCommentPath(commentPath);
                comment.setParentPath(path + RegistryConstants.URL_SEPARATOR + "comments");
                comment.setCommmentID(results.getLong("REG_ID"));
                commentList.add(comment);
            }

        } catch (SQLException e) {

            String msg = "Failed to get comments on resource " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (s != null) {s.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return commentList.toArray(new Comment[commentList.size()]);
    }

    public String[] getResourcePathsOfComments(Long[] commentIDs, Connection conn) throws RegistryException {

        if(commentIDs.length == 0) return new String[0];
        StringBuffer sqlBuf = new StringBuffer();
        sqlBuf.append("SELECT DISTINCT RC.REG_COMMENT_ID, RC.REG_VERSION FROM " +
            "REG_RESOURCE_COMMENT RC WHERE (");
        for (int i = 0; i < commentIDs.length; i++) {
            if (i > 0) sqlBuf.append(" OR ");
            sqlBuf.append("RC.REG_COMMENT_ID=?");
        }
        sqlBuf.append(")");

        List<String> commentPathList = new ArrayList<String>();

        ResultSet results = null;
        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sqlBuf.toString());

            for (int i = 0; i < commentIDs.length; i++) {
                s.setLong(i + 1, commentIDs[i]);
            }

            results = s.executeQuery();
            while (results.next()) {
                long commentID = results.getLong("REG_COMMENT_ID");
                long version = results.getLong("REG_VERSION");
                if(version > 0) {
                    String path = resourceDAO.getPath(version);
                    if (path != null) {
                        String commentPath = path + RegistryConstants.URL_SEPARATOR + "comments:" + commentID;
                        commentPathList.add(commentPath);
                    }
                }
            }

        } catch (SQLException e) {


            String msg = "Failed to get the resource for the set of comment ids."
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return commentPathList.toArray(new String[commentPathList.size()]);
    }

    public ResourceImpl getResourceWithMinimumData(String path) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        if (resourceID == null) {
            return null;
        }

        ResourceImpl resourceImpl;
        if (resourceID.isCollection()) {
            resourceImpl = new CollectionImpl();
        }
        else {
            resourceImpl = new ResourceImpl();
        }
        long version = resourceDAO.getVersion(resourceID);
        resourceImpl.setVersionNumber(version);
        resourceImpl.setPath(path);
        return resourceImpl;
    }

    public void moveComments(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        // this is non-versioned specific function.
        // do nothing when the comments versioniong on in configuration
    }

    public void moveCommentPaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        // this is non-versioned specific function.
        // do nothing when the comments versioniong on in configuration
    }
}
