/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Connection;
import java.sql.SQLException;

import net.sf.ehcache.Cache;
import net.sf.ehcache.Element;
import net.sf.ehcache.Statistics;
import net.sf.ehcache.CacheManager;

public class PathCache {

    private static PathCache pathCacheObject;
    private Cache pathCache = null;
    private CacheManager cacheManager = null;

    private PathCache(int maxElementsInMemory, boolean eternal, long timeToLiveSeconds) {
        // making this singelton
        cacheManager = new CacheManager();
        pathCache = new Cache(RegistryConstants.PATH_CACHE_ID,
                maxElementsInMemory, false, eternal, timeToLiveSeconds, 0);
        cacheManager.addCache(pathCache);
    }

    public static synchronized PathCache createPathCache(int maxElementsInMemory,
                                                         boolean eternal,
                                                         long timeToLiveSeconds) {

        return new PathCache(maxElementsInMemory, eternal, timeToLiveSeconds);
    }

    public static PathCache getPathCache() {
        if(pathCacheObject == null) {
            pathCacheObject = createPathCache(RegistryConstants.DEFAULT_PATH_CACHE_SIZE,
                                              RegistryConstants.DEFAULT_PATH_CACHE_IS_ETERNAL,
                                              RegistryConstants.DEFAULT_PATH_CACHE_TIMETO_LIVE);
        }
        return pathCacheObject;
    }

    private static class CacheEntry {
        CacheEntry(int pathId) {
            this.pathId = pathId;
        }
        int pathId;
    }

    private static final Log log = LogFactory.getLog(ResourceDAO.class);

    public double hitRate() {
        Statistics stats = pathCache.getStatistics();
        return (double) stats.getCacheHits() / ((double)(stats.getCacheHits() + stats.getCacheMisses()));
    }

    public int addEntry(String path, int parentPathId) throws SQLException {
        ResultSet results = null;
        PreparedStatement ps = null;
        Connection conn = RegistryContext.getSingleton().getDataSource().getConnection();
        if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
            conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
        }
        conn.setAutoCommit(false);
        boolean success = false;
        int pathId = 0;


        try {
            String sql = "INSERT INTO REG_PATH(REG_PATH_VALUE, REG_PATH_PARENT_ID) VALUES (?, ?)";
            ps = conn.prepareStatement(sql, new String[] {"REG_PATH_ID"});
            ps.setString(1, path);
            ps.setInt(2, parentPathId);
            ps.executeUpdate();

            results = ps.getGeneratedKeys();
            if (results.next()) {
                pathId = results.getInt(1);
                if(pathId > 0) {
                    success = true;
                    return pathId;
                }
            }

        } catch (SQLException e) {
            // we have to be expecting an exception with the duplicate value for the path value
            // which can be further checked from here..
            // will print the stacktrace as well
            String msg = "Failed to insert resource to " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw e;
        } finally {
            if (success) {
                try {
                    conn.commit();
                    CacheEntry e = new CacheEntry(pathId);
                    Element cacheEle = new Element(path, e);
                    pathCache.put(cacheEle);

                } catch (SQLException e) {
                    String msg = "Failed to commit transaction. Inserting " + path + ". " + e.getMessage();
                    log.error(msg, e);
                    throw e;

                } finally {
                    try {
                        if (results != null) {results.close();}
                        if (ps != null) {ps.close();}
                        if (conn != null) {conn.close();}
                    } catch (SQLException e) {
                        String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + e.getMessage();
                        log.error(msg,e);
                        throw e;
                    }
                }
            } else {
                try {
                    conn.rollback();

                } catch (SQLException e) {
                    String msg = "Failed to rollback transaction. Inserting " + path + ". " + e.getMessage();
                    log.error(msg, e);
                    throw e;
                } finally {
                    try {
                        if (results != null) {results.close();}
                        if (ps != null) {ps.close();}
                        if (conn != null) {conn.close();}
                    } catch (SQLException e) {
                        String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + e.getMessage();
                        log.error(msg,e);
                        throw e;
                    }
                }
            }
        }

        return -1;
    }

    public int getPathID(Connection conn, String path) throws SQLException{

        Element cacheEle = pathCache.get(path);
        if (cacheEle != null) {
            CacheEntry result = (CacheEntry)cacheEle.getObjectValue();
            if (result != null) {
                return result.pathId;
            }
        } else {
            ResultSet results = null;
            PreparedStatement ps = null;
            try {
                String sql = "SELECT REG_PATH_ID FROM REG_PATH WHERE REG_PATH_VALUE=?";
                ps = conn.prepareStatement(sql);
                ps.setString(1, path);
                results = ps.executeQuery();

                int pathId = 0;
                if (results.next()) {
                    pathId = results.getInt("REG_PATH_ID");

                    if(pathId > 0) {
                        CacheEntry e = new CacheEntry(pathId);
                        cacheEle = new Element(path, e);
                        pathCache.put(cacheEle);
                        return pathId;
                    }
                }

            } catch (SQLException e) {
                String msg = "Failed to retrieving resource from " + path + ". " + e.getMessage();
                log.error(msg, e);
                throw e;
            } finally {
                try {
                    if (results != null) {results.close();}
                    if (ps != null) {ps.close();}
                } catch (SQLException e) {
                    String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + e.getMessage();
                    log.error(msg,e);
                    throw e;
                }
            }
        }
        return -1;
    }
}