/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.ResourceImpl;
import org.wso2.carbon.registry.core.ResourceIDImpl;
import org.wso2.carbon.registry.core.CollectionImpl;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.RatingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

public class RatingsVersionDAO extends RatingsDAO{

    private static final Log log = LogFactory.getLog(RatingsVersionDAO.class);
    private ResourceDAO resourceDAO = new ResourceDAO();

    public void addRating(ResourceImpl resourceImpl, String userID, int rating)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null, ps2 = null;
        ResultSet result = null;
        try {

            String sql = "INSERT INTO REG_RATING (REG_RATING, REG_USER_ID, REG_RATED_TIME) VALUES (?,?,?)";

            ps = conn.prepareStatement(sql,  new String[] {"REG_ID"});
            ps.setInt(1, rating);
            ps.setString(2, userID);
            long now = System.currentTimeMillis();
            ps.setDate(3, new Date(now));

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if(result.next()) {
                int rateID = result.getInt(1);

                String sql2 = "INSERT INTO REG_RESOURCE_RATING (REG_RATING_ID, REG_VERSION) VALUES(?,?)";
                ps2 = conn.prepareStatement(sql2);

                ps2.setInt(1, rateID);
                ps2.setLong(2, resourceImpl.getVersionNumber());

                ps2.executeUpdate();

            }

        } catch (SQLException e) {

            String msg = "Failed to rate resource " + resourceImpl.getVersionNumber() +
                    " with rating " + rating + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
                if (ps2 != null) {ps2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void copyRatings(ResourceImpl fromResource, ResourceImpl toResource)
            throws RegistryException {

        RatingDO[] ratingDOs = getResourceRatingDO(fromResource);
        addRatings(toResource, ratingDOs);
    }

    public int getRateID(ResourceImpl resourceImpl, String userID) throws RegistryException {

        RatingDO ratingDO = getRatingDO(resourceImpl, userID);
        return ratingDO.getRatingID();
    }

    public float getAverageRating(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        int sumRating;
        int countRating;

        PreparedStatement ps = null;
        ResultSet result1 = null;
        try {
            String sql = "SELECT SUM(R.REG_RATING) FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                "WHERE RR.REG_VERSION=? AND " +
                "RR.REG_RATING_ID=R.REG_ID";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());

            result1 = ps.executeQuery();

            sumRating = 0;
            if (result1.next()) {
                sumRating = result1.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get sum of all ratings on resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result1 != null) {result1.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        ps = null;
        ResultSet result2 = null;
        try {
            String sql = "SELECT COUNT(R.REG_RATING) FROM REG_RATING R, REG_RESOURCE_RATING RR " +
               "WHERE RR.REG_VERSION=? AND " +
               "RR.REG_RATING_ID=R.REG_ID";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());

            result2 = ps.executeQuery();

            countRating = 0;
            if (result2.next()) {
                countRating = result2.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get ratings count on resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result2 != null) {result2.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        float averageRating = 0;
        if (countRating > 0) {
            averageRating = (float)sumRating / countRating;
        }

        return averageRating;
    }

    public RatingDO getRatingDO(ResourceImpl resourceImpl, String userName) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        ResultSet result = null;
        RatingDO ratingDO = new RatingDO();
        try {
            String sql = "SELECT R.REG_ID, R.REG_RATING, R.REG_RATED_TIME " +
                "FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                "WHERE RR.REG_VERSION=? AND " +
                "RR.REG_RATING_ID=R.REG_ID AND R.REG_USER_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());
            ps.setString(2, userName);

            result = ps.executeQuery();
            if (result.next()) {
                ratingDO.setRating(result.getInt("REG_RATING"));
                ratingDO.setRatedTime(new java.util.Date(
                        result.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime()));
                ratingDO.setRatedUserName(userName);
                ratingDO.setRatingID(result.getInt("REG_ID"));
            }
        } catch (SQLException e) {

            String msg = "Failed to get rating on resource " + resourceImpl.getPath() +
                    " done by user " + userName + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
            }
            catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
        return ratingDO;
    }


    public RatingDO[] getResourceRatingDO(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        ResultSet result = null;
        ArrayList<RatingDO> ratingDOs = new ArrayList<RatingDO>();
        try {
            String sql = "SELECT R.REG_ID, R.REG_RATING, R.REG_USER_ID, R.REG_RATED_TIME " +
                "FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                "WHERE RR.REG_VERSION=? AND " +
                "RR.REG_RATING_ID=R.REG_ID";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());


            result = ps.executeQuery();
            while (result.next()) {
                RatingDO ratingDO = new RatingDO();
                ratingDO.setRating(result.getInt("REG_RATING"));
                ratingDO.setRatedTime(new java.util.Date(
                        result.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime()));
                ratingDO.setRatedUserName(result.getString("REG_USER_ID"));
                ratingDO.setRatingID(result.getInt("REG_ID"));

                ratingDOs.add(ratingDO);
            }
        } catch (SQLException e) {

            String msg = "Failed to get rating on resource " + resourceImpl.getPath() +
                    ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
            }
            catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
        return ratingDOs.toArray(new RatingDO[ratingDOs.size()]);
    }

    public String[] getRatedUserNames(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List userNameList = new ArrayList();

        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            String sql = "SELECT R.REG_USER_ID FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                "WHERE RR.REG_VERSION=? AND " +
                "RR.REG_RATING_ID=R.REG_ID";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());


            results = ps.executeQuery();
            while (results.next()) {
                String ratedUserName = results.getString(DatabaseConstants.USER_ID_FIELD);
                userNameList.add(ratedUserName);
            }

        } catch (SQLException e) {

            String msg = "Failed to users who have rated the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        String[] userNames = new String[userNameList.size()];
        for (int i = 0; i < userNames.length; i++) {
            userNames[i] = ((String)userNameList.get(i));
        }

        return userNames;
    }

    public RatingDO getRating(long ratingID) throws SQLException, RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT RR.REG_VERSION, " +
                "R.REG_USER_ID, R.REG_RATING, R.REG_RATED_TIME " +
                "FROM REG_RATING R, REG_RESOURCE_RATING RR WHERE R.REG_ID =? AND " +
                "R.REG_ID = RR.REG_RATING_ID";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, ratingID);

        ResultSet results = s.executeQuery();
        if (results.next()) {

            java.util.Date ratedTime = new java.util.Date(
                    results.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime());

            RatingDO ratingDAO = new RatingDO();
            ratingDAO.setRatedUserName(results.getString(DatabaseConstants.USER_ID_FIELD));
            ratingDAO.setRatedTime(ratedTime);
            ratingDAO.setRating(results.getInt(DatabaseConstants.RATING_FIELD));
            // TODO:
            //ratingDAO.setResourceID(results.getString(DatabaseConstants.AID_FIELD));

            String resourcePath = null;
            long version = results.getLong("REG_VERSION");
            if(version > 0) {
                resourcePath = resourceDAO.getPath(version);
            }
            if (resourcePath != null) {
                ratingDAO.setResourcePath(resourcePath);
            }
            return ratingDAO;
        }

        return null;
    }


    public ResourceImpl getResourceWithMinimumData(String path) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        if (resourceID == null) {
            return null;
        }

        ResourceImpl resourceImpl;
        if (resourceID.isCollection()) {
            resourceImpl = new CollectionImpl();
        }
        else {
            resourceImpl = new ResourceImpl();
        }
        long version = resourceDAO.getVersion(resourceID);
        resourceImpl.setVersionNumber(version);
        resourceImpl.setPath(path);
        return resourceImpl;
    }

    public void moveRatings(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        // this is non-versioned specific function.
        // do nothing when the rating versioniong on in configuration
    }

    public void moveRatingPaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        // this is non-versioned specific function.
        // do nothing when the rating versioniong on in configuration
    }
}