/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.StaticConfiguration;

import java.io.InputStream;
import java.sql.*;
import java.util.*;
import java.util.Date;

public class ResourceDAO {

    private static final Log log = LogFactory.getLog(ResourceDAO.class);

    public static final int NULL = -1;
    public static final int RESOURCE = 0;
    public static final int COLLECTION = 1;

    /**
     * Returns the resource ID (RID) of the resource refered by the given path.
     * NOTE: This doesn't garuentee that the path exist in the resource. It only garuentees
     * if it is a collection: the path exist in the path table.
     * if it is a resource: the parent path exist in the path table. In order to make sure the existance of the resource
     * pelase use the resourceExists function
     * Note that if the same path is used for a collection and resource, this returns the
     * resourceID of the collection
     * 
     * @param path Pure path of the resource
     * @return RID of resource exists in path cache. null if the resource does not exists.
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException something went wrong
     */
    public ResourceIDImpl getResourceID(String path) throws RegistryException {
        // NEW
        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            // step1 : need to check whether it is a resource or collection while retrieving path id
            PathCache pathCache = PathCache.getPathCache();
            int pathID = pathCache.getPathID(conn, path);
            boolean isCollection = true;
            if (pathID == -1) {
                isCollection = false;
            }

            // ---------------------------------------------------------------------------------------------------------
            // step2: get the resource name + path id for non-collection
            // (for collection this is alreday done from the step1
            String resourceName = null;
            if (!isCollection) {
                // we have to reget the path id for the parent path..
                String parentPath = RegistryUtils.getParentPath(path);
                resourceName = RegistryUtils.getResourceName(path);

                pathID = pathCache.getPathID(conn, parentPath);
            }

            if (pathID != -1) {
                ResourceIDImpl resourceID = new ResourceIDImpl();
                resourceID.setCollection(isCollection);
                resourceID.setName(resourceName);
                resourceID.setPathID(pathID);
                resourceID.setPath(path);
                return resourceID;
            }


        } catch (SQLException e) {
            String msg = "Failed to get ID of the resource at path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return null;
    }


    /**
     * Returns the resource ID (RID) of the resource refered by the given path. We use this overloaded function
     * when we know our resource is a collection or non-collection
     * NOTE: This doesn't garuentee that the path exist in the resource. It only garuentees
     * if it is a collection: the path exist in the path table.
     * if it is a resource: the parent path exist in the path table. In order to make sure the existance of the resource
     * pelase use the resourceExists function
     *
     * @param path Pure path of the resource
     * @param isCollection true if it is a collection
     * @return RID of resource exists. null if the resource does not exists.
     * @throws java.sql.SQLException something went wrong
     */
    public ResourceIDImpl getResourceID(String path, boolean isCollection) throws RegistryException {
        // NEW
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            // step1 : need to check whether it is a resource or collection while retrieving path id
            PathCache pathCache = PathCache.getPathCache();

            int pathID = -1;
            String resourceName = null;
            if (isCollection) {
                pathID = pathCache.getPathID(conn, path);
            } else {
                // we have to reget the path id for the parent path..
                String parentPath;
                if (path.equals(RegistryConstants.ROOT_PATH)) {
                    parentPath = null; 
                } else {
                    parentPath = RegistryUtils.getParentPath(path);
                }
                resourceName = RegistryUtils.getResourceName(path);

                pathID = pathCache.getPathID(conn, parentPath);
            }

            if (pathID != -1) {
                ResourceIDImpl resourceID = new ResourceIDImpl();
                resourceID.setCollection(isCollection);
                resourceID.setName(resourceName);
                resourceID.setPathID(pathID);
                resourceID.setPath(path);
                return resourceID;
            }


        } catch (SQLException e) {
            String msg = "Failed to get ID of the resource at path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return null;
    }

    public boolean resourceExists(ResourceIDImpl resourceID) throws RegistryException {
        return getVersion(resourceID) != -1;
    }

    public boolean resourceExists(String path) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID == null) {
            // resource definitely doesn't exists
            return false;
        }
        if (resourceExists(resourceID)) {
            // resource definitely exist
            return true;
        }
        // if resourceID is a collection we have to check resourceID with non-collection as well..
        if (resourceID.isCollection()) {
            return resourceExists(path, false);
        }
        return false;
    }

    public boolean resourceExists(String path, boolean isCollection) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path, isCollection);
        return resourceID != null && resourceExists(resourceID);    
    }

    public long getVersion(ResourceIDImpl resourceID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        long version = -1;
        try {
            if (resourceID.isCollection()) {
                String sql = "SELECT REG_VERSION FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME IS NULL";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());

                result = ps.executeQuery();
                if (result.next()) {
                    version = result.getLong("REG_VERSION");
                }
            }
            else {
                String sql = "SELECT REG_VERSION FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());

                result = ps.executeQuery();
                if (result.next()) {
                    version = result.getLong("REG_VERSION");
                }
            }
        } catch (SQLException e) {
            String msg = "Failed to check the existence of the resource " +
                    resourceID.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return version;         
    }

    /**
     * Returns the resource in the given path filled with metadata and access to the content.
     * If a resource does not exist in the given path, null is returned.
     *
     * @param path Path of the resource.
     *
     * @return ResourceImpl filled with resource metadata and access to the resource content.
     * @throws org.wso2.carbon.registry.core.exceptions.RegistryException something went wrong
     */
    public ResourceImpl get(String path) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID == null) {
            return null;
        }
        return get(resourceID);
    }

    public ResourceImpl get(ResourceIDImpl resourceID) throws RegistryException {
        ResourceImpl resourceImpl = getResource(resourceID);
        if(resourceImpl == null) {
            // it is possible the resource doesn't exist
            return null;
        }
        fillResourceProperties(resourceImpl);
        return resourceImpl; 
    }


    public ResourceImpl getWithNoUpdate(ResourceIDImpl resourceID) throws RegistryException {
        ResourceImpl resourceImpl = getResourceWithNoUpdate(resourceID);
        if(resourceImpl == null) {
            // it is possible the resource doesn't exist
            return null;
        }
        fillResourcePropertiesWithNoUpdate(resourceImpl);
        return resourceImpl;
    }

    public CollectionImpl get(String path, int start, int pageLen) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID == null) {
            return null;
        }
        return get(resourceID, start, pageLen);
    }


    public CollectionImpl get(ResourceIDImpl resourceID, int start, int pageLen) throws RegistryException {
        CollectionImpl collection = (CollectionImpl)getResource(resourceID);
        if(collection == null) {
            // it is possible the resource doesn't exist
            return null;
        }
        fillChildren(collection, start, pageLen);
        fillResourceProperties(collection);
        return collection;
    }

    /**
     * Fill the content (for non-collection) and the properties for a resource that already filled meta data
     * @param resourceImpl
     * @throws RegistryException
     */
    public void fillResource(ResourceImpl resourceImpl) throws RegistryException {
        if (resourceImpl != null && !(resourceImpl instanceof CollectionImpl) &&
                        resourceImpl.getDbBasedContentID() > 0) {
            fillResourceContentWithNoUpdate(resourceImpl);
        }
        fillResourcePropertiesWithNoUpdate(resourceImpl);
    }

    /**
     * Fill the childrent and the properties for a resource that already filled meta data
     * @param resourceImpl
     * @param start
     * @param pageLen
     * @throws RegistryException
     */
    public void fillResource(ResourceImpl resourceImpl, int start, int pageLen) throws RegistryException {
        CollectionImpl collection = (CollectionImpl)resourceImpl;
        fillChildren(collection, start, pageLen);
        fillResourcePropertiesWithNoUpdate(resourceImpl);
    }


    public void fillResourcePropertiesWithNoUpdate(ResourceImpl resourceImpl) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            /*String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                    "WHERE P.REG_RID=RP.REG_PROPERTY_ID AND RP.REG_VERSION=?"; */
            if (StaticConfiguration.isVersioningProperties()) {
                String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                        "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_VERSION=?";
                ps = conn.prepareStatement(propSQL);
                ps.setLong(1, resourceImpl.getVersionNumber());
                results = ps.executeQuery();
            }
            else if (resourceImpl instanceof CollectionImpl) {
                String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                        "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                results = ps.executeQuery();
            }
            else {
                String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                        "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                results = ps.executeQuery();
            }

            while (results.next()) {
                String name = results.getString("REG_NAME");
                String value = results.getString("REG_VALUE");
                resourceImpl.addPropertyWithNoUpdate(name, value);
            }
        } catch (SQLException e) {
            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }


    public void fillResourceProperties(ResourceImpl resourceImpl) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            /*String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                    "WHERE P.REG_RID=RP.REG_PROPERTY_ID AND RP.REG_VERSION=?"; */
            if (StaticConfiguration.isVersioningProperties()) {
                String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                        "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_VERSION=?";
                ps = conn.prepareStatement(propSQL);
                ps.setLong(1, resourceImpl.getVersionNumber());
                results = ps.executeQuery();
            }
            else if (resourceImpl instanceof CollectionImpl) {
                String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                        "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                results = ps.executeQuery();
            }
            else {
                String propSQL = "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, REG_RESOURCE_PROPERTY RP " +
                        "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                results = ps.executeQuery();
            }

            while (results.next()) {
                String name = results.getString("REG_NAME");
                String value = results.getString("REG_VALUE");
                resourceImpl.addProperty(name, value);
            }
        } catch (SQLException e) {
            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
        resourceImpl.setPropertiesModified(false);
    }

    public void addRoot(String path, ResourceImpl resourceImpl)
            throws RegistryException {

        createAndApplyResourceID(path, null, resourceImpl);

        addResource(resourceImpl, false);
        addProperties(resourceImpl);
    }

    public void add(String path, ResourceIDImpl parentID, ResourceImpl resourceImpl)
            throws RegistryException {

        if (!(resourceImpl instanceof org.wso2.carbon.registry.core.Collection)) {
            addContent(resourceImpl);
        }

        // creating the resourceID
        createAndApplyResourceID(path, parentID, resourceImpl);

        addResource(resourceImpl, false);

        addProperties(resourceImpl);
    }

    public void createAndApplyResourceID(String path, ResourceIDImpl parentID, ResourceImpl resourceImpl)
                                throws RegistryException{
        ResourceIDImpl resourceID = createResourceID(path, parentID, resourceImpl instanceof CollectionImpl);
        resourceImpl.setName(resourceID.getName());
        resourceImpl.setPath(resourceID.getPath());
        resourceImpl.setPathID(resourceID.getPathID());
    }


    public ResourceIDImpl createResourceID(String path, ResourceIDImpl parentID, boolean isCollection)
                                throws RegistryException{
        ResourceIDImpl resourceID = new ResourceIDImpl();
        try {
            int parentPathID = -1;
            if(parentID != null) {
                parentPathID = parentID.getPathID();
            }
            if (isCollection) {
                resourceID.setName(null);
                resourceID.setPath(path);

                Connection conn = Transaction.getConnection();
                int pathID = PathCache.getPathCache().getPathID(conn, path);
                if (pathID == -1) {
                    pathID = PathCache.getPathCache().addEntry(path, parentPathID);
                }
                resourceID.setPathID(pathID);
            }
            else {
                String resourceName = RegistryUtils.getResourceName(path);
                resourceID.setName(resourceName);
                resourceID.setPath(path);
                resourceID.setPathID(parentPathID);
            }
        } catch (SQLException e) {
            String msg = "Failed to create the resource id for the resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        return resourceID;
    }


    public void deleteContentStream(int contentID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String resourceContentSQL = "DELETE FROM REG_CONTENT WHERE REG_CONTENT_ID = ?";
            ps  = conn.prepareStatement(resourceContentSQL);
            ps.setLong(1, contentID);
            ps.executeUpdate();

        }
        catch (SQLException ex)  {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg,ex);
            throw new RegistryException(msg,ex);
        }
        finally {
            try {
                if (ps != null) {ps.close();}
            } catch(SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }

    public InputStream getContentStream(int contentID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String resourceContentSQL = "SELECT REG_CONTENT_DATA  FROM  REG_CONTENT WHERE REG_CONTENT_ID = ?";
            ps  = conn.prepareStatement(resourceContentSQL);
            ps.setLong(1, contentID);
            result = ps.executeQuery();
            if (result.next()){
                InputStream rawInputStream = result.getBinaryStream("REG_CONTENT_DATA");
                if (rawInputStream != null) {
                    InputStream regData = RegistryUtils.getMemoryStream(rawInputStream);
                    return regData;
                }
            }

        }
        catch (SQLException ex)  {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg,ex);
            throw new RegistryException(msg,ex);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
            } catch(SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
        return null;
    }

    public void update(ResourceImpl resourceImpl) throws RegistryException {

        if (!(resourceImpl instanceof CollectionImpl)) {
            // adding content new
            addContent(resourceImpl);
        }
        // there is no difference of adding a resource and updating as it is always
        // a new entry to the resource table
        addResource(resourceImpl, true);

        // adding properties..
        addProperties(resourceImpl);
        
    }

    public int getChildCount(CollectionImpl collection, Connection conn)
            throws RegistryException {

        ResourceIDImpl resourceID = collection.getResourceIDImpl();
        ResultSet results1 = null;
        PreparedStatement ps1 = null;
        ResultSet results2 = null;
        PreparedStatement ps2 = null;


        int childCount = 0;
        try {
            // ---------------------------------------------------------------------------------------------------------
            // step1: get the child resources.
            String sql = "SELECT COUNT(R.REG_NAME) AS REG_RES_COUNT FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? ";

            ps1 = conn.prepareStatement(sql);
            ps1.setInt(1, resourceID.getPathID());
            results1 = ps1.executeQuery();

            if (results1.next()) {
                // here -1 is to avoid the parent collection itself
                childCount += results1.getInt("REG_RES_COUNT");
            }

            // step2: get the child collections
            sql = "SELECT COUNT(P.REG_PATH_ID) AS REG_RES_COUNT FROM REG_PATH P, REG_RESOURCE R " +
                    "WHERE P.REG_PATH_PARENT_ID=? AND R.REG_PATH_ID=P.REG_PATH_ID";
            ps2 = conn.prepareStatement(sql);
            ps2.setInt(1, resourceID.getPathID());

            results2 = ps2.executeQuery();

            if (results2.next()) {
                childCount += results2.getInt("REG_RES_COUNT");
            }

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    collection.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results1 != null) {results1.close();}
                if (ps1 != null) {ps1.close();}
                if (results2 != null) {results2.close();}
                if (ps2 != null) {ps2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
        return childCount;
    }


    /**
     * Return the author (the user who creates the resource) if a resource exists at given path.
     * If a resource does not exist, null is returned.
     *
     * @param path Path of the resource.
     * @return Author if the resourc exists. Null otherwiese.
     * @throws RegistryException On any database related exception.
     */
    public String getResourceAuthor(String path) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID != null) {
            ResourceDO resourceDO = getResourceDO(resourceID);
            if (resourceDO != null)
                return resourceDO.getAuthor();
        }
        return null;
    }

    public void fillChildren(CollectionImpl collection, int start, int pageLen)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        fillChildren(collection, start, pageLen, conn);
    }

    public void fillChildren(CollectionImpl collection, int start, int pageLen, Connection conn)
            throws RegistryException {
        String[] childPaths = getChildren(collection, start, pageLen, conn);
        collection.setContent(childPaths);
    }    

    public String[] getChildren(CollectionImpl collection, int start, int pageLen, Connection conn)
            throws RegistryException {

        ResourceIDImpl resourceID = collection.getResourceIDImpl();
        ResultSet results1 = null;
        PreparedStatement ps1 = null;
        ResultSet results2 = null;
        PreparedStatement ps2 = null;

        ResourcePath parentResourcePath = new ResourcePath(resourceID.getPath());
        String parentPath = parentResourcePath.getPath();
        try {
            // ---------------------------------------------------------------------------------------------------------
            // step1: get the child resources.
            String sql = "SELECT R.REG_NAME FROM REG_RESOURCE R WHERE R.REG_PATH_ID=?";

            ps1 = conn.prepareStatement(sql);
            ps1.setInt(1, resourceID.getPathID());
            results1 = ps1.executeQuery();

            List <String> childResourcesList = new ArrayList();
            List <String> childCollectionsList = new ArrayList();
            while (results1.next()) {
                String childName = results1.getString("REG_NAME");
                if(childName == null) {
                    // skip the null named resource (which is the parent resource
                    continue;
                }
                String childPath = parentPath +
                        (parentPath.equals(RegistryConstants.PATH_SEPARATOR)? "": RegistryConstants.PATH_SEPARATOR) +
                        childName;

                ResourceIDImpl childResourceID = new ResourceIDImpl();
                childResourceID.setPath(childPath);
                childResourceID.setCollection(false);
                childResourceID.setName(childName);
                childResourceID.setPathID(resourceID.getPathID());
                if (AuthorizationUtils.authorize(childPath, ActionConstants.GET)) {
                    childResourcesList.add(childPath);
                }
            }

            int end = start + pageLen;
            if (childResourcesList.size() < end || pageLen == -1) {
                // step2: get the child collections
                sql = "SELECT P.REG_PATH_ID, P.REG_PATH_VALUE FROM REG_PATH P, REG_RESOURCE R " +
                        "WHERE P.REG_PATH_PARENT_ID=? AND R.REG_PATH_ID=P.REG_PATH_ID AND R.REG_NAME IS NULL";
                ps2 = conn.prepareStatement(sql);
                ps2.setInt(1, resourceID.getPathID());

                results2 = ps2.executeQuery();

                while (results2.next()) {
                    String childPath = results2.getString("REG_PATH_VALUE");
                    int childPathId = results2.getInt("REG_PATH_ID");

                    ResourceIDImpl childResourceID = new ResourceIDImpl();
                    childResourceID.setPath(childPath);
                    childResourceID.setCollection(true);
                    childResourceID.setName(null);
                    childResourceID.setPathID(childPathId);
                    if (AuthorizationUtils.authorize(childPath, ActionConstants.GET)) {
                        childCollectionsList.add(childPath);
                    }
                }
            }

            int totalSize = childResourcesList.size() + childCollectionsList.size();
            if (totalSize < start) {
                throw new RegistryException("Didn't have enough results to start at #" + start);
            }
            Collections.sort(childResourcesList);
            Collections.sort(childCollectionsList);

            if(pageLen == -1) {
                pageLen = totalSize - start;
                end = totalSize-1;
            }

            int resultMaxSize = pageLen < (totalSize - start)? pageLen:(totalSize - start) ;
            String[] childPaths = new String[resultMaxSize];
            // copying child resources
            int current = 0;
            int i;
            for (i = 0; i < childResourcesList.size(); i ++) {
                if (current >= resultMaxSize) {
                    break;
                }
                if (i >= start && i <= end) {
                    childPaths[current] = childResourcesList.get(i);
                    current ++;
                }
            }
            // copying child collections
            for(int j = 0; j < childCollectionsList.size(); j ++, i ++) {
                if (current >= resultMaxSize) {
                    break;
                }
                if (i >= start && i <= end) {
                    childPaths[current] = childCollectionsList.get(j);
                    current ++;
                }
            }
            return childPaths;

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    collection.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results1 != null) {results1.close();}
                if (ps1 != null) {ps1.close();}
                if (results2 != null) {results2.close();}
                if (ps2 != null) {ps2.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }


    public ResourceImpl getResourceMetaData(String path) throws RegistryException {
        ResourceIDImpl resourceIDImpl = getResourceID(path);
        ResourceImpl resourceImpl = null;
        if (resourceIDImpl != null) {
            resourceImpl = getResourceMetaData(resourceIDImpl);

            if (resourceImpl == null && resourceIDImpl.isCollection()) {
                // we should check the resourceID for a non-collection too.
                resourceIDImpl = getResourceID(path, false);
                if (resourceIDImpl != null) {
                    resourceImpl = getResourceMetaData(resourceIDImpl);
                }
            }
        }
        return resourceImpl;
    }


    public ResourceImpl getResourceMetaData(ResourceIDImpl resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResourceImpl resourceImpl = null;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            boolean isCollection = resourceID.isCollection();
            String sql;
            int pathID = resourceID.getPathID();
            String resourceName = resourceID.getName();
            if (isCollection) {
               // collection
                sql = "SELECT REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                        "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_VERSION, REG_DESCRIPTION, REG_CONTENT_ID " +
                        "FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, pathID);
            }
            else {
                // non collection
                sql = "SELECT REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                        "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_VERSION, REG_DESCRIPTION, REG_CONTENT_ID " +
                        "FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME = ? ";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, pathID);
                ps.setString(2, resourceName);
            }
            
            result = ps.executeQuery();

            if (result.next()) {

                if (isCollection) {
                    resourceImpl = new CollectionImpl();
                } else {
                    resourceImpl = new ResourceImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setPathID(resourceID.getPathID());
                resourceImpl.setPath(resourceID.getPath());
                resourceImpl.setName(resourceID.getName());
                resourceImpl.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceImpl.setAuthorUserName(result.getString("REG_CREATOR"));
                resourceImpl.setCreatedTime(result.getTimestamp("REG_CREATED_TIME"));
                resourceImpl.setLastUpdaterUserName(result.getString("REG_LAST_UPDATOR"));
                resourceImpl.setLastModified(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceImpl.setVersionNumber(result.getInt("REG_VERSION"));
                resourceImpl.setDescription(result.getString("REG_DESCRIPTION"));
                resourceImpl.setDbBasedContentID(result.getInt("REG_CONTENT_ID"));
            }
        } catch (SQLException e) {
            String msg = "Failed to get the resource at path " + resourceID.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }

        return resourceImpl;
    }

    private ResourceImpl getResource(ResourceIDImpl resourceID) throws RegistryException {
        ResourceImpl resourceImpl = getResourceMetaData(resourceID);

        if (resourceImpl != null && !(resourceImpl instanceof CollectionImpl) &&
                        resourceImpl.getDbBasedContentID() > 0) {
            fillResourceContent(resourceImpl);
        }
        return resourceImpl;
    }

    public ResourceImpl getResourceWithNoUpdate(ResourceIDImpl resourceID) throws RegistryException {
        ResourceImpl resourceImpl = getResourceMetaData(resourceID);

        if (resourceImpl != null && !(resourceImpl instanceof CollectionImpl) &&
                        resourceImpl.getDbBasedContentID() > 0) {
            fillResourceContentWithNoUpdate(resourceImpl);
        }
        return resourceImpl;
    }

    public void fillResourceContent(ResourceImpl resourceImpl) throws RegistryException {
        int contentId = resourceImpl.getDbBasedContentID();
        InputStream contentStream = getContentStream(contentId);
        if (contentStream == null) {
            resourceImpl.setContent(null);
        }
        else {
            resourceImpl.setContentStream(contentStream);
        }
    }


    public void fillResourceContentWithNoUpdate(ResourceImpl resourceImpl) throws RegistryException {
        int contentId = resourceImpl.getDbBasedContentID();
        InputStream contentStream = getContentStream(contentId);
        if (contentStream == null) {
            resourceImpl.setContentWithNoUpdate(null);
        }
        else {
            resourceImpl.setContentStreamWithNoUpdate(contentStream);
        }
    }

    public void addResource(ResourceImpl resourceImpl, boolean isUpdatingExisting)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
        try {
            String sql = "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, REG_CREATOR, " +
                                "REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                                "REG_CONTENT_ID) " +
                                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";


            long now = System.currentTimeMillis();

            ps = conn.prepareStatement(sql,  new String[] {"REG_VERSION"});

            ps.setInt(1, resourceID.getPathID());
            ps.setString(2, resourceID.getName());
            ps.setString(3, resourceImpl.getMediaType());
            if (isUpdatingExisting) {
                String authorName = resourceImpl.getAuthorUserName();
                if(authorName == null) {
                    authorName = CurrentSession.getUser();
                    resourceImpl.setAuthorUserName(authorName);
                }
                ps.setString(4, authorName);

                Date createdTime = resourceImpl.getCreatedTime();
                Timestamp createdTimestamp;
                if (createdTime == null) {
                    createdTimestamp = new Timestamp(now);
                }
                else {
                    createdTimestamp = new Timestamp(createdTime.getTime());
                }
                ps.setTimestamp(5, createdTimestamp);
            }
            else {
                ps.setString(4, CurrentSession.getUser());
                resourceImpl.setAuthorUserName(CurrentSession.getUser());
                ps.setTimestamp(5, new Timestamp(now));
            }
            ps.setString(6, CurrentSession.getUser());
            ps.setTimestamp(7, new Timestamp(now));
            ps.setString(8, resourceImpl.getDescription());
            if(resourceImpl.getDbBasedContentID() > 0) {
                ps.setInt(9, resourceImpl.getDbBasedContentID());
            }
            else {
                ps.setNull(9, Types.INTEGER);
            }

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if(result.next()) {
                long version = result.getLong(1);
                resourceImpl.setVersionNumber(version);
            }

            // Update state of the ResourceImpl
            resourceImpl.setLastModified(new Date(now));

        } catch (SQLException e) {
            String msg = "Failed to add resource to path " + resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }



    public void addResourceWithNoUpdate(ResourceImpl resourceImpl)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
        try {
            String sql = "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, REG_CREATOR, " +
                                "REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                                "REG_CONTENT_ID) " +
                                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";

            ps = conn.prepareStatement(sql,  new String[] {"REG_VERSION"});

            ps.setInt(1, resourceID.getPathID());
            ps.setString(2, resourceID.getName());
            ps.setString(3, resourceImpl.getMediaType());
            ps.setString(4, resourceImpl.getAuthorUserName());

            Date createdTime = resourceImpl.getCreatedTime();
            ps.setTimestamp(5, new Timestamp(createdTime.getTime()));

            ps.setString(6, resourceImpl.getLastUpdaterUserName());

            Date updatedTime = resourceImpl.getLastModified();
            ps.setTimestamp(7, new Timestamp(updatedTime.getTime()));
            
            ps.setString(8, resourceImpl.getDescription());
            if(resourceImpl.getDbBasedContentID() > 0) {
                ps.setInt(9, resourceImpl.getDbBasedContentID());
            }
            else {
                ps.setNull(9, Types.INTEGER);
            }

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if(result.next()) {
                long version = result.getLong(1);
                resourceImpl.setVersionNumber(version);
            }

        } catch (SQLException e) {
            String msg = "Failed to add resource to path " + resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void addResourceDO(ResourceDO resourceDO)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;

        try {
            String sql = "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, REG_CREATOR, " +
                                "REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                                "REG_CONTENT_ID) " +
                                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";

            ps = conn.prepareStatement(sql,  new String[] {"REG_VERSION"});

            ps.setInt(1, resourceDO.getPathID());
            ps.setString(2, resourceDO.getName());
            ps.setString(3, resourceDO.getMediaType());
            ps.setString(4, resourceDO.getAuthor());
            ps.setTimestamp(5, resourceDO.getCreatedOn());
            ps.setString(6, resourceDO.getLastUpdator());
            ps.setTimestamp(7, resourceDO.getLastUpdatedOn());
            ps.setString(8, resourceDO.getDescription());
            if(resourceDO.getContentID() > 0) {
                ps.setInt(9, resourceDO.getContentID());
            }
            else {
                ps.setNull(9, Types.INTEGER);
            }

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if(result.next()) {
                long version = result.getLong(1);
                resourceDO.setVersion(version);
            }

            ps.close();

        } catch (SQLException e) {
            String msg = "Failed to add resource to version " + resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void deleteResource(ResourceDO resourceDO) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (resourceDO.getName() == null) {
                String sql = "DELETE FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME IS NULL";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());

                ps.executeUpdate();
            }
            else {
                String sql = "DELETE FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
                ps.setString(2, resourceDO.getName());

                ps.executeUpdate();
            }

        } catch (SQLException e) {

            String msg = "Failed to delete the resource with id " + resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {ps.close();}
            }
            catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }

    public void addProperties(ResourceImpl resource) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps1 = null;
        PreparedStatement ps2 = null;
        try {

            String sql1, sql2;
            boolean isVersioningProperties = StaticConfiguration.isVersioningProperties();

            sql1 = "INSERT INTO REG_PROPERTY (REG_NAME, REG_VALUE) VALUES (?, ?)";
            if (isVersioningProperties) {
                sql2 = "INSERT INTO REG_RESOURCE_PROPERTY (REG_PROPERTY_ID, REG_VERSION) VALUES (?, ?)";
            }
            else {
                sql2 = "INSERT INTO REG_RESOURCE_PROPERTY (REG_PROPERTY_ID, REG_PATH_ID, REG_RESOURCE_NAME) VALUES (?, ?, ?)";
            }
            
            ResourceIDImpl resourceID = resource.getResourceIDImpl();
            Properties props = resource.getProperties();

            if (props != null) {
                Iterator i = props.keySet().iterator();
                while (i.hasNext()) {
                    String name = (String)i.next();

                    List propValues = (List) props.get(name);
                    if (propValues != null) {
                        Iterator<String> iValues = propValues.iterator();
                        while (iValues.hasNext()) {
                            String value = iValues.next();

                            ps1 = conn.prepareStatement(sql1,  new String[] {"REG_ID"});
                            ps2 = conn.prepareStatement(sql2);
                            
                            // prepare to execute query1 for the property
                            ps1.setString(1, name);
                            ps1.setString(2, value);
                            ps1.executeUpdate();

                            ResultSet resultset1 = ps1.getGeneratedKeys();
                            if (resultset1.next()) {
                                // setting the property id
                                int propertyId = resultset1.getInt(1);
                                ps2.setInt(1, propertyId);
                                if (isVersioningProperties) {
                                    ps2.setLong(2, resource.getVersionNumber());
                                }
                                else {
                                    ps2.setLong(2, resourceID.getPathID());
                                    if (resourceID.getName() == null) {
                                        ps2.setNull(3, Types.VARCHAR);
                                    }
                                    else {
                                        ps2.setString(3, resourceID.getName());
                                    }
                                }
                                ps2.executeUpdate();
                            }
                            if (ps1 != null) {
                                ps1.close();
                                ps1 = null;
                            }
                            if (ps2 != null) {
                                ps2.close();
                                ps2 = null;
                            }
                        }
                    }
                }
            }
        } catch (SQLException e) {

            String msg = "Failed to add properties to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }  finally {
            try {
                if (ps1 != null) {ps1.close();}
                if (ps2 != null) {ps2.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }

    private Integer[] getPropertyIds(ResourceDO resourceDO) throws RegistryException{

        Connection conn = Transaction.getConnection();

        List<Integer> propertyIds = new ArrayList<Integer>();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            String sql;
            if (StaticConfiguration.isVersioningProperties()) {
                sql = "SELECT REG_PROPERTY_ID FROM REG_RESOURCE_PROPERTY WHERE "+
                      "REG_VERSION=?";

                ps = conn.prepareStatement(sql);
                ps.setLong(1, resourceDO.getVersion());

            }
            else if(resourceDO.getName() == null) {
                // a collection
                sql = "SELECT REG_PROPERTY_ID FROM REG_RESOURCE_PROPERTY WHERE "+
                      "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
            }
            else {
                sql = "SELECT REG_PROPERTY_ID FROM REG_RESOURCE_PROPERTY WHERE "+
                      "REG_PATH_ID=? AND REG_RESOURCE_NAME=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
                ps.setString(2, resourceDO.getName());
            }

            results = ps.executeQuery();
            while (results.next()) {
                int propertyId = results.getInt("REG_PROPERTY_ID");
                propertyIds.add(propertyId);
            }

        } catch (SQLException e) {

            String msg = "Failed to get properties on resource version" +
                    resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {results.close();}
                if (ps != null) {ps.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }

        return propertyIds.toArray(new Integer[propertyIds.size()]);
    }


    public void removeProperties(ResourceDO resourceDO) throws RegistryException {
            Connection conn = Transaction.getConnection();

        PreparedStatement ps1 = null, ps2= null;

        Integer[] propertyIds = getPropertyIds(resourceDO);
        if(propertyIds == null) {
            return;
        }
        try {

            String sql = "DELETE FROM REG_RESOURCE_PROPERTY WHERE REG_PROPERTY_ID= ?";
            ps1 = conn.prepareStatement(sql);

            sql = "DELETE FROM REG_PROPERTY WHERE REG_ID= ?";
            ps2 = conn.prepareStatement(sql);

            for(Integer propertyId: propertyIds) {
                ps1.setInt(1, propertyId);
                ps2.setInt(1, propertyId);
                ps1.addBatch();
                ps2.addBatch();
            }

            if(propertyIds.length > 0) {
                try {
                    ps1.executeBatch();
                    ps2.executeBatch();
                } catch (SQLException e) {
                    ps1.clearBatch();
                    ps2.clearBatch();
                    // the exception will be handled in the next catch block
                    throw e;
                }
            }
        } catch (SQLException e) {

            String msg = "Failed to remove properties from resource version " +
                        resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) {
                    ps1.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void addContent(ResourceImpl resourceImpl) throws RegistryException {

        if (resourceImpl.getContent() == null || !(resourceImpl.getContent() instanceof byte[])) {
            if (log.isDebugEnabled()) {
                String msg = "Content of the resource " + resourceImpl.getPath() +
                        " is null or not a byte array. Content will not be persisted to the database.";
                log.debug(msg);
            }
            resourceImpl.setDbBasedContentID(0);
            return;
        }

        InputStream contentStream = resourceImpl.getContentStream();
        int contentID = addContentBytes(contentStream);
        resourceImpl.setDbBasedContentID(contentID);
    }

    public int addContentBytes(InputStream contentStream) throws RegistryException {

        Connection conn = Transaction.getConnection();
        int contentID = -1;

        try {
            String sql = "INSERT INTO REG_CONTENT (REG_CONTENT_DATA) VALUES (?)";

            int size = contentStream.available();
            PreparedStatement ps = conn.prepareStatement(sql,  new String[] {"REG_CONTENT_ID"});
            ps.setBinaryStream(1, contentStream, size);
            ps.executeUpdate();

            ResultSet result = ps.getGeneratedKeys();
            if (result.next()) {
                contentID = result.getInt(1);
            }
            ps.close();

        } catch (Exception e) {

            String msg = "Failed to write resource content to the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        return contentID;
    }

    public ResourceDO getResourceDO(long version) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {

            String sql = "SELECT REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                    "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID " +
                    "FROM REG_RESOURCE WHERE REG_VERSION = ?";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, version);

            ResourceDO resourceDO = null;
            result = ps.executeQuery();
            if (result.next()) {
                resourceDO = new ResourceDO();

                resourceDO.setPathID(result.getInt("REG_PATH_ID"));
                resourceDO.setName(result.getString("REG_NAME"));
                resourceDO.setVersion(version);
                resourceDO.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceDO.setAuthor(result.getString("REG_CREATOR"));
                resourceDO.setCreatedOn(result.getTimestamp("REG_CREATED_TIME"));
                resourceDO.setLastUpdator(result.getString("REG_LAST_UPDATOR"));
                resourceDO.setLastUpdatedOn(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceDO.setDescription(result.getString("REG_DESCRIPTION"));
                resourceDO.setContentID(result.getInt("REG_CONTENT_ID"));
            }

            return resourceDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource version:" +
                    version + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }


    public ResourceDO getResourceDO(ResourceIDImpl resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql = null;
            if(resourceID.isCollection()) {
                sql = "SELECT  R.REG_PATH_ID, R.REG_NAME, R.REG_VERSION, R.REG_MEDIA_TYPE, " +
                    "R.REG_CREATOR, R.REG_CREATED_TIME, R.REG_LAST_UPDATOR, R.REG_LAST_UPDATED_TIME, " +
                    "R.REG_DESCRIPTION, R.REG_CONTENT_ID " +
                    "FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? AND R.REG_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());

            }
            else {
                sql = "SELECT  R.REG_PATH_ID, R.REG_NAME, R.REG_VERSION, R.REG_MEDIA_TYPE, " +
                    "R.REG_CREATOR, R.REG_CREATED_TIME, R.REG_LAST_UPDATOR, R.REG_LAST_UPDATED_TIME, " +
                    "R.REG_DESCRIPTION, R.REG_CONTENT_ID " +
                    "FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? AND R.REG_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
            }

            ResourceDO resourceDO = null;
            result = ps.executeQuery();
            if (result.next()) {
                resourceDO = new ResourceDO();

                resourceDO.setPathID(result.getInt("REG_PATH_ID"));
                resourceDO.setName(result.getString("REG_NAME"));
                resourceDO.setVersion(result.getInt("REG_VERSION"));
                resourceDO.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceDO.setAuthor(result.getString("REG_CREATOR"));
                resourceDO.setCreatedOn(result.getTimestamp("REG_CREATED_TIME"));
                resourceDO.setLastUpdator(result.getString("REG_LAST_UPDATOR"));
                resourceDO.setLastUpdatedOn(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceDO.setDescription(result.getString("REG_DESCRIPTION"));
                resourceDO.setContentID(result.getInt("REG_CONTENT_ID"));
            }

            return resourceDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally{
        	if(result != null){
        		try {
					result.close();
				} catch (SQLException e) {
					log.error(RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR,e);
				}
        	}
        	if(ps != null){
        		try {
					ps.close();
				} catch (SQLException e) {
					log.error(RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR,e);
				}
        	}
        }
    }

    public ArrayList <ResourceIDImpl> getChildPathIds(ResourceIDImpl resourceID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        ArrayList <ResourceIDImpl> childPathList = new ArrayList();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            // ---------------------------------------------------------------------------------------------------------
            // step1: get the child resources.
            String sql = "SELECT R.REG_NAME FROM REG_RESOURCE R WHERE R.REG_PATH_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setInt(1, resourceID.getPathID());
            results = ps.executeQuery();


            int current = 0;
            while (results.next()) {
                String childName = results.getString("REG_NAME");
                if(childName == null) {
                    // skip the parent resource..
                    continue;
                }
                String parentPath = resourceID.getPath();
                String childPath = parentPath +
                        (parentPath.equals(RegistryConstants.PATH_SEPARATOR)? "": RegistryConstants.PATH_SEPARATOR) +
                        childName;
                ResourceIDImpl childResourceID = new ResourceIDImpl();
                childResourceID.setPath(childPath);
                childResourceID.setCollection(false);
                childResourceID.setName(childName);
                childResourceID.setPathID(resourceID.getPathID());
                childPathList.add(childResourceID);
            }
            if (results != null) {
                results.close();
                results = null;
            }
            if (ps != null) {
                ps.close();
                ps = null;
            }

            sql = "SELECT P.REG_PATH_ID, P.REG_PATH_VALUE FROM REG_PATH P, REG_RESOURCE R " +
                    "WHERE P.REG_PATH_PARENT_ID=? AND R.REG_PATH_ID=P.REG_PATH_ID AND R.REG_NAME IS NULL";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, resourceID.getPathID());

            results = ps.executeQuery();

            while (results.next()) {
                String childPath = results.getString("REG_PATH_VALUE");
                int childPathId = results.getInt("REG_PATH_ID");

                ResourceIDImpl childResourceID = new ResourceIDImpl();
                childResourceID.setPath(childPath);
                childResourceID.setCollection(true);
                childResourceID.setName(null);
                childResourceID.setPathID(childPathId);
                childPathList.add(childResourceID);
            }
            if (results != null) {
                results.close();
                results = null;
            }
            if (ps != null) {
                ps.close();
                ps = null;
            }

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceID.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally{
        	if(results != null){
        		try {
					results.close();
				} catch (SQLException e) {
					log.error(RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR,e);
				}
        	}
        	if(ps != null){
        		try {
					ps.close();
				} catch (SQLException e) {
					log.error(RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR,e);
				}
        	}
        }
        return childPathList;
    }

    public String getPathFromId(int pathId) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql = "SELECT REG_PATH_VALUE FROM REG_PATH WHERE REG_PATH_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setInt(1, pathId);
            
            result = ps.executeQuery();
            if(result.next()) {
                String path = result.getString("REG_PATH_VALUE");

                return path;
            }
            
         } catch (SQLException e) {

            String msg = "Failed to get the path for the path id " +
                    pathId + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return null;
    }

    public String getPath(long version) throws RegistryException{
        ResourceDO resourceDO = getResourceDO(version);
        if(resourceDO == null) {
            return null;
        }
        return getPath(resourceDO.getPathID(), resourceDO.getName(), false);
    }

    public String getPath(int pathId, String name, boolean checkExistance) throws RegistryException{
        String pathCollection = getPathFromId(pathId);
        if(pathCollection == null) {
            return null;
        }
        String resourceName = name;
        String correctPath = null;
        if(resourceName == null) {
            // this is the parent path..
            correctPath = pathCollection;
        }
        else if (pathCollection.equals(RegistryConstants.ROOT_PATH)) {
            correctPath = pathCollection + resourceName;   
        }
        else {
            correctPath = pathCollection + RegistryConstants.PATH_SEPARATOR + resourceName;
        }

        // need to check the existance of the path, otherwise return null,
        if (checkExistance) {
            ResourceIDImpl resourceID = new ResourceIDImpl();
            resourceID.setCollection(resourceName == null);
            resourceID.setName(resourceName);
            resourceID.setPath(correctPath);
            resourceID.setPathID(pathId);
            ResourceDO resourceDO = getResourceDO(resourceID);
            if (resourceDO == null) {
                return null;
            }
        }
        return correctPath;
    }


    public void moveResources(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE SET REG_PATH_ID=? WHERE " +
                             "REG_PATH_ID=? AND REG_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.executeUpdate();
            }
            else {
                String sql = "UPDATE REG_RESOURCE SET REG_PATH_ID=?, REG_NAME =? WHERE " +
                             "REG_PATH_ID=? AND REG_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move resource from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }


        /**
     * this function will move the paths from one path id to anthor regardless of the resource name
     * @param source
     * @param target
     * @throws RegistryException
     */
    public void moveResourcePaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE SET REG_PATH_ID=? WHERE REG_PATH_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move resource paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }

    public void moveProperties(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        if (StaticConfiguration.isVersioningProperties()) {
            return;
        }
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE_PROPERTY SET REG_PATH_ID=? WHERE " +
                             "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.executeUpdate();
            }
            else {
                String sql = "UPDATE REG_RESOURCE_PROPERTY SET REG_PATH_ID=?, REG_RESOURCE_NAME=? WHERE " +
                             "REG_PATH_ID=? AND REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move properties from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

        /**
     * this function will move the paths from one path id to anthor regardless of the resource name
     * @param source
     * @param target
     * @throws RegistryException
     */
    public void movePropertyPaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        if (StaticConfiguration.isVersioningProperties()) {
            return;
        }
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE_PROPERTY SET REG_PATH_ID=? WHERE REG_PATH_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move property paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }
}
