/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.jdbc.utils.VersionRetriever;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.StaticConfiguration;

import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.sql.*;
import java.util.*;

public class ResourceVersionDAO {

    private static Log log = LogFactory.getLog(ResourceVersionDAO.class);
    private ResourceDAO resourceDAO;
    private CommentsDAO commentsDAO;
    private RatingsDAO ratingsDAO;
    private TagsDAO tagsDAO;

    public ResourceVersionDAO() {
        resourceDAO = new ResourceDAO();
        commentsDAO = StaticConfiguration.getCommentsDAO();
        ratingsDAO = StaticConfiguration.getRatingsDAO();
        tagsDAO = StaticConfiguration.getTagsDAO();
    }

    public Long[] getSnapshotIDs(String resourcePath) throws RegistryException {
        Connection conn = Transaction.getConnection();

        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            ResourceIDImpl resourceIdImpl = resourceDAO.getResourceID(resourcePath);

            if(resourceIdImpl.isCollection()) {
                String sql = "SELECT REG_SNAPSHOT_ID FROM REG_SNAPSHOT WHERE " +
                        "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceIdImpl.getPathID());
            }
            else {
                String sql = "SELECT REG_SNAPSHOT_ID FROM REG_SNAPSHOT WHERE " +
                        "REG_PATH_ID=? AND REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceIdImpl.getPathID());
                ps.setString(2, resourceIdImpl.getName());
            }

            results = ps.executeQuery();
            List<Long> snapshotIDs = new ArrayList();
            while (results.next()) {
                long snapshotNumber = results.getLong("REG_SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }
            Collections.sort(snapshotIDs, Collections.reverseOrder());

            return snapshotIDs.toArray(new Long[snapshotIDs.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }


    public void fillResourceContentArchived(ResourceImpl resourceImpl) throws RegistryException {
        Connection conn = Transaction.getConnection();
        ResultSet result1 = null;
        ResultSet result2 = null;
        PreparedStatement ps1 = null;
        PreparedStatement ps2 = null;
        try {
            // we can't use the UNION sql opreator as some databases (derby) doesn't support
            // UNION with BLOB data
            // first check the current resource table
            String resourceContentSQL = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_HISTORY WHERE REG_CONTENT_ID = ? ";
            ps1  = conn.prepareStatement(resourceContentSQL);
            int contentId = resourceImpl.getDbBasedContentID();
            ps1.setInt(1, contentId);
            result1 = ps1.executeQuery();
            if (result1.next()){
                resourceImpl.setContentStreamWithNoUpdate(
                        RegistryUtils.getMemoryStream(result1.getBinaryStream("REG_CONTENT_DATA")));
            }
        }
        catch (SQLException ex)  {
            String msg = "Failed in filling resource content for resource " + resourceImpl.getPath() + ex.getMessage();
            log.error(msg,ex);
            throw new RegistryException(msg,ex);
         }
        finally {
            try {
                if (result1 != null) {result1.close();}
                if (ps1 != null) {ps1.close();}
                if (result2 != null) {result2.close();}
                if (ps2 != null) {ps2.close();}
            }
            catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }

    public ResourceImpl get(ResourceIDImpl resourceID, long snapshotID) throws RegistryException {
        VersionRetriever versionRetriever = getVersionList(snapshotID);
        ResourceDO resourceDO = null;
        int versionIndex = 0;
        while (true) {
            long version = versionRetriever.getVersion(versionIndex);
            if (version == -1) {
                // stream is over..
                break;
            }
            resourceDO = getResourceDOArchived(version);
            if(resourceDO.getPathID() == resourceID.getPathID() &&
                ((resourceID.isCollection() && resourceDO.getName() == null) ||
                 (resourceID.getName() != null && resourceID.getName().equals(resourceDO.getName())))) {
                break;
            }
            resourceDO = null;
            versionIndex ++;
        }
        if(resourceDO == null) {
            String msg = "The resource coudn't find for . " +
                    resourceID.getPath() + " for the snapshot " + snapshotID + ".";
            log.debug(msg);
            return null;
        }
        ResourceImpl resourceImpl;
        if(resourceID.isCollection()) {
            CollectionVersionImpl collectionImpl = new CollectionVersionImpl(resourceID.getPath(), resourceDO);
            collectionImpl.setVersionListIndex(versionIndex);
            collectionImpl.setVersionList(versionRetriever);
            resourceImpl = collectionImpl;
        }
        else {
            resourceImpl = new ResourceImpl(resourceID.getPath(), resourceDO);
        }
        return resourceImpl;
    }

    public VersionRetriever getVersionList(ResourceIDImpl resourceID, long snapshotID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        VersionRetriever versionRetriever = null;
        try {
            if(resourceID.isCollection()) {
                String sql = "SELECT REG_PATH_ID, REG_RESOURCE_VIDS FROM REG_SNAPSHOT WHERE REG_SNAPSHOT_ID=? " +
                                "AND REG_PATH_ID = ? AND REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setLong(1, snapshotID);
                ps.setInt(2, resourceID.getPathID());
            }
            else {
                String sql = "SELECT REG_PATH_ID, REG_RESOURCE_VIDS FROM REG_SNAPSHOT WHERE REG_SNAPSHOT_ID=? " +
                        "AND REG_PATH_ID = ? AND REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setLong(1, snapshotID);
                ps.setInt(2, resourceID.getPathID());
                ps.setString(3, resourceID.getName());
            }

            result = ps.executeQuery();

            if (result.next()){
                InputStream resourceVIDStream =  RegistryUtils.getMemoryStream(result.getBinaryStream("REG_RESOURCE_VIDS"));
                versionRetriever = new VersionRetriever(resourceVIDStream);
            }
        } catch (Exception e) {

            String msg = "Failed to get version of resource " + resourceID.getPath() +
                    " of snapshot " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return versionRetriever;
    }


    public VersionRetriever getVersionList(long snapshotID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        VersionRetriever versionRetriever = null;
        try {
            String sql = "SELECT REG_PATH_ID, REG_RESOURCE_VIDS FROM REG_SNAPSHOT WHERE REG_SNAPSHOT_ID=? ";
            ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);

            result = ps.executeQuery();

            if (result.next()){
                InputStream resourceVIDStream =  RegistryUtils.getMemoryStream(
                            result.getBinaryStream("REG_RESOURCE_VIDS"));
                versionRetriever = new VersionRetriever(resourceVIDStream);
            }
        } catch (Exception e) {

            String msg = "Failed to get version of the snapshot " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return versionRetriever;
    }


    public CollectionImpl get(ResourceIDImpl resourceID, long snapshotID,
                                    int start, int pageLen) throws RegistryException {
        if (!resourceID.isCollection()) {
            String msg = "Child resource range can only be specified for collections. " +
                    resourceID.getPath() + " is not a collection.";
            log.error(msg);
            throw new RegistryException(msg);
        }
        
        VersionRetriever versionRetriever = getVersionList(snapshotID);
        int versionIndex = 0;
        ResourceDO resourceDO = null;
        while (true) {
            long version = versionRetriever.getVersion(versionIndex);
            if (version == -1) {
                // no more stream
                break;
            }
            resourceDO = getResourceDOArchived(version);
            if(resourceDO.getPathID() == resourceID.getPathID() &&
                ((resourceID.isCollection() && resourceDO.getName() == null) ||
                 (resourceID.getName() != null && resourceID.getName().equals(resourceDO.getName())))) {
                break;
            }
            resourceDO = null;
            versionIndex ++;
        }
        if(resourceDO == null) {
            String msg = "The resource coudn't find for . " +
                    resourceID.getPath() + " for the snapshot " + snapshotID + ".";
            log.debug(msg);
            return null;
        }
        CollectionVersionImpl collectionImpl = new CollectionVersionImpl(resourceID.getPath(), resourceDO);
        collectionImpl.setVersionListIndex(versionIndex);
        collectionImpl.setVersionList(versionRetriever);

        fillChildren(collectionImpl, versionRetriever, versionIndex, start, pageLen, snapshotID);
        resourceDAO.fillResourceProperties(collectionImpl);
        
        return collectionImpl;
    }

    public void fillChildren(CollectionImpl collectionImpl, VersionRetriever versionRetriever,
                             int parentVersionIndex, int start, int pageLen, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        String[] childPaths = getChildPaths(collectionImpl.getResourceIDImpl(),
                                versionRetriever, parentVersionIndex, start, pageLen, snapshotID, conn);
        collectionImpl.setContent(childPaths);
    }

    public int getChildCount(String resourceID, long versionNumber, Connection conn)
            throws RegistryException {

        ResultSet results = null;
        PreparedStatement ps = null;
        try {

            int childCount = 0;

            String sql = "SELECT REG_CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE REG_PARENT_RID=? AND REG_PARENT_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, versionNumber);

            results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("REG_CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    childCount++;
                }
            }

            return childCount;

        } catch (SQLException e) {

            String msg = "Failed to get child count of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }


    public String[] getChildPaths(ResourceIDImpl resourceID, VersionRetriever versionRetriever, int parentVersionIndex,
                     int start, int pageLen,
                     long snapshotID, Connection conn) throws RegistryException {

        List <String> childPathList = new ArrayList();
        PreparedStatement ps = null;
        PreparedStatement ps2 = null;
        ResultSet result = null;
        ResultSet result2 = null;

        String parentPath = getCurrentPath(resourceID.getPath());


        // we have the versionRetriever of the descendents collection, need to figure out upto which one the collections
        // are immediate child of the collectionImpl
        // skipping the index 0 to skip the itself
        try {
            int current = 0;
            int end = start + pageLen;
            boolean isValidPath = false;
            int versionIndex = parentVersionIndex + 1;
            while (true) {
                // make sure we are within limit
                if ((pageLen != -1 && current > end)) {
                    break;
                }
                
                long version = versionRetriever.getVersion(versionIndex);
                if(version == -1) {
                    // stream is over..
                    break;
                }
                String sql = "SELECT REG_PATH_ID, REG_NAME FROM REG_RESOURCE WHERE REG_VERSION=? " +
                             "UNION " +
                             "SELECT REG_PATH_ID, REG_NAME FROM REG_RESOURCE_HISTORY WHERE REG_VERSION=? ";

                ps = conn.prepareStatement(sql);
                ps.setLong(1, version);
                ps.setLong(2, version);
                result = ps.executeQuery();

                if (result.next()) {
                    int pathId = result.getInt("REG_PATH_ID");
                    String resourceName = result.getString("REG_NAME");
                    if (pathId == resourceID.getPathID() && resourceName != null) {
                        // this should be a child resource
                        String childPath = parentPath + 
                        (parentPath.equals(RegistryConstants.PATH_SEPARATOR)? "": RegistryConstants.PATH_SEPARATOR) +
                                resourceName +
                                ";version:" + snapshotID;
                        if(current >= start) {
                            childPathList.add(childPath);
                        }
                        isValidPath = true;
                        current ++;
                    }
                    else if (resourceName == null){
                        // could be child resources, was replaced to this from
                        // SELECTED REG_PATH_VALUE FROM REG_PATH WHERE REG_PATH_ID=? AND REG_PARENT_PATH_ID=?"
                        // to avoid another cobination of indexes
                        sql = "SELECT REG_PATH_PARENT_ID, REG_PATH_VALUE FROM REG_PATH WHERE REG_PATH_ID=?";

                        ps2 = conn.prepareStatement(sql);
                        ps2.setLong(1, pathId);
                        result2 = ps2.executeQuery();
                        if (result2.next()) {
                            int parentPathId = result2.getInt("REG_PATH_PARENT_ID");
                            if (parentPathId == resourceID.getPathID()) {
                                // so we confirm that this is a child of our collection
                                String childPath = result2.getString("REG_PATH_VALUE")+
                                    ";version:" + snapshotID;
                                if(current >= start) {
                                    childPathList.add(childPath);
                                }
                                isValidPath = true;
                                current ++;
                            }
                        }
                    }
                }
                if (!isValidPath && current > 0) {
                    break;
                }
                versionIndex ++;
            }
        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceID.getPath() + "for version. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
                if (result2 != null) {
                    result2.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return childPathList.toArray(new String[childPathList.size()]);
    }


    private String getCurrentPath(String resourcePath) {
        String currentPath = resourcePath;
        if (resourcePath.indexOf("?") > 0) {
            currentPath = resourcePath.split("\\?")[0];
        } else if (resourcePath.indexOf(RegistryConstants.URL_SEPARATOR) > 0) {
            currentPath = resourcePath.split("\\;")[0];
        }

        return currentPath;
    }

    public String addContentVersionAsBytes(byte[] contentBytes) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String contentID = null;

        if (contentBytes != null) {

            contentID = UUID.randomUUID().toString();

            PreparedStatement ps = null;
            try {

                String sql = "INSERT INTO REG_CONTENT_VERSION " +
                        "(REG_CONTENT_VERSION_ID, REG_CONTENT_DATA) VALUES (?, ?)";

                ps = conn.prepareStatement(sql);
                ps.setString(1, contentID);
                ps.setBytes(2, contentBytes);
                ps.executeUpdate();

            } catch (SQLException e) {

                String msg = "Failed to write resource content to the database. " + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);

            } finally {
                try {
                    if (ps != null) {
                        ps.close();
                    }
                }
                catch (SQLException ex) {
                    String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                    log.error(msg, ex);
                    throw new RegistryException(msg, ex);
                }
            }
        }

        return contentID;
    }

    public byte[] getContentBytesFromDB(String contentID, Connection conn)
            throws RegistryException {

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            byte[] contentBytes = null;

            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_VERSION WHERE REG_CONTENT_VERSION_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            result = ps.executeQuery();

            if (result.next()) {
                contentBytes = result.getBytes("REG_CONTENT_DATA");
            }

            return contentBytes;

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public byte[] getResourceContentBytes(String contentID, Connection conn)
            throws RegistryException {

        byte[] contentBytes = null;

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_VERSION WHERE REG_CONTENT_VERSION_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            result = ps.executeQuery();
            if (result.next()) {
                contentBytes = result.getBytes("REG_CONTENT_DATA");
            }

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return contentBytes;
    }

    public void createSnapshot(int pathId, String name, InputStream versionsStream) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "INSERT INTO REG_SNAPSHOT (REG_PATH_ID, REG_RESOURCE_NAME, REG_RESOURCE_VIDS) VALUES (?, ?, ?)";

            int size = versionsStream.available();
            ps = conn.prepareStatement(sql);
            ps.setInt(1, pathId);
            ps.setString(2, name);
            ps.setBinaryStream(3, versionsStream, size);
            ps.executeUpdate();

        } catch (Exception e) {

            String msg = "Failed to write resource content to the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    /**
     * Check wether the resource is already in the history with the give version
     * @param version
     * @return
     */
    public boolean isResourceHistoryExist(long version) throws RegistryException{
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_PATH_ID FROM REG_RESOURCE_HISTORY WHERE REG_VERSION=?";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, version);
            ResultSet result = ps.executeQuery();
            if (result.next()) {
                return true;
            }

        } catch (SQLException e) {
            String msg = "Failed reading the history table for resource version  " +
                          version + " . " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return false;
    }

    /**
     * Check wether the resource is already in the history with the give path
     * @param path
     * @return
     */
    public boolean isResourceHistoryExist(String path) throws RegistryException{

        // first assuming it is a collection
        ResourceIDImpl resourceID =  resourceDAO.getResourceID(path, true);
        if (resourceID != null) {
           return isResourceHistoryExist(resourceID);
        }

        // second assuming it is a resource
        resourceID = resourceDAO.getResourceID(path, false);
        return resourceID == null || isResourceHistoryExist(resourceID);

    }

    /**
     * Check wether the resource is already in the history with the give resourceID
     * @param resourceID
     * @return
     */
    public boolean isResourceHistoryExist(ResourceIDImpl resourceID) throws RegistryException{
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            if (resourceID.isCollection()) {
                String sql = "SELECT REG_PATH_ID FROM REG_RESOURCE_HISTORY WHERE REG_PATH_ID=? AND REG_NAME IS NULL";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                result = ps.executeQuery();
                if (result.next()) {
                    return true;
                }
            }
            else {
                String sql = "SELECT REG_PATH_ID FROM REG_RESOURCE_HISTORY WHERE REG_PATH_ID=? AND REG_NAME = ?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                result = ps.executeQuery();
                if (result.next()) {
                    return true;
                }
            }

        } catch (SQLException e) {
            String msg = "Failed reading the history table for resource path " +
                          resourceID.getPath() + " . " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return false;
    }

    /**
     * Check wether the content is already in the history
     * @param contentId
     * @return true if it exist in history, false otherwise
     */
    public boolean isContentHistoryExist(int contentId) throws RegistryException{
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_CONTENT_DATA FROM REG_CONTENT_HISTORY WHERE REG_CONTENT_ID = ?";

            ps = conn.prepareStatement(sql);
            ps.setInt(1, contentId);
            result = ps.executeQuery();
            if (result.next()) {
                return true;
            }
        } catch (SQLException e) {
            String msg = "Failed reading the history table for content " +
                         contentId + " . " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return false;
    }

    public void copyAllContentToHistory() throws RegistryException {
        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            // get all the table in the REG_CONTENT
            String resourceContentSQL = "SELECT REG_CONTENT_ID, REG_CONTENT_DATA FROM  REG_CONTENT";
            ps  = conn.prepareStatement(resourceContentSQL);
            result = ps.executeQuery();

            while (result.next()) {
                int contentId = result.getInt("REG_CONTENT_DATA");
                InputStream dataStream = RegistryUtils.getMemoryStream(result.getBinaryStream("REG_CONTENT_DATA"));

                // now copy them to the resource
                if (!isContentHistoryExist(contentId)) {
                    putContentStreamToHistory(contentId, dataStream);
                }
            }
        } catch (SQLException e) {
            String msg = "Failed to copy the contents to the history table. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void copyAllResourceToHistory() throws RegistryException {
        Connection conn = Transaction.getConnection();

        ResourceDO resourceDO = null;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            // get all the table in the REG_RESOURCE
            String sql = "SELECT REG_PATH_ID, REG_NAME, REG_VERSION, REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                        "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID " +
                        "FROM REG_RESOURCE";

            ps = conn.prepareStatement(sql);
            result = ps.executeQuery();
            if (result.next()) {
                resourceDO = new ResourceDO();

                // this is always the current version of the resource
                resourceDO.setPathID(result.getInt("REG_PATH_ID"));
                resourceDO.setName(result.getString("REG_NAME"));
                resourceDO.setVersion(result.getInt("REG_VERSION"));
                resourceDO.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceDO.setAuthor(result.getString("REG_CREATOR"));
                resourceDO.setCreatedOn(result.getTimestamp("REG_CREATED_TIME"));
                resourceDO.setLastUpdator(result.getString("REG_LAST_UPDATOR"));
                resourceDO.setLastUpdatedOn(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceDO.setDescription(result.getString("REG_DESCRIPTION"));
                resourceDO.setContentID(result.getInt("REG_CONTENT_ID"));

                // now copy them to the resource
                if (!isResourceHistoryExist(resourceDO.getVersion())) {
                    putResourceToHistory(resourceDO);
                }
            }
        } catch (SQLException e) {
            String msg = "Failed to get the resource at version" + resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public String restoreResources(long version) throws RegistryException {
        // get the archieved resource
        ResourceDO resourceDO = getResourceDOArchived(version);
        String resourcePath = resourceDAO.getPath(resourceDO.getPathID(), resourceDO.getName(), false);

        // create the resourceId
        ResourceImpl oldResource;
        if (resourceDO.getName() == null) {
            // this is a collection
            oldResource = new CollectionImpl(resourcePath, resourceDO);
        }
        else {
            oldResource = new ResourceImpl(resourcePath, resourceDO);
        }

        int oldContentID = resourceDO.getContentID();
        int newContentID = 0;
        if (oldContentID > 0) {
            // if the non-collection restore content
            // get the archived content
            InputStream contentData = getContentArchived(oldContentID);
            if(contentData != null) {
                newContentID = resourceDAO.addContentBytes(contentData);
                resourceDO.setContentID(newContentID);
            }
        }
        resourceDAO.addResourceDO(resourceDO);
        // copy comments, taggings, ratings to new version
        ResourceImpl newResource;
        if (resourceDO.getName() == null) {
            newResource = new CollectionImpl(resourcePath, resourceDO);
        }
        else {
            newResource = new ResourceImpl(resourcePath, resourceDO);
        }

        if (StaticConfiguration.isVersioningProperties()) {
            resourceDAO.fillResourceProperties(oldResource);
            newResource.setProperties(oldResource.getProperties());
            resourceDAO.addProperties(newResource);
        }

        commentsDAO.copyComments(oldResource, newResource);
        tagsDAO.copyTags(oldResource, newResource);
        ratingsDAO.copyRatings(oldResource, newResource);

        // finally return the resource path
        return resourcePath;
    }


    private ResourceDO getResourceDOArchived(long version) throws RegistryException {
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT REG_PATH_ID, REG_NAME, REG_VERSION, REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                        "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, REG_CONTENT_ID " +
                        "FROM REG_RESOURCE_HISTORY WHERE REG_VERSION =?";
            ps = conn.prepareStatement(sql);
            ps.setLong(1, version);

            result = ps.executeQuery();

            if(result.next()) {
                ResourceDO resourceDO = new ResourceDO();

                // this is always the current version of the resource
                resourceDO.setPathID(result.getInt("REG_PATH_ID"));
                resourceDO.setName(result.getString("REG_NAME"));
                resourceDO.setVersion(result.getInt("REG_VERSION"));
                resourceDO.setMediaType(result.getString("REG_MEDIA_TYPE"));
                resourceDO.setAuthor(result.getString("REG_CREATOR"));
                resourceDO.setCreatedOn(result.getTimestamp("REG_CREATED_TIME"));
                resourceDO.setLastUpdator(result.getString("REG_LAST_UPDATOR"));
                resourceDO.setLastUpdatedOn(result.getTimestamp("REG_LAST_UPDATED_TIME"));
                resourceDO.setDescription(result.getString("REG_DESCRIPTION"));
                resourceDO.setContentID(result.getInt("REG_CONTENT_ID"));

                return resourceDO;
            }
        } catch (SQLException e) {

            String msg = "Failed to get the resource version " +
                    version + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return null;
    }

    private InputStream getContentArchived(int contentID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {

            String resourceContentSQL = "SELECT REG_CONTENT_DATA  FROM  REG_CONTENT_HISTORY WHERE REG_CONTENT_ID = ?";
            ps  = conn.prepareStatement(resourceContentSQL);
            ps.setLong(1, contentID);
            result = ps.executeQuery();
            if (result.next()) {
                InputStream rawStream = result.getBinaryStream("REG_CONTENT_DATA");
                if (rawStream != null) {
                    InputStream regData = RegistryUtils.getMemoryStream(rawStream);
                    return regData;
                }
            }
        } catch (SQLException e) {

            String msg = "Failed to get the archived content " +
                    contentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
        return null;
    }

    public void versionResource(ResourceDO resourceDO, boolean keepProperties) throws RegistryException {
        if (resourceDO.getName() != null) {
            // this implies this is a resource
            int contentID = resourceDO.getContentID();
            if (contentID > 0) {
                // the content stream is deleted later within this function to avoid violation
                // of forign key constrains
                versionContent(contentID);
            }
        }
        // now write the query to copy the old to the history table.
        if (!isResourceHistoryExist(resourceDO.getVersion())) {
            putResourceToHistory(resourceDO);
        }

        // remove the modified non-versioned resources
        if(!StaticConfiguration.isVersioningProperties() && !keepProperties) {
            resourceDAO.removeProperties(resourceDO);
        }

        // delete the old entry from the resource table
        resourceDAO.deleteResource(resourceDO);

        // version the content as well
        if (resourceDO.getName() != null) {

            int contentID = resourceDO.getContentID();
            if (contentID > 0) {
                // delete the old content stream from the latest table
                resourceDAO.deleteContentStream(contentID);
            }
        }
    }

    public void putResourceToHistory(ResourceDO resourceDO) throws RegistryException{
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "INSERT INTO REG_RESOURCE_HISTORY (REG_PATH_ID, REG_NAME, REG_VERSION, REG_MEDIA_TYPE, " +
                    "REG_CREATOR, REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                    "REG_CONTENT_ID) " +
                    "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
            ps = conn.prepareStatement(sql);

            ps.setInt(1, resourceDO.getPathID());
            ps.setString(2, resourceDO.getName());
            ps.setLong(3, resourceDO.getVersion());
            ps.setString(4, resourceDO.getMediaType());
            ps.setString(5, resourceDO.getAuthor());
            ps.setTimestamp(6, new Timestamp(resourceDO.getCreatedOn().getTime()));
            ps.setString(7, resourceDO.getLastUpdator());
            ps.setTimestamp(8, new Timestamp(resourceDO.getLastUpdatedOn().getTime()));
            ps.setString(9, resourceDO.getDescription());
            if(resourceDO.getContentID() > 0) {
                ps.setInt(10, resourceDO.getContentID());
            }
            else {
                ps.setNull(10, Types.INTEGER);
            }
            ps.executeUpdate();

        }
        catch (SQLException ex)  {
            String msg = "Failed to copy resource version" +  resourceDO.getVersion() +
                        " to the history table " + ex.getMessage();
            log.error(msg,ex);
            throw new RegistryException(msg,ex);
        }
        finally {
            try {
                if (ps != null) {ps.close();}
            } catch(SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }

    private void versionContent(int contentID) throws RegistryException {

        if(!isContentHistoryExist(contentID)) {
            InputStream contentStream = resourceDAO.getContentStream(contentID);
            if (contentStream == null) {
                // create an empty input stream
                contentStream = new ByteArrayInputStream("".getBytes());
            }
            // copy the content to the content_history table
            putContentStreamToHistory(contentID, contentStream);
        }
    }

    public void putContentToHistory(int contentId) throws RegistryException {
        Connection conn = Transaction.getConnection();
        if (isContentHistoryExist(contentId)) {
            return;
        }

        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            // get all the table in the REG_CONTENT
            String resourceContentSQL = "SELECT REG_CONTENT_DATA FROM  REG_CONTENT WHERE REG_CONTENT_ID =? ";
            ps  = conn.prepareStatement(resourceContentSQL);
            ps.setInt(1, contentId);
            result = ps.executeQuery();

            if (result.next()) {
                InputStream dataStream = RegistryUtils.getMemoryStream(result.getBinaryStream("REG_CONTENT_DATA"));

                // now copy them to the resource
                putContentStreamToHistory(contentId, dataStream);
            }
        } catch (SQLException e) {
            String msg = "Failed to copy the contents to the history table. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    public void putContentStreamToHistory(int contentID, InputStream contentStream) throws RegistryException{
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {

            String sql = "INSERT INTO REG_CONTENT_HISTORY (REG_CONTENT_ID, REG_CONTENT_DATA) VALUES (?, ?)";

            int size = contentStream.available();
            ps = conn.prepareStatement(sql);
            ps.setInt(1, contentID);
            ps.setBinaryStream(2, contentStream, size);
            ps.executeUpdate();

        }
        catch (Exception ex)  {
            String msg = "Failed to put the content into history with the content id " +
                        contentID + ". " + ex.getMessage();
            log.error(msg,ex);
            throw new RegistryException(msg,ex);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (ps != null) {ps.close();}
            } catch(SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }
}
