/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.TaggingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class TagsDAO {

    private static final Log log = LogFactory.getLog(TagsDAO.class);
    private ResourceDAO resourceDAO = new ResourceDAO();

    public void addTagging(String tagName, ResourceImpl resource, String userID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        PreparedStatement ps2 = null;
        ResultSet result = null;
        try {
            String sql = "INSERT INTO REG_TAG (REG_TAG_NAME, REG_USER_ID, REG_TAGGED_TIME) VALUES (?,?,?)";

            long now = System.currentTimeMillis();

            ps = conn.prepareStatement(sql,  new String[] {"REG_ID"});
            ps.setString(1, tagName);
            ps.setString(2, userID);
            ps.setDate(3, new Date(now));

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                int tagId = result.getInt(1);

                String sql2 = "INSERT INTO REG_RESOURCE_TAG (REG_TAG_ID, REG_PATH_ID, REG_RESOURCE_NAME) " +
                           "VALUES(?,?,?)";
                ps2 = conn.prepareStatement(sql2);

                ps2.setInt(1, tagId);
                ps2.setInt(2, resource.getPathID());
                ps2.setString(3, resource.getName());

                ps2.executeUpdate();
            }

        } catch (SQLException e) {

            String msg = "Failed to add tag " + tagName + " to resource " + resource.getPath() +
                    " by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) { ps.close(); }
                if (ps2 != null) { ps2.close(); }
                if (result != null) { result.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void addTaggings(ResourceImpl resource, TaggingDO[] taggingDOs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        long now = System.currentTimeMillis();
        PreparedStatement ps = null, ps2 = null;
        ResultSet result = null;

        try {
            for (TaggingDO taggingDO : taggingDOs) {
                String sql = "INSERT INTO REG_TAG (REG_TAG_NAME, REG_USER_ID, REG_TAGGED_TIME) VALUES (?,?,?)";

                ps = conn.prepareStatement(sql,  new String[] {"REG_ID"});
                ps.setString(1, taggingDO.getTagName());
                ps.setString(2, taggingDO.getTaggedUserName());
                ps.setDate(3, new Date(now));

                ps.executeUpdate();

                result = ps.getGeneratedKeys();
                if (result.next()) {
                    int tagId = result.getInt(1);

                    String sql2 = "INSERT INTO REG_RESOURCE_TAG (REG_TAG_ID, REG_PATH_ID, REG_RESOURCE_NAME) " +
                               "VALUES(?,?,?)";
                    ps2 = conn.prepareStatement(sql2);

                    ps2.setInt(1, tagId);
                    ps2.setInt(2, resource.getPathID());
                    ps2.setString(3, resource.getName());

                    ps2.executeUpdate();
                }
                result.close();
                ps.close();
                ps2.close();
            }
            ps = null;
            ps2 = null;
            result = null;

        } catch (SQLException e) {

            String msg = "Failed to add tags to resource " + resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) { result.close(); }
                if (ps != null) { ps.close(); }
                if (ps2 != null) { ps2.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void copyTags(ResourceImpl fromResource, ResourceImpl toResource)
            throws RegistryException {

        if(fromResource.getPath() == toResource.getPath()) {
            // no special copying needed
            return;
        }
        List<TaggingDO> tagList = getTagDOs(fromResource);

        addTaggings(toResource, tagList.toArray(new TaggingDO[tagList.size()]));
    }

    public boolean taggingExists(String tagName, ResourceImpl resourceImpl, String userID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;

        try {
            if(resourceImpl instanceof CollectionImpl) {
                String sql =
                    "SELECT T.REG_ID FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "LOWER(T.REG_TAG_NAME)=LOWER(?) AND T.REG_USER_ID =? AND " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setString(1, tagName);
                ps.setString(2, userID);
                ps.setInt(3, resourceImpl.getPathID());
            }
            else {
                String sql =
                    "SELECT T.REG_ID FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "LOWER(T.REG_TAG_NAME)=LOWER(?) AND T.REG_USER_ID =? AND " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setString(1, tagName);
                ps.setString(2, userID);
                ps.setInt(3, resourceImpl.getPathID());
                ps.setString(4, resourceImpl.getName());
            }
            result = ps.executeQuery();

            boolean tagExists = false;
            if (result.next()) {
                tagExists = true;
            }

            return tagExists;

        } catch (SQLException e) {

            String msg = "Failed to check the existence of the tag " + tagName +
                    " on resource " + resourceImpl.getPath() + " by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) { ps.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

    }

    public void removeTags(ResourceImpl resource, String tag)
            throws RegistryException {

        removeTags(resource, tag, "*");
    }

    public void removeTags(ResourceImpl resource, String tag, String userName)
            throws RegistryException {
        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null, ps2= null;

        TaggingDO[] taggingDOs = getTagging(resource, tag, userName);
        if(taggingDOs == null || taggingDOs.length == 0) {
            return;
        }
        try {

            for(TaggingDO taggingDO: taggingDOs) {
                String sql = "DELETE FROM REG_RESOURCE_TAG WHERE REG_TAG_ID= ?";
                ps2 = conn.prepareStatement(sql);
                ps2.setLong(1, taggingDO.getTagID());
                ps2.executeUpdate();

                sql = "DELETE FROM REG_TAG WHERE REG_ID= ?";
                ps = conn.prepareStatement(sql);
                ps.setLong(1, taggingDO.getTagID());
                ps.executeUpdate();
            }
        } catch (SQLException e) {

            String msg = "Failed to remove tag " + tag + " from the resource " +
                    resource.getPath() + "with user id. " + userName + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }
    
    public void removeTags(ResourceImpl resource)
            throws RegistryException {
        Connection conn = Transaction.getConnection();

        PreparedStatement ps1 = null, ps2= null;

        TaggingDO[] taggingDOs = getTagging(resource);
        if(taggingDOs == null) {
            return;
        }
        try {

            String sql = "DELETE FROM REG_RESOURCE_TAG WHERE REG_TAG_ID= ?";
            ps1 = conn.prepareStatement(sql);

            sql = "DELETE FROM REG_TAG WHERE REG_ID= ?";
            ps2 = conn.prepareStatement(sql);

            for(TaggingDO taggingDO: taggingDOs) {
                ps1.setLong(1, taggingDO.getTagID());
                ps2.setLong(1, taggingDO.getTagID());
                ps1.addBatch();
                ps2.addBatch();
            }

            if(taggingDOs.length > 0) {
                try {
                    ps1.executeBatch();
                    ps2.executeBatch();
                } catch (SQLException e) {
                    ps1.clearBatch();
                    ps2.clearBatch();
                    // the exception will be handled in the next catch block
                    throw e;
                }
            }

        } catch (SQLException e) {

            String msg = "Failed to remove tagging from " + resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) {
                    ps1.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }


    public String[] getPathsWithTag(String tag, Connection conn) throws SQLException, RegistryException {

        String sql = "SELECT  RT.REG_PATH_ID, RT.REG_RESOURCE_NAME, RT.REG_VERSION FROM " +
            "REG_RESOURCE_TAG RT, REG_TAG T WHERE T.REG_TAG_NAME = ? AND RT.REG_TAG_ID=T.REG_ID ";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, tag);

        ResultSet results = s.executeQuery();
        List pathList = new ArrayList();
        while (results.next()) {
            long version = results.getLong("REG_VERSION");
            if(version > 0) {
                String path = resourceDAO.getPath(version);
                if (path != null) {
                    pathList.add(path);
                }
            }
            else {
                int pathId = results.getInt("REG_PATH_ID");
                String resourceName = results.getString("REG_RESOURCE_NAME");
                String path = resourceDAO.getPath(pathId, resourceName, true);
                if (path != null) {
                    pathList.add(path);
                }
            }
        }

        try {
            if (results != null) {
                results.close();
            }
            if (s != null) {
                s.close();
            }
        }
        catch (SQLException ex) {
            String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
            log.error(msg, ex);
            throw new RegistryException(msg, ex);
        }

        return (String[]) pathList.toArray(new String[pathList.size()]);
    }
    

    public String[] getTags(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ResultSet results = null;
        PreparedStatement ps = null;
        try {

            if(resourceImpl instanceof CollectionImpl) {
                String sql =
                    "SELECT T.REG_TAG_NAME FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
            }
            else {
                String sql =
                    "SELECT T.REG_TAG_NAME FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
            }
            results = ps.executeQuery();

            List tagList = new ArrayList();
            while (results.next()) {
                tagList.add(results.getString(DatabaseConstants.TAG_NAME_FIELD));
            }

            return (String[]) tagList.toArray(new String[tagList.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get tags associated with the resource path " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }



    public List<TaggingDO> getTagDOs(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ResultSet results = null;
        PreparedStatement ps = null;
        List<TaggingDO> tagList = new ArrayList<TaggingDO>();
        try {

            if(resourceImpl instanceof CollectionImpl) {
                String sql =
                    "SELECT T.REG_TAG_NAME, T.REG_USER_ID, T.REG_TAGGED_TIME " +
                    "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
            }
            else {
                String sql =
                    "SELECT T.REG_TAG_NAME, T.REG_USER_ID, T.REG_TAGGED_TIME " +
                    "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
            }
            results = ps.executeQuery();

            while (results.next()) {
                TaggingDO taggingDO = new TaggingDO();
                taggingDO.setTagName(results.getString(1));
                taggingDO.setTaggedUserName(results.getString(2));
                taggingDO.setTaggedTime(results.getDate(3));
                tagList.add(taggingDO);
            }

            return tagList;

        } catch (SQLException e) {

            String msg = "Failed to get tags associated with the resource path " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public List getPathsWithAnyTag(String[] tags) throws RegistryException {

        if (tags == null || tags.length == 0) return null;

        Connection conn = Transaction.getConnection();

        StringBuffer sqlBuf = new StringBuffer();
        sqlBuf.append("SELECT DISTINCT RT.REG_PATH_ID, RT.REG_RESOURCE_NAME FROM " +
            "REG_RESOURCE_TAG RT, REG_TAG T WHERE RT.REG_TAG_ID=T.REG_ID ");
        sqlBuf.append(" AND (");
        for (int i = 0; i < tags.length; i++) {
            if (i > 0) sqlBuf.append(" OR ");
            sqlBuf.append("lower(T.REG_TAG_NAME)=lower(?)");
        }
        sqlBuf.append(")");

        List resourcePaths = new ArrayList();

        ResultSet results = null;
        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sqlBuf.toString());

            for (int i = 0; i < tags.length; i++) {
                s.setString(i + 1, tags[i]);
            }

            results = s.executeQuery();
            while (results.next()) {
                int pathId = results.getInt("REG_PATH_ID");
                String resourceName = results.getString("REG_RESOURCE_NAME");
                String path = resourceDAO.getPath(pathId, resourceName, true);
                if (path != null) {
                    resourcePaths.add(path);
                }
            }

        } catch (SQLException e) {

            String msg = "Failed to resource paths with any of the tags " +
                    Arrays.toString(tags) + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return resourcePaths;
    }

    public long getTagCount(ResourceImpl resourceImpl, String tag) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            if(resourceImpl instanceof CollectionImpl) {
                String sql =
                    "SELECT COUNT(T.REG_TAG_NAME) " +
                    "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "lower(T.REG_TAG_NAME)=lower(?) AND " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME IS NULL " +
                    "GROUP BY RT.REG_PATH_ID";
                ps = conn.prepareStatement(sql);
                ps.setString(1, tag);
                ps.setInt(2, resourceImpl.getPathID());
            }
            else {
                String sql =
                    "SELECT COUNT(T.REG_TAG_NAME) " +
                    "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                    "lower(T.REG_TAG_NAME)=lower(?) AND " +
                    "T.REG_ID=RT.REG_TAG_ID AND RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME=? " +
                    "GROUP BY RT.REG_PATH_ID, RT.REG_RESOURCE_NAME";
                ps = conn.prepareStatement(sql);
                ps.setString(1, tag);
                ps.setInt(2, resourceImpl.getPathID());
                ps.setString(3, resourceImpl.getName());
            }
            result = ps.executeQuery();

            long tagCount = 0;
            if (result.next()) {
                tagCount = result.getLong(1);
            }

            return tagCount;

        } catch (SQLException e) {

            String msg = "Failed to get tag count of tag " + tag +
                    " on resource " + resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public Tag[] getTagsWithCount(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List tagList = new ArrayList();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            if(resourceImpl instanceof CollectionImpl) {
                String sql = "SELECT T.REG_TAG_NAME, COUNT(T.REG_ID) FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                        "WHERE RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME IS NULL AND "  +
                        "T.REG_ID=RT.REG_TAG_ID GROUP BY T.REG_TAG_NAME";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
            }
            else {
                String sql = "SELECT T.REG_TAG_NAME, COUNT(T.REG_ID) FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                        "WHERE RT.REG_PATH_ID=? AND RT.REG_RESOURCE_NAME=? AND " +
                        "T.REG_ID=RT.REG_TAG_ID GROUP BY T.REG_TAG_NAME";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
            }
            result = ps.executeQuery();
            while (result.next()) {
                Tag tag = new Tag();
                tag.setTagName(result.getString(DatabaseConstants.TAG_NAME_FIELD));
                tag.setTagCount(result.getLong(2));
                tagList.add(tag);
            }

        } catch (SQLException e) {

            String msg = "Failed to get tags and tag counts of the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return (Tag[]) tagList.toArray(new Tag[tagList.size()]);
    }

    public TaggingDO[] getTagging(ResourceImpl resource, String tag, String userName)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;

        List<TaggingDO> taggingDOs = new ArrayList<TaggingDO>();
        try {
            TaggingDO taggingDO = null;

            if(resource instanceof CollectionImpl) {
                String sql = "SELECT T.REG_ID, T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                         " WHERE RT.REG_PATH_ID = ? AND RT.REG_RESOURCE_NAME IS NULL AND RT.REG_TAG_ID=T.REG_ID " +
                         " AND T.REG_TAG_NAME=? ";
                if (!userName.equals("*")) {
                    sql = sql + "AND T.REG_USER_ID=?";
                }
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resource.getPathID());
                ps.setString(2, tag);
                if (!userName.equals("*")) {
                    ps.setString(3, userName);
                }
            }
            else {
                String sql = "SELECT T.REG_ID, T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                         " WHERE RT.REG_PATH_ID = ? AND RT.REG_RESOURCE_NAME = ? AND RT.REG_TAG_ID=T.REG_ID " +
                         " AND T.REG_TAG_NAME=? ";
                if (!userName.equals("*")) {
                    sql = sql + "AND T.REG_USER_ID=?";
                }

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resource.getPathID());
                ps.setString(2, resource.getName());
                ps.setString(3, tag);
                if (!userName.equals("*")) {
                    ps.setString(4, userName);
                }
            }

            result = ps.executeQuery();
            while (result.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                taggingDO.setResourcePath(resource.getPath());
                taggingDO.setTagName(tag);
                taggingDO.setTaggedTime(taggedTime);
                taggingDO.setTaggedUserName(userName);
                taggingDO.setTagID(result.getLong("REG_ID"));

                taggingDOs.add(taggingDO);
            }

            return taggingDOs.toArray(new TaggingDO[taggingDOs.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get tagging information for tag " + tag +
                    " on resource " + resource.getPath() + " by user " + userName + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public TaggingDO[] getTagging(ResourceImpl resource)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<TaggingDO> taggingDOs = new ArrayList<TaggingDO>();
        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            TaggingDO taggingDO;

            if(resource instanceof CollectionImpl) {
                String sql = "SELECT T.REG_ID, T.REG_TAG_NAME, T.REG_USER_ID, " +
                         "T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                         "WHERE RT.REG_PATH_ID = ? AND RT.REG_RESOURCE_NAME IS NULL AND RT.REG_TAG_ID=T.REG_ID";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resource.getPathID());
            }
            else {
                String sql = "SELECT T.REG_ID, T.REG_TAG_NAME, T.REG_USER_ID, " +
                         "T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                         "WHERE RT.REG_PATH_ID = ? AND RT.REG_RESOURCE_NAME = ? AND RT.REG_TAG_ID=T.REG_ID ";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resource.getPathID());
                ps.setString(2, resource.getName());
            }

            results = ps.executeQuery();
            while (results.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        results.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                taggingDO.setResourcePath(resource.getPath());
                taggingDO.setTagName(results.getString("REG_TAG_NAME"));
                taggingDO.setTaggedTime(taggedTime);
                taggingDO.setTaggedUserName(results.getString("REG_USER_ID"));
                taggingDO.setTagID(results.getLong("REG_ID"));
                taggingDOs.add(taggingDO);
            }


        } catch (SQLException e) {

            String msg = "Failed to get tagging information for the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return taggingDOs.toArray(new TaggingDO[taggingDOs.size()]);
    }

    public TaggingDO getTagging(long taggingID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement s = null;
        try {
            String sql = "SELECT T.REG_USER_ID, T.REG_TAG_NAME, T.REG_TAGGED_TIME, " +
                    "RT.REG_PATH_ID, RT.REG_RESOURCE_NAME " +
                    "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE T.REG_ID=? AND T.REG_ID=RT.REG_TAG_ID";

            s = conn.prepareStatement(sql);
            s.setLong(1, taggingID);

            TaggingDO taggingDO = null;
            result = s.executeQuery();
            if (result.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                // TODO
                //taggingDO.setResourceID(result.getString(DatabaseConstants.AID_FIELD));
                taggingDO.setTagName(result.getString(DatabaseConstants.TAG_NAME_FIELD));
                taggingDO.setTaggedUserName(result.getString(DatabaseConstants.USER_ID_FIELD));
                taggingDO.setTaggedTime(taggedTime);

                String resourcePath;
                int pathId = result.getInt("REG_PATH_ID");
                String resourceName = result.getString("REG_RESOURCE_NAME");
                resourcePath = resourceDAO.getPath(pathId, resourceName, true);
                if (resourcePath != null) {
                    taggingDO.setResourcePath(resourcePath);
                }
            }

            return taggingDO;

        } catch (SQLException e) {

            String msg = "Failed to get tagging information for tag ID " +
                    taggingID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public ResourceImpl getResourceWithMinimumData(String path) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        if (resourceID == null) {
            return null;
        }
        ResourceImpl resourceImpl;
        if (resourceID.isCollection()) {
            resourceImpl = new CollectionImpl();
        }
        else {
            resourceImpl = new ResourceImpl();
        }
        resourceImpl.setName(resourceID.getName());
        resourceImpl.setPathID(resourceID.getPathID());
        resourceImpl.setPath(resourceID.getPath());
        return resourceImpl;
    }


    public void moveTags(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE_TAG SET REG_PATH_ID=? WHERE " +
                             "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.executeUpdate();
            }
            else {
                String sql = "UPDATE REG_RESOURCE_TAG SET REG_PATH_ID=?, REG_RESOURCE_NAME=? WHERE " +
                             "REG_PATH_ID=? AND REG_RESOURCE_NAME=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move tags from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
              }
              catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
              }
        }
    }

    /**
     * this function will move the paths from one path id to anthor regardless of the resource name
     * @param source
     * @param target
     * @throws RegistryException
     */
    public void moveTagPaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE_TAG SET REG_PATH_ID=? WHERE REG_PATH_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move tag paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }finally {
            try {
                if (ps != null) {ps.close();}
             }
             catch (SQLException ex)  {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg,ex);
                throw new RegistryException(msg,ex);
            }
        }
    }
}
