/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.TaggingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class TagsVersionDAO extends TagsDAO{

    private static final Log log = LogFactory.getLog(TagsVersionDAO.class);
    private ResourceDAO resourceDAO = new ResourceDAO();

    public void addTagging(String tagName, ResourceImpl resource, String userID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        PreparedStatement ps2 = null;
        ResultSet result = null;
        try {
            String sql = "INSERT INTO REG_TAG (REG_TAG_NAME, REG_USER_ID, REG_TAGGED_TIME) VALUES (?,?,?)";

            long now = System.currentTimeMillis();

            ps = conn.prepareStatement(sql,  new String[] {"REG_ID"});
            ps.setString(1, tagName);
            ps.setString(2, userID);
            ps.setDate(3, new Date(now));

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                int tagId = result.getInt(1);
                String sql2 = "INSERT INTO REG_RESOURCE_TAG (REG_TAG_ID, REG_VERSION) VALUES(?,?)";
                ps2 = conn.prepareStatement(sql2);

                ps2.setInt(1, tagId);
                ps2.setLong(2, resource.getVersionNumber());

                ps2.executeUpdate();
            }

        } catch (SQLException e) {

            String msg = "Failed to add tag " + tagName + " to resource " + resource.getPath() +
                    " by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) { ps.close(); }
                if (ps2 != null) { ps2.close(); }
                if (result != null) { result.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void addTaggings(ResourceImpl resource, TaggingDO[] taggingDOs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        long now = System.currentTimeMillis();
        PreparedStatement ps = null, ps2 = null;
        ResultSet result = null;

        try {
            for (TaggingDO taggingDO : taggingDOs) {
                String sql = "INSERT INTO REG_TAG (REG_TAG_NAME, REG_USER_ID, REG_TAGGED_TIME) VALUES (?,?,?)";

                ps = conn.prepareStatement(sql,  new String[] {"REG_ID"});
                ps.setString(1, taggingDO.getTagName());
                ps.setString(2, taggingDO.getTaggedUserName());
                ps.setDate(3, new Date(now));

                ps.executeUpdate();

                result = ps.getGeneratedKeys();
                if (result.next()) {
                    int tagId = result.getInt(1);
                    String sql2 = "INSERT INTO REG_RESOURCE_TAG (REG_TAG_ID, REG_VERSION) VALUES(?,?)";
                    ps2 = conn.prepareStatement(sql2);

                    ps2.setInt(1, tagId);
                    ps2.setLong(2, resource.getVersionNumber());

                    ps2.executeUpdate();
                }
                result.close();
                ps.close();
                ps2.close();
            }
            ps = null;
            ps2 = null;
            result = null;

        } catch (SQLException e) {

            String msg = "Failed to add tags to resource " + resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) { result.close(); }
                if (ps != null) { ps.close(); }
                if (ps2 != null) { ps2.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public void copyTags(ResourceImpl fromResource, ResourceImpl toResource)
            throws RegistryException {

        List<TaggingDO> tagList = getTagDOs(fromResource);

        addTaggings(toResource, tagList.toArray(new TaggingDO[tagList.size()]));
    }

    public boolean taggingExists(String tagName, ResourceImpl resourceImpl, String userID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;

        try {
            String sql =
                "SELECT T.REG_ID FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                "LOWER(T.REG_TAG_NAME)=LOWER(?) AND T.REG_USER_ID =? AND " +
                "T.REG_ID=RT.REG_TAG_ID AND RT.REG_VERSION=?";
            ps = conn.prepareStatement(sql);
            ps.setString(1, tagName);
            ps.setString(2, userID);
            ps.setLong(3, resourceImpl.getVersionNumber());
            result = ps.executeQuery();

            boolean tagExists = false;
            if (result.next()) {
                tagExists = true;
            }

            return tagExists;

        } catch (SQLException e) {

            String msg = "Failed to check the existence of the tag " + tagName +
                    " on resource " + resourceImpl.getPath() + " by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) { ps.close(); }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

    }


    public String[] getTags(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            String sql =
                "SELECT T.REG_TAG_NAME FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                "T.REG_ID=RT.REG_TAG_ID AND RT.REG_VERSION=?";
            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());
            results = ps.executeQuery();

            List tagList = new ArrayList();
            while (results.next()) {
                tagList.add(results.getString(DatabaseConstants.TAG_NAME_FIELD));
            }

            return (String[]) tagList.toArray(new String[tagList.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get tags associated with the resource path " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }



    public List<TaggingDO> getTagDOs(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ResultSet results = null;
        PreparedStatement ps = null;
        List<TaggingDO> tagList = new ArrayList<TaggingDO>();
        try {
            String sql =
                "SELECT T.REG_TAG_NAME, T.REG_USER_ID, T.REG_TAGGED_TIME " +
                "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                "T.REG_ID=RT.REG_TAG_ID AND RT.REG_VERSION=?";
            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());
            results = ps.executeQuery();

            while (results.next()) {
                TaggingDO taggingDO = new TaggingDO();
                taggingDO.setTagName(results.getString(1));
                taggingDO.setTaggedUserName(results.getString(2));
                taggingDO.setTaggedTime(results.getDate(3));
                tagList.add(taggingDO);
            }

            return tagList;

        } catch (SQLException e) {

            String msg = "Failed to get tags associated with the resource path " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public List getPathsWithAnyTag(String[] tags) throws RegistryException {

        if (tags == null || tags.length == 0) return null;

        Connection conn = Transaction.getConnection();

        StringBuffer sqlBuf = new StringBuffer();
        sqlBuf.append("SELECT DISTINCT RT.REG_VERSION FROM " +
            "REG_RESOURCE_TAG RT, REG_TAG T WHERE RT.REG_TAG_ID=T.REG_ID ");
        sqlBuf.append(" AND (");
        for (int i = 0; i < tags.length; i++) {
            if (i > 0) sqlBuf.append(" OR ");
            sqlBuf.append("lower(T.REG_TAG_NAME)=lower(?)");
        }
        sqlBuf.append(")");

        List resourcePaths = new ArrayList();

        ResultSet results = null;
        PreparedStatement s = null;
        try {
            s = conn.prepareStatement(sqlBuf.toString());

            for (int i = 0; i < tags.length; i++) {
                s.setString(i + 1, tags[i]);
            }

            results = s.executeQuery();
            while (results.next()) {
                long version = results.getLong("REG_VERSION");
                if(version > 0) {
                    String path = resourceDAO.getPath(version);
                    if (path != null) {
                        resourcePaths.add(path);
                    }
                }
            }

        } catch (SQLException e) {

            String msg = "Failed to resource paths with any of the tags " +
                    Arrays.toString(tags) + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return resourcePaths;
    }

    public long getTagCount(ResourceImpl resourceImpl, String tag) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql =
                "SELECT COUNT(T.REG_TAG_NAME) " +
                "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE " +
                "lower(T.REG_TAG_NAME)=lower(?) AND " +
                "T.REG_ID=RT.REG_TAG_ID AND RT.REG_VERSION=? GROUP BY RT.REG_VERSION";
            ps = conn.prepareStatement(sql);
            ps.setString(1, tag);
            ps.setLong(2, resourceImpl.getVersionNumber());
            result = ps.executeQuery();

            long tagCount = 0;
            if (result.next()) {
                tagCount = result.getLong(1);
            }

            return tagCount;

        } catch (SQLException e) {

            String msg = "Failed to get tag count of tag " + tag +
                    " on resource " + resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public Tag[] getTagsWithCount(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List tagList = new ArrayList();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String sql = "SELECT T.REG_TAG_NAME, COUNT(T.REG_ID) FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                    "WHERE RT.REG_VERSION=? AND T.REG_ID=RT.REG_TAG_ID GROUP BY T.REG_TAG_NAME";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, resourceImpl.getVersionNumber());

            result = ps.executeQuery();
            while (result.next()) {
                Tag tag = new Tag();
                tag.setTagName(result.getString(DatabaseConstants.TAG_NAME_FIELD));
                tag.setTagCount(result.getLong(2));
                tagList.add(tag);
            }

        } catch (SQLException e) {

            String msg = "Failed to get tags and tag counts of the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return (Tag[]) tagList.toArray(new Tag[tagList.size()]);
    }

    public TaggingDO[] getTagging(ResourceImpl resource, String tag, String userName)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement ps = null;

        List<TaggingDO> taggingDOs = new ArrayList<TaggingDO>();
        try {
            TaggingDO taggingDO = null;

            String sql = "SELECT T.REG_ID, T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                     " WHERE RT.REG_VERSION = ? AND RT.REG_TAG_ID=T.REG_ID " +
                     " AND T.REG_TAG_NAME=? ";
            if (!userName.equals("*")) {
                sql = sql + "AND T.REG_USER_ID=?";
            }
            ps = conn.prepareStatement(sql);
            ps.setLong(1, resource.getVersionNumber());
            ps.setString(2, tag);
            if (!userName.equals("*")) {
                ps.setString(3, userName);
            }

            result = ps.executeQuery();
            while (result.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                taggingDO.setResourcePath(resource.getPath());
                taggingDO.setTagName(tag);
                taggingDO.setTaggedTime(taggedTime);
                taggingDO.setTaggedUserName(userName);
                taggingDO.setTagID(result.getLong("REG_ID"));

                taggingDOs.add(taggingDO);
            }

            return taggingDOs.toArray(new TaggingDO[taggingDOs.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get tagging information for tag " + tag +
                    " on resource " + resource.getPath() + " by user " + userName + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }

    public TaggingDO[] getTagging(ResourceImpl resource)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<TaggingDO> taggingDOs = new ArrayList<TaggingDO>();
        ResultSet results = null;
        PreparedStatement ps = null;
        try {
            TaggingDO taggingDO;

            String sql = "SELECT T.REG_ID, T.REG_TAG_NAME, T.REG_USER_ID, " +
                     "T.REG_TAGGED_TIME FROM REG_TAG T, REG_RESOURCE_TAG RT " +
                     "WHERE RT.REG_VERSION =? AND RT.REG_TAG_ID=T.REG_ID";
            ps = conn.prepareStatement(sql);
            ps.setLong(1, resource.getVersionNumber());


            results = ps.executeQuery();
            while (results.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        results.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                taggingDO.setResourcePath(resource.getPath());
                taggingDO.setTagName(results.getString("REG_TAG_NAME"));
                taggingDO.setTaggedTime(taggedTime);
                taggingDO.setTaggedUserName(results.getString("REG_USER_ID"));
                taggingDO.setTagID(results.getLong("REG_ID"));
                taggingDOs.add(taggingDO);
            }


        } catch (SQLException e) {

            String msg = "Failed to get tagging information for the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }

        return taggingDOs.toArray(new TaggingDO[taggingDOs.size()]);
    }

    public TaggingDO getTagging(long taggingID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResultSet result = null;
        PreparedStatement s = null;
        try {
            String sql = "SELECT T.REG_USER_ID, T.REG_TAG_NAME, T.REG_TAGGED_TIME, " +
                    "RT.REG_VERSION " +
                    "FROM REG_TAG T, REG_RESOURCE_TAG RT WHERE T.REG_ID=? AND T.REG_ID=RT.REG_TAG_ID";

            s = conn.prepareStatement(sql);
            s.setLong(1, taggingID);

            TaggingDO taggingDO = null;
            result = s.executeQuery();
            if (result.next()) {

                java.util.Date taggedTime = new java.util.Date(
                        result.getTimestamp(DatabaseConstants.TAGGED_TIME_FIELD).getTime());

                taggingDO = new TaggingDO();
                // TODO
                //taggingDO.setResourceID(result.getString(DatabaseConstants.AID_FIELD));
                taggingDO.setTagName(result.getString(DatabaseConstants.TAG_NAME_FIELD));
                taggingDO.setTaggedUserName(result.getString(DatabaseConstants.USER_ID_FIELD));
                taggingDO.setTaggedTime(taggedTime);

                long version = result.getLong("REG_VERSION");
                String resourcePath = null;
                if(version > 0) {
                    resourcePath = resourceDAO.getPath(version);
                }
                if (resourcePath != null) {
                    taggingDO.setResourcePath(resourcePath);
                }
            }

            return taggingDO;

        } catch (SQLException e) {

            String msg = "Failed to get tagging information for tag ID " +
                    taggingID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (s != null) {
                    s.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULTSET_PREPAREDSTATEMENT_CLOSE_ERROR + ex.getMessage();
                log.error(msg, ex);
                throw new RegistryException(msg, ex);
            }
        }
    }


    public ResourceImpl getResourceWithMinimumData(String path) throws RegistryException {
        ResourceIDImpl resourceID = resourceDAO.getResourceID(path);
        if (resourceID == null) {
            return null;
        }

        ResourceImpl resourceImpl;
        if (resourceID.isCollection()) {
            resourceImpl = new CollectionImpl();
        }
        else {
            resourceImpl = new ResourceImpl();
        }
        long version = resourceDAO.getVersion(resourceID);
        resourceImpl.setVersionNumber(version);
        resourceImpl.setPath(path);
        return resourceImpl;
    }

    public void moveTags(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        // this is non-versioned specific function.
        // do nothing when the tags versioniong on in configuration
    }
    
    public void moveTagPaths(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        // this is non-versioned specific function.
        // do nothing when the tags versioniong on in configuration
    }
}