/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.handlers;

import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.handlers.filters.Filter;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;

import java.util.LinkedHashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Manages the handlers and thier invocations. Handlers, which are required to be invoked must be
 * registered with this class using the addHanlder(...) method. Handlers are allowed for a subset
 * of Registry API methods (see Java doc of Handler class for more detrails). Handler authors can
 * further specify exactly for which methods thier handlers should engage. For example, handler
 * author may specify his handler should be engaged only for GET and DELETE operations. This class
 * maintains separate maps for each supported operation and puts handlers to corresponding maps
 * once registered.
 *
 * Handlers are registered with a Filter instance. Filter decides (based on the request) whether or
 * not to invoke its associated handler.
 *
 * This class has a separate method for each supported operation. These methods are invoked by
 * BasicRegistry once such method is invoked in the Registry API. Then this class iterates through
 * the map corresponding to the invoked method and evalautes filters of registered handlers. For
 * each filter which evaluates to true, its associated handler will be invoked. This process
 * continues either till there is no more handlers or till the processingComplete parameter of
 * the RequestContext is set to true.
 *
 * There is only one instance of this class exists per registry instance.
 */
public class HandlerManager {

    private static final Log log = LogFactory.getLog(HandlerManager.class);

    /**
     * Maps of <Filter, Handler> pairs. A map exists for each supported operation.
     */
    private Map <Filter, Handler> getHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> putHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> deleteHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> importHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> putChildHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> importChildHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> invokeAspectHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> moveHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> copyHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> renameHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> createLinkHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> removeLinkHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> addAssociationHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> removeAssociationHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getAssociationsHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getAllAssociationsHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> applyTagHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getResourcePathsWithTagHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getTagsHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> removeTagHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> addCommentHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> editCommentHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getCommentsHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> rateResourceHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getAverageRatingHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getRatingHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> createVersionHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> getVersionsHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> restoreVersionHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> executeQueryHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> searchContentHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> resourceExistsHandlerMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> dumpMap = new LinkedHashMap <Filter, Handler> ();
    private Map <Filter, Handler> restoreMap = new LinkedHashMap <Filter, Handler> ();


    /**
     * Registers handlers with the handler manager. Each handler should be registered with a
     * Filter. If a handler should be engaged only to a subset of allowed methods, those methods
     * can be specified as a string array.
     *
     * @param methods Methods for which the registered handler should be engaged. Allowed values in
     * the string array are "GET", "PUT", "IMPORT", "DELETE", "PUT_CHILD", "IMPORT_CHILD",
     * "MOVE", "COPY", "RENAME", "CREATE_LINK", "REMOVE_LINK", "ADD_ASSOCIATION", "RESOURCE_EXISTS",
     * "REMOVE_ASSOCIATION", "GET_ASSOCIATIONS", "GET_ALL_ASSOCIATIONS", "APPLY_TAG",
     * "GET_RESOURCE_PATHS_WITH_TAG", "GET_TAGS", "REMOVE_TAG", "ADD_COMMENT", "EDIT_COMMENT",
     * "GET_COMMENTS", "RATE_RESOURCE", "GET_AVERAGE_RATING", "GET_RATING", "CREATE_VERSION",
     * "GET_VERSIONS", "RESTORE_VERSION", "EXECUTE_QUERY", "SEARCH_CONTENT", and "INVOKE_ASPECT".
     * If null is given, handler will be engaged to all methods.
     *
     * @param filter Filter instance associated with the handler.
     *
     * @param handler Handler instance to be registered.
     */
    public void addHandler(String[] methods, Filter filter, Handler handler) {

        filter.setHandler(handler);

        if (methods == null || RegistryUtils.containsString(Filter.GET, methods)) {
            getHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.PUT, methods)) {
            putHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.DELETE, methods)) {
            deleteHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.IMPORT, methods)) {
            importHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.PUT_CHILD, methods)) {
            putChildHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.IMPORT_CHILD, methods)) {
            importChildHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.INVOKE_ASPECT, methods)) {
            invokeAspectHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.MOVE, methods)) {
            moveHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.COPY, methods)) {
            copyHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.RENAME, methods)) {
            renameHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.CREATE_LINK, methods)) {
            createLinkHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.REMOVE_LINK, methods)) {
            removeLinkHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.ADD_ASSOCIATION, methods)) {
            addAssociationHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.REMOVE_ASSOCIATION, methods)) {
            removeAssociationHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_ASSOCIATIONS, methods)) {
            getAssociationsHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_ALL_ASSOCIATIONS, methods)) {
            getAllAssociationsHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.APPLY_TAG, methods)) {
            applyTagHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_RESOURCE_PATHS_WITH_TAG, methods)) {
            getResourcePathsWithTagHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_TAGS, methods)) {
            getTagsHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.REMOVE_TAG, methods)) {
            removeTagHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.ADD_COMMENT, methods)) {
            addCommentHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.EDIT_COMMENT, methods)) {
            editCommentHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_COMMENTS, methods)) {
            getCommentsHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.RATE_RESOURCE, methods)) {
            rateResourceHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_AVERAGE_RATING, methods)) {
            getAverageRatingHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_RATING, methods)) {
            getRatingHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.CREATE_VERSION, methods)) {
            createVersionHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_VERSIONS, methods)) {
            getVersionsHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.RESTORE_VERSION, methods)) {
            restoreVersionHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.EXECUTE_QUERY, methods)) {
            executeQueryHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.SEARCH_CONTENT, methods)) {
            searchContentHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.RESOURCE_EXISTS, methods)) {
            resourceExistsHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.DUMP, methods)) {
            dumpMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.RESTORE, methods)) {
            restoreMap.put(filter, handler);
        }

        String methodInfo = "";
        if (methods == null) {
            methodInfo = "all";
        } else {
            for (int i = 0; i < methods.length; i++) {
                String method = methods[i];
                methodInfo = methodInfo + " " + method;
            }
        }
        if (log.isDebugEnabled()) {
            log.debug("Registered the handler " + filter.getClass().getName() +
                " --> " + handler.getClass().getName() + " for " + methodInfo + " methods.");
        }
    }

    /**
     * remove a handler from all the filters, all the methods
     * @param handler
     */
    public void removeHandler(Handler handler) {

        for (Map.Entry<Filter, Handler> entry: getHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: putHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                putHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: deleteHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                deleteHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: importHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                importHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: putChildHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                putChildHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: importChildHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                importChildHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: invokeAspectHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                invokeAspectHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: moveHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                moveHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: copyHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                copyHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: renameHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                renameHandlerMap.remove(entry.getKey());
                break;
            }
        }
        
        for (Map.Entry<Filter, Handler> entry: createLinkHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                createLinkHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: createLinkHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                createLinkHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: removeLinkHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                removeLinkHandlerMap.remove(entry.getKey());
                break;
            }
        }


        for (Map.Entry<Filter, Handler> entry: addAssociationHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                addAssociationHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: removeAssociationHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                removeAssociationHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: applyTagHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                applyTagHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: removeTagHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                removeTagHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: rateResourceHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                rateResourceHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: restoreVersionHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                restoreVersionHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: createVersionHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                createVersionHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: editCommentHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                editCommentHandlerMap.remove(entry.getKey());
                break;
            }
        }
        
        for (Map.Entry<Filter, Handler> entry: getAssociationsHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getAssociationsHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getAllAssociationsHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getAllAssociationsHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getResourcePathsWithTagHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getResourcePathsWithTagHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getTagsHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getTagsHandlerMap.remove(entry.getKey());
                break;
            }
        }
        
        for (Map.Entry<Filter, Handler> entry: addCommentHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                addCommentHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getCommentsHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getCommentsHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getRatingHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getRatingHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getAverageRatingHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getAverageRatingHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: getVersionsHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                getVersionsHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: executeQueryHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                executeQueryHandlerMap.remove(entry.getKey());
                break;
            }
        }

        for (Map.Entry<Filter, Handler> entry: searchContentHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                searchContentHandlerMap.remove(entry.getKey());
                break;
            }
        }


        for (Map.Entry<Filter, Handler> entry: resourceExistsHandlerMap.entrySet()) {
            if (handler.equals(entry.getValue())){
                resourceExistsHandlerMap.remove(entry.getKey());
                break;
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("Removed the handler " + handler.getClass().getName() + " for all methods.");
        }
    }

    /**
     * Removes handlers with the handler manager. Each handler should be registered with a
     * Filter. If a handler should be disengaged only for a subset of allowed methods, those methods
     * can be specified as a string array.
     *
     * @param methods Methods for which the registered handler should be disengaged. Allowed values in
     * the string array are "GET", "PUT", "IMPORT", "DELETE", "PUT_CHILD", "IMPORT_CHILD", and
     * "INVOKE_ASPECT". If null is given, handler will be disengaged to all methods.
     *
     * @param filter Filter instance associated with the handler. Each filter that you pass in must
     * have the associated handler set to it.
     *
     * @param handler Handler instance to be unregistered.
     *
     * @see Filter#setHandler(Handler)
     */
    public void removeHandler(String[] methods, Filter filter, Handler handler) {

        if (methods == null || RegistryUtils.containsString(Filter.GET, methods)) {
            Handler h = getHandlerMap.get(filter);
            if (h.equals(handler)) {
                getHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.PUT, methods)) {
            Handler h = putHandlerMap.get(filter);
            if (h.equals(handler)) {
                putHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.DELETE, methods)) {
            Handler h = deleteHandlerMap.get(filter);
            if (h.equals(handler)) {
                deleteHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.IMPORT, methods)) {
            Handler h = importHandlerMap.get(filter);
            if (h.equals(handler)) {
                importHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.PUT_CHILD, methods)) {
            Handler h = putChildHandlerMap.get(filter);
            if (h.equals(handler)) {
                putChildHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.IMPORT_CHILD, methods)) {
            Handler h = importChildHandlerMap.get(filter);
            if (h.equals(handler)) {
                importChildHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.INVOKE_ASPECT, methods)) {
            Handler h = invokeAspectHandlerMap.get(filter);
            if (h.equals(handler)) {
                invokeAspectHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.MOVE, methods)) {
            Handler h = moveHandlerMap.get(filter);
            if (h.equals(handler)) {
                moveHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.COPY, methods)) {
            Handler h = copyHandlerMap.get(filter);
            if (h.equals(handler)) {
                copyHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.RENAME, methods)) {
            Handler h = renameHandlerMap.get(filter);
            if (h.equals(handler)) {
                renameHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.CREATE_LINK, methods)) {
            Handler h = createLinkHandlerMap.get(filter);
            if (h.equals(handler)) {
                createLinkHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.REMOVE_LINK, methods)) {
            Handler h = removeLinkHandlerMap.get(filter);
            if (h.equals(handler)) {
                removeLinkHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.ADD_ASSOCIATION, methods)) {
            Handler h = addAssociationHandlerMap.get(filter);
            if (h.equals(handler)) {
                addAssociationHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.REMOVE_ASSOCIATION, methods)) {
            Handler h = removeAssociationHandlerMap.get(filter);
            if (h.equals(handler)) {
                removeAssociationHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.APPLY_TAG, methods)) {
            Handler h = applyTagHandlerMap.get(filter);
            if (h.equals(handler)) {
                applyTagHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.REMOVE_TAG, methods)) {
            Handler h = removeTagHandlerMap.get(filter);
            if (h.equals(handler)) {
                removeTagHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.RATE_RESOURCE, methods)) {
            Handler h = rateResourceHandlerMap.get(filter);
            if (h.equals(handler)) {
                rateResourceHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.RESTORE_VERSION, methods)) {
            Handler h = restoreVersionHandlerMap.get(filter);
            if (h.equals(handler)) {
                restoreVersionHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.CREATE_VERSION, methods)) {
            Handler h = createVersionHandlerMap.get(filter);
            if (h.equals(handler)) {
                createVersionHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.EDIT_COMMENT, methods)) {
            Handler h = editCommentHandlerMap.get(filter);
            if (h.equals(handler)) {
                editCommentHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_ASSOCIATIONS, methods)) {
            Handler h = getAssociationsHandlerMap.get(filter);
            if (h.equals(handler)) {
                getAssociationsHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_ALL_ASSOCIATIONS, methods)) {
            Handler h = getAllAssociationsHandlerMap.get(filter);
            if (h.equals(handler)) {
                getAllAssociationsHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_RESOURCE_PATHS_WITH_TAG, methods)) {
            Handler h = getResourcePathsWithTagHandlerMap.get(filter);
            if (h.equals(handler)) {
                getResourcePathsWithTagHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_TAGS, methods)) {
            Handler h = getTagsHandlerMap.get(filter);
            if (h.equals(handler)) {
                getTagsHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.ADD_COMMENT, methods)) {
            Handler h = addCommentHandlerMap.get(filter);
            if (h.equals(handler)) {
                addCommentHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_COMMENTS, methods)) {
            Handler h = getCommentsHandlerMap.get(filter);
            if (h.equals(handler)) {
                getCommentsHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_RATING, methods)) {
            Handler h = getRatingHandlerMap.get(filter);
            if (h.equals(handler)) {
                getRatingHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_AVERAGE_RATING, methods)) {
            Handler h = getAverageRatingHandlerMap.get(filter);
            if (h.equals(handler)) {
                getAverageRatingHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.GET_VERSIONS, methods)) {
            Handler h = getVersionsHandlerMap.get(filter);
            if (h.equals(handler)) {
                getVersionsHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.EXECUTE_QUERY, methods)) {
            Handler h = executeQueryHandlerMap.get(filter);
            if (h.equals(handler)) {
                executeQueryHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.SEARCH_CONTENT, methods)) {
            Handler h = searchContentHandlerMap.get(filter);
            if (h.equals(handler)) {
                searchContentHandlerMap.remove(filter);
            }
        }

        if (methods == null || RegistryUtils.containsString(Filter.RESOURCE_EXISTS, methods)) {
            Handler h = resourceExistsHandlerMap.get(filter);
            if (h.equals(handler)) {
                resourceExistsHandlerMap.remove(filter);
            }
        }

        String methodInfo = "";
        if (methods == null) {
            methodInfo = "all";
        } else {
            for (int i = 0; i < methods.length; i++) {
                String method = methods[i];
                methodInfo = methodInfo + " " + method;
            }
        }
        if (log.isDebugEnabled()) {
            log.debug("Removed the handler " + filter.getClass().getName() +
                " --> " + handler.getClass().getName() + " for " + methodInfo + " methods.");
        }
    }

    /**
     * Manages the handler invocations of EDIT_COMMENT method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void editComment(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = editCommentHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleEditComment(requestContext)) {
                editCommentHandlerMap.get(filter).editComment(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of CREATE_VERSION method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void createVersion(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = createVersionHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleCreateVersion(requestContext)) {
                createVersionHandlerMap.get(filter).createVersion(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of RESTORE_VERSION method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void restoreVersion(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = restoreVersionHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleRestoreVersion(requestContext)) {
                restoreVersionHandlerMap.get(filter).restoreVersion(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of RATE_RESOURCE method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void rateResource(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = rateResourceHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleRateResource(requestContext)) {
                rateResourceHandlerMap.get(filter).rateResource(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of REMOVE_TAG method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void removeTag(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = removeTagHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleRemoveTag(requestContext)) {
                removeTagHandlerMap.get(filter).removeTag(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of APPLY_TAG method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void applyTag(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = applyTagHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleApplyTag(requestContext)) {
                applyTagHandlerMap.get(filter).applyTag(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of REMOVE_ASSOCIATION method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void removeAssociation(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = removeAssociationHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleRemoveAssociation(requestContext)) {
                removeAssociationHandlerMap.get(filter).removeAssociation(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of ADD_ASSOCIATION method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void addAssociation(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = addAssociationHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleAddAssociation(requestContext)) {
                addAssociationHandlerMap.get(filter).addAssociation(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of GET_ALL_ASSOCIATIONS method.
     *
     * @param requestContext Details of the request.
     * @return All associations.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Association[] getAllAssociations(RequestContext requestContext) throws RegistryException {

        Association[] associations = null;
        Iterator<Filter> filters = getAllAssociationsHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetAllAssociations(requestContext)) {
                associations = getAllAssociationsHandlerMap.get(filter).getAllAssociations(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return associations;
    }

    /**
     * Manages the handler invocations of GET_ASSOCIATIONS method.
     *
     * @param requestContext Details of the request.
     * @return Associations of matching type.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Association[] getAssociations(RequestContext requestContext) throws RegistryException {
        Association[] assocations = null;
        Iterator<Filter> filters = getAssociationsHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetAssociations(requestContext)) {
                assocations = getAssociationsHandlerMap.get(filter).getAssociations(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return assocations;
    }

    /**
     * Manages the handler invocations of GET_RESOURCE_PATHS_WITH_TAG method.
     *
     * @param requestContext Details of the request.
     * @return Tagged resource paths.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public TaggedResourcePath[] getResourcePathsWithTag(RequestContext requestContext) throws RegistryException {
        TaggedResourcePath[] taggedPaths = null;
        Iterator<Filter> filters = getResourcePathsWithTagHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetResourcePathsWithTag(requestContext)) {
                taggedPaths = getResourcePathsWithTagHandlerMap.get(filter).getResourcePathsWithTag(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return taggedPaths;
    }

   /**
     * Manages the handler invocations of GET_TAGS method.
     *
     * @param requestContext Details of the request.
     * @return Array of tags.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Tag[] getTags(RequestContext requestContext) throws RegistryException {

        Tag[] tags = null;
        Iterator<Filter> filters = getTagsHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetTags(requestContext)) {
                tags = getTagsHandlerMap.get(filter).getTags(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return tags;
    }

    /**
     * Manages the handler invocations of GET_COMMENTS method.
     *
     * @param requestContext Details of the request.
     * @return Array of comments.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Comment[] getComments(RequestContext requestContext) throws RegistryException {
        Comment[] comments = null;
        Iterator<Filter> filters = getCommentsHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetComments(requestContext)) {
                comments = getCommentsHandlerMap.get(filter).getComments(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return comments;
    }

    /**
     * Manages the handler invocations of GET_AVERAGE_RATING method.
     *
     * @param requestContext Details of the request.
     * @return Average rating.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public float getAverageRating(RequestContext requestContext) throws RegistryException {
        float rating = -1;
        Iterator<Filter> filters = getAverageRatingHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetAverageRating(requestContext)) {
                rating = getAverageRatingHandlerMap.get(filter).getAverageRating(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return rating;
    }

    /**
     * Manages the handler invocations of GET_RATING method.
     *
     * @param requestContext Details of the request.
     * @return Rating given by user.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public int getRating(RequestContext requestContext) throws RegistryException {
        int rating = -1;
        Iterator<Filter> filters = getRatingHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetRating(requestContext)) {
                rating = getRatingHandlerMap.get(filter).getRating(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return rating;
    }

    /**
     * Manages the handler invocations of GET_VERSIONS method.
     *
     * @param requestContext Details of the request.
     * @return Versions.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String[] getVersions(RequestContext requestContext) throws RegistryException {
        String[] versions = null;
        Iterator<Filter> filters = getVersionsHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGetVersions(requestContext)) {
                versions = getVersionsHandlerMap.get(filter).getVersions(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return versions;
    }

    /**
     * Manages the handler invocations of EXECUTE_QUERY method.
     *
     * @param requestContext Details of the request.
     * @return Collection of results.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Collection executeQuery(RequestContext requestContext) throws RegistryException {
        Collection collection = null;
        Iterator<Filter> filters = executeQueryHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleExecuteQuery(requestContext)) {
                collection = executeQueryHandlerMap.get(filter).executeQuery(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return collection;
    }

    /**
     * Manages the handler invocations of SEARCH_CONTENT method.
     *
     * @param requestContext Details of the request.
     * @return Collection of content.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Collection searchContent(RequestContext requestContext) throws RegistryException {
        Collection collection = null;
        Iterator<Filter> filters = searchContentHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleSearchContent(requestContext)) {
                collection = searchContentHandlerMap.get(filter).searchContent(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return collection;
    }

    /**
     * Manages the handler invocations of ADD_COMMENT method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String addComment(RequestContext requestContext) throws RegistryException {
        String comment = null;
        Iterator<Filter> filters = addCommentHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleAddComment(requestContext)) {
                comment = addCommentHandlerMap.get(filter).addComment(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return comment;
    }


    /**
     * Manages the handler invocations of GET method.
     *
     * @param requestContext Details of the request.
     * @return Resource requested from the GET operation.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Resource get(RequestContext requestContext) throws RegistryException {

        Resource resource = null;

        Iterator<Filter> filters = getHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGet(requestContext)) {
                Handler handler = getHandlerMap.get(filter);
                resource = handler.get(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return resource;
    }

    /**
     * Manages the handler invocations of PUT method.
     *
     * @param requestContext Details of the request.
     * @return Path where the resource is actually stored. Subsequent accesses to the resource
     * should use this path.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String put(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = putHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handlePut(requestContext)) {
                putHandlerMap.get(filter).put(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        if (requestContext.isProcessingComplete()) {
            return requestContext.getActualPath();
        } else {
            return null;
        }
    }

    /**
     * Manages the handler invocations of IMPORT method.
     *
     * @param requestContext Details of the request.
     * @return Path where the resource is actually stored. Subsequent accesses to the resource
     * should use this path.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String importResource(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = importHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleImportResource(requestContext)) {
                importHandlerMap.get(filter).importResource(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        if (requestContext.isProcessingComplete()) {
            return requestContext.getActualPath();
        } else {
            return null;
        }
    }

    /**
     * Manages the handler invocations of DELETE method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void delete(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = deleteHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleDelete(requestContext)) {
                deleteHandlerMap.get(filter).delete(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of PUT_CHILD method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void putChild(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = putChildHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handlePutChild(requestContext)) {
                putChildHandlerMap.get(filter).putChild(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of IMPORT_CHILD method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void importChild(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = importChildHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleImportChild(requestContext)) {
                importChildHandlerMap.get(filter).importChild(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of INVOKE_ASPECT method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void invokeAspect(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = invokeAspectHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleInvokeAspect(requestContext)) {
                invokeAspectHandlerMap.get(filter).invokeAspect(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of COPY method.
     *
     * @param requestContext Details of the request.
     * @return the actual path for the new resource.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String copy(RequestContext requestContext) throws RegistryException {

        String copiedPath = null;
        Iterator<Filter> filters = copyHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleCopy(requestContext)) {
                copiedPath = copyHandlerMap.get(filter).copy(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
        return copiedPath;
    }

    /**
     * Manages the handler invocations of MOVE method.
     *
     * @param requestContext Details of the request.
     * @return the actual path for the new resource.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String move(RequestContext requestContext) throws RegistryException {
        String movedPath = null;
        Iterator<Filter> filters = moveHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleMove(requestContext)) {
                movedPath = moveHandlerMap.get(filter).move(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
        return movedPath;
    }

    /**
     * Manages the handler invocations of RENAME method.
     *
     * @param requestContext Details of the request.
     * @return the actual path for the new resource.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String rename(RequestContext requestContext) throws RegistryException {
        String renamedPath = null;
        Iterator<Filter> filters = renameHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleRename(requestContext)) {
                renamedPath = renameHandlerMap.get(filter).rename(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
        return renamedPath;
    }

    /**
     * Manages the handler invocations of CREATE_LINK method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void createLink(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = createLinkHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleCreateLink(requestContext)) {
                createLinkHandlerMap.get(filter).createLink(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of REMOVE_LINK method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void removeLink(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = removeLinkHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleRemoveLink(requestContext)) {
                removeLinkHandlerMap.get(filter).removeLink(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of RESOURCE_EXISTS method.
     *
     * @param requestContext Details of the request.
     * @return whether the resource exists
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public boolean resourceExists(RequestContext requestContext) throws RegistryException {
        boolean resourceExist = false;
        Iterator<Filter> filters = resourceExistsHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleResourceExists(requestContext)) {
                resourceExist = resourceExistsHandlerMap.get(filter).resourceExists(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return resourceExist;
    }


    /**
     * Manages the handler invocations of DUMP method.
     *
     * @param requestContext Details of the request.
     * @return the dumped element
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public OMElement dump(RequestContext requestContext) throws RegistryException {
        OMElement dumpedElement = null;
        Iterator<Filter> filters = dumpMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleDump(requestContext)) {
                dumpMap.get(filter).dump(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
        return dumpedElement;
    }


    /**
     * Manages the handler invocations of RESTORE method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void restore(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = restoreMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleDump(requestContext)) {
                restoreMap.get(filter).restore(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }
}
