/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.handlers;

import org.wso2.carbon.registry.core.jdbc.Repository;
import org.wso2.carbon.registry.core.jdbc.VersionRepository;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.apache.axiom.om.OMElement;

import java.util.HashMap;
import java.util.Map;
import java.io.Writer;
import java.io.Reader;

/**
 * Objects of this class contains the information about the current request to the registry.
 * All attributes of such objects should be either null or should contain a *valid" value. That is,
 * if an attribute contains a non-null value it is safe to assume that attribute contains a valid
 * value as mentioned by the javadoc of that attribute.
 */
public class RequestContext {

    /**
     * Specifies whether the processing of the current request is completed or not. If this is true,
     * any further processing of the request is not done.
     */
    private boolean processingComplete;

    /**
     * Path of the currently processing resource as given by the client of the Registry API
     */
    private ResourcePath resourcePath;

    /**
     * Path of the currently processing resource, which may have been modified by Registry impls to
     * represent the actual path in which resource resides.
     */
    private String actualPath;

    /**
     * Currently processing resource. This is the resource referred by the actual path.
     */
    private Resource resource;

    /**
     * URL where the current resource content is located.
     */
    private String sourceURL;

    /**
     * Path of the parent collection of the current resource.
     */
    private String parentPath;

    /**
     * Parent collection of the current resource.
     */
    private Collection parentCollection;

    private Map<String, Object> properties = new HashMap<String, Object>();

    private Registry registry;
    private Repository repository;
    private VersionRepository versionRepository;

    private Aspect aspect;
    private String action;

    private String sourcePath;
    private String targetPath;
    private String targetSubPath;

    private String userName;
    private String tag;
    private String associationType;
    private int rating;
    private Comment comment;
    private String versionPath;
    private Map queryParameters;
    private String keywords;
    private Reader dumpingReader;
    private Writer dumpingWriter;

    private RegistryContext registryContext;

    public RequestContext(
            Registry registry, Repository repository, VersionRepository versionRepository) {
        this.registry = registry;
        this.repository = repository;
        this.versionRepository = versionRepository;
    }

    public Registry getRegistry() {
        return registry;
    }

    public Repository getRepository() {
        return repository;
    }

    public VersionRepository getVersionRepository() {
        return versionRepository;
    }

    public boolean isProcessingComplete() {
        return processingComplete;
    }

    public void setProcessingComplete(boolean processingComplete) {
        this.processingComplete = processingComplete;
    }

    public ResourcePath getResourcePath() {
        if (resourcePath == null && resource != null) {
            resourcePath = new ResourcePath(resource.getPath());
        }
        return resourcePath;
    }

    public void setResourcePath(ResourcePath resourcePath) {
        this.resourcePath = resourcePath;
    }

    public String getSourceURL() {
        return sourceURL;
    }

    public void setSourceURL(String sourceURL) {
        this.sourceURL = sourceURL;
    }

    public String getActualPath() {
        return actualPath;
    }

    public void setActualPath(String actualPath) {
        this.actualPath = actualPath;
    }

    public Resource getResource() {
        return resource;
    }

    public void setResource(Resource resource) {
        this.resource = resource;
    }

    public String getParentPath() {
        return parentPath;
    }

    public void setParentPath(String parentPath) {
        this.parentPath = parentPath;
    }

    public Aspect getAspect() {
        return aspect;
    }

    public void setAspect(Aspect aspect) {
        this.aspect = aspect;
    }

    public String getAction() {
        return action;
    }

    public void setAction(String action) {
        this.action = action;
    }

    public Collection getParentCollection() {
        return parentCollection;
    }

    public void setParentCollection(Collection parentCollection) {
        this.parentCollection = parentCollection;
    }

    public void setProperty(String name, Object value) {
        properties.put(name, value);
    }

    public Object getProperty(String name) {
        return properties.get(name);
    }

    public String getSourcePath() {
        return sourcePath;
    }

    public void setSourcePath(String sourcePath) {
        this.sourcePath = sourcePath;
    }

    public String getTargetPath() {
        return targetPath;
    }

    public void setTargetPath(String targetPath) {
        this.targetPath = targetPath;
    }

    public String getTargetSubPath() {
        return targetSubPath;
    }

    public void setTargetSubPath(String targetSubPath) {
        this.targetSubPath = targetSubPath;
    }

    public String getUserName() {
        return userName;
    }

    public void setUserName(String userName) {
        this.userName = userName;
    }

    public String getTag() {
        return tag;
    }

    public void setTag(String tag) {
        this.tag = tag;
    }

    public String getAssociationType() {
        return associationType;
    }

    public void setAssociationType(String associationType) {
        this.associationType = associationType;
    }

    public int getRating() {
        return rating;
    }

    public void setRating(int rating) {
        this.rating = rating;
    }

    public Comment getComment() {
        return comment;
    }

    public void setComment(Comment comment) {
        this.comment = comment;
    }

    public String getVersionPath() {
        return versionPath;
    }

    public void setVersionPath(String versionPath) {
        this.versionPath = versionPath;
    }

    public Map getQueryParameters() {
        return queryParameters;
    }

    public void setQueryParameters(Map queryParameters) {
        this.queryParameters = queryParameters;
    }

    public String getKeywords() {
        return keywords;
    }

    public void setKeywords(String keywords) {
        this.keywords = keywords;
    }

    public Reader getDumpingReader() {
        return dumpingReader;
    }

    public void setDumpingReader(Reader dumpingReader) {
        this.dumpingReader = dumpingReader;
    }


    public Writer getDumpingWriter() {
        return dumpingWriter;
    }

    public void setDumpingWriter(Writer dumpingWriter) {
        this.dumpingWriter = dumpingWriter;
    }


    public RegistryContext getRegistryContext() {
        return registryContext;
    }

    public void setRegistryContext(RegistryContext registryContext) {
        this.registryContext = registryContext;
    }
}
