/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.handlers.builtin;

import org.wso2.carbon.registry.app.RemoteRegistry;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.handlers.Handler;
import org.wso2.carbon.registry.core.jdbc.handlers.RequestContext;
import org.wso2.carbon.registry.core.jdbc.handlers.HandlerManager;
import org.wso2.carbon.registry.core.jdbc.dao.AssociationDAO;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;

import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLInputFactory;
import java.io.*;

public class MountHandler extends Handler {

    private static final Log log = LogFactory.getLog(ResourceImpl.class);

    private String id;
    private String conURL;
    private String mountPoint;
    private String userName;
    private String password;
    private String subPath;

    public void put(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            Resource resource = requestContext.getResource();

            // removes the following properties if they are present
            resource.removeProperty(RegistryConstants.REGISTRY_LINK);
            resource.removeProperty(RegistryConstants.REGISTRY_USER);
            resource.removeProperty(RegistryConstants.REGISTRY_MOUNT);
            resource.removeProperty(RegistryConstants.REGISTRY_REAL_PATH);

            // sets the actual path of the resource

            ((ResourceImpl) resource).setPath(actualPath);
            remoteRegistry.put(actualPath, resource);
            requestContext.setProcessingComplete(true);
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public boolean resourceExists(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        boolean resourceExists = false;
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            resourceExists = remoteRegistry.resourceExists(actualPath);
            requestContext.setProcessingComplete(true);
        } catch (Exception e) {
            log.error(e);
//            throw new RegistryException(e.getMessage());
        }
        return resourceExists;
    }

    public Resource get(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        Resource tempResource;
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            tempResource = remoteRegistry.get(actualPath);

            if (tempResource instanceof Collection) {
                String[] paths = (String[]) tempResource.getContent();
                for (int i = 0; i < paths.length; i++) {
                    if (subPath != null && subPath.length() != 0) {
                        paths[i] = this.mountPoint + paths[i].substring(subPath.length());
                    } else {
                        paths[i] = this.mountPoint + paths[i];
                    }
                }
                tempResource.setContent(paths);
            }
            ((ResourceImpl) tempResource).setPath(fullPath);
            ((ResourceImpl) tempResource).setAuthorUserName(CurrentSession.getUser());
            ((ResourceImpl) tempResource).setUserName(CurrentSession.getUser());
        } catch (Exception e) {
            log.error("Could not mount remote instance " + this.conURL + " " + e.getMessage());
            tempResource = requestContext.getRepository().get(this.mountPoint);
            tempResource.setProperty("registry.absent","true");
            tempResource.setDescription("Couldn't mount the remote instance. Content can't be displayed.");
        }

        tempResource.addProperty("registry.link", "true");
        tempResource.addProperty("registry.mount", "true");
        tempResource.addProperty("registry.user", this.userName);
        tempResource.addProperty("registry.realpath", this.conURL + "/resourceContent?path=" + actualPath);
        requestContext.setProcessingComplete(true);
        return tempResource;
    }

    public void delete(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        RegistryContext registryContext = requestContext.getRegistryContext();
        if (fullPath.equals(this.mountPoint)) {
            requestContext.getRegistry().removeLink(requestContext.getResource().getPath());
            if (registryContext != null) {
                HandlerManager hm = registryContext.getHandlerManager();
                hm.removeHandler(this);
            }
        } else {
            String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
            if (subPath != null) {
                actualPath = subPath + actualPath;
            }
            if (actualPath.length() == 0) {
                actualPath = "/";
            }
            try {
                setSystemProperties();
                RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
                remoteRegistry.delete(actualPath);
            } catch (Exception e) {
                String msg = "Could not delete the remote resource.";
                log.error(msg);
                throw new RegistryException(msg, e);
            }
        }
        requestContext.setProcessingComplete(true);
    }

    public String rename(RequestContext requestContext) throws RegistryException {
        String fullResourcePath = requestContext.getSourcePath();
        String fullTargetPath = requestContext.getTargetPath();
        RegistryContext registryContext = requestContext.getRegistryContext();

        String newPath;
        if (fullResourcePath.equals(this.mountPoint)) {
            newPath = requestContext.getRepository().rename(new ResourcePath(fullResourcePath), fullTargetPath);
            requestContext.getRegistry().createLink(fullTargetPath, this.id);
            requestContext.getRegistry().removeLink(this.mountPoint);
            if (registryContext != null) {
                HandlerManager hm = registryContext.getHandlerManager();
                hm.removeHandler(this);
            }
        } else {
            String actualResourcePath = fullResourcePath.substring(
                    this.mountPoint.length(), fullResourcePath.length());
            String actualTargetPath = fullTargetPath.substring(
                    this.mountPoint.length(), fullTargetPath.length());
            if (subPath != null) {
                actualResourcePath = subPath + actualResourcePath;
                actualTargetPath = subPath + actualTargetPath;
            }
            if (actualResourcePath.length() == 0) {
                actualResourcePath = "/";
            }
            if (actualTargetPath.length() == 0) {
                actualTargetPath = "/";
            }

            try {
                setSystemProperties();
                RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
                newPath = remoteRegistry.rename(actualResourcePath, actualTargetPath);
            } catch (Exception e) {
                String msg = "Could not rename the remote resource.";
                log.error(msg);
                throw new RegistryException(msg, e);
            }
        }
        requestContext.setProcessingComplete(true);

        return newPath;
    }

    public String move(RequestContext requestContext) throws RegistryException {
        String fullResourcePath = requestContext.getSourcePath();
        String fullTargetPath = requestContext.getTargetPath();
        RegistryContext registryContext = requestContext.getRegistryContext();

        try {
            if (fullResourcePath.equals(this.mountPoint)) {
                requestContext.getRepository().move(new ResourcePath(fullResourcePath), fullTargetPath);
                requestContext.getRegistry().createLink(fullTargetPath, this.id);
                requestContext.getRegistry().removeLink(this.mountPoint);
                if (registryContext != null) {
                    HandlerManager hm = registryContext.getHandlerManager();
                    hm.removeHandler(this);
                }
            } else if (fullResourcePath.startsWith(this.mountPoint)) {
                String actualResourcePath = fullResourcePath.substring(
                        this.mountPoint.length(), fullResourcePath.length());
                String actualTargetPath;
                if (subPath != null) {
                    actualResourcePath = subPath + actualResourcePath;
                }

                setSystemProperties();
                RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
                if (fullTargetPath.startsWith(this.mountPoint)) {
                    actualTargetPath = fullTargetPath.substring(
                            this.mountPoint.length(), fullTargetPath.length());
                    if (subPath != null) {
                        actualTargetPath = subPath + actualTargetPath;
                    }
                    remoteRegistry.move(actualResourcePath, actualTargetPath);
                } else {
                    actualTargetPath = fullTargetPath;

                    StringWriter out = new StringWriter();
                    requestContext.getRegistry().dump(actualResourcePath, out);
                    Reader input = new StringReader(out.toString());
                    remoteRegistry.restore(actualTargetPath, input);
                    requestContext.getRegistry().delete(actualResourcePath);
                }
            } else if (fullTargetPath.contains(this.mountPoint)) {
                String actualTargetPath = fullTargetPath.substring(
                        this.mountPoint.length(), fullTargetPath.length());
                String actualResourcePath;
                if (subPath != null) {
                    actualTargetPath = subPath + actualTargetPath;
                }
                setSystemProperties();
                RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);

                actualResourcePath = fullResourcePath;
                StringWriter out = new StringWriter();

                requestContext.getRegistry().dump(actualResourcePath, out);
                Reader input = new StringReader(out.toString());
                remoteRegistry.restore(actualTargetPath, input);
                requestContext.getRegistry().delete(actualResourcePath);
            }
        } catch (Exception e) {
            String msg = "Could not move the remote resource " + fullResourcePath + ".";
            log.error(msg);
            throw new RegistryException(msg, e);
        }
        requestContext.setProcessingComplete(true);

        return fullTargetPath;
    }

    public String copy(RequestContext requestContext) throws RegistryException {
        String fullResourcePath = requestContext.getSourcePath();
        String fullTargetPath = requestContext.getTargetPath();

        try {
            if (fullResourcePath.equals(this.mountPoint)) {
                requestContext.getRegistry().createLink(fullTargetPath, this.id);
            } else if (fullResourcePath.startsWith(this.mountPoint)) {
                String actualResourcePath = fullResourcePath.substring(
                        this.mountPoint.length(), fullResourcePath.length());
                String actualTargetPath;
                if (subPath != null) {
                    actualResourcePath = subPath + actualResourcePath;
                }

                setSystemProperties();
                RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
                if (fullTargetPath.startsWith(this.mountPoint)) {
                    actualTargetPath = fullTargetPath.substring(
                            this.mountPoint.length(), fullTargetPath.length());
                    if (subPath != null) {
                        actualTargetPath = subPath + actualTargetPath;
                    }
                    remoteRegistry.copy(actualResourcePath, actualTargetPath);
                } else {
                    actualTargetPath = fullTargetPath;

                    StringWriter out = new StringWriter();
                    remoteRegistry.dump(actualResourcePath, out);
                    Reader input = new StringReader(out.toString());
                    requestContext.getRegistry().restore(actualTargetPath, input);

                }
            } else if (fullTargetPath.contains(this.mountPoint)) {
                String actualTargetPath = fullTargetPath.substring(
                        this.mountPoint.length(), fullTargetPath.length());
                String actualResourcePath;
                if (subPath != null) {
                    actualTargetPath = subPath + actualTargetPath;
                }
                setSystemProperties();
                RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);

                actualResourcePath = fullResourcePath;
                StringWriter out = new StringWriter();


                requestContext.getRegistry().dump(actualResourcePath, out);
                Reader input = new StringReader(out.toString());
                remoteRegistry.restore(actualTargetPath, input);
                requestContext.getRegistry().delete(actualResourcePath);
            }
        } catch (Exception e) {
            String msg = "Could not copy the remote resource " + fullResourcePath + ".";
            log.error(msg);
            throw new RegistryException(msg, e);
        }
        requestContext.setProcessingComplete(true);

        return fullTargetPath;
    }

    public float getAverageRating(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        float rating;
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            rating = remoteRegistry.getAverageRating(actualPath);
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("Could not get average ratings from" + this.conURL);
            }
            return 0;
        }
        requestContext.setProcessingComplete(true);

        return rating;
    }

    public int getRating(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        int rating;
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            rating = remoteRegistry.getRating(actualPath, this.userName);
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("Could not get ratings from " + this.conURL);
            }
            return 0;
        }
        requestContext.setProcessingComplete(true);

        return rating;
    }

    public void rateResource(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        CurrentSession.setUser(this.userName);

        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            remoteRegistry.rateResource(actualPath, requestContext.getRating());
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
        requestContext.setProcessingComplete(true);
    }

    public Comment[] getComments(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        Comment[] comments;
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            comments = remoteRegistry.getComments(actualPath);
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("Could not get comments from" + this.conURL);
            }
            return new Comment[1];
        }
        requestContext.setProcessingComplete(true);

        return comments;
    }

    public String addComment(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        String commentPath;
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            commentPath = remoteRegistry.addComment(actualPath, requestContext.getComment());
        } catch (Exception e) {
            String msg = "Could not add comment to the resource in " + this.conURL;
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
            throw new RegistryException(msg, e);
        }
        requestContext.setProcessingComplete(true);

        return commentPath;
    }

    public Tag[] getTags(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        Tag[] tags;
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            tags = remoteRegistry.getTags(actualPath);
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("Could not get tags from" + this.conURL);
            }
            return new Tag[1];
        }
        requestContext.setProcessingComplete(true);

        return tags;
    }

    public void applyTag(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            remoteRegistry.applyTag(actualPath, requestContext.getTag());
        } catch (Exception e) {
            String msg = "Could not apply tag to the resource in " + this.conURL;
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
            throw new RegistryException(msg, e);
        }
        requestContext.setProcessingComplete(true);

    }

    public Association[] getAllAssociations(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }

        Association[] associations;
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            associations = remoteRegistry.getAllAssociations(actualPath);
        } catch (Exception e) {
            String msg = "Could not get associations of " + this.conURL;
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
            return new Association[1];
        }
        String sourcePath;
        String destPath;
        for (int i = 0; i < associations.length; i++) {
            sourcePath = associations[i].getSourcePath();
            destPath = associations[i].getDestinationPath();
            if (subPath != null && subPath.length() != 0) {
                associations[i].setSourcePath(this.mountPoint + sourcePath.substring(subPath.length()));
                associations[i].setDestinationPath(this.mountPoint + destPath.substring(subPath.length()));
            } else {
                associations[i].setSourcePath(this.mountPoint + sourcePath);
                associations[i].setDestinationPath(this.mountPoint + destPath);
            }
        }
        requestContext.setProcessingComplete(true);
        return associations;
    }

    public Association[] getAssociations(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }

        Association[] associations;
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            associations = remoteRegistry.getAssociations(actualPath, requestContext.getAssociationType());
        } catch (Exception e) {
            String msg = "Could not get associations of " + this.conURL;
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
            return new Association[1];
        }
        String sourcePath;
        String destPath;
        for (int i = 0; i < associations.length; i++) {
            sourcePath = associations[i].getSourcePath();
            destPath = associations[i].getDestinationPath();
            if (subPath != null && subPath.length() != 0) {
                associations[i].setSourcePath(this.mountPoint + sourcePath.substring(subPath.length()));
                associations[i].setDestinationPath(this.mountPoint + destPath.substring(subPath.length()));
            } else {
                associations[i].setSourcePath(this.mountPoint + sourcePath);
                associations[i].setDestinationPath(this.mountPoint + destPath);
            }
        }
        requestContext.setProcessingComplete(true);
        return associations;
    }

    public void addAssociation(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getSourcePath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            remoteRegistry.addAssociation(actualPath, requestContext.getTargetPath(),
                    requestContext.getAssociationType());
        } catch (Exception e) {
            String msg = "Could not add associations for " + this.conURL;
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
        }
        requestContext.setProcessingComplete(true);
    }

    public void removeAssociation(RequestContext requestContext) throws RegistryException{
        String fullPath = requestContext.getSourcePath();
        String targetPath = requestContext.getTargetPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        targetPath = targetPath.substring(this.mountPoint.length(), targetPath.length());

        if (subPath != null) {
            actualPath = subPath + actualPath;
            targetPath = subPath + targetPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }

        if(targetPath.length() == 0){
            targetPath = "/";
        }

        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            remoteRegistry.removeAssociation(actualPath, targetPath,
                    requestContext.getAssociationType());
        } catch (Exception e) {
            String msg = "Could not remove associations for " + this.conURL;
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
        }
        requestContext.setProcessingComplete(true);
    }

    public void importResource(RequestContext requestContext)
            throws RegistryException {

        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());

        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            remoteRegistry.importResource(actualPath, requestContext.getSourceURL(),
                    requestContext.getResource());
        } catch (Exception e) {
            String msg = "Could not import resource from " + requestContext.getSourceURL();
            if (log.isWarnEnabled()) {
                log.warn(msg);
            }
        }
        requestContext.setProcessingComplete(true);
    }

    public void dump(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        Writer writer = requestContext.getDumpingWriter();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);

            remoteRegistry.dump(actualPath, writer);
            requestContext.setProcessingComplete(true);
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }


    public void restore(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        if (subPath != null) {
            actualPath = subPath + actualPath;
        }
        if (actualPath.length() == 0) {
            actualPath = "/";
        }
        try {
            setSystemProperties();
            RemoteRegistry remoteRegistry = new RemoteRegistry(this.conURL, this.userName, this.password);
            remoteRegistry.restore(actualPath, requestContext.getDumpingReader());
            requestContext.setProcessingComplete(true);
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public String getId() {
        return id;
    }

    public void setId(String id) {
        this.id = id;
    }

    public String getConURL() {
        return conURL;
    }

    public void setConURL(String conURL) {
        if (conURL.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            this.conURL = conURL.substring(0, conURL.lastIndexOf(
                    RegistryConstants.PATH_SEPARATOR) - 1);
        }
        this.conURL = conURL;
    }

    public String getMountPoint() {
        return mountPoint;
    }

    public void setMountPoint(String mountPoint) {
        this.mountPoint = mountPoint;
    }

    public String getUserName() {
        return userName;
    }

    public void setUserName(String userName) {
        this.userName = userName;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public String getSubPath() {
        return subPath;
    }

    public void setSubPath(String subPath) {
        if (subPath.equals(RegistryConstants.ROOT_PATH)) {
            // skip just the ROOT PATH
            this.subPath = "";
        } else if (subPath.length() > 1 &&
                subPath.charAt(subPath.length() - 1) == RegistryConstants.PATH_SEPARATOR.charAt(0)) {
            // skip the PATH seperator at the end 
            this.subPath = subPath.substring(0, subPath.length() - 1);
        } else {
            this.subPath = subPath;
        }
    }

    private void setSystemProperties() {
        System.setProperty("javax.net.ssl.trustStore", System.getProperty("carbon.home") + "/resources/security/client-truststore.jks");
        System.setProperty("javax.net.ssl.trustStorePassword", "wso2carbon");
        System.setProperty("javax.net.ssl.trustStoreType", "JKS");
    }
}
