/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.handlers.builtin;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dao.AssociationDAO;
import org.wso2.carbon.registry.core.jdbc.handlers.Handler;
import org.wso2.carbon.registry.core.jdbc.handlers.RequestContext;
import org.wso2.carbon.registry.core.jdbc.handlers.HandlerManager;
import org.wso2.carbon.registry.core.session.CurrentSession;

public class SymLinkHandler extends Handler {

    private static final Log log = LogFactory.getLog(ResourceImpl.class);

    private String mountPoint;
    private String targetPoint;

    public void put(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);
        try {
            Resource resource = requestContext.getResource();

            // removes the following properties if they are present
            resource.removeProperty(RegistryConstants.REGISTRY_MOUNT_POINT);
            resource.removeProperty(RegistryConstants.REGISTRY_TARGET_POINT);
            resource.removeProperty(RegistryConstants.REGISTRY_LINK);

            // sets the actual path of the resource
            requestContext.getRegistry().put(actualPath, requestContext.getResource());
            requestContext.setProcessingComplete(true);
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }



    public boolean resourceExists(RequestContext requestContext) throws RegistryException {
        boolean resourceExists = false;
        String fullPath = requestContext.getResourcePath().getPath();
        String subPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        String actualPath = this.targetPoint+subPath;
        try {
            resourceExists = requestContext.getRegistry().resourceExists(actualPath);
        } catch (Exception e) {
            if(log.isWarnEnabled()) {
                log.warn("Could not create the symbolic link. Target " + actualPath + "not found.");
            }
        }
        requestContext.setProcessingComplete(true);
        return resourceExists;
    }

    public Resource get(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String subPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
        String actualPath = this.targetPoint+subPath;
        Resource tempResource = null;
        try {
            tempResource = requestContext.getRegistry().get(actualPath);
            if (tempResource instanceof Collection) {
                String[] paths = (String[]) tempResource.getContent();
                for (int i = 0; i < paths.length; i++) {
                    if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                        paths[i] = this.mountPoint + paths[i];
                    } else {
                        paths[i] = this.mountPoint + paths[i].substring(this.targetPoint.length(), paths[i].length());
                    }
                }
                tempResource.setContent(paths);
            }
            ((ResourceImpl) tempResource).setPath(fullPath);
            ((ResourceImpl) tempResource).setAuthorUserName(CurrentSession.getUser());
            ((ResourceImpl) tempResource).setUserName(CurrentSession.getUser());
        } catch (Exception e) {
            if(log.isWarnEnabled()) {
                log.warn("Could not create the symbolic link. Target " + actualPath + "not found.");
            }
            tempResource = requestContext.getRepository().get(this.mountPoint);
            tempResource.setProperty("registry.absent","true");
            tempResource.setDescription("Couldn't create the symbolic link. Content can't be displayed.");
        }
        tempResource.addProperty("registry.link", "true");
        tempResource.addProperty("registry.mountpoint", this.mountPoint);
        tempResource.addProperty("registry.targetpoint", this.targetPoint);
        //Used to store paths when there are recursive calls
        tempResource.addProperty("registry.path", fullPath);
        requestContext.setProcessingComplete(true);
        return tempResource;
    }

    public void delete(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        RegistryContext registryContext = requestContext.getRegistryContext();
        if (fullPath.equals(this.mountPoint)) {
            requestContext.getRegistry().removeLink(fullPath);
            if (registryContext != null) {
                HandlerManager hm = registryContext.getHandlerManager();
                hm.removeHandler(this);
            }
        } else {
            String subPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
            String actualPath = this.targetPoint+subPath;
            requestContext.getRegistry().delete(actualPath);
        }
        requestContext.setProcessingComplete(true);
    }

    public String rename(RequestContext requestContext) throws RegistryException {
        String fullResourcePath = requestContext.getSourcePath();
        String fullTargetPath = requestContext.getTargetPath();
        RegistryContext registryContext = requestContext.getRegistryContext();
        if (fullResourcePath.equals(this.mountPoint)) {
            requestContext.getRegistry().createLink(fullTargetPath, this.targetPoint);
            requestContext.getRegistry().removeLink(this.mountPoint);
            if (registryContext != null) {
                HandlerManager hm = registryContext.getHandlerManager();
                hm.removeHandler(this);
            }
        } else {
            String subPath = fullResourcePath.substring(this.mountPoint.length(), fullResourcePath.length());
            String actualResourcePath = this.targetPoint + subPath;
            subPath = fullTargetPath.substring(this.mountPoint.length(), fullTargetPath.length());
            String actualTargetPath = this.targetPoint + subPath;
            requestContext.getRegistry().rename(actualResourcePath, actualTargetPath);
        }
        requestContext.setProcessingComplete(true);

        return fullTargetPath;
    }

    public String move(RequestContext requestContext) throws RegistryException {
        String fullResourcePath = requestContext.getSourcePath();
        String fullTargetPath = requestContext.getTargetPath();
        RegistryContext registryContext = requestContext.getRegistryContext();
        if (fullResourcePath.equals(this.mountPoint)) {
            requestContext.getRegistry().createLink(fullTargetPath, this.targetPoint);
            requestContext.getRegistry().removeLink(this.mountPoint);
            if (registryContext != null) {
                HandlerManager hm = registryContext.getHandlerManager();
                hm.removeHandler(this);
            }
        } else if (fullResourcePath.startsWith(this.mountPoint)) {
            String subPath = fullResourcePath.substring(this.mountPoint.length(), fullResourcePath.length());
            String actualResourcePath;
            String actualTargetPath = fullTargetPath;

            if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                actualResourcePath = subPath;
            } else {
                actualResourcePath = this.targetPoint + subPath;
            }
            requestContext.getRegistry().move(actualResourcePath, actualTargetPath);
        } else if (fullTargetPath.startsWith(this.mountPoint)) {
            String subPath = fullTargetPath.substring(this.mountPoint.length(), fullTargetPath.length());
            String actualResourcePath = fullResourcePath;
            String actualTargetPath;

            if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                actualTargetPath = subPath;
            } else {
                actualTargetPath = this.targetPoint + subPath;
            }
            requestContext.getRegistry().move(actualResourcePath, actualTargetPath);
        }
        requestContext.setProcessingComplete(true);

        return fullTargetPath;
    }

    public String copy(RequestContext requestContext) throws RegistryException {
        String fullResourcePath = requestContext.getSourcePath();
        String fullTargetPath = requestContext.getTargetPath();
        if (fullResourcePath.equals(this.mountPoint)) {
            requestContext.getRegistry().createLink(fullTargetPath, this.targetPoint);
        } else if (fullResourcePath.startsWith(this.mountPoint)) {
            String subPath = fullResourcePath.substring(this.mountPoint.length(), fullResourcePath.length());
            String actualResourcePath;
            String actualTargetPath = fullTargetPath;

            if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                actualResourcePath = subPath;
            } else {
                actualResourcePath = this.targetPoint + subPath;
            }
            requestContext.getRegistry().copy(actualResourcePath, actualTargetPath);
        } else if (fullTargetPath.startsWith(this.mountPoint)) {
            String subPath = fullTargetPath.substring(this.mountPoint.length(), fullTargetPath.length());
            String actualResourcePath = fullResourcePath;
            String actualTargetPath;

            if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                actualTargetPath = subPath;
            } else {
                actualTargetPath = this.targetPoint + subPath;
            }
            requestContext.getRegistry().copy(actualResourcePath, actualTargetPath);
        }
        requestContext.setProcessingComplete(true);

        return fullTargetPath;
    }

    public float getAverageRating(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        float rating = requestContext.getRegistry().getAverageRating(actualPath);
        requestContext.setProcessingComplete(true);

        return rating;
    }

    public int getRating(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        int rating = requestContext.getRegistry().getRating(actualPath, requestContext.getUserName());
        requestContext.setProcessingComplete(true);

        return rating;
    }

    public void rateResource(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        requestContext.getRegistry().rateResource(actualPath, requestContext.getRating());
        requestContext.setProcessingComplete(true);
    }

    public Comment[] getComments(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        Comment [] comments = requestContext.getRegistry().getComments(actualPath);
        requestContext.setProcessingComplete(true);

        return comments;
    }

    public String addComment(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        String commentPath = requestContext.getRegistry().addComment(actualPath,
                requestContext.getComment());

        requestContext.setProcessingComplete(true);
        return commentPath;
    }

    public Tag[] getTags(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        Tag [] tags = requestContext.getRegistry().getTags(actualPath);
        requestContext.setProcessingComplete(true);

        return tags;
    }

    public void applyTag(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        requestContext.getRegistry().applyTag(actualPath,
        requestContext.getTag());
        requestContext.setProcessingComplete(true);    
    }

    public Association[] getAllAssociations(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        Association [] associations = new AssociationDAO().getAllAssociations(actualPath);
        String sourcePath;
        String destinationPath;
        for (int i=0; i<associations.length; i++) {
            sourcePath = associations[i].getSourcePath();
            destinationPath = associations[i].getDestinationPath();
            // eaither source path or destination path should be equal to the actual path
            // we are changing that path to the rewritten (symlinking) path
            if (sourcePath.equals(actualPath)) {
                associations[i].setSourcePath(fullPath);
            }
            if (destinationPath.equals(actualPath)) {
                associations[i].setDestinationPath(fullPath);
            }
        }
        requestContext.setProcessingComplete(true);
        return associations;
    }

    public Association[] getAssociations(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        Association [] associations = requestContext.getRegistry().getAssociations(actualPath,
                requestContext.getAssociationType());
        String sourcePath;
        for (int i=0; i<associations.length; i++) {
            sourcePath = associations[i].getSourcePath();
            if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                        associations[i].setSourcePath(this.mountPoint + sourcePath);
                    } else {
                        associations[i].setSourcePath(this.mountPoint + sourcePath.substring(this.targetPoint.length(), sourcePath.length()));
                    }
        }
        requestContext.setProcessingComplete(true);
        return associations;
    }

    public void addAssociation(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getSourcePath();
        String actualPath = getActualPath(fullPath);

        requestContext.getRegistry().addAssociation(actualPath, requestContext.getTargetPath(),
                requestContext.getAssociationType());
        requestContext.setProcessingComplete(true);
    }

    public void importResource(RequestContext requestContext)
            throws RegistryException {

        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        requestContext.getRegistry().importResource(actualPath, requestContext.getSourceURL(),
                requestContext.getResource());

        requestContext.setProcessingComplete(true);
    }

    public void createLink(RequestContext requestContext) throws RegistryException {
        String fullPath = requestContext.getResourcePath().getPath();
        String actualPath = getActualPath(fullPath);

        requestContext.getRegistry().createLink(actualPath, requestContext.getTargetPath());
        requestContext.setProcessingComplete(true);
    }


    public String getMountPoint() {
        return mountPoint;
    }

    public void setMountPoint(String mountPoint) {
        this.mountPoint = mountPoint;
    }

    public String getTargetPoint() {
        return targetPoint;
    }

    public void setTargetPoint(String targetPoint) {
        this.targetPoint = targetPoint;
    }

    private String getActualPath(String fullPath) {
        String actualPath;
        if (fullPath.equals(this.mountPoint)) {
            actualPath = this.targetPoint;
        } else {
            String subPath = fullPath.substring(this.mountPoint.length(), fullPath.length());
            if (this.targetPoint.equals(RegistryConstants.PATH_SEPARATOR)) {
                actualPath = subPath;
            } else {
                actualPath = this.targetPoint + subPath;
            }
        }

        return actualPath;
    }

}
