/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.handlers.filters;

import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.handlers.RequestContext;

/**
 * This is a built-in Filter implementation to match against the URL (path) of the resources. This
 * can match different URLs for different handler operations. URLs to match for necessary handler
 * operations can be given as regular expressions. If a URL for particular operation is not
 * specified, it will always evaluate to false.
 *
 * Handler authors can use this Filter in thier configurations if the filtering requirement is only
 * to match against URLs of the resources.
 */
public class URLMatcher extends Filter {

    /**
     * URLs to match against resource path for handler oprations. They should be in the form of
     * regular expressions.
     */
    private String getPattern;
    private String putPattern;
    private String importPattern;
    private String deletePattern;
    private String putChildPattern;
    private String importChildPattern;
    private String invokeAspectPattern;
    private String movePattern;
    private String copyPattern;
    private String renamePattern;
    private String createLinkPattern;
    private String removeLinkPattern;
    private String resourceExistsPattern;
    private String addAssociationPattern;
    private String removeAssociationPattern;
    private String getAllAssociationsPattern;
    private String getAssociationsPattern;
    private String applyTagPattern;
    private String getTagsPattern;
    private String removeTagPattern;
    private String addCommentPattern;
    private String editCommentPattern;
    private String getCommentsPattern;
    private String rateResourcePattern;
    private String getAverageRatingPattern;
    private String getRatingPattern;
    private String createVersionPattern;
    private String getVersionsPattern;
    private String restoreVersionPattern;
    private String dumpPattern;
    private String restorePattern;

    /**
     * Compares this MediaTypeMatcher to the specified object.  The result is 
     * {@code true} if and only if the argument is not {@code null} and is a 
     * {@code MediaTypeMatcher} object that contains the same values for the 
     * fields as this object.
     *
     * @param   other The object to compare the {@code MediaTypeMatcher} against
     * @return  {@code true} if the given object represents a {@code 
     *          MediaTypeMatcher} equivalent to this instance, {@code false} 
     *          otherwise.
     */
    public boolean equals(Object other) {
        if (this == other) {
            return true;
        } else if (other == null) {
            return false;
        }
        if (other instanceof URLMatcher) {
            URLMatcher otherURLMatcher = (URLMatcher)other;
            if (!this.getPattern.equals(otherURLMatcher.getPattern)) {
                return false;
            } else if (!this.deletePattern.equals(otherURLMatcher.deletePattern)) {
                return false;
            } else if (!this.putPattern.equals(otherURLMatcher.putPattern)) {
                return false;
            } else if (!this.importPattern.equals(otherURLMatcher.importPattern)) {
                return false;
            } else if (!this.putChildPattern.equals(otherURLMatcher.putChildPattern)) {
                return false;
            } else if (!this.importChildPattern.equals(otherURLMatcher.importChildPattern)) {
                return false;
            } else if (!this.invokeAspectPattern.equals(otherURLMatcher.invokeAspectPattern)) {
                return false;
            } else if (!this.movePattern.equals(otherURLMatcher.movePattern)) {
                return false;
            } else if (!this.copyPattern.equals(otherURLMatcher.copyPattern)) {
                return false;
            } else if (!this.renamePattern.equals(otherURLMatcher.renamePattern)) {
                return false;
            } else if (!this.createLinkPattern.equals(otherURLMatcher.createLinkPattern)) {
                return false;
            } else if (!this.removeLinkPattern.equals(otherURLMatcher.removeLinkPattern)) {
                return false;
            } else if (!this.resourceExistsPattern.equals(otherURLMatcher.resourceExistsPattern)) {
                return false;
            } else if (!this.addAssociationPattern.equals(otherURLMatcher.addAssociationPattern)) {
                return false;
            } else if (!this.removeAssociationPattern.equals(otherURLMatcher.removeAssociationPattern)) {
                return false;
            } else if (!this.getAllAssociationsPattern.equals(otherURLMatcher.getAllAssociationsPattern)) {
                return false;
            } else if (!this.getAssociationsPattern.equals(otherURLMatcher.getAssociationsPattern)) {
                return false;
            } else if (!this.applyTagPattern.equals(otherURLMatcher.applyTagPattern)) {
                return false;
            } else if (!this.getTagsPattern.equals(otherURLMatcher.getTagsPattern)) {
                return false;
            } else if (!this.removeTagPattern.equals(otherURLMatcher.removeTagPattern)) {
                return false;
            } else if (!this.addCommentPattern.equals(otherURLMatcher.addCommentPattern)) {
                return false;
            } else if (!this.editCommentPattern.equals(otherURLMatcher.editCommentPattern)) {
                return false;
            } else if (!this.getCommentsPattern.equals(otherURLMatcher.getCommentsPattern)) {
                return false;
            } else if (!this.rateResourcePattern.equals(otherURLMatcher.rateResourcePattern)) {
                return false;
            } else if (!this.getAverageRatingPattern.equals(otherURLMatcher.getAverageRatingPattern)) {
                return false;
            } else if (!this.getRatingPattern.equals(otherURLMatcher.getRatingPattern)) {
                return false;
            } else if (!this.createVersionPattern.equals(otherURLMatcher.createVersionPattern)) {
                return false;
            } else if (!this.getVersionsPattern.equals(otherURLMatcher.getVersionsPattern)) {
                return false;
            } else if (!this.restoreVersionPattern.equals(otherURLMatcher.restoreVersionPattern)) {
                return false;
            } else if (!this.dumpPattern.equals(otherURLMatcher.dumpPattern)) {
                return false;
            } else if (!this.restorePattern.equals(otherURLMatcher.restorePattern)) {
                return false;
            } else if (this.invert != otherURLMatcher.invert) {
                return false;
            } else if (!this.handler.equals(otherURLMatcher.handler)) {
                return false;
            }
            return true;
        }
        return false;
    }

    public boolean handleGet(RequestContext requestContext) throws RegistryException {
        return getPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getPattern));
    }

    public boolean handlePut(RequestContext requestContext) throws RegistryException {
        return putPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(putPattern));
    }

    public boolean handleImportResource(RequestContext requestContext) throws RegistryException {
        return importPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(importPattern));
    }

    public boolean handleDelete(RequestContext requestContext) throws RegistryException {
        return deletePattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(deletePattern));
    }

    public boolean handlePutChild(RequestContext requestContext) throws RegistryException {
        return putChildPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(putChildPattern));
    }

    public boolean handleImportChild(RequestContext requestContext) throws RegistryException {
        return importChildPattern != null && (invert !=
               requestContext.getResourcePath().getCompletePath().matches(importChildPattern));
    }

    public boolean handleInvokeAspect(RequestContext requestContext) throws RegistryException {
        return invokeAspectPattern != null && (invert !=
               requestContext.getResourcePath().getCompletePath().matches(invokeAspectPattern));
    }

    public boolean handleCopy(RequestContext requestContext) throws RegistryException {
        return copyPattern != null &&  ((invert !=
               requestContext.getSourcePath().matches(copyPattern)) || (invert !=
               requestContext.getTargetPath().matches(copyPattern)));
    }

    public boolean handleMove(RequestContext requestContext) throws RegistryException {
        return movePattern != null &&  ((invert !=
               requestContext.getSourcePath().matches(movePattern))|| (invert !=
               requestContext.getTargetPath().matches(movePattern)));
    }

    public boolean handleRename(RequestContext requestContext) throws RegistryException {
        return renamePattern != null && (invert !=
               requestContext.getSourcePath().matches(renamePattern));
    }

    public boolean handleCreateLink(RequestContext requestContext) throws RegistryException {
        return createLinkPattern != null && (invert !=
               requestContext.getResourcePath().getCompletePath().matches(createLinkPattern));
    }

    public boolean handleRemoveLink(RequestContext requestContext) throws RegistryException {
        return removeLinkPattern != null && (invert !=
               requestContext.getResourcePath().getCompletePath().matches(removeLinkPattern));
    }

    public boolean handleResourceExists(RequestContext requestContext) throws RegistryException {
        return resourceExistsPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(resourceExistsPattern));
    }

    public boolean handleAddAssociation(RequestContext requestContext) throws RegistryException {
        return addAssociationPattern != null && (invert !=
                requestContext.getSourcePath().matches(addAssociationPattern));
    }

    public boolean handleRemoveAssociation(RequestContext requestContext) throws RegistryException {
        return removeAssociationPattern != null && (invert !=
                requestContext.getSourcePath().matches(removeAssociationPattern));
    }

    public boolean handleGetAllAssociations(RequestContext requestContext) throws RegistryException {
        return getAllAssociationsPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getAllAssociationsPattern));
    }

    public boolean handleGetAssociations(RequestContext requestContext) throws RegistryException {
        return getAssociationsPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getAssociationsPattern));
    }

    public boolean handleApplyTag(RequestContext requestContext) throws RegistryException {
        return applyTagPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(applyTagPattern));
    }

    public boolean handleGetTags(RequestContext requestContext) throws RegistryException {
        return getTagsPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getTagsPattern));
    }

    public boolean handleRemoveTag(RequestContext requestContext) throws RegistryException {
        return removeTagPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(removeTagPattern));
    }

    public boolean handleAddComment(RequestContext requestContext) throws RegistryException {
        return addCommentPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(addCommentPattern));
    }

    public boolean handleEditComment(RequestContext requestContext) throws RegistryException {
        return editCommentPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(editCommentPattern));
    }

    public boolean handleGetComments(RequestContext requestContext) throws RegistryException {
        return getCommentsPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getCommentsPattern));
    }

    public boolean handleRateResource(RequestContext requestContext) throws RegistryException {
        return rateResourcePattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(rateResourcePattern));
    }

    public boolean handleGetAverageRating(RequestContext requestContext) throws RegistryException {
        return getAverageRatingPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getAverageRatingPattern));
    }

    public boolean handleGetRating(RequestContext requestContext) throws RegistryException {
        return getRatingPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getRatingPattern));
    }

    public boolean handleCreateVersion(RequestContext requestContext) throws RegistryException {
        return createVersionPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(createVersionPattern));
    }

    public boolean handleGetVersions(RequestContext requestContext) throws RegistryException {
        return getVersionsPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(getVersionsPattern));
    }

    public boolean handleRestoreVersion(RequestContext requestContext) throws RegistryException {
        return restoreVersionPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(restoreVersionPattern));
    }

    public boolean handleDump(RequestContext requestContext) throws RegistryException {
        return dumpPattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(dumpPattern));
    }

    public boolean handleRestore(RequestContext requestContext) throws RegistryException {
        return restorePattern != null && (invert !=
                requestContext.getResourcePath().getCompletePath().matches(restorePattern));
    }

    public String getGetPattern() {
        return getPattern;
    }

    public void setGetPattern(String getPattern) {
        this.getPattern = getPattern;
    }

    public String getPutPattern() {
        return putPattern;
    }

    public void setPutPattern(String putPattern) {
        this.putPattern = putPattern;
    }

    public String getImportPattern() {
        return importPattern;
    }

    public void setImportPattern(String importPattern) {
        this.importPattern = importPattern;
    }

    public String getDeletePattern() {
        return deletePattern;
    }

    public void setDeletePattern(String deletePattern) {
        this.deletePattern = deletePattern;
    }

    public String getPutChildPattern() {
        return putChildPattern;
    }

    public void setPutChildPattern(String putChildPattern) {
        this.putChildPattern = putChildPattern;
    }

    public String getImportChildPattern() {
        return importChildPattern;
    }

    public void setImportChildPattern(String importChildPattern) {
        this.importChildPattern = importChildPattern;
    }

    public String getInvokeAspectPattern() {
        return invokeAspectPattern;
    }

    public void setInvokeAspectPattern(String invokeAspectPattern) {
        this.invokeAspectPattern = invokeAspectPattern;
    }

    public String getMovePattern() {
        return movePattern;
    }

    public void setMovePattern(String movePattern) {
        this.movePattern = movePattern;
    }

    public String getCopyPattern() {
        return copyPattern;
    }

    public void setCopyPattern(String copyPattern) {
        this.copyPattern = copyPattern;
    }

    public String getRenamePattern() {
        return renamePattern;
    }

    public void setRenamePattern(String renamePattern) {
        this.renamePattern = renamePattern;
    }

    public String getCreateLinkPattern() {
        return createLinkPattern;
    }

    public void setCreateLinkPattern(String createLinkPattern) {
        this.createLinkPattern = createLinkPattern;
    }

    public String getRemoveLinkPattern() {
        return removeLinkPattern;
    }

    public void setRemoveLinkPattern(String removeLinkPattern) {
        this.removeLinkPattern = removeLinkPattern;
    }

    public String getResourceExistsPattern() {
        return resourceExistsPattern;
    }

    public void setResourceExistsPattern(String resourceExistsPattern) {
        this.resourceExistsPattern = resourceExistsPattern;
    }

    public void setAddAssociationPattern(String addAssociationPattern) {
        this.addAssociationPattern = addAssociationPattern;
    }

    public String getAddAssociationPattern() {
        return addAssociationPattern;
    }

    public void setRemoveAssociationPattern(String removeAssociationPattern) {
        this.removeAssociationPattern = removeAssociationPattern;
    }

    public String getRemoveAssociationPattern() {
        return removeAssociationPattern;
    }

    public void setGetAllAssociationsPattern(String getAllAssociationsPattern) {
        this.getAllAssociationsPattern = getAllAssociationsPattern;
    }

    public String getGetAllAssociationsPattern() {
        return getAllAssociationsPattern;
    }

    public void setGetAssociationsPattern(String getAssociationsPattern) {
        this.getAssociationsPattern = getAssociationsPattern;
    }

    public String getGetAssociationsPattern() {
        return getAssociationsPattern;
    }

    public void setApplyTagPattern(String applyTagPattern) {
        this.applyTagPattern = applyTagPattern;
    }

    public String getApplyTagPattern() {
        return applyTagPattern;
    }

    public void setGetTagsPattern(String getTagsPattern) {
        this.getTagsPattern = getTagsPattern;
    }

    public String getGetTagsPattern() {
        return getTagsPattern;
    }

    public void setRemoveTagPattern(String removeTagPattern) {
        this.removeTagPattern = removeTagPattern;
    }

    public String getRemoveTagPattern() {
        return removeTagPattern;
    }

    public void setAddCommentPattern(String addCommentPattern) {
        this.addCommentPattern = addCommentPattern;
    }

    public String getAddCommentPattern() {
        return addCommentPattern;
    }

    public void setEditCommentPattern(String editCommentPattern) {
        this.editCommentPattern = editCommentPattern;
    }

    public String getEditCommentPattern() {
        return editCommentPattern;
    }

    public void setGetCommentsPattern(String getCommentsPattern) {
        this.getCommentsPattern = getCommentsPattern;
    }

    public String getGetCommentsPattern() {
        return getCommentsPattern;
    }

    public void setRateResourcePattern(String rateResourcePattern) {
        this.rateResourcePattern = rateResourcePattern;
    }

    public String getRateResourcePattern() {
        return rateResourcePattern;
    }

    public void setGetAverageRatingPattern(String getAverageRatingPattern) {
        this.getAverageRatingPattern = getAverageRatingPattern;
    }

    public String getGetAverageRatingPattern() {
        return getAverageRatingPattern;
    }

    public void setGetRatingPattern(String getRatingPattern) {
        this.getRatingPattern = getRatingPattern;
    }

    public String getGetRatingPattern() {
        return getRatingPattern;
    }

    public void setCreateVersionPattern(String createVersionPattern) {
        this.createVersionPattern = createVersionPattern;
    }

    public String getCreateVersionPattern() {
        return createVersionPattern;
    }

    public void setGetVersionsPattern(String getVersionsPattern) {
        this.getVersionsPattern = getVersionsPattern;
    }

    public String getGetVersionsPattern() {
        return getVersionsPattern;
    }

    public void setRestoreVersionPattern(String restoreVersionPattern) {
        this.restoreVersionPattern = restoreVersionPattern;
    }

    public String getRestoreVersionPattern() {
        return restoreVersionPattern;
    }

    public void setDumpPattern(String dumpPattern) {
        this.dumpPattern = dumpPattern;
    }

    public String getDumpPattern() {
        return dumpPattern;
    }

    public void setRestorePattern(String restorePattern) {
        this.restorePattern = restorePattern;
    }

    public String getRestorePattern() {
        return restorePattern;
    }
}
