/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.queries;

import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.config.QueryProcessorConfiguration;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.user.core.UserRealm;

import javax.sql.DataSource;
import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class QueryProcessorManager {

    private Map queryProcessors = new HashMap();

    public QueryProcessorManager(DataSource dataSource, UserRealm realm, Registry registry)
            throws RegistryException {
        queryProcessors.put(RegistryConstants.SQL_QUERY_MEDIA_TYPE,
                            new SQLQueryProcessor(dataSource, realm));

        // add run-time query processors configured using registry.xml

        RegistryContext registryContext = registry.getRegistryContext();

        if (registryContext != null) {
            Iterator iQueryProcessors = registryContext.getQueryProcessors().iterator();
            while (iQueryProcessors.hasNext()) {
                QueryProcessorConfiguration configuration =
                        (QueryProcessorConfiguration)iQueryProcessors.next();

                try {
                    Class qpClass = Class.forName(configuration.getProcessorClassName());
                    Constructor constructor =
                            qpClass.getConstructor(new Class[]{DataSource.class, UserRealm.class});

                    QueryProcessor queryProcessor =
                            (QueryProcessor)constructor.newInstance(dataSource, realm);

                    queryProcessors.put(configuration.getQueryType(), queryProcessor);

                } catch (ClassNotFoundException e) {
                    String msg = "Could not find the query processor class for query type: " +
                                 configuration.getQueryType();
                    throw new RegistryException(msg, e);

                } catch (Exception e) {
                    String msg = "Failed to initiate query processor for query type: " +
                                 configuration.getQueryType();
                    throw new RegistryException(msg, e);

                }
            }
        }
    }

    public QueryProcessor getQueryProcessor(String queryType) {
        return (QueryProcessor)queryProcessors.get(queryType);
    }

    public Collection executeQuery(Registry registry, Resource queryResource, Map parameters) throws RegistryException {

        if (queryResource.getMediaType() == null || queryResource.getMediaType().length() == 0) {
            String msg = "Failed to execute query at path: " + queryResource.getPath() +
                         ". Query resources should have a media type to map to a query processor.";
            throw new RegistryException(msg);
        }

        QueryProcessor queryProcessor =
                (QueryProcessor)queryProcessors.get(queryResource.getMediaType());
        if (queryProcessor == null) {
            String msg = "Failed to execute query at path: " + queryResource.getPath() +
                         ". No query processor is associated with the query type: " +
                         queryResource.getMediaType();
            throw new RegistryException(msg);
        }

        return queryProcessor.executeQuery(registry, queryResource, parameters);
    }
}
