/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.queries;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.jdbc.dao.RatingsDAO;
import org.wso2.carbon.registry.core.jdbc.dao.TagsDAO;
import org.wso2.carbon.registry.core.jdbc.dao.ResourceDAO;
import org.wso2.carbon.registry.core.jdbc.dao.CommentsDAO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.RatingDO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.TaggingDO;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.StaticConfiguration;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.user.core.UserRealm;

import javax.sql.DataSource;
import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

public class SQLQueryProcessor extends QueryProcessor {

    private static final Log log = LogFactory.getLog(SQLQueryProcessor.class);
    private ResourceDAO resourceDAO;
    private CommentsDAO commentsDAO;
    private RatingsDAO ratingsDAO;
    private TagsDAO tagsDAO;

    public SQLQueryProcessor(DataSource dataSource, UserRealm realm) {
        super(dataSource, realm);
        resourceDAO = new ResourceDAO();
        commentsDAO = StaticConfiguration.getCommentsDAO();
        ratingsDAO = StaticConfiguration.getRatingsDAO();
        tagsDAO = StaticConfiguration.getTagsDAO();
    }

    public Collection executeQuery(Registry registry, Resource query, Map parameters) throws RegistryException {

        Collection resultCollection = null;

        Connection conn = null;
        String sqlString = "";
        ResultSet results = null;
        PreparedStatement s = null;

        try {

            Object obj = query.getContent();
            if(obj instanceof String){
                sqlString = (String) obj;
            } else if (obj instanceof byte[]) {
                sqlString = new String((byte[]) obj);
            }else{
                throw new RegistryException("Unable to execute query at "+query.getPath()
                        +".Found resource content of type "+(obj == null?"null":obj.getClass().getName())
                        +".Expected java.lang.String or byte[]");
            }

            /*
            conn = dataSource.getConnection();
            DatabaseMetaData metaData = conn.getMetaData();
            if (metaData.getDatabaseProductName().matches("(?i).*derby.*")) {
                conn.setTransactionIsolation(Connection.TRANSACTION_READ_UNCOMMITTED);
            } else {
                conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
            }
            */
            conn = Transaction.getConnection();

            s = conn.prepareStatement(sqlString, ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);

            if (parameters != null) {
                parameters.remove("content");
                Iterator iParams = parameters.keySet().iterator();
                while (iParams.hasNext()) {
                    String paramNumber = (String)iParams.next();
                    Object paramValue = parameters.get(paramNumber);
                    s.setObject(Integer.parseInt(paramNumber), paramValue);
                }
            }

            results = s.executeQuery();

            String resultType = query.getProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME);

            if (resultType == null || resultType.equals(RegistryConstants.RESOURCES_RESULT_TYPE)) {

                // Result is a normal resource, which is stored in the Resources table or a collection
                // of normal resources.

                resultCollection = fillResourcesCollection(results);

            } else if (resultType.equals(RegistryConstants.COMMENTS_RESULT_TYPE)) {

                resultCollection = fillCommentsCollection(results, conn);

            } else if (resultType.equals(RegistryConstants.RATINGS_RESULT_TYPE)) {

                resultCollection = fillRatingsCollection(results, conn);

            } else if (resultType.equals(RegistryConstants.TAGS_RESULT_TYPE)) {

                resultCollection = fillTagsCollection(results);

            } else if (resultType.equals("TagsWithCount")) {

                // Result is in the form of (tag, number of taggings) tuples
                // Fill TagCount objects with results and add them as an array


            } else if (resultType.equals("ResourcePathsWithTagCount")) {

                // Result is in the form of (resource path, number of taggings) tuples
                // Fill ResourceTagCount objects with results and add them as an array

            }

            if (resultCollection == null) {
                String msg = "Unknown result type: " + resultType + " defined for the query: " +
                        sqlString + " located in path: " + query.getPath();
                log.error(msg);
                throw new RegistryException(msg);
            }

        } catch (SQLException e) {
            throw new RegistryException(e.getMessage());

        } finally {
            if (results != null) {
                try {
                    results.close();
                } catch (SQLException e) {
                    String msg = "Failed to close the result set. " + e.getMessage();
                    throw new RegistryException(msg, e);
                }
            }

            if (s != null) {
                try {
                    s.close();
                } catch (SQLException e) {
                    throw new RegistryException("Failed to close the statement. " + e.getMessage());
                }
            }

            /*
            try {
                conn.close();
            } catch (SQLException e) {
                throw new RegistryException("Failed to close the connection. " + e.getMessage());
            }*/
        }

        if (resultCollection != null && RegistryConstants.RESOURCES_RESULT_TYPE.equals(
                query.getProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME))) {

            List<String> filteredResults = new ArrayList<String>();
            String[] resultPaths = (String[]) resultCollection.getContent();
            for (String resultPath: resultPaths) {
                if (AuthorizationUtils.authorize(resultPath, ActionConstants.GET)){
                    filteredResults.add(resultPath);
                }
            }

            String[] filteredContent = filteredResults.toArray(new String[filteredResults.size()]);
            resultCollection.setContent(filteredContent);
        }

        return resultCollection;
    }

    private Collection fillResourcesCollection(ResultSet results)
            throws SQLException, RegistryException {

        // Result is a normal resource, which is stored in the Resources table or a collection
        // of normal resources.

        List pathList = new ArrayList();
        while (results.next()) {
            int pathId = results.getInt("REG_PATH_ID");
            String resourceName = results.getString("REG_NAME");
            String path = resourceDAO.getPath(pathId, resourceName, true);
            if (path != null && (!pathList.contains(path))) {
                pathList.add(path);
            }
        }

        String[] paths = (String[])pathList.toArray(new String[pathList.size()]);
        return new CollectionImpl(paths);
    }

    private Collection fillCommentsCollection(ResultSet results, Connection conn)
            throws SQLException, RegistryException {

        // SQL query should restrun a list of DatabaseConstants.COMMENT_ID_FIELDs.
        // resultArtifact contains a String[] of URLs for comments.

        // URL for a comment <resource_path>?comment<comment_id>
        // e.g. /p1/r1?comment12

        List<Long> commentIDs = new ArrayList<Long>();

        while (results.next()) {
            long commentID = results.getLong("REG_COMMENT_ID");
            commentIDs.add(commentID);
        }

        String[] commentPaths =commentsDAO.getResourcePathsOfComments(
                    commentIDs.toArray(new Long[commentIDs.size()]), conn);

        return new CollectionImpl(commentPaths);
    }

    private Collection fillRatingsCollection(ResultSet results, Connection conn)
            throws SQLException, RegistryException {

        // SQL query should restrun a list of DatabaseConstants.RATING_ID_FIELDs
        // resultArtifact contains a String[] of URLs for ratings.

        // URL for a comment <resource_path>;ratings:<userName>
        // e.g. /p1/r1;ratings:ruwan

        List ratingPathList = new ArrayList();

        while (results.next()) {

            long ratingID = results.getLong(DatabaseConstants.RATING_ID_FIELD);

            RatingDO ratingDAO = ratingsDAO.getRating(ratingID);

            String ratingPath =
                    ratingDAO.getResourcePath() + RegistryConstants.URL_SEPARATOR + "ratings:" +
                            ratingDAO.getRatedUserName();
            ratingPathList.add(ratingPath);
        }

        String[] ratingPaths =
                (String[])ratingPathList.toArray(new String[ratingPathList.size()]);

        return new CollectionImpl(ratingPaths);
    }

    private Collection fillTagsCollection(ResultSet results)
            throws SQLException, RegistryException {

        // URL for a tag /p1/r1;tags:tagName:userName

        List tagPathList = new ArrayList();
        while (results.next()) {

            long taggingID = results.getLong(DatabaseConstants.TAGGING_ID_FIELD);

            TaggingDO taggingDO = tagsDAO.getTagging(taggingID);

            //String userName = results.getString(DatabaseConstants.USER_ID_FIELD);
            //long artifactID = results.getLong(DatabaseConstants.ARTIFACT_ID_FIELD);
            //long tagID = results.getLong(DatabaseConstants.TAG_ID_FIELD);
            //
            //String artifactSQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            //PreparedStatement s1 = conn.prepareStatement(artifactSQL);
            //s1.setLong(1, artifactID);
            //ResultSet result1 = s1.executeQuery();
            //
            //String taggedPath = "";
            //if (result1.next()) {
            //    taggedPath = result1.getString(DatabaseConstants.PATH_FIELD);
            //}
            //
            //String tagSQL = "SELECT TAG_NAME FROM TAGS WHERE TAG_ID=?";
            //PreparedStatement s2 = conn.prepareStatement(tagSQL);
            //s2.setLong(1, tagID);
            //ResultSet result2 = s2.executeQuery();
            //
            //String tag = "";
            //if (result2.next()) {
            //    tag = result2.getString(DatabaseConstants.TAG_NAME_FIELD);
            //}

            String tagPath = taggingDO.getResourcePath() + RegistryConstants.URL_SEPARATOR +
                    "tags:" + taggingDO.getTagName() + ":" + taggingDO.getTaggedUserName();
            tagPathList.add(tagPath);
        }

        String[] tagPaths = (String[])tagPathList.toArray(new String[tagPathList.size()]);

        return new CollectionImpl(tagPaths);
    }
}
