/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.realm;

import org.wso2.carbon.user.core.*;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.InvalidSyntaxException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class CarbonRealm implements UserRealm {

    private static final Log log = LogFactory.getLog(CarbonRealm.class);

    private ServiceTracker defaultRealmTracker;

    private ServiceTracker delegatingRealmTracker;

    private BundleContext bundleContext;

    public CarbonRealm(ServiceTracker realmTracker, BundleContext bundleContext) {
        this.defaultRealmTracker = realmTracker;
        this.delegatingRealmTracker = null;
        this.bundleContext = bundleContext;
    }

    public Object getRealmConfiguration() throws UserStoreException {
        String msg = "getRealmConfiguration is not required for external users in the Carbon environment. " +
                "Therefore this method is not supported.";
        throw new UnsupportedOperationException(msg);
    }

    public void init(Object o) throws UserStoreException {
        String msg = "init is not required for external users in the Carbon environment. " +
                "Therefore this method is not supported.";
        throw new UnsupportedOperationException(msg);
    }

    public Authenticator getAuthenticator() throws UserStoreException {

        UserRealm delegatingRealm = getDelegatingRealm();
        if (delegatingRealm != null) {
            return delegatingRealm.getAuthenticator();
        }

        return getDefaultRealm().getAuthenticator();
    }

    public Authorizer getAuthorizer() throws UserStoreException {

        UserRealm delegatingRealm = getDelegatingRealm();
        if (delegatingRealm != null) {
            return delegatingRealm.getAuthorizer();
        }

        return getDefaultRealm().getAuthorizer();
    }

    public UserStoreReader getUserStoreReader() throws UserStoreException {

        UserRealm delegatingRealm = getDelegatingRealm();
        if (delegatingRealm != null) {
            return delegatingRealm.getUserStoreReader();
        }

        return getDefaultRealm().getUserStoreReader();
    }

    public UserStoreAdmin getUserStoreAdmin() throws UserStoreException {
        // accessing the user store admin of the delegating realm has no meaning as it is impossible to modify users
        // in external user stores.
        return getDefaultRealm().getUserStoreAdmin();
    }

    public AccessControlAdmin getAccessControlAdmin() throws UserStoreException {

        UserRealm delegatingRealm = getDelegatingRealm();
        if (delegatingRealm != null) {
            return delegatingRealm.getAccessControlAdmin();
        }

        return getDefaultRealm().getAccessControlAdmin();
    }

    private UserRealm getDefaultRealm() throws UserStoreException {

        UserRealm realm = (UserRealm) defaultRealmTracker.getService();
        if (realm == null) {
            String msg = "Realm service is not available. Make sure that the required " +
                    "version of the User Manager component is properly intalled.";
            log.error(msg);
            throw new UserStoreException(msg);
        }

        return realm;
    }

    private UserRealm getDelegatingRealm() throws UserStoreException {

        if (delegatingRealmTracker != null) {
            return (UserRealm) delegatingRealmTracker.getService();
        }

        try {
            ServiceReference[] serviceRefs = bundleContext.
                    getServiceReferences(UserRealm.class.getName(), "(" + UserCoreConstants.REALM_GENRE + "=" + UserCoreConstants.DELEGATING_REALM + ")");
            if (serviceRefs != null && serviceRefs.length > 0) {
                ServiceTracker realmTracker = new ServiceTracker(bundleContext, serviceRefs[0], null);
                realmTracker.open();
                this.delegatingRealmTracker = realmTracker;
                return (UserRealm) delegatingRealmTracker.getService();

            } else {
                String msg = "Could not get the service reference to the delegating realm. Using the default realm for " +
                    "this request. Please ignore this message, if this occurs in the start up of the server as " +
                    "delegating realm may not have been initialized at this moment.";

                if (log.isDebugEnabled()) {
                    log.debug(msg);
                }
            }
        } catch (InvalidSyntaxException e) {
            String msg = "Error occured while retrieving the service reference of the delegating realm. " +
                    "Switching to the default realm for this request. " + e.getMessage();
            log.error(msg, e);
        }

        return null;
    }
    
    public void cleanUp() throws UserStoreException{
	
    }
    
    public UserClaimsAdmin getClaimsAdmin(String storeName) throws UserStoreException {
        // TODO Auto-generated method stub
        return null;
    }

    public UserProfileAdmin getUserProfileAdmin(String storeName) throws UserStoreException {
        // TODO Auto-generated method stub
        return null;
    }

    public void setClaimsAdmin(UserClaimsAdmin claimsAdmin) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void setUserProfileAdmin(UserProfileAdmin profileAdmin) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }
    
    
}
