/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.realm;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.user.core.*;

/**
 * This class makes sure that the user store of the registry is always in consistent state. It wraps
 * the given realm and peforms validations for all required actions before delegating the
 * requests to the wrapped realm.
 *
 * Realm is only used for getting other user store related classes. Therefore, registry provides
 * a wrappers for all such classes whereever a validation is necessary.
 */
public class RegistryRealm implements UserRealm {

    private static final Log log = LogFactory.getLog(RegistryRealm.class);

    private UserStoreAdmin userStoreAdmin;
    private AccessControlAdmin accessControlAdmin;
    private UserRealm coreRealm;

    /**
     * This contructor is provided for extending the RegistryRealm.
     */
    public RegistryRealm() {}

    public RegistryRealm(UserRealm userRealm) throws RegistryException {
        initialize(userRealm);
    }

    private void initialize(UserRealm userRealm) throws RegistryException {

        this.coreRealm = userRealm;

        try {
            if (log.isTraceEnabled()) {
                log.trace("Retrieving user manager components from the realm ...");
            }
            this.userStoreAdmin = new RegistryUserStoreAdmin(userRealm);
            this.accessControlAdmin =
                    new RegistryAccessControlAdmin(userRealm);

            if (log.isTraceEnabled()) {
                log.trace("Populating the initial user store ...");
            }
            AuthorizationUtils.populateUserStore(this);

        } catch (UserStoreException e) {
            String msg = "Failed to initialize the registry user store. " + e.getMessage();
            log.fatal(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void cleanUp() throws UserStoreException{
	return;
    }
    
    public Object getRealmConfiguration() throws UserStoreException {
        return getRealm().getRealmConfiguration();
    }

    public void init(Object configBean) throws UserStoreException {
        getRealm().init(configBean);
    }

    public UserStoreAdmin getUserStoreAdmin() throws UserStoreException {
        return userStoreAdmin;
    }

    public AccessControlAdmin getAccessControlAdmin() throws UserStoreException {
        return accessControlAdmin;
    }

    public Authenticator getAuthenticator() throws UserStoreException {
        return getRealm().getAuthenticator();
    }

    public Authorizer getAuthorizer() throws UserStoreException {
        return accessControlAdmin;
    }

    public UserStoreReader getUserStoreReader() throws UserStoreException {
        return getRealm().getUserStoreReader();
    }

    private UserRealm getRealm() throws UserStoreException {
        if (this.coreRealm == null) {
            String msg = "Realm service is not available. Make sure that the required " +
                    "version of the User Manager component is properly intalled.";
            log.error(msg);
            throw new UserStoreException(msg);
        }

        return coreRealm;
    }
    
    public UserProfileAdmin getUserProfileAdmin(String storeName) throws UserStoreException{
        return getRealm().getUserProfileAdmin(storeName);
    }

    public void setUserProfileAdmin(UserProfileAdmin profileAdmin) throws UserStoreException{
        getRealm().setUserProfileAdmin(profileAdmin);
    }

    public UserClaimsAdmin getClaimsAdmin(String storeName) throws UserStoreException{
        return getRealm().getClaimsAdmin(storeName);
    }

    public void setClaimsAdmin(UserClaimsAdmin claimsAdmin) throws UserStoreException{
        this.setClaimsAdmin(claimsAdmin);
    }
}
