/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.servlet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.i18n.Messages;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.secure.RegistryUMConstants;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.def.DefaultRealm;
import org.wso2.carbon.user.core.def.DefaultRealmConfig;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.sql.DataSource;

public class RegistryRealmFactory {

    private static final Log log = LogFactory.getLog(RegistryRealmFactory.class);

    public static UserRealm createRegistryRealm(ServletConfig config) throws RegistryException {

        UserRealm realm;

        String dataSourceName = config.getInitParameter(RegistryUMConstants.UM_DATASOURCE_NAME);
        String dbURL = config.getInitParameter(RegistryUMConstants.UM_DATABASE_CONNECTION_URL);
        String dbDriverName = config.getInitParameter(RegistryUMConstants.UM_DATABASE_DRIVER_NAME);
        String dbUserName = config.getInitParameter(RegistryUMConstants.UM_DATABASE_USER_NAME);
        String dbPassword = config.getInitParameter(RegistryUMConstants.UM_DATABASE_PASSWORD);

        if (dataSourceName != null) {
            realm = createDataSourceAwareRegistryRealm(dataSourceName);

        } else if (dbURL != null) {
            realm = createURLBasedRegistryRealm(dbURL, dbDriverName, dbUserName, dbPassword);

        } else {
            realm = createInMemoryRegistryRealm();
        }

        return realm;
    }

    public static UserRealm createInMemoryRegistryRealm() throws RegistryException {
        //new HSQLDBInitializer().createHSQLTables();
        try {
            DefaultRealm realm = new DefaultRealm();
            DefaultRealmConfig config = (DefaultRealmConfig)realm.getRealmConfiguration();
            config.setConnectionURL(RegistryUMConstants.HSQL_DB_URL);
            config.setDriverName(DatabaseConstants.HSQL_DRIVER_NAME);
            realm.init(config);

            // define built in users and roles
            AuthorizationUtils.populateUserStore(realm);
            return realm;

        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public static UserRealm createURLBasedRegistryRealm(String dbURL,
                                                    String driverName,
                                                    String userName,
                                                    String password) throws RegistryException {
        //new HSQLDBInitializer().createHSQLTables();
        try {
            DefaultRealm realm = new DefaultRealm();
            DefaultRealmConfig config = (DefaultRealmConfig)realm.getRealmConfiguration();
            config.setConnectionURL(dbURL);
            config.setDriverName(driverName);
            config.setConnectionUserName(userName);
            config.setConnectionPassword(password);
            realm.init(config);

            // define built in users and roles
            AuthorizationUtils.populateUserStore(realm);
            return realm;

        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public static UserRealm createDataSourceAwareRegistryRealm(String dataSourceName)
            throws RegistryException {
        DataSource dataSource = null;
        try {
            if (log.isTraceEnabled()) {
                log.trace("Trying to find the user manager data source " + dataSourceName + "...");
            }
            Context context = new InitialContext();
            dataSource = (DataSource)context.lookup(dataSourceName);
            if (log.isTraceEnabled()) {
                log.trace("Found the user manager data source " + dataSourceName + ".");
            }
        } catch (NamingException e) {
            log.error(Messages.
                    getMessage("datasource.not.found", dataSourceName));
        }
        try {
            DefaultRealm realm = new DefaultRealm();
            DefaultRealmConfig config = (DefaultRealmConfig)realm.getRealmConfiguration();
            config.setDataSource(dataSource);
            realm.init(config);
            // define built in roles
            AuthorizationUtils.populateUserStore(realm);
            return realm;

        } catch (Exception e) {
            e.printStackTrace();
            throw new RegistryException(e.getMessage());
        }
    }
}
