/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.session;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.realm.pathbased.PathBasedRegistryRealm;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.Authenticator;
import org.wso2.carbon.user.core.AuthenticatorException;
import org.wso2.carbon.user.core.authorizing.AuthorizingRealm;
import org.wso2.carbon.user.core.authorizing.AuthorizingRealmConfig;

import javax.sql.DataSource;
import java.util.Date;
import java.util.Map;
import java.sql.Connection;
import java.io.OutputStream;
import java.io.Writer;
import java.io.Reader;

/**
 * Maintains information required for particular user. When registry functionality is invoked
 * through this objects of this class, user's information is passes along the thread of
 * invocation. Therefore, all the actions are recorded against that user's name and all
 * authorizations are checked using that user's realm. Note that, although this class maintains
 * user details, it does not perform any authorization. Underlying registry or other entities
 * in the thread of invocation should perform authorizations as needed using the realm given by
 * this class.
 */
public class UserRegistry implements Registry {

    public static final Log log = LogFactory.getLog(UserRegistry.class);

    /**
     * User name of the user. All actions done using this registry will be recorded against this
     * user name.
     */
    private String userName;

    /**
     * Realm of the user. All actions done using this registry will be authorized using this realm.
     */
    private UserRealm userRealm;

    /**
     * Path based registry realm to be used by API users.
     */
    private PathBasedRegistryRealm pathBasedRegistryRealm;

    /**
     * Core registry to delegate the requests. All UserRegistry instances share this core
     * registry.
     */
    private Registry coreRegistry;

    /**
     * Database connection is stored in this variable, if there is an open transaction. UserRegistry should provide
     * the connection object for uncommited transactions by setting it as a thread local variable. This is required as
     * there is no garauntee that the same thread will serve all method calls of a single
     * transaction (e.g. in servlets).
     */
    //private Connection connection;

    public UserRegistry (String userName, String password, Registry coreRegistry,
                         UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {

        try {
            Authenticator authenticator = defaultRealm.getAuthenticator();

            if (!authenticator.authenticate(userName, password)) {
                String msg = "Attempted to authenticate invalid user.";
                log.error(msg);
                throw new RegistryException(msg);
            }

        } catch (AuthenticatorException e) {
            String msg = "Could not check authentication. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } catch (UserStoreException e){
            String msg = "Could not obtain the authenticator for authenticating the user " +
                    userName + ". \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        init(userName, coreRegistry, defaultRealm, dataSource);
    }

    public UserRegistry (String userName, Registry coreRegistry, UserRealm defaultRealm) throws RegistryException {

        this.userName = userName;

        this.coreRegistry = coreRegistry;        

        UserRealm userRealm;

        //TODO why are we doing this ?? Changing the parameters
        try {
            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userName);
            config.setCurrentUserReadable(true);
            config.setCurrentUserEditable(true);
            userRealm.init(config);
            this.userRealm = userRealm;

        } catch (UserStoreException e){
            String msg = "Could not create the realm for user " + userName + ". \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }


    public UserRegistry (
            String userName, Registry coreRegistry, UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {

        init(userName, coreRegistry, defaultRealm, dataSource);
    }

    private void init(
            String userName, Registry coreRegistry, UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {

        this.userName = userName;

        this.coreRegistry = coreRegistry;

        UserRealm userRealm;

        try {
            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userName);
            config.setCurrentUserReadable(true);
            config.setCurrentUserEditable(true);
            userRealm.init(config);
            this.userRealm = userRealm;

        } catch (UserStoreException e){
            String msg = "Could not create the realm for user " + userName + ". \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        pathBasedRegistryRealm = new PathBasedRegistryRealm(userRealm, dataSource);

        //this.connection = null;
    }

    public void beginTransaction() throws RegistryException {
        try {
            setSessionInformation();

            coreRegistry.beginTransaction();
            //connection = Transaction.getConnection();
        } finally {
            clearSessionInformation();
        }
    }

    public void commitTransaction() throws RegistryException {
        try {
            setSessionInformation();

            coreRegistry.commitTransaction();
            //connection = Transaction.getConnection();
        } finally {
            clearSessionInformation();
        }
    }

    public void rollbackTransaction() throws RegistryException {
        try {
            setSessionInformation();
            
            coreRegistry.rollbackTransaction();
            //connection = Transaction.getConnection();
        } finally {
            clearSessionInformation();
        }
    }

    public Resource newResource() throws RegistryException {
        try {
            setSessionInformation();
            return coreRegistry.newResource();
        } finally {
            clearSessionInformation();
        }
    }

    public Collection newCollection() throws RegistryException {
        try {
            setSessionInformation();
            return coreRegistry.newCollection();
        } finally {
            clearSessionInformation();
        }
    }

    public RegistryContext getRegistryContext() {
        try {
            setSessionInformation();
            return coreRegistry.getRegistryContext();
        } finally {
            clearSessionInformation();
        }
    }

    public String getUserName() {
        return userName;
    }

    /**
     * Registry API users access the user realm using this method. Registry authorizations are
     * stored based on resource UUIDs. But UUIDs are meaningless to the API users. API users
     * are expected to deal only with resource paths. Therefore, this method, returns a user realm,
     * whcih converts paths to resource UUIDs before, any resource related operation.
     *
     * @return Resource path based registry realm.
     */
    public UserRealm getUserRealm() {
        if(pathBasedRegistryRealm != null){
            return pathBasedRegistryRealm;
        }
        else{
            return userRealm;
        }
    }

    public Resource get(String path) throws RegistryException {

        try {
            setSessionInformation();
            ResourceImpl resource = (ResourceImpl) coreRegistry.get(path);
            resource.setUserName(userName);
            resource.setUserRealm(userRealm);
            return resource;

        } finally {
            clearSessionInformation();
        }
    }

    public Resource getMetaData(String path) throws RegistryException {

        try {
            setSessionInformation();
            ResourceImpl resource = (ResourceImpl) coreRegistry.getMetaData(path);
            resource.setUserName(userName);
            resource.setUserRealm(userRealm);
            return resource;

        } finally {
            clearSessionInformation();
        }
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {

        try {
            setSessionInformation();
            Collection collection = coreRegistry.get(path, start, pageSize);

            // collection implementation extends from the resource implementation.
            ResourceImpl resourceImpl = (ResourceImpl) collection;
            resourceImpl.setUserName(userName);
            resourceImpl.setUserRealm(userRealm);

            return collection;

        } finally {
            clearSessionInformation();
        }
    }

    public boolean resourceExists(String path) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.resourceExists(path);
        } finally {
            clearSessionInformation();
        }
    }

    public String put(String suggestedPath, Resource resource) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.put(suggestedPath, resource);
        } finally {
            clearSessionInformation();
        }
    }

    public void delete(String path) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.delete(path);
        } finally {
            clearSessionInformation();
        }
    }

    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.importResource(suggestedPath, sourceURL, metadata);
        } finally {
            clearSessionInformation();
        }
    }

    public String rename(String currentPath, String newPath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.rename(currentPath, newPath);
        } finally {
            clearSessionInformation();
        }
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.move(currentPath, newPath);
        } finally {
            clearSessionInformation();
        }
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.copy(sourcePath, targetPath);
        } finally {
            clearSessionInformation();
        }
    }

    public void createVersion(String path) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.createVersion(path);
        } finally {
            clearSessionInformation();
        }
    }

    public String[] getVersions(String path) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getVersions(path);
        } finally {
            clearSessionInformation();
        }
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.restoreVersion(versionPath);
        } finally {
            clearSessionInformation();
        }
    }

    ////////////////////////////////////////////////////////
    // Associations
    ////////////////////////////////////////////////////////
    public void addAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.addAssociation(sourcePath, targetPath, associationType);
        } finally {
            clearSessionInformation();
        }
    }

    public void removeAssociation(String sourcePath, String associationPaths,
                                  String associationType)
            throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.removeAssociation(sourcePath, associationPaths, associationType);
        } finally {
            clearSessionInformation();
        }
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getAllAssociations(resourcePath);
        } finally {
            clearSessionInformation();
        }
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getAssociations(resourcePath, associationType);
        } finally {
            clearSessionInformation();
        }
    }


    public void applyTag(String resourcePath, String tag) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.applyTag(resourcePath, tag);
        } finally {
            clearSessionInformation();
        }
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getResourcePathsWithTag(tag);
        } finally {
            clearSessionInformation();
        }
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getTags(resourcePath);
        } finally {
            clearSessionInformation();
        }
    }

    public void removeTag(String path, String tag) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.removeTag(path, tag);
        } finally {
            clearSessionInformation();
        }
    }

    public String addComment(String resourcePath, Comment comment) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.addComment(resourcePath, comment);
        } finally {
            clearSessionInformation();
        }
    }

    public void editComment(String commentPath, String text) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.editComment(commentPath, text);
        } finally {
            clearSessionInformation();
        }
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getComments(resourcePath);
        } finally {
            clearSessionInformation();
        }
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.rateResource(resourcePath, rating);
        } finally {
            clearSessionInformation();
        }
    }

    public float getAverageRating(String resourcePath) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getAverageRating(resourcePath);
        } finally {
            clearSessionInformation();
        }
    }

    public int getRating(String path, String userName) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getRating(path, userName);
        } finally {
            clearSessionInformation();
        }
    }

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        try {
            setSessionInformation();
            Collection collection = coreRegistry.executeQuery(path, parameters);

            ResourceImpl resourceImpl = (ResourceImpl) collection;
            resourceImpl.setUserName(userName);
            resourceImpl.setUserRealm(userRealm);

            return collection;

        } finally {
            clearSessionInformation();
        }
    }

    public LogEntry[] getLogs(
            String resourcePath,
            int action,
            String userName,
            Date from,
            Date to,
            boolean recentFirst)
            throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getLogs(resourcePath, action, userName, from, to, recentFirst);
        } finally {
            clearSessionInformation();
        }
    }


    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getLogCollection(resourcePath, action, userName, from, to, recentFirst);
        } finally {
            clearSessionInformation();
        }
    }

    public String[] getAvailableAspects() {

        try {
            setSessionInformation();
            return coreRegistry.getAvailableAspects();
        } finally {
            clearSessionInformation();
        }
    }

    public void associateAspect(String resourcePath, String aspect) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.associateAspect(resourcePath, aspect);
        } finally {
            clearSessionInformation();
        }
    }

    public void invokeAspect(String resourcePath, String aspectName, String action) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.invokeAspect(resourcePath, aspectName, action);
        } finally {
            clearSessionInformation();
        }
    }

    public String[] getAspectActions(String resourcePath, String aspectName) throws RegistryException {

        try {
            setSessionInformation();
            return coreRegistry.getAspectActions(resourcePath, aspectName);
        } finally {
            clearSessionInformation();
        }
    }

    public Collection searchContent(String keywords) throws RegistryException {

        try {
            setSessionInformation();
            Collection collection = coreRegistry.searchContent(keywords);

            ResourceImpl resourceImpl = (ResourceImpl) collection;
            resourceImpl.setUserName(userName);
            resourceImpl.setUserRealm(userRealm);

            return collection;
            
        } finally {
            clearSessionInformation();
        }
    }

    public void createLink(String path, String target) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.createLink(path, target);
        } finally {
            clearSessionInformation();
        }
    }

    public void createLink(String path, String target,
                           String targetSubPath) throws RegistryException {
        try {
            setSessionInformation();
            coreRegistry.createLink(path, target, targetSubPath);
        } finally {
            clearSessionInformation();
        }
    }

    public void removeLink(String path) throws RegistryException {

        try {
            setSessionInformation();
            coreRegistry.removeLink(path);
        } finally {
            clearSessionInformation();
        }
    }

    private void setSessionInformation() {

        /*if (connection != null) {
            Transaction.setConnection(connection);
        }*/

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);
    }

    private void clearSessionInformation() {

        /*if (Transaction.isStarted()) {
            Transaction.setStarted(false);
            Transaction.setConnection(null);
        }*/

        CurrentSession.removeUser();
        CurrentSession.removeRealm();
    }

    // check in, check out functionality
    public void restore(String path, Reader reader) throws RegistryException {
        try {
            setSessionInformation();
            coreRegistry.restore(path, reader);
        } finally {
            clearSessionInformation();
        }
    }

    public void dump(String path, Writer writer) throws RegistryException {
        try {
            setSessionInformation();
            coreRegistry.dump(path, writer);
        } finally {
            clearSessionInformation();
        }
    }

    public String getEventingServiceURL(String path) throws RegistryException {
        String eventingServiceURL = null;
        try {
            setSessionInformation();
            eventingServiceURL = coreRegistry.getEventingServiceURL(path);
        } finally {
            clearSessionInformation();
        }
        return eventingServiceURL;
    }

    public void setEventingServiceURL(String path, String eventingServiceURL) throws RegistryException {
        try {
            setSessionInformation();
            coreRegistry.setEventingServiceURL(path, eventingServiceURL);
        } finally {
            clearSessionInformation();
        }
    }

}
