/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.utils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.ActionConstants;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.ResourceIDImpl;
import org.wso2.carbon.registry.core.config.RegistryConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.internal.RegistryCoreServiceComponent;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.user.core.*;
import org.wso2.carbon.user.core.authorizing.AccessControlConstants;

public class AuthorizationUtils {

    public static final Log log = LogFactory.getLog(AuthorizationUtils.class);

    public static boolean authorize(String resourcePath, String action)
            throws RegistryException {
        UserRealm userRealm = CurrentSession.getRealm();
        String userName = CurrentSession.getUser();

        try {
            if (!userRealm.getAuthorizer().isUserAuthorized(userName, resourcePath, action)) {
                return false;
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return true;
    }

    public static void copyAuthorizations(String sourcePath, String targetPath)
            throws RegistryException {

        UserRealm userRealm = CurrentSession.getRealm();

        try {
            userRealm.getAccessControlAdmin().copyAuthorizations(sourcePath, targetPath);

        } catch (UserStoreException e) {

            String msg = "Could not copy authorizations. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public static void clearAuthorizations(String resourcePath)
            throws RegistryException {

        UserRealm userRealm = CurrentSession.getRealm();

        try {
            userRealm.getAccessControlAdmin().clearResourceAuthorizations(resourcePath);

        } catch (UserStoreException e) {

            String msg = "Could not clear authorizations. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public static void setRootAuthorizations(String rootPath, UserRealm systemRealm)
            throws RegistryException {

        if (systemRealm == null) {
            return;
        }

        try {
            AccessControlAdmin accessControlAdmin = systemRealm.getAccessControlAdmin();
            RegistryConfiguration regConfig = RegistryCoreServiceComponent.getRegistryConfig();
            String username = regConfig.getValue(RegistryConfiguration.SYSTEM_USER_NAME);
            
            accessControlAdmin.authorizeUser(username, rootPath, ActionConstants.GET);
            accessControlAdmin.authorizeUser(username, rootPath, ActionConstants.PUT);
            accessControlAdmin.authorizeUser(username, rootPath, ActionConstants.DELETE);
            accessControlAdmin.authorizeUser(username, rootPath, AccessControlConstants.AUTHORIZE);

            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    rootPath, ActionConstants.GET);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    rootPath, ActionConstants.PUT);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    rootPath, ActionConstants.DELETE);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    rootPath, AccessControlConstants.AUTHORIZE);

            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    rootPath, ActionConstants.GET);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    rootPath, ActionConstants.PUT);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    rootPath, ActionConstants.DELETE);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    rootPath, AccessControlConstants.AUTHORIZE);

            accessControlAdmin.authorizeRole(
                    RegistryConstants.EVERYONE_ROLE, rootPath,
                    ActionConstants.GET);


        } catch (UserStoreException e) {
            String msg = "Could not set authorizations for the root. \nCaused by: "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }
    }

    public static void setAdminAuthorizations(String resourcePath, UserRealm systemRealm)
            throws RegistryException {

        if (systemRealm == null) {
            return;
        }

        try {
            AccessControlAdmin accessControlAdmin = systemRealm
                    .getAccessControlAdmin();

            accessControlAdmin.clearResourceAuthorizations(resourcePath);

            RegistryConfiguration regConfig = RegistryCoreServiceComponent.getRegistryConfig();
            String username = regConfig.getValue(RegistryConfiguration.SYSTEM_USER_NAME);
            
            accessControlAdmin.authorizeUser(username, resourcePath, ActionConstants.GET);
            accessControlAdmin.authorizeUser(username, resourcePath, ActionConstants.PUT);
            accessControlAdmin.authorizeUser(username, resourcePath, ActionConstants.DELETE);
            accessControlAdmin.authorizeUser(username, resourcePath,
                    AccessControlConstants.AUTHORIZE);

            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    resourcePath, ActionConstants.GET);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    resourcePath, ActionConstants.PUT);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    resourcePath, ActionConstants.DELETE);
            accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                    resourcePath, AccessControlConstants.AUTHORIZE);

            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    resourcePath, ActionConstants.GET);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    resourcePath, ActionConstants.PUT);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    resourcePath, ActionConstants.DELETE);
            accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                    resourcePath, AccessControlConstants.AUTHORIZE);

        } catch (UserStoreException e) {
            String msg = "Could not set authorizations for the resource " + resourcePath  + ". Caused by: "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);
        }
    }

    /**
     * Populates all necessary users, roles and authorizations related user store. Note that the
     * authorizations related to resource store is not populated by this method.
     *
     * @param realm Realm for which data has to be populated
     * @throws UserStoreException
     */
    public static void populateUserStore(UserRealm realm)
            throws UserStoreException {

        if (realm == null) {
            return;
        }

        UserStoreReader userStoreReader = realm.getUserStoreReader();
        UserStoreAdmin userStoreAdmin = realm.getUserStoreAdmin();
        AccessControlAdmin accessControlAdmin = realm.getAccessControlAdmin();

        if (!userStoreReader.isExistingRole(RegistryConstants.ADMIN_ROLE)) {
            userStoreAdmin.addRole(RegistryConstants.ADMIN_ROLE);
        }

        // adding the admin role and granting permissions are two independent tasks. There can be
        // scenarios where the admin role exists without any permissions (e.g. admin role defined
        // in external user store)
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.DELETE);

        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.DELETE);

        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeRole(RegistryConstants.ADMIN_ROLE,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.DELETE);

        if (!userStoreReader.isExistingRole(RegistryConstants.EVERYONE_ROLE)) {
            userStoreAdmin.addRole(RegistryConstants.EVERYONE_ROLE);
        }
        if (!userStoreReader.isExistingRole(RegistryConstants.GUESTS_ROLE)) {
            userStoreAdmin.addRole(RegistryConstants.GUESTS_ROLE);
        }

        accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeRole(RegistryConstants.EVERYONE_ROLE,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.READ);

        // define built in users
        RegistryConfiguration regConfig = RegistryCoreServiceComponent.getRegistryConfig();
        String username = regConfig.getValue(RegistryConfiguration.SYSTEM_USER_NAME);
        String password = UUIDGenerator.generateUUID();
        
        if (!userStoreReader.isExistingUser(username)) {
            userStoreAdmin.addUser(username, password);
        }

        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.DELETE);

        accessControlAdmin.authorizeUser(username, AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.DELETE);

        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeUser(username, AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.DELETE);

        if (!userStoreReader.isExistingUser(RegistryConstants.ADMIN_USER)) {

            userStoreAdmin.addUser(RegistryConstants.ADMIN_USER,
                    RegistryConstants.ADMIN_PASSWORD);
        }

        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_RESOURCE,
                AccessControlConstants.DELETE);

        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.ROLE_RESOURCE,
                AccessControlConstants.DELETE);

        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.READ);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.ADD);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.EDIT);
        accessControlAdmin.authorizeUser(RegistryConstants.ADMIN_USER,
                AccessControlConstants.USER_PERMISSION_RESOURCE,
                AccessControlConstants.DELETE);

        if (!userStoreReader.isExistingUser(RegistryConstants.ANONYMOUS_USER)) {
            userStoreAdmin.addUser(RegistryConstants.ANONYMOUS_USER,
                    RegistryConstants.ANONYMOUS_PASSWORD);
        }
        if (!userStoreReader.isExistingUser(RegistryConstants.GUEST_USER)) {
            userStoreAdmin.addUser(RegistryConstants.GUEST_USER,
                    UUIDGenerator.generateUUID());
        }
    }

    /**
     * Path of a resource given to the Registry interface may contain extensions
     * to refer metadata about resources. But we always store the authorization
     * for resources against the pure resource path, stored in the ARTIFACTS
     * table. This methods extracts that pure resource path from a given path.
     *
     * @param resourcePath
     *            A path string, which may contain extensions
     * @return pure resource path for the given path
     */
    public static String getAuthorizationPath(String resourcePath) {

        // if the user has permission to the current version, he will get
        // permission to all
        // previous versions of the same resource

        String preparedPath = resourcePath;
        if (resourcePath.indexOf("?") > 0) {
            preparedPath = resourcePath.split("\\?")[0];
        } else if (resourcePath.indexOf(RegistryConstants.URL_SEPARATOR) > 0) {
            preparedPath = resourcePath.split("\\;")[0];
        }

        if (preparedPath.equals(RegistryConstants.ROOT_PATH)) {
            return preparedPath;

        } else {

            if (!preparedPath.startsWith(RegistryConstants.ROOT_PATH)) {
                preparedPath = RegistryConstants.ROOT_PATH + preparedPath;
            }

            if (preparedPath.endsWith(RegistryConstants.PATH_SEPARATOR)) {
                preparedPath = preparedPath.substring(0,
                        preparedPath.length() - 1);
            }
        }

        return preparedPath;
    }
}
