/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.test.jdbc;

import org.wso2.carbon.registry.core.jdbc.EmbeddedRegistry;
import org.wso2.carbon.registry.core.jdbc.InMemoryEmbeddedRegistry;
import org.wso2.carbon.registry.core.jdbc.realm.RegistryRealm;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.test.utils.BaseTestCase;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import java.util.Properties;
import java.io.StringWriter;
import java.io.Reader;
import java.io.StringReader;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.xmlbeans.xml.stream.XMLInputStream;

import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;


public class DumpTest extends BaseTestCase {

    protected static EmbeddedRegistry embeddedRegistry = null;
    protected static Registry registry = null;
    protected static RegistryRealm realm = null;

    public void setUp() {
        super.setUp();
        if (embeddedRegistry != null) {
            return;
        }
        try {
            embeddedRegistry = new InMemoryEmbeddedRegistry();
            registry = embeddedRegistry.getUserRegistry(
                    RegistryConstants.ADMIN_USER, RegistryConstants.ADMIN_PASSWORD);
        } catch (RegistryException e) {
                fail("Failed to initialize the registry. Caused by: " + e.getMessage());
        }
    }

    public void testResourceDump() throws RegistryException, XMLStreamException {
        Resource r = registry.newResource();
        r.setProperty("key1", "value1");
        r.setProperty("key2", "value2");
        r.setContent("content 1");
        registry.put("/testDump", r);

        registry.addComment("/testDump", new Comment("comment1"));
        registry.addComment("/testDump", new Comment("comment2"));

        registry.applyTag("/testDump", "tag1");
        registry.applyTag("/testDump", "tag2");

        registry.rateResource("/testDump", 3);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testDump", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/testDumpDup", input);
        r = registry.get("/testDumpDup");

        assertEquals(new String((byte[])r.getContent()), "content 1");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value1");
        assertEquals(r.getProperty("key2"), "value2");

        // getting the comments
        Comment[] comments = registry.getComments("/testDumpDup");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment1");
        assertEquals(comments[1].getText(), "comment2");

        // getting the tags
        Tag[] tags = registry.getTags("/testDumpDup");
        assertEquals(tags[0].getTagName(), "tag1");
        assertEquals(tags[1].getTagName(), "tag2");

        // getting the ratings
        int rate = registry.getRating("/testDumpDup", "admin");
        assertEquals(rate, 3);
    }


    public void testCollectionDump() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value1");
        r.setProperty("key2", "value2");
        registry.put("/testDumpC", r);

        // adding childs
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testDumpC/child1C", r1);

        Resource r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/testDumpC/child2R", r2);

        registry.addComment("/testDumpC", new Comment("comment1"));
        registry.addComment("/testDumpC", new Comment("comment2"));

        registry.applyTag("/testDumpC", "tag1");
        registry.applyTag("/testDumpC", "tag2");

        registry.rateResource("/testDumpC", 3);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testDumpC", writer);
        Reader input = new StringReader(writer.toString());

        // now restoring and retriving the dumped element
        registry.restore("/testDumpDupC", input);
        r = registry.get("/testDumpDupC");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value1");
        assertEquals(r.getProperty("key2"), "value2");

        // getting the comments
        Comment[] comments = registry.getComments("/testDumpDupC");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment1");
        assertEquals(comments[1].getText(), "comment2");

        // getting the tags
        Tag[] tags = registry.getTags("/testDumpDupC");
        assertEquals(tags[0].getTagName(), "tag1");
        assertEquals(tags[1].getTagName(), "tag2");

        // getting the ratings
        int rate = registry.getRating("/testDumpDupC", "admin");
        assertEquals(rate, 3);

        // getting the chils
        r1 = registry.get("/testDumpDupC/child1C");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/testDumpDupC/child2R");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }


    public void testRootDump() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value1");
        r.setProperty("key2", "value2");
        registry.put("/", r);

        // adding childs
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/child1CX", r1);

        Resource r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/child2RX", r2);

        registry.addComment("/", new Comment("comment1"));
        registry.addComment("/", new Comment("comment2"));

        registry.applyTag("/", "tag1");
        registry.applyTag("/", "tag2");

        registry.rateResource("/", 3);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/testDumpDupR", input);
        r = registry.get("/testDumpDupR");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value1");
        assertEquals(r.getProperty("key2"), "value2");

        // getting the comments
        Comment[] comments = registry.getComments("/testDumpDupR");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment1");
        assertEquals(comments[1].getText(), "comment2");

        // getting the tags
        Tag[] tags = registry.getTags("/testDumpDupR");
        assertEquals(tags[0].getTagName(), "tag1");
        assertEquals(tags[1].getTagName(), "tag2");

        // getting the ratings
        int rate = registry.getRating("/testDumpDupR", "admin");
        assertEquals(rate, 3);
        
        // getting the chils
        r1 = registry.get("/testDumpDupR/child1CX");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/testDumpDupR/child2RX");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }

    public void testRootRestore() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value3");
        r.setProperty("key2", "value4");
        registry.put("/testSomewhereElse", r);

        // adding childs
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY", r1);

        Resource r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/testSomewhereElse/child2RY", r2);

        registry.addComment("/testSomewhereElse", new Comment("comment3"));
        registry.addComment("/testSomewhereElse", new Comment("comment4"));

        registry.applyTag("/testSomewhereElse", "tag3");
        registry.applyTag("/testSomewhereElse", "tag4");

        registry.rateResource("/testSomewhereElse", 2);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testSomewhereElse", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/", input);
        r = registry.get("/");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value3");
        assertEquals(r.getProperty("key2"), "value4");

        // getting the comments
        Comment[] comments = registry.getComments("/");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment3");
        assertEquals(comments[1].getText(), "comment4");

        // getting the tags
        Tag[] tags = registry.getTags("/");
        assertEquals(tags[0].getTagName(), "tag3");
        assertEquals(tags[1].getTagName(), "tag4");

        // getting the ratings
        int rate = registry.getRating("/", "admin");
        assertEquals(rate, 2);

        // getting the chils
        r1 = registry.get("/child1CY");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/child2RY");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }



    public void testNewRestore() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value3");
        r.setProperty("key2", "value4");
        registry.put("/testSomewhereElse", r);

        // adding childs
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY", r1);
        
        r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY/xxxx", r1);
        
        r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY/yyyy", r1);

        Resource r2 = registry.newCollection();
        r2.setContent("content child2R");
        registry.put("/testSomewhereElse/newwwC1", r2);

        r2 = registry.newCollection();
        r2.setContent("content child2R");
        registry.put("/testSomewhereElse/newwwC2", r2);

        r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/testSomewhereElse/child2RY", r2);

        registry.addComment("/testSomewhereElse", new Comment("comment3"));
        registry.addComment("/testSomewhereElse", new Comment("comment4"));

        registry.applyTag("/testSomewhereElse", "tag3");
        registry.applyTag("/testSomewhereElse", "tag4");

        registry.rateResource("/testSomewhereElse", 2);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testSomewhereElse", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/", input);
        r = registry.get("/");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value3");
        assertEquals(r.getProperty("key2"), "value4");

        // getting the comments
        Comment[] comments = registry.getComments("/");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment3");
        assertEquals(comments[1].getText(), "comment4");

        // getting the tags
        Tag[] tags = registry.getTags("/");
        assertEquals(tags[0].getTagName(), "tag3");
        assertEquals(tags[1].getTagName(), "tag4");

        // getting the ratings
        int rate = registry.getRating("/", "admin");
        assertEquals(rate, 2);

        // getting the chils
        r1 = registry.get("/child1CY");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");
        
        r2 = registry.get("/child1CY/xxxx");
        assertTrue((r2 instanceof CollectionImpl));
        r2 = registry.get("/child1CY/yyyy");
        assertTrue((r2 instanceof CollectionImpl));

        r2 = registry.get("/child2RY");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }    
}