/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.test.jdbc;

import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.realm.RegistryRealm;
import org.wso2.carbon.registry.core.jdbc.EmbeddedRegistry;
import org.wso2.carbon.registry.core.jdbc.InMemoryEmbeddedRegistry;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.test.utils.BaseTestCase;

public class TransactionsTest extends BaseTestCase {

    protected static EmbeddedRegistry embeddedRegistry = null;
    protected static Registry registry = null;
    protected static RegistryRealm realm = null;

    public void setUp() {
        super.setUp();
        if (embeddedRegistry != null) {
            return;
        }
        try {
            embeddedRegistry = new InMemoryEmbeddedRegistry();
            registry = embeddedRegistry.getUserRegistry(
                    RegistryConstants.ADMIN_USER, RegistryConstants.ADMIN_PASSWORD);
        } catch (RegistryException e) {
                fail("Failed to initialize the registry. Caused by: " + e.getMessage());
        }
    }

    public void testSuccessfullSimpleTransaction() throws RegistryException {

        registry.beginTransaction();

        Resource r1 = registry.newResource();
        r1.setProperty("test", "t1");
        r1.setContent("some content");
        registry.put("/t1/r1", r1);

        registry.commitTransaction();

        Resource r1b = registry.get("/t1/r1");
        assertEquals("Completed transaction resource should have a " +
                "property named 'test' with value 't1'", r1b.getProperty("test"), "t1");
    }

     public void testFailedSimpleTransaction() throws RegistryException {

        registry.beginTransaction();

        Resource r2 = registry.newResource();
        r2.setProperty("test", "t2");
        r2.setContent("some content");
        registry.put("/t1/r2", r2);

        registry.rollbackTransaction();

        try {
            registry.get("/t1/r2");
            fail("Resource added by the incomplete transaction should be deleted.");
        } catch (RegistryException e) {}
    }

    public void testSuccessfullMultioperationTransaction() throws RegistryException {

        registry.beginTransaction();

        Resource r1 = registry.newResource();
        r1.setProperty("test", "t2");
        r1.setContent("some content");
        registry.put("t2/r1", r1);

        registry.addComment("/t2/r1", new Comment("this is a good one"));

        registry.rateResource("/t2/r1", 3);

        registry.applyTag("/t2/r1", "cool");

        registry.commitTransaction();

        Resource r1b = registry.get("/t2/r1");
        assertEquals("Completed transaction resource should have a " +
                "property named 'test' with value 't2'", r1b.getProperty("test"), "t2");

        Comment[] comments = registry.getComments("/t2/r1");
        assertEquals("Resource should have a comment", comments[0].getText(), "this is a good one");

        Tag[] tags = registry.getTags("/t2/r1");
        assertEquals("Resource should have a tag", tags[0].getTagName(), "cool");

        assertEquals("Average rating of /t2/r1 should be 3",
                registry.getAverageRating("/t2/r1"), 3f);
    }

    public void testFailedMultioperationTransaction() throws RegistryException {

        registry.beginTransaction();

        Resource r1 = registry.newResource();
        r1.setProperty("test", "t2");
        r1.setContent("some content");
        registry.put("t3/r1", r1);

        registry.addComment("/t3/r1", new Comment("this is a good one"));

        registry.rateResource("/t3/r1", 3);

        registry.applyTag("/t3/r1", "cool");

        registry.rollbackTransaction();

        try {
            registry.get("/t3/r1");
            fail("Resource /t3/r1 should be deleted after transaction is rolled back.");
        } catch (RegistryException e) {}

        Comment[] comments = registry.getComments("/t3/r1");
        assertEquals("Resource should not have comments", comments.length, 0);

        Tag[] tags = registry.getTags("/t3/r1");
        assertEquals("Resource should not have tags", tags.length, 0);

        assertEquals("Average rating of /t3/r1 should be 0",
                registry.getAverageRating("/t3/r1"), 0f);
    }
}
