/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.wso2.carbon.securevault.internal;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.commons.security.SecurityConstants;
import org.apache.synapse.commons.security.secret.SecretCallbackHandler;
import org.apache.synapse.commons.security.secret.SecretCallbackHandlerFactory;
import org.apache.synapse.commons.security.secret.SecretManager;
import org.osgi.framework.ServiceRegistration;
import org.osgi.service.component.ComponentContext;
import org.wso2.carbon.securevault.SecretCallbackHandlerService;
import org.wso2.carbon.securevault.SecretCallbackHandlerServiceImpl;
import org.wso2.carbon.securevault.SecretManagerSecretCallbackHandler;
import org.wso2.carbon.utils.ServerConstants;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;


/**
 * @scr.component name="secret.manager.initializer.component" immediate="true"
 */
public class SecretManagerInitializerComponent {

    private SecretManager secretManager = SecretManager.getInstance();
    private static final Log log = LogFactory.getLog(SecretManagerInitializerComponent.class);
    private String SECRET_CONF = "secret-conf.properties";
    private static String CONF_DIR = "conf";
    private static String GLOBAL_PREFIX = "carbon.";

    protected void activate(ComponentContext ctxt) {

        if (log.isDebugEnabled()) {
            log.debug("Starting 'SecretManagerInitializerComponent'");
        }

        Properties properties = loadProperties();

        if (secretManager.isInitialized()) {
            if (log.isDebugEnabled()) {
                log.debug("SecretManager already has been initialized.");
            }
        } else {
            secretManager.init(properties);
        }

        SecretCallbackHandlerServiceImpl serviceImpl = null;

        if (!secretManager.isInitialized()) {

            SecretCallbackHandler passwordProvider =
                    SecretCallbackHandlerFactory.createSecretCallbackHandler(properties,
                            GLOBAL_PREFIX + SecurityConstants.PROP_SECRET_PROVIDER);

            if (passwordProvider != null) {
                serviceImpl = new SecretCallbackHandlerServiceImpl();
                serviceImpl.setSecretCallbackHandler(passwordProvider);

            }
        }

        if (serviceImpl == null) {
            serviceImpl = new SecretCallbackHandlerServiceImpl();
            serviceImpl.setSecretCallbackHandler(
                    new SecretManagerSecretCallbackHandler(secretManager));
        }

        ServiceRegistration registration = ctxt.getBundleContext().registerService(
                SecretCallbackHandlerService.class.getName(),
                serviceImpl, null);

    }

    protected void deactivate(ComponentContext ctxt) {

        if (log.isDebugEnabled()) {
            log.debug("Stopping 'SecretManagerInitializerComponent'");

        }
        secretManager.shoutDown();
    }

    private Properties loadProperties() {
        Properties properties = new Properties();
        String carbonHome = System.getProperty(ServerConstants.CARBON_HOME);
        String filePath = carbonHome + File.separator + CONF_DIR + File.separator +
                SECRET_CONF;

        File dataSourceFile = new File(filePath);
        if (!dataSourceFile.exists()) {
            return properties;
        }

        InputStream in = null;
        try {
            in = new FileInputStream(dataSourceFile);
            properties.load(in);
        } catch (IOException e) {
            String msg = "Error loading properties from a file at :" + filePath;
            log.warn(msg, e);
            return properties;
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException ignored) {

                }
            }
        }
        return properties;
    }
}
