/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.ui.transports.fileupload;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.utils.ServerConstants;
import org.wso2.carbon.utils.ConfigurationContextService;
import org.wso2.carbon.CarbonException;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.ui.CarbonUIUtil;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.servlet.ServletConfig;
import javax.xml.namespace.QName;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class FileUploadExecutorManager {

    private static Log log = LogFactory.getLog(FileUploadExecutorManager.class);

    private Map<String, AbstractFileUploadExecutor> executorMap = new HashMap<String, AbstractFileUploadExecutor>();

    private BundleContext context;

    private ConfigurationContext configContext;

    private String webContext;

    public FileUploadExecutorManager(BundleContext bundleContext, ConfigurationContext configCtx, String webContext)
            throws CarbonException {
        this.context = bundleContext;
        this.configContext = configCtx;
        this.webContext = webContext;
        this.executorMapFilter();
    }

    public boolean execute(HttpServletRequest request, HttpServletResponse response) throws IOException,
            CarbonException {
        
        HttpSession session = request.getSession();
        String cookie = (String) session.getAttribute(ServerConstants.ADMIN_SERVICE_COOKIE);
        request.setAttribute(CarbonConstants.ADMIN_SERVICE_COOKIE, cookie);
        request.setAttribute(CarbonConstants.WEB_CONTEXT, webContext);
        request.setAttribute(CarbonConstants.SERVER_URL, CarbonUIUtil.getServerURL(
                request.getSession().getServletContext(), request.getSession()));

        //for generic Executor to work
        String requestURI = request.getRequestURI();        
        boolean executed = false;
        //TODO - fileupload is hardcoded
        int indexToSplit = requestURI.indexOf("fileupload/") + "fileupload/".length();
        String actionString = requestURI.substring(indexToSplit);
        for (Iterator iterator = executorMap.keySet().iterator(); iterator.hasNext();) {
            String key = (String) iterator.next();

            if (key.equals(actionString)) {
                AbstractFileUploadExecutor obj = executorMap.get(key);
                obj.executeGeneric(request, response, configContext);
                executed = true;
                break;
            }
        }
        try {
            ServiceReference[] serviceReferences =
                    context.getServiceReferences(AbstractFileUploadExecutor.class.getName(), null);
            if (serviceReferences != null) {
                for (ServiceReference reference : serviceReferences) {
                    String action = (String) reference.getProperty("action");
                    if (action != null && requestURI.indexOf(action) > -1) {
                        AbstractFileUploadExecutor uploadExecutor =
                                (AbstractFileUploadExecutor) context.getService(reference);
                        uploadExecutor.executeGeneric(request, response, configContext);
                        executed = true;
                        break;
                    }
                }
            }
        } catch (InvalidSyntaxException e) {
            String msg = "Error while loading FileUploadExecutor(s) references";
            log.error(msg, e);
            throw new CarbonException(e);
        }
        if (!executed) {
            // this is the place '*'  handles if present
            Object obj = executorMap.get("*");
            if (obj == null) {
                log.warn("Reached 'All' section but Could not find the Implementation Class");
                return false;
            }
            AnyFileUploadExecutor executor = (AnyFileUploadExecutor) obj;
            executor.executeGeneric(request, response, configContext);
        }
        return true;
    }

    private void executorMapFilter() throws CarbonException {
        ServerConfiguration serverConfiguration = ServerConfiguration.getInstance();
        OMElement documentElement = serverConfiguration.getDocumentElement();
        OMElement fileUploadConfigElement =
                documentElement.getFirstChildWithName(
                        new QName(ServerConstants.CARBON_SERVER_XML_NAMESPACE, "FileUploadConfig"));
        for (Iterator iterator = fileUploadConfigElement.getChildElements(); iterator.hasNext();) {
            OMElement mapppingElement = (OMElement) iterator.next();
            if (mapppingElement.getLocalName().equalsIgnoreCase("Mapping")) {
                OMElement actionsElement =
                        mapppingElement.getFirstChildWithName(
                                new QName(ServerConstants.CARBON_SERVER_XML_NAMESPACE, "Actions"));

                if (actionsElement == null) {
                    String msg = "The mandatory FileUploadConfig/Actions entry " +
                            "does not exist or is empty in the CARBON_HOME/conf/carbon.xml " +
                            "file. Please fix this error in the  carbon.xml file and restart.";
                    log.error(msg);
                    throw new CarbonException(msg);
                }
                Iterator actionElementIterator =
                        actionsElement.getChildrenWithName(
                                new QName(ServerConstants.CARBON_SERVER_XML_NAMESPACE, "Action"));

                if (!actionElementIterator.hasNext()) {
                    String msg = "A FileUploadConfig/Mapping entry in the " +
                            "CARBON_HOME/conf/carbon.xml should have at least on Action " +
                            "defined. Please fix this error in the carbon.xml file and " +
                            "restart.";
                    log.error(msg);
                    throw new CarbonException(msg);
                }

                OMElement classElement = mapppingElement.getFirstChildWithName(
                        new QName(ServerConstants.CARBON_SERVER_XML_NAMESPACE, "Class"));

                if (classElement == null || classElement.getText() == null) {
                    String msg = "The mandatory FileUploadConfig/Mapping/Class entry " +
                            "does not exist or is empty in the CARBON_HOME/conf/carbon.xml " +
                            "file. Please fix this error in the  carbon.xml file and restart.";
                    log.error(msg);
                    throw new CarbonException(msg);
                }


                AbstractFileUploadExecutor object;
                String className = classElement.getText().trim();

                try {
                    Class clazz = context.getBundle().loadClass(className);
                    Constructor constructor =
                            clazz.getConstructor();
                    object = (AbstractFileUploadExecutor) constructor
                            .newInstance();

                } catch (Exception e) {
                    String msg = "Error occurred while trying to instantiate the " + className +
                            " class specified as a FileUploadConfig/Mapping/class element in " +
                            "the CARBON_HOME/conf/carbon.xml file. Please fix this error in " +
                            "the carbon.xml file and restart.";
                    log.error(msg, e);
                    throw new CarbonException(msg, e);
                }

                while (actionElementIterator.hasNext()) {
                    OMElement actionElement = (OMElement) actionElementIterator.next();
                    if (actionElement.getText() == null) {
                        String msg = "A FileUploadConfig/Mapping/Actions/Action element in the " +
                                "CARBON_HOME/conf/carbon.xml file is empty. Please include " +
                                "the correct value in this file and restart.";
                        log.error(msg);
                        throw new CarbonException(msg);
                    }
                    executorMap.put(actionElement.getText().trim(), object);
                }
            }
        }
    }

    public void addExecutor(String action, String executorClass) throws CarbonException {
        if (action == null) {
            String msg = "A FileUploadConfig/Mapping/Actions/Action element is null ";
            log.error(msg);
        }

        if (executorClass == null || executorClass.equals("")) {
            String msg = "Provided FileUploadExecutor object is invalid ";
            log.error(msg);
        }

        AbstractFileUploadExecutor object;

        try {
            Class clazz = context.getBundle().loadClass(executorClass);
            Constructor constructor =
                    clazz.getConstructor();
            object = (AbstractFileUploadExecutor) constructor
                    .newInstance();
            executorMap.put(action, object);
        } catch (Exception e) {
            String msg = "Error occurred while trying to instantiate the " + executorClass +
                    " class specified as a FileUploadConfig/Mapping/class element";
            log.error(msg, e);
            throw new CarbonException(msg, e);
        }


    }

    public void removeExecutor(String action) {
        if (action == null) {
            String msg = "A FileUploadConfig/Mapping/Actions/Action element is null ";
            log.error(msg);
        }

        executorMap.remove(action);
    }
}
