/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.authorizing;

import java.util.List;
import java.util.Map;

import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserStoreAdmin;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.i18n.Messages;

/**
 * Manipulates data in the User Store
 * 
 * @see org.wso2.usermanager.UserStoreAdmin
 */
public class ACLUserStoreAdmin
        extends ACLUserStoreReader implements UserStoreAdmin {

    private Authorizer authorizer = null;

    private UserStoreAdmin usAdmin = null;

    public ACLUserStoreAdmin(Authorizer athzr, UserStoreAdmin admin,
            AuthorizingRealmConfig authConfig) {
        super(athzr, admin, authConfig);
        this.authorizer = athzr;
        this.usAdmin = admin;
    }

    /**
     * Add user to the user store
     */
    public void addUser(String userName, Object credential) throws UserStoreException {
        if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.USER_RESOURCE, AccessControlConstants.ADD)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.USER_RESOURCE, AccessControlConstants.ADD }));
        }
        usAdmin.addUser(userName, credential);
    }

    /**
     * Update user in the user store
     */
    public void updateUser(String userName, Object newCredential, Object oldCredential)
            throws UserStoreException {

        if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.USER_RESOURCE, AccessControlConstants.EDIT)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.USER_RESOURCE, AccessControlConstants.EDIT }));
        }
        usAdmin.updateUser(userName, newCredential, oldCredential);
    }

    public void updateUser(String userName, Object newCredential) throws UserStoreException {
        if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.USER_RESOURCE, AccessControlConstants.EDIT)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.USER_RESOURCE, AccessControlConstants.EDIT }));
        }
        usAdmin.updateUser(userName, newCredential);
    }

    /**
     * Delete user from user store
     */
    public void deleteUser(String userName) throws UserStoreException {
        if (config.getAuthenticatedUserName().equals(userName)) {
            return;
        } else if (ACLAdminChecker.isAdminUser(userName)) {
            /*
             * TODO : Is there a requirement to delete the user if the caller is
             * a Admin
             */
            return;
        } else if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.USER_RESOURCE, AccessControlConstants.DELETE)) {
            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.USER_RESOURCE, AccessControlConstants.DELETE }));
        }
        usAdmin.deleteUser(userName);
    }

    /**
     * Set user properties in the user store
     */
    public void setUserProperties(String userName, Map properties, String profileName) throws UserStoreException {
        if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (config.getAuthenticatedUserName().equals(userName)
                && config.isCurrentUserEditable()) {
            // do nothing - authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.USER_RESOURCE, AccessControlConstants.EDIT)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.USER_RESOURCE, AccessControlConstants.EDIT }));
        }
        usAdmin.setUserProperties(userName, properties, profileName);
    }

    /**
     * Add role to user store
     */
    public void addRole(String roleName) throws UserStoreException {
        if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.ADD)) {
            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.ADD }));
        }
        usAdmin.addRole(roleName);
    }

    /**
     * Delete role from user store
     */
    public void deleteRole(String roleName) throws UserStoreException {
        if (config.isEnableAdminBehavior() && config.getAdminRoleName().equals(roleName)) {
            return;
        } else if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE }));
        }
        usAdmin.deleteRole(roleName);
    }

    /**
     * Set role properties in the user store
     */
    public void setRoleProperties(String roleName, Map properties) throws UserStoreException {
        if (config.isEnableAdminBehavior() && config.getAdminRoleName().equals(roleName)) {
            return;
        } else if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.EDIT)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.EDIT }));
        }
        usAdmin.setRoleProperties(roleName, properties);
    }

    /**
     * Add user to role
     */
    public void addUserToRole(String userName, String roleName) throws UserStoreException {
        if (config.isEnableAdminBehavior() && config.getAdminRoleName().equals(roleName)) {
            /*
             * TODO : is there a requirement to perform this action if the
             * caller is a Admin
             */
            return;
        } else if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.ADD)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.ADD }));
        }
        usAdmin.addUserToRole(userName, roleName);
    }

    /**
     * Delete user from role
     */
    public void removeUserFromRole(String userName, String roleName) throws UserStoreException {
        if (config.isEnableAdminBehavior() && config.getAdminRoleName().equals(roleName)) {
            /*
             * TODO : is there a requirement to perform this action if the
             * caller is a Admin
             */
            return;
        } else if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE }));
        }
        usAdmin.removeUserFromRole(userName, roleName);
    }

  

    public void updateUserProperties(String userName, Map<String, String> properties,
            String profileName) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }
    
    public void updateRoleName(String roleName, String newRoleName) throws UserStoreException{
	if (config.isEnableAdminBehavior() && config.getAdminRoleName().equals(roleName)) {
            /*
             * caller is a Admin
             */
            return;
        } else if (config.isEnableAdminBehavior() && isAdmin) {
            // authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE }));
        }
	usAdmin.updateRoleName(roleName, newRoleName);
    }
    
    public void updateUserName(String userName, String newUserName) throws UserStoreException{
	if (config.isEnableAdminBehavior() && isAdmin) {
	            // authorized
	} else if (config.getAuthenticatedUserName().equals(userName)
	                && config.isCurrentUserEditable()) {
	            // do nothing - authorized
        } else if (!authorizer.isUserAuthorized(config.getAuthenticatedUserName(),
                AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE)) {

            throw new UnauthorizedException(Messages.getMessage("unAuthorized", new String[] {
                    AccessControlConstants.ROLE_RESOURCE, AccessControlConstants.DELETE }));
        }
	usAdmin.updateUserName(userName, newUserName);
    }

    public void deleteBinaryContent(String userName, String contentName) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void deleteUserProperties(String userName, String[] properties)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void deleteUserProperty(String userName, String property) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

   public void setUserBinaryContent(String userName, String contentName, byte[] data)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void setUserProperty(String username, String attributeId, String attributeValue)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public byte[] getUserBinaryContent(String userName, String contentName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        return null;
    }

    public String getUserPropertyValue(String userName, String propertyName) throws UserStoreException {
        // TODO Auto-generated method stub
        return null;
    }

    public void deleteBinaryContent(String userName, String contentName, String profileName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void deleteUserClaimValue(String userName, String claimName, String claimValue,
            String profileName) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void deleteUserClaimValues(String userName, String[] claims, String profileName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void deleteUserProperties(String userName, String[] properties, String profileName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void deleteUserProperty(String userName, String property, String profileName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void setProfilePicture(String userName, byte[] picture, String profileName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    
    public void setUserBinaryContent(String userName, String contentName, byte[] data,
            String profileName) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void setUserClaimValue(String userName, String claimURI, String claimValue,
            String profileName) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void setUserClaimValues(String userName, Map<String, String> claims, String profileName)
            throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    public void setUserProperty(String username, String attributeId, String attributeValue,
            String profileName) throws UserStoreException {
        // TODO Auto-generated method stub
        
    }

    
}
