/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.def;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.def.util.DefaultSQLHelperImpl;
import org.wso2.carbon.user.core.i18n.Messages;
import org.wso2.carbon.user.core.util.DatabaseUtil;

/**
 *
 */
public class DefaultAuthorizer implements Authorizer {

    private static Log log = LogFactory.getLog(DefaultAuthorizer.class);
    
    /**
     * DBCP datasource
     */
    protected DataSource dataSource = null;

    /**
     * Algorithm to authorize users
     */
    protected String algorithm = null;

    protected SQLHelper sqlHelper = new DefaultSQLHelperImpl();

    protected CacheManager cacheManager = null;

    protected String systemUserName;
    /**
     * The root node of the tree
     */
    protected static final TreeNode root = new TreeNode("/");

    public DefaultAuthorizer(String systemUserName) {
        cacheManager = new CacheManager();
        Cache cache1 = new Cache(UserCoreConstants.AUTHZ_CACHE,
                UserCoreConstants.MAX_OBJECTS_IN_CACHE, false, false, UserCoreConstants.TTL_CACHE,
                UserCoreConstants.IDLE_TIME_IN_CACHE);
        cacheManager.addCache(cache1);
        this.systemUserName = systemUserName;
    }

    public DefaultAuthorizer(DataSource dataSource, String algo, String systemUserName) {
        this(systemUserName);
        this.dataSource = dataSource;
        this.algorithm = algo;
    }

    public DefaultAuthorizer(DataSource dataSource, String algo, SQLHelper sqlHelper, String systemUserName) {
        this(systemUserName);
        this.dataSource = dataSource;
        this.algorithm = algo;
        this.sqlHelper = sqlHelper;
    }

    // --------- Authorizer ------------
    public boolean isUserAuthorized(String userName, String resourceId, String action)
            throws UserStoreException {
        
        if(systemUserName.equals(userName)){
            return true;
        }
        
        Cache cache = cacheManager.getCache(UserCoreConstants.AUTHZ_CACHE);
        Element element = cache.get(userName + resourceId + action);
        
        if (element != null) {
            Boolean bool = (Boolean) element.getObjectValue();
            return bool.booleanValue();
        }

        SearchResult sr = getUserPermission(userName, Util.actionToPermission(action), null, root,
                Util.toComponenets(resourceId));
        if (sr.getLastNodeAllowedAccess() == Boolean.TRUE) {
            return true;
        }

        String[] userRoles = getUserRoles(userName);
        boolean userAllowed = this.isUserAuthorizedFromRoles(userRoles, userName,
                resourceId, action);
        return userAllowed;
    }

    public boolean isRoleAuthorized(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getRolePermission(roleName, Util.actionToPermission(action),
                null, root, Util.toComponenets(resourceId));
        if (sr.getLastNodeAllowedAccess() == Boolean.TRUE) {
            return true;
        }
        return false;
    }

    public String[] getAllowedUsersForResource(String resourceId, String action)
        throws UserStoreException {
        TreeNode.Permission p = Util.actionToPermission(action);
        SearchResult sr = this.getAllowedUsersForResource(null, root, p,
                Util.toComponenets(resourceId));

        return sr.getAllowedEntities().toArray(new String[sr.getAllowedEntities().size()]);
    }

    public String[] getDeniedUsersForResource(String resourceId, String action)
        throws UserStoreException {
        TreeNode.Permission p = Util.actionToPermission(action);
        SearchResult sr = this.getDeniedUsersForResource(null, root, p,
                Util.toComponenets(resourceId));

        return sr.getDeniedEntities().toArray(new String[sr.getDeniedEntities().size()]);
    }

    public String[] getAllowedRolesForResource(String resourceId, String action)
        throws UserStoreException {
        TreeNode.Permission p = Util.actionToPermission(action);
        SearchResult sr = this.getAllowedRolesForResource(null, root, p,
                Util.toComponenets(resourceId));

        return sr.getAllowedEntities().toArray(new String[sr.getAllowedEntities().size()]);
    }

    public String[] getDeniedRolesForResource(String resourceId, String action)
        throws UserStoreException {
        TreeNode.Permission p = Util.actionToPermission(action);
        SearchResult sr = this.getDeniedRolesForResource(null, root, p,
                Util.toComponenets(resourceId));

        return sr.getDeniedEntities().toArray(new String[sr.getDeniedEntities().size()]);
    }
    
    /**
     * User authorization is checked considering the given set of roles.
     * @param userName
     * @param userRoles
     * @param resourceId
     * @param action
     * @return
     * @throws UserStoreException
     */
    public boolean isUserAuthorizedFromGivenRoles(String userName, String[] userRoles, 
            String resourceId, String action) throws UserStoreException {

        SearchResult sr = getUserPermission(userName, Util.actionToPermission(action), null, root,
                Util.toComponenets(resourceId));
        if (sr.getLastNodeAllowedAccess() == Boolean.TRUE) {
            return true;
        }
        return isUserAuthorizedFromRoles(userRoles, userName, resourceId, action);
    }
    
    protected boolean isUserAuthorizedFromRoles(String[] userRoles, String userName,
            String resourceId, String action) throws UserStoreException{
        boolean userAllowed = false;
        List<String> allowedRoles = Arrays.asList(getAllowedRolesForResource(resourceId, action));

        for (String role : userRoles) {
            if (allowedRoles.contains(role)) {
                userAllowed = true;
                break;
            }
        }
        
        return userAllowed;
    }
    
    // ------- trivial utility methods ---------
    public void printTree() {
        printNode(root);
    }

    private void printNode(TreeNode tn) {
        System.out.println(tn.getName());
        for (TreeNode t : tn.getChildren().values()) {
            printNode(t);
        }
    }

    /**
     * Find a node on the tree, starting from the given nodes, and using the list of path segments
     *
     * @param node      the start node to begin the search
     * @param pathParts a List of path segments - i.e. collection/resource names
     * @return the result as a SearchResult
     * @see SearchResult
     */
    protected SearchResult getNode(TreeNode node, List<String> pathParts) {
        if (pathParts == null || pathParts.isEmpty()) {
            return new SearchResult(node, null);

        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    if (!pathParts.isEmpty()) {
                        return getNode(child, pathParts);
                    } else {
                        return new SearchResult(child, null);
                    }
                }
            }
            return new SearchResult(node, pathParts);
        }
    }
    
    private void clearRoleAuthorization(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, Util.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() == null) {
            TreeNode.Permission p = Util.actionToPermission(action);

            Map<String, BitSet> allowRoles = sr.getLastNode().getRoleAllowPermissions();
            BitSet bs = allowRoles.get(roleName);
            if (bs != null) {
                bs.clear(p.ordinal());
            }

            Map<String, BitSet> denyRoles = sr.getLastNode().getRoleDenyPermissions();
            bs = denyRoles.get(roleName);
            if (bs != null) {
                bs.clear(p.ordinal());
            }
        }
    }

    /**
     * Find permission for given 'role' starting from the TreeNode 'node' and using the path
     * segments given in the List as pathParts
     *
     * @param role      the role for which the permissions are searched
     * @param p         the permission checked
     * @param sr        the SearchResult that merges the permissions as the tree is traversed
     * @param node      the current node
     * @param pathParts the list of path segments to traverse
     * @return the final SearchResult that merges the permissions
     */
    private SearchResult getRolePermission(String role, TreeNode.Permission p, SearchResult sr,
                                           TreeNode node, List<String> pathParts) {

        if (sr == null) {
            sr = new SearchResult();
        }

        Boolean currentNodeAllows = node.isRoleAuthorized(role, p);
        if (currentNodeAllows == Boolean.TRUE) {
            sr.setLastNodeAllowedAccess(Boolean.TRUE);
        } else if (currentNodeAllows == Boolean.FALSE) {
            sr.setLastNodeAllowedAccess(Boolean.FALSE);
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;

        } else {

            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getRolePermission(role, p, sr, child, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find permission for given 'user' starting from the TreeNode 'node' and using the path
     * segments given in the List as pathParts
     *
     * @param user      the user for which the permissions are searched
     * @param p         the permission checked
     * @param sr        the SearchResult that merges the permissions as the tree is traversed
     * @param node      the current node
     * @param pathParts the list of path segments to traverse
     * @return the final SearchResult that merges the permissions
     */
    private SearchResult getUserPermission(String user, TreeNode.Permission p, SearchResult sr,
                                           TreeNode node, List<String> pathParts) {

        if (sr == null) {
            sr = new SearchResult();
        }

        Boolean currentNodeAllows = node.isUserAuthorized(user, p);
        if (currentNodeAllows == Boolean.TRUE) {
            sr.setLastNodeAllowedAccess(Boolean.TRUE);
        } else if (currentNodeAllows == Boolean.FALSE) {
            sr.setLastNodeAllowedAccess(Boolean.FALSE);
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;

        } else {

            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getUserPermission(user, p, sr, child, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the allowed Roles for a given resource by traversing the whole pemission tree.
     *
     * @param sr - search result to contain allowed roles
     * @param node - current node
     * @param p - permission
     * @param pathParts - list of path segments to traverse
     * @return - search result with allowed roles
     */

    private SearchResult getAllowedRolesForResource(SearchResult sr, TreeNode node,
                                                    TreeNode.Permission p, List<String> pathParts) {
        if (sr == null) {
            sr = new SearchResult();
        }

        /**
         * Add allowed roles of the current node to our list in the sr
         */
        Map<String, BitSet> allowRoles = node.getRoleAllowPermissions();
        for (String role : allowRoles.keySet()) {
            BitSet bs = allowRoles.get(role);
            if (bs.get(p.ordinal())) {
                if(!sr.getAllowedEntities().contains(role)){
                    sr.getAllowedEntities().add(role);
                }
            }
        }

        /**
         * Remove denied roles of the current node from our list in the sr
         */
        Map<String, BitSet> denyRoles = node.getRoleDenyPermissions();
        for (String role : denyRoles.keySet()) {
            BitSet bs = denyRoles.get(role);
            if (bs.get(p.ordinal()) && sr.getAllowedEntities().contains(role)) {
                sr.getAllowedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getAllowedRolesForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the denied Roles for a given resource by traversing the whole pemission tree.
     *
     * @param sr - search result to contain denied roles
     * @param node - current node
     * @param p - permission
     * @param pathParts - list of path segments to traverse
     * @return - search result with denied roles
     */

    private SearchResult getDeniedRolesForResource(SearchResult sr, TreeNode node,
                                                   TreeNode.Permission p, List<String> pathParts) {
        if (sr == null) {
            sr = new SearchResult();
        }

        /**
         * Add denied roles of the current node to our list in the sr
         */
        Map<String, BitSet> denyRoles = node.getRoleDenyPermissions();
        for (String role : denyRoles.keySet()) {
            BitSet bs = denyRoles.get(role);
            if (bs.get(p.ordinal())) {
                if(!sr.getDeniedEntities().contains(role)){
                    sr.getDeniedEntities().add(role);
                }
            }
        }

        /**
         * Remove allowed roles of the current node from our list in the sr
         */
        Map<String, BitSet> allowRoles = node.getRoleAllowPermissions();
        for (String role : allowRoles.keySet()) {
            BitSet bs = allowRoles.get(role);
            if (bs.get(p.ordinal()) && sr.getDeniedEntities().contains(role)) {
                sr.getDeniedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getDeniedRolesForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the allowed Users for a given resource by traversing the whole pemission tree.
     *
     * @param sr - search result to contain allowed users
     * @param node - current node
     * @param p - permission
     * @param pathParts - list of path segments to traverse
     * @return - search result with allowed users
     */

    private SearchResult getAllowedUsersForResource(SearchResult sr, TreeNode node,
                                                    TreeNode.Permission p, List<String> pathParts) {
        if (sr == null) {
            sr = new SearchResult();
        }

        /**
         * Add allowed users of the current node to our list in the sr
         */
        Map<String, BitSet> allowUsers = node.getUserAllowPermissions();
        for (String role : allowUsers.keySet()) {
            BitSet bs = allowUsers.get(role);
            if (bs.get(p.ordinal())) {
                if(!sr.getAllowedEntities().contains(role)){
                    sr.getAllowedEntities().add(role);
                }
            }
        }

        /**
         * Remove denied users of the current node from our list in the sr
         */
        Map<String, BitSet> denyUsers = node.getUserDenyPermissions();
        for (String role : denyUsers.keySet()) {
            BitSet bs = denyUsers.get(role);
            if (bs.get(p.ordinal()) && sr.getAllowedEntities().contains(role)) {
                sr.getAllowedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getAllowedUsersForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the denied users for a given resource by traversing the whole pemission tree.
     *
     * @param sr - search result to contain denied users
     * @param node - current node
     * @param p - permission
     * @param pathParts - list of path segments to traverse
     * @return - search result with denied users
     */

    private SearchResult getDeniedUsersForResource(SearchResult sr, TreeNode node,
                                                   TreeNode.Permission p, List<String> pathParts) {
        if (sr == null) {
            sr = new SearchResult();
        }

        /**
         * Add denied users of the current node to our list in the sr
         */
        Map<String, BitSet> denyUsers = node.getUserDenyPermissions();
        for (String role : denyUsers.keySet()) {
            BitSet bs = denyUsers.get(role);
            if (bs.get(p.ordinal())) {
                if(!sr.getDeniedEntities().contains(role)){
                    sr.getDeniedEntities().add(role);
                }
            }
        }

        /**
         * Remove allowed users of the current node from our list in the sr
         */
        Map<String, BitSet> allowUsers = node.getUserAllowPermissions();
        for (String role : allowUsers.keySet()) {
            BitSet bs = allowUsers.get(role);
            if (bs.get(p.ordinal()) && sr.getDeniedEntities().contains(role)) {
                sr.getDeniedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getDeniedUsersForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }


    public Boolean getUserAuthorized(String userName, String resourceId, String action,
                                     Connection conn) throws UserStoreException {
        //TODO implement
        return true;
    }

    /**
     * The SearchResult holder
     */
    protected class SearchResult {
        /**
         * This is an exact TreeNode match, or the last node found while traversing the tree
         */
        private TreeNode lastNode;
        /**
         * The List of path segments not found on the tree, during the search
         */
        private List<String> unprocessedPaths;
        /**
         * The accumulated state of explicit permissions to the last node
         */
        private Boolean lastNodeAllowedAccess = Boolean.FALSE;

        private List<String> allowedEntities = new ArrayList<String>();
        private List<String> deniedEntities = new ArrayList<String>();

        SearchResult() {
        }

        SearchResult(TreeNode tn, List<String> up) {
            this.lastNode = tn;
            this.unprocessedPaths = up;
        }

        public TreeNode getLastNode() {
            return lastNode;
        }

        public void setLastNode(TreeNode lastNode) {
            this.lastNode = lastNode;
        }

        public List<String> getUnprocessedPaths() {
            return unprocessedPaths;
        }

        public void setUnprocessedPaths(List<String> unprocessedPaths) {
            this.unprocessedPaths = unprocessedPaths;
        }

        public Boolean getLastNodeAllowedAccess() {
            return lastNodeAllowedAccess;
        }

        public void setLastNodeAllowedAccess(Boolean lastNodeAllowedAccess) {
            this.lastNodeAllowedAccess = lastNodeAllowedAccess;
        }

        public List<String> getAllowedEntities() {
            return allowedEntities;
        }

        public void setAllowedEntities(List<String> allowedEntities) {
            this.allowedEntities = allowedEntities;
        }

        public List<String> getDeniedEntities() {
            return deniedEntities;
        }

        public void setDeniedEntities(List<String> deniedEntities) {
            this.deniedEntities = deniedEntities;
        }
    }

    public String[] getUserRoles(String username)  throws UserStoreException {
        String[] names = new String[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_USER_ROLES));
            prepStmt.setString(1,username);
            rs = prepStmt.executeQuery();
            List lst = new LinkedList();
            String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME);
            while (rs.next()) {
                lst.add(rs.getString(colName));
            }
            if (lst.size() > 0) {
                names = (String[]) lst.toArray(new String[lst.size()]);
            }
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return names;
    }
    
    /**
     * On the server startup, all permissions are populated from the DB and the permission tree
     * is built in memory..
     * @throws UserStoreException - SQL exceptions
     */

    public void populatePermissionTreeFromDB() throws UserStoreException {
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        try {
            //Populating role permissions
            dbConnection = getConnection(dbConnection);
            prepStmt1 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_EXISTING_ROLE_PERMISSIONS));
            rs = prepStmt1.executeQuery();
            String roleNameColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME);
            String resourceIDColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_RESOURCE_ID);
            String actionColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ACTION);
            String allowedColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED);

            while (rs.next()) {
                short allowed = rs.getShort(allowedColumn);
                if (allowed == UserCoreConstants.ALLOW) {
                    authorizeRoleInTree(rs.getString(roleNameColumn), rs.getString(resourceIDColumn), rs.getString(actionColumn));
                } else if (allowed == UserCoreConstants.DENY) {
                    denyRoleInTree(rs.getString(roleNameColumn), rs.getString(resourceIDColumn), rs.getString(actionColumn));
                }
            }

            //Populating user permissions
            prepStmt2 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_EXISTING_USER_PERMISSIONS));
            rs = prepStmt2.executeQuery();

            String userNameColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_USER_NAME);
            resourceIDColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_RESOURCE_ID);
            actionColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ACTION);
            allowedColumn = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_IS_ALLOWED);

            while (rs.next()) {
                short allowed = rs.getShort(allowedColumn);
                if (allowed == UserCoreConstants.ALLOW) {
                    authorizeUserInTree(rs.getString(userNameColumn),rs.getString(resourceIDColumn),rs.getString(actionColumn));
                } else if (allowed == UserCoreConstants.DENY) {
                    denyUserInTree(rs.getString(userNameColumn),rs.getString(resourceIDColumn),rs.getString(actionColumn));
                }
            }

        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt1, prepStmt2);
        }
    }
    
    public void authorizeUserInTree(String userName, String resourceId, String action){
        SearchResult sr = getNode(root, Util.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.authorizeUser(userName, Util.actionToPermission(action));

        } else {
            sr.getLastNode().authorizeUser(userName, Util.actionToPermission(action));
        }
    }
    
    public void denyUserInTree(String userName, String resourceId, String action)
            throws UserStoreException {
        SearchResult sr = getNode(root, Util.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.denyUser(userName, Util.actionToPermission(action));
        } else {
            sr.getLastNode().denyUser(userName, Util.actionToPermission(action));
        }
    }
    
    public void authorizeRoleInTree(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, Util.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.authorizeRole(roleName, Util.actionToPermission(action));

        } else {
            sr.getLastNode().authorizeRole(roleName, Util.actionToPermission(action));
        }
    }

    public void denyRoleInTree(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, Util.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.denyRole(roleName, Util.actionToPermission(action));

        } else {
            sr.getLastNode().denyRole(roleName, Util.actionToPermission(action));
        }
    }
    
    private Connection getConnection(Connection dbConnection) throws SQLException,
            UserStoreException {
        dbConnection = dataSource.getConnection();
        if (dbConnection == null) {
            throw new UserStoreException(Messages.getMessage("null_connection"));
        }
        dbConnection.setAutoCommit(false);
        return dbConnection;
    }

}
