/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.def;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.sql.DataSource;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;

import org.apache.axiom.om.util.Base64;
import org.apache.axiom.om.util.UUIDGenerator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.AuthenticatorException;
import org.wso2.carbon.user.core.UserClaimsAdmin;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserProfileAdmin;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreAdmin;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.def.util.DefaultCommons;
import org.wso2.carbon.user.core.i18n.Messages;
import org.wso2.carbon.user.core.util.DatabaseUtil;

/**
 * Manipulates data in the User Store
 * 
 */
public class DefaultUserStoreAdmin extends DefaultUserStoreReader implements UserStoreAdmin {

    private static Log log = LogFactory.getLog(DefaultUserStoreAdmin.class);

    public DefaultUserStoreAdmin(DataSource dataSource, String systemUserName) {
        super(dataSource, systemUserName);
    }

    public DefaultUserStoreAdmin(DataSource dataSource, SQLHelper sqlHelper, String systemUserName) {
        super(dataSource, sqlHelper, systemUserName);
    }

    public DefaultUserStoreAdmin(DataSource dataSource, SQLHelper sqlHelper,
            UserClaimsAdmin claimsAdmin, UserProfileAdmin profileAdmin, DefaultRealmConfig config,
            UserRealm realm) {
        super(dataSource, sqlHelper, claimsAdmin, profileAdmin, config, realm);
    }

    public void addUser(String userName, Object credential) throws UserStoreException {

        if (userName == null || credential == null) {
            throw new UserStoreException("Invalid data provided");
        }

        if (!(credential instanceof String)) {
            throw new UserStoreException("Credential Type Not Supported");
        }

        if (credential.toString().trim().length() == 0) {
            throw new UserStoreException("Invalid data provided");
        }

        userName = userName.trim();
        if (userName.length() == 0) {
            throw new UserStoreException("Invalid data provided");
        }

        if (isExistingUser(userName)) {
            throw new UserStoreException("Duplicate user name.");
        }

        Connection dbConnection = null;
        PreparedStatement addUserStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            
            String password = credential.toString().trim();
            String digestInput = password;

            if (config.isSaltedPassword()) {
                addUserStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.ADD_USER));
                String salt = UUIDGenerator.getUUID().substring(9);
                digestInput = password + salt;
                addUserStmt.setString(3, salt);
            } else {
                addUserStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.ADD_USER_OLD));
            }

            addUserStmt.setString(1, userName);
            
            MessageDigest dgst = MessageDigest.getInstance(config.getDigestFunction());
            byte[] digestedPasword = dgst.digest(digestInput.getBytes());
            String finalValue = Base64.encode(digestedPasword);
            addUserStmt.setString(2, finalValue);

            addUserStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } catch (NoSuchAlgorithmException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, addUserStmt);
        }

    }

    public void updateUser(String userName, Object newCredential, Object oldCredential)
            throws UserStoreException {
        if (oldCredential == null || newCredential == null) {
            String msg = "Old or new credential is null";
            log.error(msg);
            throw new UserStoreException(msg);
        }

        if (newCredential.toString().trim().length() == 0) {
            log.error("Invalid data providded");
            throw new UserStoreException("Invalid data provided");
        }
        
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement updateUserStmt = null;
        try {
            if (this.realm.getAuthenticator().authenticate(userName, oldCredential) == false) {
                throw new UserStoreException(Messages.getMessage("oldPasswordIncorrect"));
            }
            dbConnection = getConnection(dbConnection);
            String digestInput = ((String) newCredential).trim();
            if (config.isSaltedPassword()) {
                String salt = UUIDGenerator.getUUID().substring(9);
                digestInput = digestInput + salt;
                updateUserStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.UPDATE_USER));
                updateUserStmt.setString(2, salt);
                updateUserStmt.setString(3, userName);
            } else {
                updateUserStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.UPDATE_USER_OLD));
                updateUserStmt.setString(2, userName);
            }

            MessageDigest dgst = MessageDigest.getInstance(config.getDigestFunction());
            byte[] digestedPasword = dgst.digest(digestInput.getBytes());
            String finalValue = Base64.encode(digestedPasword);
            updateUserStmt.setString(1, finalValue);

            
            updateUserStmt.executeUpdate();
            dbConnection.commit();
        } catch (AuthenticatorException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } catch (NoSuchAlgorithmException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, updateUserStmt);
        }
    }

    public void updateUser(String userName, Object newCredential) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement updateUserStmt = null;
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        if (newCredential.toString().trim().length() == 0) {
            log.error("Invalid data providded");
            throw new UserStoreException("Invalid data provided");
        }
        try {
            dbConnection = getConnection(dbConnection);
            

            String digestInput = ((String) newCredential).trim();
            if (config.isSaltedPassword()) {
                String salt = UUIDGenerator.getUUID().substring(9);
                updateUserStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.UPDATE_USER));
                digestInput = digestInput + salt;
                updateUserStmt.setString(2, salt);
                updateUserStmt.setString(3, userName);
            } else {
                updateUserStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.UPDATE_USER_OLD));
                updateUserStmt.setString(2, userName);
            }

            MessageDigest dgst = MessageDigest.getInstance(config.getDigestFunction());
            byte[] digestedPasword = dgst.digest(digestInput.getBytes());
            String finalValue = Base64.encode(digestedPasword);
            updateUserStmt.setString(1, finalValue);
            
            updateUserStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } catch (NoSuchAlgorithmException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(Messages.getMessage("errorCreatingPasswordDigest"), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, updateUserStmt);
        }
    }

    public void deleteUser(String userName) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        PreparedStatement prepStmt3 = null;
        PreparedStatement prepStmt4 = null;
        PreparedStatement prepStmt6 = null;
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        try {
            this.realm.getAccessControlAdmin().clearUserAuthorization(userName);
            
            dbConnection = getConnection(dbConnection);
            DefaultCommons commons = new DefaultCommons();
            int id = commons.getUserId(sqlHelper, dbConnection, userName);
            if (id == -1) {
                // todo throw
                throw new UserStoreException();
            }
            prepStmt1 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ON_DELETE_USER_UM_USER_ROLES));
            prepStmt1.setInt(1, id);
            prepStmt1.executeUpdate();

            prepStmt2 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ON_DELETE_USER_UM_USER_DATA));
            prepStmt2.setInt(1, id);
            prepStmt2.executeUpdate();
            prepStmt3 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ON_DELETE_USER_UM_BIN_DATA));
            prepStmt3.setInt(1, id);
            prepStmt3.executeUpdate();
            prepStmt4 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ON_DELETE_USER_UM_USER_ATTRIBUTES));
            prepStmt4.setInt(1, id);
            prepStmt4.executeUpdate();
            
            prepStmt6 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.DELETE_USER));
            prepStmt6.setString(1, userName);
            prepStmt6.executeUpdate();
            
            //clear userId cache
    		CacheManager cacheManager = CacheManager.create();
    		Cache cache = cacheManager.getCache(UserCoreConstants.USER_ID_CACHE);
    		if(cache != null){
        		Object obj = cache.get(userName);
        		if (obj != null) {
        			//if username found, remove it from cache
        			if(log.isDebugEnabled()){
        				log.debug("Found cache entry for : "+userName+". Removing it from cache..");
        			}
        			cache.remove(userName);
        		}    			
    		}
    		
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt1, prepStmt2, prepStmt3,
                    prepStmt4, prepStmt6);
        }
    }

    public void setRoleProperties(String roleName, Map properties) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement deleteRolePropsStmt = null;
        PreparedStatement setRolePropertiesStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            DefaultCommons commons = new DefaultCommons();
            int roleid = commons.getRoleId(sqlHelper, dbConnection, roleName);
            if (roleid == -1) {
                throw new UserStoreException(Messages.getMessage("nullRole"));
            }

            deleteRolePropsStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.DELETE_ROLE_ATTRIBUTE));
            deleteRolePropsStmt.setInt(1, roleid);
            deleteRolePropsStmt.executeUpdate();

            setRolePropertiesStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ADD_ROLE_ATTRIBUTE));

            Iterator ite = properties.entrySet().iterator();
            while (ite.hasNext()) {
                Entry entry = (Entry) ite.next();
                String key = (String) entry.getKey();
                String value = (String) entry.getValue();
                if (value != null) {
                    setRolePropertiesStmt.setString(1, key);
                    setRolePropertiesStmt.setString(2, value);
                    setRolePropertiesStmt.setInt(3, roleid);
                    setRolePropertiesStmt.executeUpdate();
                }
            }
            dbConnection.commit();

        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, deleteRolePropsStmt, setRolePropertiesStmt);
        }
    }

    public void addRole(String roleName) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement addRoleStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            roleName = roleName.trim();

            if (isExistingRole(roleName)) {
                throw new UserStoreException(Messages.getMessage("duplicateRole"));
            }

            dbConnection.setAutoCommit(false);
            addRoleStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ADD_ROLE));
            addRoleStmt.setString(1, roleName);
            addRoleStmt.execute();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, addRoleStmt);
        }
    }

    public void deleteRole(String roleName) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        PreparedStatement prepStmt3 = null;
        PreparedStatement prepStmt4 = null;
        try {
            
            this.realm.getAccessControlAdmin().clearRoleAuthorization(roleName);
            
            dbConnection = getConnection(dbConnection);
            DefaultCommons commons = new DefaultCommons();
            int id = commons.getRoleId(sqlHelper, dbConnection, roleName);
            if (id == -1) {
                // todo throw
                throw new UserStoreException("Invalid role.");
            }
            
            String sql = sqlHelper.getSQL(DefaultRealmConstants.ON_DELETE_ROLE_UM_ROLE_ATTRIBUTES);
            if (sql != null) { // this check is for hybrid case
                prepStmt1 = dbConnection.prepareStatement(sql);
                prepStmt1.setInt(1, id);
                prepStmt1.executeUpdate();
            }
            
            prepStmt3 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ON_DELETE_ROLE_UM_USER_ROLES));
            prepStmt3.setInt(1, id);
            prepStmt3.executeUpdate();
            
            prepStmt4 = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.DELETE_ROLE));
            prepStmt4.setString(1, roleName);
            prepStmt4.executeUpdate();

            //clear role cache
    		CacheManager cacheManager = CacheManager.create();
    		Cache cache = cacheManager.getCache(UserCoreConstants.ROLE_ID_CACHE);
    		if(cache != null){
        		Object obj = cache.get(roleName);
        		if (obj != null) {
        			//if roleName found, remove it from cache
        			if(log.isDebugEnabled()){
        				log.debug("Found cache entry for : "+roleName+". Removing it from cache..");
        			}
        			cache.remove(roleName);
        		}    			
    		}
            
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt1, prepStmt2, prepStmt3, prepStmt4);
        }
    }

    public void addUserToRole(String userName, String roleName) throws UserStoreException {

        Connection dbConnection = null;
        PreparedStatement addUserToRoleStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            DefaultCommons commons = new DefaultCommons();
            int userid = commons.getUserId(sqlHelper, dbConnection, userName);
            int roleid = commons.getRoleId(sqlHelper, dbConnection, roleName);

            if (roleid == -1 || userid == -1) {
                throw new UserStoreException(Messages.getMessage("nullData"));
            }

            addUserToRoleStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.ADD_USER_ROLE));
            addUserToRoleStmt.setInt(1, userid);
            addUserToRoleStmt.setInt(2, roleid);
            addUserToRoleStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(Messages.getMessage("errorAddingUserToRole"), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, addUserToRoleStmt);
        }
    }

    public void removeUserFromRole(String userName, String roleName) throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        Connection dbConnection = null;
        PreparedStatement removeUserFromRoleStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            removeUserFromRoleStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.DELETE_USER_ROLE));
            removeUserFromRoleStmt.setString(1, userName);
            removeUserFromRoleStmt.setString(2, roleName);
            removeUserFromRoleStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, removeUserFromRoleStmt);
        }
    }

    public void updateRoleName(String roleName, String newRoleName) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.UPDATE_ROLE_NAME));
            prepStmt.setString(1, newRoleName);
            prepStmt.setString(2, roleName);
            prepStmt.executeUpdate();
            dbConnection.commit();
            log.info("Changed role name from :: " + roleName + " to " + newRoleName);
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    public void updateUserName(String userName, String newUserName) throws UserStoreException {
        throw new UserStoreException("This operation not allowed");
    }

    // **************************** property stuff
    // ********************************

    public void setUserProperty(String username, String attributeId, String attributeValue,
            String profileName) throws UserStoreException {
        if (systemUserName.equals(username)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        boolean isBasic = this.isBasic(attributeId);
        if (isBasic) {
            setBasicProperty(username, attributeId, attributeValue, profileName);
        } else {
            setCustomProperty(username, attributeId, attributeValue, profileName);
        }
    }

    public void setUserProperties(String userName, Map<String, String> properties,
            String profileName) throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        Iterator<Entry<String, String>> ite = properties.entrySet().iterator();
        Map<String, String> mapCustomProps = new HashMap<String, String>();
        boolean isExistingUserData = false;
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getConnection(dbConnection);

            // TODO: For DimuthuL - Table/View 'UM_USER_DATA' does not exist

            /*
             * String query =
             * sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants
             * .IS_USER_DATA); prepStmt = dbConnection.prepareStatement(query);
             * prepStmt.setString(1, profileName); prepStmt.setString(2,
             * userName); ResultSet rs = prepStmt.executeQuery(); if (rs.next())
             * { int count = rs.getInt(1); if (count > 0) { isExistingUserData =
             * true; } }
             */

            String theSQL = null;
            if (isExistingUserData) {
                theSQL = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.UPDATE_USER_DATA);
            } else {
                theSQL = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.ADD_USER_DATA);
            }

            String updateStr = "";
            String insertStr1 = "";
            String insertStr2 = "";
            List<String> valueList = new ArrayList<String>();
            while (ite.hasNext()) {
                Entry<String, String> entry = ite.next();
                String name = entry.getKey();
                String value = entry.getValue();
                boolean isBasic = this.isBasic(name);
                if (isBasic) {
                    if (isExistingUserData) {
                        // update
                        updateStr = updateStr + name + "=" + "?,";
                        valueList.add(value);
                    } else {
                        // insert
                        insertStr1 = insertStr1 + name + ",";
                        insertStr2 = insertStr2 + "?,";
                        valueList.add(value);
                    }
                } else {
                    mapCustomProps.put(name, value);
                }
            }

            if (updateStr.length() > 1) {
                updateStr = updateStr.substring(0, updateStr.length() - 1);
                theSQL = MessageFormat.format(theSQL, updateStr);
                this.setBasicProperties(userName, theSQL, valueList, dbConnection, profileName);
            } else if (insertStr1.length() > 1) {
                insertStr1 = insertStr1 + "UM_PROFILE_ID, UM_USER_ID";
                insertStr2 = insertStr2 + "?,?";
                theSQL = MessageFormat.format(theSQL, insertStr1, insertStr2);
                this.setBasicProperties(userName, theSQL, valueList, dbConnection, profileName);
            }

            if (mapCustomProps.size() > 0) {
                this.setCustomProperties(userName, mapCustomProps, dbConnection, profileName);
            }
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }

    }

    public void deleteUserProperty(String userName, String property, String profileName)
            throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        boolean isBasic = this.isBasic(property);
        if (isBasic) {
            setBasicProperty(userName, property, null, profileName);
        } else {
            deleteCustomProperty(userName, property, profileName);
        }
    }

    public void deleteUserProperties(String userName, String[] properties, String profileName)
            throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }

        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        List<String> customProps = new ArrayList<String>();
        String theSQL = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.UPDATE_USER_DATA);
        String updateStr = "";
        for (String name : properties) {
            boolean isBasic = this.isBasic(name);
            if (isBasic) {
                updateStr = updateStr + name + "=" + "NULL,";
            } else {
                customProps.add(name);
            }
        }
        PreparedStatement prepStmt = null;
        Connection dbConnection = null;
        try {
            dbConnection = getConnection(dbConnection);
            if (updateStr.length() > 1) {
                updateStr = updateStr.substring(0, updateStr.length() - 1);
                theSQL = MessageFormat.format(theSQL, updateStr);
                setBasicProperties(userName, theSQL, new ArrayList<String>(), dbConnection,
                        profileName);
            }

            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.DELETE_CUSTOM_PROPERTY);
            prepStmt = dbConnection.prepareStatement(sql);
            Iterator<String> ite = customProps.iterator();
            while (ite.hasNext()) {
                String property = ite.next();
                prepStmt.setString(1, property);
                prepStmt.setString(2, profileName);
                prepStmt.setString(3, userName);
                prepStmt.addBatch();
            }

            if (prepStmt != null) {
                int[] deled = prepStmt.executeBatch();
            }
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    public void setUserBinaryContent(String userName, String contentName, byte[] data,
            String profileName) throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        Connection dbConnection = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        try {
            dbConnection = getConnection(dbConnection);
            String sql1 = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.COUNT_BINARY_CONTENT);
            prepStmt1 = dbConnection.prepareStatement(sql1);
            prepStmt1.setString(1, contentName);
            prepStmt1.setString(2, profileName);
            prepStmt1.setString(3, userName);
            ResultSet rs = prepStmt1.executeQuery();
            boolean isExisting = false;
            while (rs.next()) {
                int count = rs.getInt(1);
                if (count > 0) {
                    isExisting = true;
                }
            }
            if (isExisting) {
                prepStmt2 = dbConnection.prepareStatement(sqlHelper
                        .getUserProfileRelatedSQL(DefaultRealmConstants.UPDATE_BINARY_CONTENT));
                prepStmt2.setBytes(1, data);
                prepStmt2.setString(2, profileName);
                prepStmt2.setString(3, userName);
            } else {
                DefaultCommons defCommons = new DefaultCommons();
                int userid = defCommons.getUserId(sqlHelper, dbConnection, userName);
                prepStmt2 = dbConnection.prepareStatement(sqlHelper
                        .getUserProfileRelatedSQL(DefaultRealmConstants.ADD_BINARY_CONTENT));
                prepStmt2.setInt(1, userid);
                prepStmt2.setString(2, contentName);
                prepStmt2.setString(3, profileName);
                prepStmt2.setBytes(4, data);
            }
            prepStmt2.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt1, prepStmt2);
        }
    }

    public void deleteBinaryContent(String userName, String contentName, String profileName)
            throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.DELETE_BINARY_CONTENT));
            prepStmt.setString(1, contentName);
            prepStmt.setString(2, userName);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    // **************************** claim stuff
    // **********************************8

    public void setUserClaimValue(String userName, String claimURI, String claimValue,
            String profileName) throws UserStoreException {
        String property = claimAdmin.getAttributeName(claimURI);
        if (property == null) {
            property = claimURI;
        }
        this.setUserProperty(userName, property, claimValue, profileName);
    }

    public void setUserClaimValues(String userName, Map<String, String> claims, String profileName)
            throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        Map<String, String> propMap = new HashMap<String, String>();

        if (claims.get(UserCoreConstants.PROFILE_CONFIGURATION) == null) {
            claims.put(UserCoreConstants.PROFILE_CONFIGURATION,
                    UserCoreConstants.DEFAULT_PROFILE_CONFIGURATION);
        }

        Iterator<Map.Entry<String, String>> ite = claims.entrySet().iterator();

        while (ite.hasNext()) {
            Map.Entry<String, String> entry = ite.next();
            String prop = claimAdmin.getAttributeName(entry.getKey());
            String propvalue = entry.getValue();
            if (prop != null) {
                propMap.put(prop, propvalue);
            } else {
                propMap.put(entry.getKey(), propvalue);
            }
        }
        this.setUserProperties(userName, propMap, profileName);
    }

    public void deleteUserClaimValue(String userName, String claimName, String claimValue,
            String profileName) throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        String propName = claimAdmin.getAttributeName(claimName);
        this.deleteUserProperty(userName, propName, profileName);
    }

    public void deleteUserClaimValues(String userName, String[] claims, String profileName)
            throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        String[] propNames = new String[claims.length];
        int i = 0;
        for (String claim : claims) {
            String prop = claimAdmin.getAttributeName(claim);
            propNames[i] = prop;
            i++;
        }
        this.deleteUserProperties(userName, propNames, profileName);
    }

    public void setProfilePicture(String userName, byte[] picture, String profileName)
            throws UserStoreException {
        if (systemUserName.equals(userName)) {
            log.error("A user is playing around with system username");
            // returning a general error message
            throw new UserStoreException("Invalid data provided");
        }
        this.setUserBinaryContent(userName, UserCoreConstants.PICTURE_CLAIM, picture, profileName);
    }

    // ********************** End claim stuff

    // *********************************** protected methods
    // *****************************

    protected void setCustomProperty(String userName, String attributeId, String attributeValue,
            String profileName) throws UserStoreException {
        String sql1 = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.GET_USER_PROPERTY);
        Connection dbConnection = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        ResultSet rs = null;
        try {
            dbConnection = getConnection(dbConnection);
            DefaultCommons commons = new DefaultCommons();
            int userid = commons.getUserId(sqlHelper, dbConnection, userName);
            prepStmt1 = dbConnection.prepareStatement(sql1);
            prepStmt1.setString(1, attributeId);
            prepStmt1.setString(2, profileName);
            prepStmt1.setString(3, userName);
            rs = prepStmt1.executeQuery();

            if (rs.next()) {
                // update
                String sql2 = sqlHelper
                        .getUserProfileRelatedSQL(DefaultRealmConstants.UPDATE_USER_PROPERTY);
                prepStmt2 = dbConnection.prepareStatement(sql2);
                prepStmt2.setString(1, attributeValue);
                prepStmt2.setString(2, attributeId);
                prepStmt2.setString(3, profileName);
                prepStmt2.setInt(4, userid);
            } else {
                // add
                String sql2 = sqlHelper
                        .getUserProfileRelatedSQL(DefaultRealmConstants.ADD_USER_PROPERTY);
                prepStmt2 = dbConnection.prepareStatement(sql2);
                prepStmt2.setString(1, attributeId);
                prepStmt2.setString(2, attributeValue);
                prepStmt2.setString(3, profileName);
                prepStmt2.setInt(4, userid);
            }
            prepStmt2.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt1, prepStmt2);
        }
    }

    protected void setBasicProperty(String userName, String attributeId, String attributeValue,
            String profileName) throws UserStoreException {
        String sql1 = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.IS_USER_DATA);
        Connection dbConnection = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        try {
            dbConnection = getConnection(dbConnection);
            DefaultCommons commons = new DefaultCommons();
            int userid = commons.getUserId(sqlHelper, dbConnection, userName);
            prepStmt1 = dbConnection.prepareStatement(sql1);
            prepStmt1.setString(1, profileName);
            prepStmt1.setString(2, userName);
            ResultSet rs = prepStmt1.executeQuery();
            if (rs.next()) {
                // update
                String sql2 = sqlHelper
                        .getUserProfileRelatedSQL(DefaultRealmConstants.UPDATE_SINGLE_USER_DATA);
                sql2 = MessageFormat.format(sql2, attributeId);
                prepStmt2 = dbConnection.prepareStatement(sql2);
                if (attributeValue != null) {
                    prepStmt2.setString(1, attributeValue);
                } else {
                    prepStmt2.setNull(1, java.sql.Types.VARCHAR);
                }
                prepStmt2.setString(2, profileName);
                prepStmt2.setInt(3, userid);

            } else {
                // add
                String sql2 = sqlHelper
                        .getUserProfileRelatedSQL(DefaultRealmConstants.ADD_SINGLE_USER_DATA);
                sql2 = MessageFormat.format(sql2, attributeId);
                prepStmt2 = dbConnection.prepareStatement(sql2);
                prepStmt2.setString(1, attributeValue);
                prepStmt2.setString(2, profileName);
                prepStmt2.setInt(3, userid);
            }

            prepStmt2.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            e.printStackTrace();
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt1, prepStmt2);
        }
    }

    /*
     * 
     */
    protected void setCustomProperties(String userName, Map<String, String> properties,
            Connection dbConnection, String profileName) throws UserStoreException {
        String sql1 = sqlHelper
                .getUserProfileRelatedSQL(DefaultRealmConstants.GET_USER_PROPERTY_NAMES);
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        PreparedStatement prepStmt3 = null;
        try {
            DefaultCommons commons = new DefaultCommons();
            int userid = commons.getUserId(sqlHelper, dbConnection, userName);
            prepStmt1 = dbConnection.prepareStatement(sql1);
            prepStmt1.setString(1, profileName);
            prepStmt1.setString(2, userName);
            ResultSet rs = prepStmt1.executeQuery();
            List<String> existingAttrs = new ArrayList<String>();
            while (rs.next()) {
                existingAttrs.add(rs.getString(1));
            }
            String sql2 = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.UPDATE_USER_PROPERTY);
            String sql3 = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.ADD_USER_PROPERTY);
            prepStmt2 = dbConnection.prepareStatement(sql2);
            prepStmt3 = dbConnection.prepareStatement(sql3);
            Iterator<Entry<String, String>> ite = properties.entrySet().iterator();
            while (ite.hasNext()) {
                Entry<String, String> entry = ite.next();
                String name = entry.getKey();
                String value = entry.getValue();
                if (existingAttrs.contains(name)) {
                    // update
                    prepStmt2.setString(1, value);
                    prepStmt2.setString(2, name);
                    prepStmt2.setString(3, profileName);
                    prepStmt2.setInt(4, userid);
                    prepStmt2.addBatch();
                } else {
                    // insert
                    prepStmt3.setString(1, name);
                    prepStmt3.setString(2, value);
                    prepStmt3.setString(3, profileName);
                    prepStmt3.setInt(4, userid);
                    prepStmt3.addBatch();
                }

            }

            int[] updated = prepStmt2.executeBatch();
            int[] added = prepStmt3.executeBatch();

            dbConnection.commit();

        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(null, prepStmt1, prepStmt2, prepStmt3);
        }
    }

    protected void setBasicProperties(String userName, String sql, List<String> values,
            Connection dbConnection, String profileName) throws UserStoreException {
        PreparedStatement prepStmt = null;
        try {
            prepStmt = dbConnection.prepareStatement(sql);
            Iterator<String> ite = values.iterator();
            int i = 1;
            while (ite.hasNext()) {
                String value = ite.next();
                if (value == null) {
                    prepStmt.setNull(i, java.sql.Types.VARCHAR);
                } else {
                    prepStmt.setString(i, value);
                }
                i++;
            }

            DefaultCommons commons = new DefaultCommons();
            int userid = commons.getUserId(sqlHelper, dbConnection, userName);
            prepStmt.setString(i, profileName);
            prepStmt.setInt(i + 1, userid);
            prepStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            e.printStackTrace();
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(null, prepStmt);
        }
    }

    protected void deleteCustomProperty(String userName, String property, String profileName)
            throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getConnection(dbConnection);
            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.DELETE_CUSTOM_PROPERTY);
            prepStmt = dbConnection.prepareStatement(sql);
            prepStmt.setString(1, property);
            prepStmt.setString(2, profileName);
            prepStmt.setString(3, userName);
            prepStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    private Connection getConnection(Connection dbConnection) throws SQLException,
            UserStoreException {
        dbConnection = dataSource.getConnection();
        if (dbConnection == null) {
            throw new UserStoreException(Messages.getMessage("null_connection"));
        }
        dbConnection.setAutoCommit(false);
        return dbConnection;
    }

}
