/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.def;

import java.io.IOException;
import java.io.InputStream;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.Claim;
import org.wso2.carbon.user.core.ClaimMapping;
import org.wso2.carbon.user.core.ProfileConfiguration;
import org.wso2.carbon.user.core.UserClaimsAdmin;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserProfileAdmin;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.UserStoreReader;
import org.wso2.carbon.user.core.def.util.DefaultCommons;
import org.wso2.carbon.user.core.def.util.DefaultSQLHelperImpl;
import org.wso2.carbon.user.core.util.DatabaseUtil;

/**
 * Manipulates data in the User Store
 * 
 */
public class DefaultUserStoreReader implements UserStoreReader {

    private static Log log = LogFactory.getLog(DefaultUserStoreReader.class);

    protected DataSource dataSource = null;

    protected SQLHelper sqlHelper = null;

    protected UserClaimsAdmin claimAdmin = null;

    protected UserProfileAdmin profileAdmin = null;
    
    protected String systemUserName;

    protected UserRealm realm = null;

    protected DefaultRealmConfig config = null;
    
    public DefaultUserStoreReader(DataSource dataSource, String systemUserName) {
        this.dataSource = dataSource;
        this.sqlHelper = new DefaultSQLHelperImpl();
        this.systemUserName = systemUserName;
    }

    public DefaultUserStoreReader(DataSource dataSource, SQLHelper sqlHelper, String systemUserName) {
        this.dataSource = dataSource;
        this.sqlHelper = sqlHelper;
        this.systemUserName = systemUserName;
    }

    public DefaultUserStoreReader(DataSource dataSource, SQLHelper sqlHelper,
            UserClaimsAdmin claimsAdmin, UserProfileAdmin profileAdmin, DefaultRealmConfig config,
            UserRealm realm) {
        this.dataSource = dataSource;
        this.sqlHelper = sqlHelper;
        this.claimAdmin = claimsAdmin;
        this.profileAdmin = profileAdmin;
        this.systemUserName = config.getSysteUserName();
        this.config = config;
        this.realm = realm;
    }
    
    public DefaultUserStoreReader(DataSource dataSource, SQLHelper sqlHelper, 
            UserClaimsAdmin claimsAdmin, UserProfileAdmin profileAdmin, String systemUserName) {
        this.dataSource = dataSource;
        this.sqlHelper = sqlHelper;
        this.claimAdmin = claimsAdmin;
        this.profileAdmin = profileAdmin;
        this.systemUserName = systemUserName;
    }

    public String[] getAllRoleNames() throws UserStoreException {
        String[] names = new String[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_ROLES_ALL));
            rs = prepStmt.executeQuery();
            List lst = new LinkedList();
            String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME);
            while (rs.next()) {
                lst.add(rs.getString(colName));
            }
            if (lst.size() > 0) {
                names = (String[]) lst.toArray(new String[lst.size()]);
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return names;
    }

    public String[] getAllUserNames() throws UserStoreException {
        String[] names = new String[0];
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        ResultSet rs = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_USERS_ALL));
            rs = prepStmt.executeQuery();
            List lst = new LinkedList();
            String colName = DefaultRealmConstants.COLUMN_NAME_USER_NAME;
            while (rs.next()) {
                String name = rs.getString(colName);
                if(!name.equals(systemUserName)){
                    lst.add(name);
                }
            }
            if (lst.size() > 0) {
                names = (String[]) lst.toArray(new String[lst.size()]);
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return names;
    }

    public Map getRoleProperties(String roleName) throws UserStoreException {
        ResultSet rs = null;
        Map props = new HashMap();
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_ROLE_ATTRIBUTES));
            prepStmt.setString(1, roleName);
            rs = prepStmt.executeQuery();
            String colAttrName = DefaultRealmConstants.COLUMN_NAME_ATTR_NAME;
            String colAttrValue = DefaultRealmConstants.COLUMN_NAME_ATTR_VALUE;
            while (rs.next()) {
                String name = rs.getString(colAttrName);
                String value = rs.getString(colAttrValue);
                props.put(name, value);
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return props;

    }

    public String[] getUserPropertyNames() throws UserStoreException {
        String[] propNames = new String[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            prepStmt = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_ATTRIBUTE_NAMES));
            rs = prepStmt.executeQuery();
            String colName = DefaultRealmConstants.COLUMN_NAME_ATTR_NAME;
            List<String> lst = new ArrayList<String>();
            while (rs.next()) {
                lst.add(rs.getString(colName));
            }
            propNames = (String[]) lst.toArray(new String[lst.size()]);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return propNames;
    }

    public String[] getUserRoles(String userName) throws UserStoreException {
        String[] names = new String[0];
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            DefaultCommons commons = new DefaultCommons();
            names = commons.getUserRoles(sqlHelper, dbConnection, userName);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeConnection(dbConnection);
        }
        return names;
    }

    public boolean isExistingUser(String userName) throws UserStoreException {
        boolean isExisting = false;
        Connection dbConnection = null;
        // we have to acknowledge the existence of the system user here
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            DefaultCommons commons = new DefaultCommons();
            if (commons.getUserId(sqlHelper, dbConnection, userName) != -1) {
                isExisting = true;
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeConnection(dbConnection);
        }
        return isExisting;
    }

    public boolean isExistingRole(String roleName) throws UserStoreException {
        boolean isExisting = false;
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            DefaultCommons commons = new DefaultCommons();
            if (commons.getRoleId(sqlHelper, dbConnection, roleName) != -1) {
                isExisting = true;
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeConnection(dbConnection);
        }

        return isExisting;
    }

    /**
     * If propertyName is null then the value is searched If propertyName
     * contains % then a likenss querry is searched
     * 
     */
    public String[] getUserNamesWithPropertyValue(String propertyName, String propetyValue)
            throws UserStoreException {
        String[] names = new String[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);

            if (propertyName != null && propertyName.contains("%")) {
                prepStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.GET_USERS_WITH_PROPERTY_LIKE));
                prepStmt.setString(1, propertyName);
                prepStmt.setString(2, propetyValue);
            } else if (propertyName != null && propertyName.length() != 0) {
                prepStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.GET_USERS_WITH_PROPERTY));
                prepStmt.setString(1, propertyName);
                prepStmt.setString(2, propetyValue);
            } else {
                prepStmt = dbConnection.prepareStatement(sqlHelper
                        .getSQL(DefaultRealmConstants.GET_USERS_WITH_PROPERTY_VALUE));
                prepStmt.setString(1, propetyValue);
            }

            rs = prepStmt.executeQuery();
            String colName = DefaultRealmConstants.COLUMN_NAME_USER_NAME;
            List lst = new ArrayList();
            while (rs.next()) {
                lst.add(rs.getString(colName));
            }
            names = (String[]) lst.toArray(new String[lst.size()]);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return names;

    }

    public String[] getUsersInRole(String roleName) throws UserStoreException {
        String[] names = new String[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement getUserInRole = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            getUserInRole = dbConnection.prepareStatement(sqlHelper
                    .getSQL(DefaultRealmConstants.GET_USERS_IN_ROLE));
            getUserInRole.setString(1, roleName);
            rs = getUserInRole.executeQuery();
            String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_USER_NAME);
            List lst = new ArrayList();
            while (rs.next()) {
                lst.add(rs.getString(colName));
            }
            names = (String[]) lst.toArray(new String[lst.size()]);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, getUserInRole);
        }
        return names;
    }

   

    public String[] listUsers(String filter, int maxItemLimit) throws UserStoreException {
        String[] users = new String[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        Statement stmt = null;
        try {
            String sqlStmt = sqlHelper.getSQL(DefaultRealmConstants.GET_USERS_ALL);
            if (filter != null && filter.trim().length() != 0) {
                filter = filter.trim();
                filter = filter.replace("*", "%");
                filter = filter.replace("?", "_");
                sqlStmt = sqlStmt + " WHERE UM_USER_NAME LIKE '" + filter + "' ORDER BY UM_USER_NAME";
            } else {
                sqlStmt = sqlStmt + " ORDER BY UM_USER_NAME";
            }

            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            dbConnection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);

            stmt = dbConnection.createStatement();
            rs = stmt.executeQuery(sqlStmt);

            List lst = new LinkedList();
            String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_USER_NAME);

            int i = 0;
            while (rs.next()) {
                if (i < maxItemLimit) {
                    String name = rs.getString(colName);
                    if(!name.equals(systemUserName)){
                        lst.add(name);
                    }
                } else {
                    break;
                }
                i++;
            }

            if (lst.size() > 0) {
                users = (String[]) lst.toArray(new String[lst.size()]);
            }

        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            try {
                if (rs != null) {
                    rs.close();
                }
                if (stmt != null) {
                    stmt.close();
                }
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(e.getMessage(), e);
            }
        }
        return users;
    }

    public byte[] getUserBinaryContent(String userName, String contentName, String profileName)
            throws UserStoreException {
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        byte[] content = new byte[0];
        Connection dbConnection = null;
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.GET_BINARY_CONTENT);
            prepStmt = dbConnection.prepareStatement(sql);
            prepStmt.setString(1, contentName);
            prepStmt.setString(2, profileName);
            prepStmt.setString(3, userName);
            rs = prepStmt.executeQuery();
            int length = 0;
            byte[] temp = new byte[DefaultRealmConstants.MAX_CONTENT_SIZE];
            while (rs.next()) {
                Blob blob = rs.getBlob(1);
                InputStream in = blob.getBinaryStream();
                length = in.read(temp);
            }
            content = new byte[length];
            for (int i = 0; i < length; i++) {
                content[i] = temp[i];
            }
        } catch (IOException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt);
        }
        return content;
    }

    public Map<String, String> getUserProperties(String userName, String[] propertyNames,
            String profileName) throws UserStoreException {
        Map<String, String> props = new HashMap<String, String>();
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            List<String> basic = new ArrayList<String>();
            List<String> custom = new ArrayList<String>();
            for (String propName : propertyNames) {
                if (isBasic(propName)) {
                    basic.add(propName);
                } else {
                    custom.add(propName);
                }
            }

            if (basic.size() > 0) {
                props = getBasicPropertyValues(userName, basic, dbConnection, profileName);
            }

            if (custom.size() > 0) {
                Map<String, String> values2 = getCustomPropertyValues(userName, custom,
                        dbConnection, profileName);
                props.putAll(values2);
            }

        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeConnection(dbConnection);
        }
        return props;
    }

    public String getUserPropertyValue(String userName, String propertyName, String profileName)
            throws UserStoreException {
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        String value = null;
        List<String> props = new ArrayList<String>();
        props.add(propertyName);
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            if (isBasic(propertyName)) {
                Map<String, String> map = getBasicPropertyValues(userName, props, dbConnection,
                        profileName);
                value = map.get(propertyName);
            } else {
                Map<String, String> map = getCustomPropertyValues(userName, props, dbConnection,
                        profileName);
                value = map.get(propertyName);
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeConnection(dbConnection);
        }
        return value;
    }

    // ********************* Profile Methods
    // ***********************************************8
    public String[] getAllProfileNames() throws UserStoreException {
        String[] profileNames = new String[0];
        ResultSet rs = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.GET_PROFILES_IN_USER_DATA);
            prepStmt1 = dbConnection.prepareStatement(sql);
            rs = prepStmt1.executeQuery();
            Set<String> names = new HashSet<String>();
            while (rs.next()) {
                names.add(rs.getString(1));
            }
            rs.close();
            sql = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.GET_PROFILES_IN_USER_ATTRIBUTES);
            prepStmt2 = dbConnection.prepareStatement(sql);
            rs = prepStmt2.executeQuery();
            names = new HashSet<String>();
            while (rs.next()) {
                names.add(rs.getString(1));
            }
            if (names.size() == 0) {
                names.add(UserCoreConstants.DEFAULT_PROFILE);
            } else {
                if (!names.contains(UserCoreConstants.DEFAULT_PROFILE)) {
                    names.add(UserCoreConstants.DEFAULT_PROFILE);
                }
            }

            profileNames = (String[]) names.toArray(new String[names.size()]);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt1, prepStmt2);
        }
        return profileNames;
    }

    public String[] getProfileNames(String username) throws UserStoreException {
        String[] profileNames = new String[0];
        ResultSet rs = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException("Cannot obtain db connection");
            }
            dbConnection.setAutoCommit(false);
            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.GET_PROFILES_IN_USER_DATA_FOR_USER);
            prepStmt1 = dbConnection.prepareStatement(sql);
            prepStmt1.setString(1, username);
            rs = prepStmt1.executeQuery();
            Set<String> names = new HashSet<String>();
            while (rs.next()) {
                names.add(rs.getString(1));
            }
            rs.close();
            sql = sqlHelper.getUserProfileRelatedSQL(DefaultRealmConstants.GET_PROFILES_IN_USER_ATTRIBUTES_FOR_USER);
            prepStmt2 = dbConnection.prepareStatement(sql);
            prepStmt2.setString(1, username);
            rs = prepStmt2.executeQuery();
            
            while (rs.next()) {
                names.add(rs.getString(1));
            }

            if (names.size() == 0) {
                names.add(UserCoreConstants.DEFAULT_PROFILE);
            } else {
                if (!names.contains(UserCoreConstants.DEFAULT_PROFILE)) {
                    names.add(UserCoreConstants.DEFAULT_PROFILE);
                }
            }

            profileNames = (String[]) names.toArray(new String[names.size()]);
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt1, prepStmt2);
        }
        return profileNames;
    }

    public byte[] getProfilePicture(String userName, String profileName) throws UserStoreException {
        return this.getUserBinaryContent(userName, UserCoreConstants.PICTURE_CLAIM, profileName);
    }

    public String getUserClaimValue(String userName, String claim, String profileName)
            throws UserStoreException {

        String profileConfiguration = null;

        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }

        profileConfiguration = this.getUserPropertyValue(userName,
                UserCoreConstants.PROFILE_CONFIGURATION, profileName);
        String property = claimAdmin.getAttributeName(claim);
        
        if (profileConfiguration == null) {
            profileConfiguration = UserCoreConstants.DEFAULT_PROFILE_CONFIGURATION;
        }

        ProfileConfiguration profileConfig = profileAdmin.getProfileConfig(profileConfiguration);
        if (profileConfig != null) {
            List<String> hidden = getMappingAttributeList(profileConfig.getHiddenClaims());
            if (hidden.contains(property)) {
                return null;
            }
        }

        String value = this.getUserPropertyValue(userName, property, profileName);
        if ((value == null || value.trim().length() == 0)
                && !UserCoreConstants.DEFAULT_PROFILE.equals(profileName) && profileConfig != null) {
            List<String> overridden = getMappingAttributeList(profileConfig.getOverriddenClaims());
            if (!overridden.contains(property)) {
                value = this.getUserPropertyValue(userName, property,
                        UserCoreConstants.DEFAULT_PROFILE);
            }
        }
        return value;
    }

    public Map<String, String> getUserClaimValues(String userName, String[] claims,
            String profileName) throws UserStoreException {

        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }
        
        String profileConfigurationName = null;
        boolean isPofileCnfigurationRequested = false;

        Set<String> propertySet = new HashSet<String>();
        for (String claim : claims) {
            ClaimMapping mapping = null;
            mapping = claimAdmin.getClaimMapping(claim);
            // There can be cases some claim values being requested for claims
            // we don't have.
            if (mapping != null) {
                propertySet.add(mapping.getMappedAttribute());
            }

            if (UserCoreConstants.PROFILE_CONFIGURATION.equals(claim)) {
                isPofileCnfigurationRequested = true;
            }
        }
                
        propertySet.add(UserCoreConstants.PROFILE_CONFIGURATION);
        
        String[] properties = propertySet.toArray(new String[propertySet.size()]);
        Map<String, String> uerProperties = this.getUserProperties(userName, properties, profileName);
        
        profileConfigurationName = uerProperties.get(UserCoreConstants.PROFILE_CONFIGURATION);
        if (profileConfigurationName == null) {
            profileConfigurationName = UserCoreConstants.DEFAULT_PROFILE_CONFIGURATION;
        }

        ProfileConfiguration profileConfig = profileAdmin.getProfileConfig(profileConfigurationName);
        
        List<String> hidden = null;
        if (profileConfig != null) {
            hidden = getMappingAttributeList(profileConfig.getHiddenClaims());
        }

        List<String> overridden = null;
        if (profileConfig != null) {
            overridden = getMappingAttributeList(profileConfig.getOverriddenClaims());
        }

        List<String> getAgain = new ArrayList<String>();
        Map<String,String> finalValues = new HashMap<String, String>();
        
        if (isPofileCnfigurationRequested) {
            finalValues.put(UserCoreConstants.PROFILE_CONFIGURATION, profileConfigurationName);
        }
                  
        for (String claim : claims) {
            ClaimMapping mapping = null;
            mapping = claimAdmin.getClaimMapping(claim);
            if (mapping != null) {
                String property = mapping.getMappedAttribute();
                String value = uerProperties.get(property);
                if (hidden!= null && hidden.contains(property)) {
                    continue;
                }

                if (profileName.equals(UserCoreConstants.DEFAULT_PROFILE)) {
                    // Check whether we have a value for the requested attribute
                    if (value != null && value.trim().length()>0) {
                        finalValues.put(claim, value);
                    }
                } else if (profileConfig != null && value == null || value.trim().length() == 0) {
                    if (overridden != null && overridden.contains(property)) {
                        finalValues.put(claim, value);
                    } else {
                        getAgain.add(claim);
                    }
                } else {
                    if (value != null && value.trim().length()>0) {
                        finalValues.put(claim, value);
                    }
                }
            }
        }

        if (getAgain.size() > 0) {
            // oh the beautiful recursion
            Map<String, String> mapClaimValues = this.getUserClaimValues(userName,
                    (String[]) getAgain.toArray(new String[getAgain.size()]),
                    UserCoreConstants.DEFAULT_PROFILE);

            Iterator<Map.Entry<String, String>> ite3 = mapClaimValues.entrySet().iterator();
            while (ite3.hasNext()) {
                Map.Entry<String, String> entry = ite3.next();
                if (entry.getValue() != null) {
                    finalValues.put(entry.getKey(), entry.getValue());
                }
            }
        }

        return finalValues;
    }

    public Claim[] getUserClaimValues(String userName, String profileName)
            throws UserStoreException {
        if (profileName == null) {
            profileName = UserCoreConstants.DEFAULT_PROFILE;
        }

        String[] claims = claimAdmin.getAllClaimUris();
        Map<String, String> values = this.getUserClaimValues(userName, claims, profileName);
        Claim[] finalValues = new Claim[values.size()];
        int i=0;
        for (Iterator<Map.Entry<String, String>> ite = values.entrySet().iterator(); ite.hasNext();) {
            Map.Entry<String, String> entry = ite.next();
            Claim claim = new Claim();
            claim.setValue(entry.getValue());
            claim.setClaimUri(entry.getKey());
            String displayTag = claimAdmin.getClaim(entry.getKey()).getDisplayTag();
            claim.setDisplayTag(displayTag);
            finalValues[i] = claim;
            i++;
        }
        return finalValues;
    }

    // **************** Utility methods ***************************************

    protected Map<String, String> getBasicPropertyValues(String userName,
            List<String> propertyNames, Connection dbConnection, String profileName)
            throws UserStoreException {
        Map<String, String> values = new HashMap<String, String>();
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {

            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.GET_USER_DATA_VALUES);
            prepStmt = dbConnection.prepareStatement(sql);
            prepStmt.setString(1, profileName);
            prepStmt.setString(2, userName);
            rs = prepStmt.executeQuery();
            while (rs.next()) {
                Iterator<String> ite = propertyNames.iterator();
                while (ite.hasNext()) {
                    String name = ite.next();
                    String value = rs.getString(name);
                    values.put(name, value);
                }
            }

        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(null, rs, prepStmt);
        }
        return values;
    }

    protected Map<String, String> getCustomPropertyValues(String userName,
            List<String> propertyNames, Connection dbConnection, String profileName)
            throws UserStoreException {
        Map<String, String> values = new HashMap<String, String>();
        ResultSet rs = null;
        PreparedStatement prepStmt = null;
        try {
            String sql = sqlHelper
                    .getUserProfileRelatedSQL(DefaultRealmConstants.GET_USER_PROPERTY_VALUE);
            prepStmt = dbConnection.prepareStatement(sql);
            prepStmt.setString(3, userName);
            prepStmt.setString(2, profileName);
            Iterator<String> ite = propertyNames.iterator();
            while (ite.hasNext()) {
                String name = ite.next();
                prepStmt.setString(1, name);
                rs = prepStmt.executeQuery();
                String value = null;
                while (rs.next()) {
                    value = rs.getString(1);
                }
                values.put(name, value);
            }
        } catch (SQLException e) {
            log.error(e.getMessage(),e);
            throw new UserStoreException(e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(null, rs, prepStmt);
        }
        return values;
    }
    
    /**
     * 
     * @param claimList
     * @return
     * @throws UserStoreException
     */
    protected List<String> getMappingAttributeList(List<String> claimList)
            throws UserStoreException {
        ArrayList<String> attributeList = null;
        Iterator<String> claimIter = null;
        
        attributeList = new ArrayList<String>();
        if (claimList == null) {
            return attributeList;
        }
        claimIter = claimList.iterator();
        while (claimIter.hasNext()) {
            attributeList.add(claimAdmin.getAttributeName(claimIter.next()));
        }
        return attributeList;
    }

    protected boolean isBasic(String attributeId) {
        if (attributeId.equals(DefaultRealmConstants.EMAIL)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.FIRST_NAME)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.LAST_NAME)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.BIRTH_DATE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.FULL_NAME)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.NAME_PREFIX)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.GENDER)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.TIME_ZONE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.COMPANY_NAME)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.JOB_TITLE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.PRIMARY_PHONE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.HOME_PHONE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.WORK_PHONE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.MOBILE_PHONE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.STREET_ADDRESS)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.CITY)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.STATE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.COUNTRY)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.POSTAL_CODE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.WEB_PAGE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.LANGUAGE)) {
            return true;
        } else if (attributeId.equals(DefaultRealmConstants.BLOG)) {
            return true;
        }
        return false;
    }

}
