/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.delegating;

import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.internal.UserCoreServiceComponent;

/**
 * Reads data from the user store
 */
public class DelegatingAuthorizer implements Authorizer {

    protected UserRealm defaultRealm = null;

    public DelegatingAuthorizer(UserRealm defaultRealm) {
        this.defaultRealm = defaultRealm;
    }

    public String[] getAllowedRolesForResource(String resourceId, String action)
            throws UserStoreException {
        String[] roles = defaultRealm.getAuthorizer()
                .getAllowedRolesForResource(resourceId, action);
        UserRealm hybrealm = UserCoreServiceComponent.getHybridRealm();
        if (hybrealm != null) {
            String[] hybRoles = hybrealm.getAuthorizer().getAllowedRolesForResource(resourceId,
                    action);
            return getResultantRoles(roles, hybRoles);
        }
        return roles;
    }

    public boolean isUserAuthorized(String userName, String resourceId, String action)
            throws UserStoreException {
        boolean isAuthed = false;
        if (defaultRealm.getUserStoreReader().isExistingUser(userName)) {
            isAuthed = defaultRealm.getAuthorizer().isUserAuthorized(userName, resourceId, action);
        } else {
            UserRealm hybrealm = UserCoreServiceComponent.getHybridRealm();
            if (hybrealm != null) {
                if (hybrealm.getUserStoreReader().isExistingUser(userName)) {
                    isAuthed = hybrealm.getAuthorizer().isUserAuthorized(userName, resourceId,
                            action);
                }
            }
        }
        return isAuthed;
    }

    public String[] getAllowedUsersForResource(String resourceId, String action)
            throws UserStoreException {
        String[] roles = defaultRealm.getAuthorizer()
                .getAllowedUsersForResource(resourceId, action);
        UserRealm hybrealm = UserCoreServiceComponent.getHybridRealm();
        if (hybrealm != null) {
            String[] hybRoles = hybrealm.getAuthorizer().getAllowedUsersForResource(resourceId,
                    action);
            return getResultantRoles(roles, hybRoles);
        }
        return roles;
    }

    public String[] getDeniedRolesForResource(String resourceId, String action)
            throws UserStoreException {
        String[] roles = defaultRealm.getAuthorizer().getDeniedRolesForResource(resourceId, action);
        UserRealm hybrealm = UserCoreServiceComponent.getHybridRealm();
        if (hybrealm != null) {
            String[] hybRoles = hybrealm.getAuthorizer().getDeniedRolesForResource(resourceId,
                    action);
            return getResultantRoles(roles, hybRoles);
        }
        return roles;

    }

    public String[] getDeniedUsersForResource(String resourceId, String action)
            throws UserStoreException {
        String[] roles = defaultRealm.getAuthorizer().getDeniedUsersForResource(resourceId, action);
        UserRealm hybrealm = UserCoreServiceComponent.getHybridRealm();
        if (hybrealm != null) {
            String[] hybRoles = hybrealm.getAuthorizer().getDeniedUsersForResource(resourceId,
                    action);
            return getResultantRoles(roles, hybRoles);
        }
        return roles;

    }

    public boolean isRoleAuthorized(String roleName, String resourceId, String action)
            throws UserStoreException {
        boolean isAuthed = false;
        if (defaultRealm.getUserStoreReader().isExistingRole(roleName)) {
            isAuthed = defaultRealm.getAuthorizer().isRoleAuthorized(roleName, resourceId, action);
        } else {
            UserRealm hybrealm = UserCoreServiceComponent.getHybridRealm();
            if (hybrealm != null) {
                if (hybrealm.getUserStoreReader().isExistingRole(roleName)) {
                    isAuthed = hybrealm.getAuthorizer().isRoleAuthorized(roleName, resourceId,
                            action);
                }
            }
        }
        return isAuthed;
    }

    private String[] getResultantRoles(String[] defaultRoles, String[] hybridRoles) {
        if (defaultRoles == null || defaultRoles.length == 0) {
            return hybridRoles;
        }
        if (hybridRoles != null && hybridRoles.length > 0) {
            String[] combinedRoles = new String[defaultRoles.length + hybridRoles.length];
            int j = 0;
            for (int i = 0; i < defaultRoles.length; i++) {
                combinedRoles[j++] = defaultRoles[i];
            }
            for (int i = 0; i < hybridRoles.length; i++) {
                combinedRoles[j++] = hybridRoles[i];
            }
            return combinedRoles;
        }
        return new String[0];
    }

}
