package org.wso2.carbon.user.mgt.service;

import java.sql.Connection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.naming.directory.DirContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.user.core.AccessControlAdmin;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserClaimsAdmin;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreAdmin;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.hybrid.HybridInternalRoleAdmin;
import org.wso2.carbon.user.core.hybrid.HybridRealm;
import org.wso2.carbon.user.core.hybrid.actdir.ActDirRealmConfig;
import org.wso2.carbon.user.core.hybrid.jdbc.JDBCRealmConfig;
import org.wso2.carbon.user.core.hybrid.ldap.LDAPRealmConfig;
import org.wso2.carbon.user.mgt.UserMgtConstants;
import org.wso2.carbon.user.mgt.UserMgtContext;

public class UserMgtServiceImpl {

    private static Log log = LogFactory.getLog(UserStoreAdmin.class);

    public UserMgtServiceImpl() {

    }

    public String[] getInternalUsers() throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            return realm.getUserStoreReader().getAllUserNames();
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] listInternalUsers(String filter) throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            return realm.getUserStoreReader()
                    .listUsers(filter, UserMgtConstants.MAX_USER_COUNT + 1);
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] getInternalRoles() throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            return realm.getUserStoreReader().getAllRoleNames();
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void deleteUserFromInternalStore(String userName) throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            realm.getUserStoreAdmin().deleteUser(userName);            
            String path = RegistryConstants.PROFILES_PATH + userName;
            Registry registry = UserMgtContext.getRegistry();
            if(registry.resourceExists(path)){
                registry.delete(path);
            }
        } catch (RegistryException e) {
            String msg =  "Error deleting user from registry, "+e.getMessage() ;
            log.error(msg, e);
            throw new UserMgtException(msg, e);
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void addUserToInternalStore(String userName, String password, String[] roles)
            throws UserMgtException {
        try {
            checkName(userName);
            UserRealm realm = UserMgtContext.getDefaultRealm();
            UserStoreAdmin admin = realm.getUserStoreAdmin();
            admin.addUser(userName, password);
            for (String roleName : roles) {
                admin.addUserToRole(userName, roleName);
            }
            admin.addUserToRole(userName, "everyone");
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void changePassword(String userName, String newPassword) throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            realm.getUserStoreAdmin().updateUser(userName, newPassword);
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void addRoleToInternalStore(String roleName, String[] userNames, String[] permissions)
            throws UserMgtException {
        try {
            Arrays.sort(userNames);
            if (Arrays.binarySearch(userNames, "guest") > -1) {
                log.error("User is trying to manupulate guest user. Possible security threat.");
                throw new UserMgtException("Invaid data"); //possible security attack
            }
            checkName(roleName);
            UserRealm realm = UserMgtContext.getDefaultRealm();
            UserStoreAdmin usAdmin = realm.getUserStoreAdmin();
            usAdmin.addRole(roleName);
            for (String userName : userNames) {
                usAdmin.addUserToRole(userName, roleName);
            }
            AccessControlAdmin acAdmin = realm.getAccessControlAdmin();
            this.addSystemPermissionsToRole(roleName, permissions);

        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void deleteRoleFromInternalStore(String roleName) throws UserMgtException {
        try {
            if(roleName.equals("everyone") || roleName.equals("admin")){
                throw new UserMgtException("Deleting everyone role is not allowed.");
            }
            UserRealm defRealm = UserMgtContext.getDefaultRealm();
            defRealm.getUserStoreAdmin().deleteRole(roleName);
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void updateUsersOfRole(String roleName, String[] usersShown, String[] usersSelected) throws UserMgtException {
        try {
            if("everyone".equals(roleName)){
                log.error("Removing users from everyone role is not allowed. A possible security threat");
                throw new UserMgtException(
                "Removing users from everyone role is not allowed.");
            }
            
            if ("guests".equals(roleName)) {
                log.error("User is trying to edit guests role. Possible security threat.");
                throw new UserMgtException("Invalid data"); // obscure error message
            }
            
            Arrays.sort(usersShown);
            Arrays.sort(usersSelected);
            if (roleName.equals("admin") && (Arrays.binarySearch(usersShown, "admin")>-1)) {
                log.error("User is trying to edit admin role. Possible security threat.");
                if (Arrays.binarySearch(usersSelected, "admin") < 0) {
                    throw new UserMgtException(
                            "Cannot perform update. Admin user cannot be removed from admin role.");
                }
            }
            
            if (Arrays.binarySearch(usersShown, "guest")>-1) {
                log.error("User is trying to add guest user. Possible security threat.");
                throw new UserMgtException("Invalid data"); // obscure error message
            }

            UserRealm defRealm = UserMgtContext.getDefaultRealm();
            UserStoreAdmin admin = defRealm.getUserStoreAdmin();
            String[] oldUserList = admin.getUsersInRole(roleName);
            Arrays.sort(oldUserList);
            
            List<String> delUsers = new ArrayList<String>();
            List<String> addUsers = new ArrayList<String>();
            
            for (String userName : usersShown) {
                int oldindex = Arrays.binarySearch(oldUserList, userName);
                int selindex = Arrays.binarySearch(usersSelected, userName);
                if(oldindex > -1 && selindex < 0){
                    //deleted
                    delUsers.add(userName);
                }else if(oldindex<0 && selindex > -1){
                    //added
                    addUsers.add(userName);
                }
            }

            Iterator<String> ite = delUsers.iterator();
            while (ite.hasNext()) {
                admin.removeUserFromRole(ite.next(), roleName);
            }

            ite = addUsers.iterator();
            while (ite.hasNext()) {
                admin.addUserToRole(ite.next(), roleName);
            }
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] getUsersInRole(String roleName) throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            return realm.getUserStoreAdmin().getUsersInRole(roleName);
        } catch (Exception e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void addRolesToResource(String[] roleNames, String resource, String action)
            throws UserMgtException {

    }

    public void updateRolesofResource(String[] roleNames, String resource, String action)
            throws UserMgtException {

    }

    public String[] getSystemPermissions() throws UserMgtException {
        return UserMgtContext.getSystemPermissionsAsArray();
    }

    public void addSystemPermissionsToRole(String roleName, String[] actions)
            throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getDefaultRealm();
            AccessControlAdmin acAdmin = realm.getAccessControlAdmin();
            for (String action : actions) {
                acAdmin.authorizeRole(roleName, UserMgtConstants.SYSTEM_RESOURCE, action);
            }
        } catch (Exception e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] getSystemPermissionsOfRole(String roleName) throws UserMgtException {
        return getSystemPermissionsOfRole(UserMgtContext.getDefaultRealm(), roleName);
    }

    public void updateSystemPermissionsOfRole(String roleName, String[] actions)
            throws UserMgtException {

        if (roleName.equals("admin")) {
            throw new UserMgtException("Cannot perform update. Admin role cannot be updated");
        }
        if (roleName.equals("admin")) {
            throw new UserMgtException("Cannot perform update. Admin role cannot be updated");
        }

        List<String> sysPermissions = UserMgtContext.getSystemPermissions();
        List<String> authed = Arrays.asList(actions);
        List<String> delList = new ArrayList<String>();
        try {
            Iterator<String> ite = sysPermissions.iterator();
            UserRealm realm = UserMgtContext.getDefaultRealm();
            AccessControlAdmin admin = realm.getAccessControlAdmin();
            while (ite.hasNext()) {
                String action = ite.next();
                boolean isContained = authed.contains(action);
                boolean isAuthed = admin.isRoleAuthorized(roleName,
                        UserMgtConstants.SYSTEM_RESOURCE, action);
                if (isContained == false && isAuthed == true) {
                    delList.add(action);
                } else if (isContained == true && isAuthed == false) {
                    admin.authorizeRole(roleName, UserMgtConstants.SYSTEM_RESOURCE, action);
                }
            }
            ite = delList.iterator();
            while (ite.hasNext()) {
                String action = ite.next();
                admin.clearRoleAuthorization(roleName, UserMgtConstants.SYSTEM_RESOURCE, action);
            }

        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void updateUserName(String userName, String newUserName) throws UserMgtException {
        throw new UnsupportedOperationException("Changing username is not allowed.");
    }

    public void updateRoleName(String roleName, String newRoleName) throws UserMgtException {
        throw new UnsupportedOperationException("Changing role name is not allowed.");
    }

    // ***************************** External User Store
    // ******************************* //

    public String[] getExternalUsers() throws UserMgtException {
        String[] users = new String[0];
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            if (realm != null) {
                users = realm.getUserStoreReader().getAllUserNames();
            }
            return users;
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] listExternalUsers(String filter) throws UserMgtException {
        String[] users = new String[0];
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            if (realm != null) {
                users = realm.getUserStoreReader().listUsers(filter,
                        UserMgtConstants.MAX_USER_COUNT + 1);
            }
            return users;
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] getExternalGroups() throws UserMgtException {
        String[] groups = new String[0];
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();

            if (realm != null) {
                String[] allGroups = realm.getUserStoreReader().getAllRoleNames();

                HybridRealm hybrealm = ((HybridRealm) realm);
                HybridInternalRoleAdmin roleAdmin = hybrealm.getInternalRoleAdmin();
                String[] internalGroups = roleAdmin.getInternalRoles();
                Arrays.sort(internalGroups);

                List<String> values = new ArrayList<String>();

                for (String groupName : allGroups) {
                    if (Arrays.binarySearch(internalGroups, groupName) < 0) {
                        values.add(groupName);
                    }
                }

                groups = values.toArray(new String[values.size()]);

            }
            return groups;
        } catch (UserStoreException e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void addExternalUserStoreConfig(String realmType, StoreProperty[] configs)
            throws UserMgtException {
        try {
            Registry registry = UserMgtContext.getRegistry();
            Resource resource = registry.newResource();
            String roleList = null;
            String userRoles = null;
            String usersInRole = null;
            String attrsSQL = null;
            String profNames = null;
            String profAttrs = null;
            for (StoreProperty prop : configs) {
                String name = prop.getName();
                if (name.equals("RoleListSQL")) {
                    roleList = prop.getValue();
                } else if (name.equals("UsersInRoleSQL")) {
                    usersInRole = prop.getValue();
                } else if (name.equals("UserRoleSQL")) {
                    userRoles = prop.getValue();
                } else if (name.equals("AttrsSQL")) {
                    attrsSQL = prop.getValue();
                } else if (name.equals("ProfileNamesSQL")) {
                    profNames = prop.getValue();
                } else if (name.equals("AttrsForProfileSQL")) {
                    profAttrs = prop.getValue();
                } else {
                    resource.addProperty(prop.getName(), prop.getValue());
                }
            }

            if (roleList != null && userRoles != null && usersInRole != null) {
                roleList = roleList.trim();
                userRoles = userRoles.trim();
                usersInRole = usersInRole.trim();

                if (roleList.length() > 0 && userRoles.length() > 0 && usersInRole.length() > 0) {
                    resource.addProperty("RoleListSQL", roleList);
                    resource.addProperty("UserRoleSQL", userRoles);
                    resource.addProperty("UsersInRoleSQL", usersInRole);
                }
            }

            if (attrsSQL != null) {
                attrsSQL = attrsSQL.trim();
                if (attrsSQL.length() > 0) {
                    resource.addProperty("AttrsSQL", attrsSQL);
                }
            }
            
            if(profNames != null && profAttrs != null){
                profNames=profNames.trim();
                profAttrs=profAttrs.trim();
                if (profNames.length()>0 && profAttrs.length()>0) {
                    resource.addProperty("ProfileNamesSQL", profNames);
                    resource.addProperty("AttrsForProfileSQL", profAttrs);
                }
            }

            resource.addProperty(UserMgtConstants.REALM_TYPE, realmType);
            registry.put(UserMgtConstants.EXTERNAL_USERSTORE, resource);
            UserMgtContext.registerHybridRealmService();

        } catch (RegistryException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        } catch (Exception e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void updateExternalUserStoreConfig(StoreProperty[] props) throws UserMgtException {
        try {
            Registry registry = UserMgtContext.getRegistry();
            Resource resource = registry.get(UserMgtConstants.EXTERNAL_USERSTORE);
            Properties oldprops = resource.getProperties();
            Set values = oldprops.entrySet();
            for (StoreProperty prop : props) {
                String name = prop.getName();
                if ("UIDSearchName".equals(name)) {
                    continue;
                }
                String value = prop.getValue();
                if (!UserMgtConstants.REALM_TYPE.equals(name)) {
                    resource.setProperty(name, value);
                }
            }
            registry.put(UserMgtConstants.EXTERNAL_USERSTORE, resource);
            UserMgtContext.unregisterHybridRealmService();
            UserMgtContext.registerHybridRealmService();
        } catch (Throwable e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public StoreProperty[] getExternalUserStoreConfig() throws UserMgtException {

        try {
            Registry registry = UserMgtContext.getRegistry();
            Resource resource = registry.get(UserMgtConstants.EXTERNAL_USERSTORE);
            Properties props = resource.getProperties();
            Set values = props.entrySet();
            StoreProperty[] storeProps = new StoreProperty[values.size()];
            Iterator ite = values.iterator();
            int i = 0;
            while (ite.hasNext()) {
                Map.Entry entry = (Map.Entry) ite.next();
                String name = (String) entry.getKey();
                String value = (String) ((List) entry.getValue()).get(0);
                StoreProperty storeProp = new StoreProperty();
                storeProp.setName(name);
                storeProp.setValue(value);
                storeProps[i] = storeProp;
                i++;
            }
            return storeProps;
        } catch (Throwable e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public boolean hasExternalUserStore() throws UserMgtException {
        boolean isExist = false;
        try {
            Registry registry = UserMgtContext.getRegistry();
            if (registry.resourceExists(UserMgtConstants.EXTERNAL_USERSTORE)) {
                isExist = true;
            }
        } catch (RegistryException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
        return isExist;
    }

    public void deleteExternalUserStore() throws UserMgtException {
        try {
            Registry registry = UserMgtContext.getRegistry();
            registry.delete(UserMgtConstants.EXTERNAL_USERSTORE);
            UserRealm realm = UserMgtContext.getHybridRealm();
            realm.cleanUp();
            UserMgtContext.unregisterHybridRealmService();
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        } catch (RegistryException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public boolean testConnection() throws UserMgtException {
        boolean isOk = false;
        Registry registry = UserMgtContext.getRegistry();

        if(log.isDebugEnabled()){
            log.debug("Testing connection started");
        }
        
        try {
            if (!registry.resourceExists(UserMgtConstants.EXTERNAL_USERSTORE)) {
                return false;
            }
            
            Resource resource = registry.get(UserMgtConstants.EXTERNAL_USERSTORE);

            String realmName = resource.getProperty("RealmType");

            if (realmName.equals("jdbc")) {
                JDBCRealmConfig config = new JDBCRealmConfig();
                config.setConnectionPassword(resource.getProperty("ConnectionPassword"));
                config.setConnectionName(resource.getProperty("ConnectionName"));
                config.setConnectionURL(resource.getProperty("ConnectionURL"));
                config.setDriverName(resource.getProperty("DriverName"));
                Connection connection = config.getDataSource().getConnection();
                if (connection != null) {
                    isOk = true;
                }
            } else if (realmName.equals("ldap")) {
                LDAPRealmConfig config = new LDAPRealmConfig();
                config.setConnectionPassword(resource.getProperty("ConnectionPassword"));
                config.setConnectionName(resource.getProperty("ConnectionName"));
                config.setConnectionURL(resource.getProperty("ConnectionURL"));
                config.setUserContextName(resource.getProperty("UserContextName"));
                config.setUidSearchName(resource.getProperty("UIDSearchName"));
                DirContext context = config.getContext();
                if (context != null) {
                    isOk = true;
                }
            } else if (realmName.equals("act-dir")) {
                ActDirRealmConfig config = new ActDirRealmConfig();
                config.setConnectionPassword(resource.getProperty("ConnectionPassword"));
                config.setConnectionName(resource.getProperty("ConnectionName"));
                config.setConnectionURL(resource.getProperty("ConnectionURL"));
                config.setSearchBase(resource.getProperty("SearchBase"));
                DirContext context = config.getContext();
                if (context != null) {
                    isOk = true;
                }
            }
        } catch (Exception e) {
            log.error(e.getMessage(),e);
        } 
        
        if(log.isDebugEnabled()){
            log.debug("Testing connection ended. The result is :: " + isOk);
        }
        return isOk;

    }

    public String[] getUsersInExternalRole(String roleName) throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            return realm.getUserStoreAdmin().getUsersInRole(roleName);
        } catch (Exception e) {
            // previously logged so logging not needed
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void updateSystemPermissionsOfExternalRole(String roleName, String[] permissions)
            throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            AccessControlAdmin acAdmin = realm.getAccessControlAdmin();
            List<String> sysPermissions = UserMgtContext.getSystemPermissions();

            Iterator<String> ite = sysPermissions.iterator();
            while (ite.hasNext()) {
                String temp = ite.next();
                boolean isAuthed = acAdmin.isRoleAuthorized(roleName,
                        UserMgtConstants.SYSTEM_RESOURCE, temp);
                if (isAuthed) {
                    acAdmin
                            .clearRoleAuthorization(roleName, UserMgtConstants.SYSTEM_RESOURCE,
                                    temp);
                }
            }

            for (String action : permissions) {
                acAdmin.authorizeRole(roleName, UserMgtConstants.SYSTEM_RESOURCE, action);
            }
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] getSystemPermissionsOfExternalRole(String roleName) throws UserMgtException {
        return getSystemPermissionsOfRole(UserMgtContext.getHybridRealm(), roleName);
    }

    // ******************************* Roles with External users

    public void addSpecialInternalRole(String roleName, String[] userNames, String[] permissions)
            throws UserMgtException {
        try {

            UserRealm realm = UserMgtContext.getHybridRealm();

            HybridRealm hybrealm = ((HybridRealm) realm);
            HybridInternalRoleAdmin roleAdmin = hybrealm.getInternalRoleAdmin();
            roleAdmin.addInternalRole(roleName);

            for (String userName : userNames) {
                roleAdmin.addUserToInternalRole(userName, roleName);
            }

            AccessControlAdmin acAdmin = realm.getAccessControlAdmin();
            for (String action : permissions) {
                acAdmin.authorizeRole(roleName, UserMgtConstants.SYSTEM_RESOURCE, action);
            }

        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }

    }

    public void deleteSpecialInternalRole(String roleName) throws UserMgtException {
        try {

            UserRealm realm = UserMgtContext.getHybridRealm();

            HybridRealm hybrealm = ((HybridRealm) realm);
            HybridInternalRoleAdmin roleAdmin = hybrealm.getInternalRoleAdmin();
            roleAdmin.deleteInternalRole(roleName);
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }

    }

    public String[] getSpecialInternalRoles() throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            if (realm != null) {
                HybridRealm hybrealm = ((HybridRealm) realm);
                HybridInternalRoleAdmin roleAdmin = hybrealm.getInternalRoleAdmin();
                return roleAdmin.getInternalRoles();
            }
            return new String[0];
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void updateSystemPermissionsOfSpecialInternalRole(String roleName, String[] permissions)
            throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            AccessControlAdmin acAdmin = realm.getAccessControlAdmin();
            List<String> sysPermissions = UserMgtContext.getSystemPermissions();

            Iterator<String> ite = sysPermissions.iterator();
            while (ite.hasNext()) {
                String temp = ite.next();
                boolean isAuthed = acAdmin.isRoleAuthorized(roleName,
                        UserMgtConstants.SYSTEM_RESOURCE, temp);
                if (isAuthed) {
                    acAdmin
                            .clearRoleAuthorization(roleName, UserMgtConstants.SYSTEM_RESOURCE,
                                    temp);
                }
            }

            for (String action : permissions) {
                acAdmin.authorizeRole(roleName, UserMgtConstants.SYSTEM_RESOURCE, action);
            }
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public void updateUsersOfSpecialInternalRole(String roleName, String[] usersShown, String[] usersSelected)
            throws UserMgtException {
        try {
            
            if(roleName.equals("everyone")){
                throw new UserMgtException(
                "Removing users from everyone role is not allowed.");
            }
            
            Arrays.sort(usersShown);
            Arrays.sort(usersSelected);
            UserRealm realm = UserMgtContext.getHybridRealm();
            HybridRealm hybrealm = ((HybridRealm) realm);
            HybridInternalRoleAdmin roleAdmin = hybrealm.getInternalRoleAdmin();

            String[] oldUserList = roleAdmin.getUsersInInternalRole(roleName);
            Arrays.sort(oldUserList);
            
            List<String> delUsers = new ArrayList<String>();
            List<String> addUsers = new ArrayList<String>();
            
            for (String userName : usersShown) {
                int oldindex = Arrays.binarySearch(oldUserList, userName);
                int selindex = Arrays.binarySearch(usersSelected, userName);
                if(oldindex > -1 && selindex < 0){
                    //deleted
                    delUsers.add(userName);
                }else if(oldindex<0 && selindex > -1){
                    //added
                    addUsers.add(userName);
                }
            }

            Iterator<String> ite = delUsers.iterator();
            while (ite.hasNext()) {
                roleAdmin.removeUserFromInternalRole(ite.next(), roleName);
            }

            ite = addUsers.iterator();
            while (ite.hasNext()) {
                roleAdmin.addUserToInternalRole(ite.next(), roleName);
            }

        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }

    }

    public String[] getUsersOfSpecialInternalRole(String roleName) throws UserMgtException {
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();

            HybridRealm hybrealm = ((HybridRealm) realm);
            HybridInternalRoleAdmin roleAdmin = hybrealm.getInternalRoleAdmin();
            return roleAdmin.getUsersInInternalRole(roleName);
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    public String[] getSystemPermissionsOfSpecialInternalRole(String roleName)
            throws UserMgtException {
        return getSystemPermissionsOfRole(UserMgtContext.getHybridRealm(), roleName);
    }

    // ***************** general *********************

    public String[] getAllRolesNames() throws UserMgtException {
        String[] allGroups = new String[0];
        String[] inGroups = this.getInternalRoles();
        allGroups = inGroups;
        try {
            UserRealm realm = UserMgtContext.getHybridRealm();
            if (realm != null) {
                String[] exGroups = realm.getUserStoreReader().getAllRoleNames();
                List<String> users = new ArrayList<String>();
                if (exGroups.length > 0) {
                    users.addAll(Arrays.asList(exGroups));
                    users.addAll(Arrays.asList(inGroups));
                    allGroups = users.toArray(new String[users.size()]);
                }
            }
        } catch (UserStoreException e) {
            // already logged - no logging
            throw new UserMgtException(e.getMessage(), e);
        }
        return allGroups;
    }
    
    public String[] getUserRoles(String userName) throws UserMgtException {
        UserRealm realm = UserMgtContext.getDefaultRealm();
        String[] userRoles = new String[0];
        try {
            if(realm.getUserStoreAdmin().isExistingUser(userName)){
                userRoles = realm.getUserStoreAdmin().getUserRoles(userName);
            }else {
                realm = UserMgtContext.getHybridRealm();
                if(realm != null){
                    if(realm.getUserStoreReader().isExistingUser(userName)){
                        userRoles = realm.getUserStoreReader().getUserRoles(userName);
                    }
                }
            }
            return userRoles;
        } catch (UserStoreException e) {
            // already logged - no logging
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    private String[] getSystemPermissionsOfRole(UserRealm realm, String roleName)
            throws UserMgtException {
        List<String> sysPermissions = UserMgtContext.getSystemPermissions();
        List<String> authed = new ArrayList<String>();
        try {
            Iterator<String> ite = sysPermissions.iterator();
            Authorizer authorizer = realm.getAuthorizer();
            while (ite.hasNext()) {
                String value = ite.next();
                if (authorizer.isRoleAuthorized(roleName, UserMgtConstants.SYSTEM_RESOURCE, value)) {
                    authed.add(value);
                }
            }
            return (String[]) authed.toArray(new String[authed.size()]);
        } catch (UserStoreException e) {
            log.error(e);
            throw new UserMgtException(e.getMessage(), e);
        }
    }

    private UserClaimsAdmin getClaimsAdmin(UserRealm realm) throws UserStoreException {
        //TODO ......... dimuthul
        return realm.getClaimsAdmin("Internal");
    }
    
    private void checkName(String name) throws UserMgtException{
        Pattern p = Pattern.compile(CarbonConstants.NAME_REGULAR_EXPRESSION);
        Matcher m = p.matcher(name);
        if (!m.matches()) {
            String msg = "Invalid name entered. Please make sure the enetered" +
                " name deos not contain special characters and it's length is between 3 and 30";
            log.debug(msg);
            throw new UserMgtException(msg);
        }
    }

}
