package org.wso2.carbon.utils.security;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.utils.WSO2Constants;

import java.util.Properties;

/**
 * Create and initialize PasswordProvider
 * Retrieve password from PasswordProvider
 */
public class PasswordManager {

    private static Log log = LogFactory.getLog(PasswordManager.class);
    private static PasswordManager ourInstance = new PasswordManager();

    private boolean initialized = false;
    private PasswordProvider passwordProvider;

    public static PasswordManager getInstance() {
        return ourInstance;
    }

    private PasswordManager() {
    }

    /**
     * initialize  SimplePasswordManager with given properties
     *
     * @param properties configuration properties
     */
    public void init(Properties properties) {

        if (initialized) {
            if (log.isDebugEnabled()) {
                log.debug("SimplePasswordManager already has been started.");
            }
            return;
        }

        String provider = getProperty(properties, WSO2Constants.PASSWORD_PROVIDER, null);

        if (provider != null && !"".equals(provider)) {

            try {
                Class aClass = Thread.currentThread().getContextClassLoader().loadClass(provider);
                Object object = aClass.newInstance();
                if (!(object instanceof PasswordProvider)) {
                    handleException("Invalid class as PasswordProvider : Class Name : " +
                            provider);
                } else {
                    passwordProvider = (PasswordProvider) object;
                }
                passwordProvider.init(properties);

            } catch (ClassNotFoundException e) {
                handleException("A PasswordProvider cannot be found for class name : " +
                        provider, e);
            } catch (IllegalAccessException e) {
                handleException("Error creating a instance from class : " + provider, e);
            } catch (InstantiationException e) {
                handleException("Error creating a instance from class : " + provider, e);
            }
            initialized = true;
        } else {
            log.warn("isPasswordEncrypted has been set to true. But there is " +
                    "no a PasswordProvider has been configured.");
        }

    }

    /**
     * Resolved given password using an instance of a PasswordProvider
     *
     * @param encryptedPassword Encrypted password
     * @return resolved password
     */
    public String resolve(String encryptedPassword) {

        if (encryptedPassword == null || "".equals(encryptedPassword)) {
            if (log.isDebugEnabled()) {
                log.debug("Given Encrypted Password is empty or null. Returning itself");
            }
            return encryptedPassword;
        }
        if (!initialized || passwordProvider == null) {
            if (log.isDebugEnabled()) {
                log.debug("SimplePasswordManager has not been initialized. Returning itself");
            }
            return encryptedPassword;
        }

        return passwordProvider.resolve(encryptedPassword);
    }

    public boolean isInitialized() {
        return initialized;
    }

    /**
     * Helper methods for handle errors.
     *
     * @param msg The error message
     * @param e   Thorwen Exception
     */
    private static void handleException(String msg, Exception e) {
        log.error(msg, e);
        throw new RuntimeException(msg, e);
    }

    /**
     * Helper methods for handle errors.
     *
     * @param msg The error message
     */
    private static void handleException(String msg) {
        log.error(msg);
        throw new RuntimeException(msg);
    }

    /**
     * Helper method to get the value of the property from a given property bag
     *
     * @param properties   The property collection
     * @param name         The name of the property
     * @param defaultValue The default value for the property
     * @return The value of the property if it is found , otherwise , default value
     */
    private static String getProperty(Properties properties, String name, String defaultValue) {

        String result = properties.getProperty(name);
        if ((result == null || result.length() == 0 || "".equals(result)) && defaultValue != null) {
            if (log.isDebugEnabled()) {
                log.debug("The name with ' " + name + " ' cannot be found. " +
                        "Using default value " + defaultValue);
            }
            result = defaultValue;
        }
        if (result != null) {
            return result.trim();
        } else {
            return defaultValue;
        }
    }

}
