/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.utils.transport;

import org.apache.axis2.AxisFault;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.SessionContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.transport.TransportListener;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.utils.NetworkUtils;
import org.wso2.carbon.utils.ServerConstants;
import org.wso2.carbon.utils.SessionContextUtil;

import java.lang.reflect.Method;
import java.net.SocketException;

/**
 * This class will provide basic functionality to write http and https listeners for WSO2 projects.
 */
public abstract class AbstractTransportListener implements TransportListener {
    private static final Log log = LogFactory.getLog(AbstractTransportListener.class);

    private static final String TRANSPORT_MANAGER =
            "org.wso2.carbon.server.transports.TransportManager";
    private Class transportManagerClass;
    private Object transportManager;
    private String transport;

    protected ConfigurationContext configurationContext;
    protected int port = -1;
    protected int proxyPort = -1;

    private boolean isStopped;

    public void init(ConfigurationContext configContext,
                     TransportInDescription transportIn) throws AxisFault {
        this.configurationContext = configContext;

        //Set the port
        if (System.getProperty(ServerConstants.STANDALONE_MODE, "false").equalsIgnoreCase("true")) {
            transport = transportIn.getName();
            try {
                transportManagerClass = Class.forName(TRANSPORT_MANAGER);
                transportManager = transportManagerClass.newInstance();
                Method method = transportManagerClass.getMethod("getPort", String.class);
                port = (Integer) method.invoke(transportManager, transport);

                Parameter portParam = new Parameter("port", String.valueOf(port));
                transportIn.addParameter(portParam);
            } catch (Exception e) {
                String msg = "Cannot get transport port";
                log.error(msg, e);
                throw new AxisFault(msg, e);
            }
        } else {
            Parameter portParam = transportIn.getParameter("port");
            if (portParam != null) {
                port = Integer.parseInt(portParam.getValue().toString().trim());
            }
        }
        String portSysProp = System.getProperty(transport + "Port");
        if (portSysProp != null) {
            try {
                port = Integer.parseInt(portSysProp);
                if (log.isDebugEnabled()) {
                    log.debug("Using " + transport + " port " + port +" defined in System property " +
                             (transport + "Port"));
                }
            } catch (NumberFormatException ignored) {
            }
        }

        // Set the proxy port
        if (System.getProperty(ServerConstants.STANDALONE_MODE, "false").equalsIgnoreCase("true")) {
            transport = transportIn.getName();
            try {
                Method method = transportManagerClass.getMethod("getProxyPort", String.class);
                proxyPort = (Integer) method.invoke(transportManager, transport);

                Parameter proxyPortParam = new Parameter("proxyPort", String.valueOf(proxyPort));
                transportIn.addParameter(proxyPortParam);
            } catch (Exception e) {
                String msg = "Cannot get transport proxy port";
                log.error(msg, e);
                throw new AxisFault(msg, e);
            }
        } else {
            Parameter portParam = transportIn.getParameter("proxyPort");
            if (portParam != null) {
                proxyPort = Integer.parseInt(portParam.getValue().toString().trim());
            }
        }
    }

    public synchronized void start() throws AxisFault {
        // On first time startup, org.wso2.carbon.core.internal.StartupFinalizerServiceComponent
        // will start the transports when intialization is completed. However, when the transport is
        // stopped & started using the TransportManagement component, this code will start the
        // TransportListener

         if (System.getProperty(ServerConstants.STANDALONE_MODE, "false").equalsIgnoreCase("true")
                 && isStopped) {
            try {
                Class<?> transportManagerClass = Class.forName(TRANSPORT_MANAGER);
                Object transportManager = transportManagerClass.newInstance();

                Class[] parametertype = new Class[] {String.class};
                Method method = transportManagerClass.getMethod("startTransport", parametertype);
                method.invoke(transportManager, new String[]{transport});
            } catch (Exception e) {
                String msg = "Cannot start " + transport + " transport";
                log.fatal(msg, e);
                return;
            }
        }
        isStopped = false;
    }

    public synchronized void stop() throws AxisFault {
        if (System.getProperty(ServerConstants.STANDALONE_MODE, "false").equalsIgnoreCase("true") &&
            transportManagerClass != null) {
           try {
                Method method = transportManagerClass.getMethod("stopTransport", String.class);
                method.invoke(transportManager, transport);
            } catch (Exception e) {
                String msg = "Cannot get stop transport";
                log.error(msg, e);
                throw new AxisFault(msg, e);
            }
        }
        isStopped = true;
    }

    public EndpointReference getEPR(String protocol,
                                    String serviceName,
                                    String ip) throws AxisFault {
        if (configurationContext == null) {
            return null;
        }
        String serviceContextPath = configurationContext.getServiceContextPath();
        if (serviceContextPath == null) {
            throw new AxisFault("couldn't find service path");
        }
        return genEpr(protocol, ip, serviceContextPath, serviceName);
    }

    public SessionContext getSessionContext(MessageContext messageContext) {
        return SessionContextUtil.createSessionContext(messageContext);
    }

    public void destroy() {
        this.configurationContext = null;
    }

    protected EndpointReference genEpr(String protocol, String ip, String serviceContextPath,
                                       String serviceName) throws AxisFault {
        try {
            if (ip == null) {
                ip = NetworkUtils.getLocalHostname();
            }
            String tmp = protocol + "://" + ip;
            if (proxyPort == 80 || proxyPort == 443) {
                tmp += serviceContextPath + "/" + serviceName;
            } else if (proxyPort != -1) {
                tmp += ":" + proxyPort + serviceContextPath + "/" + serviceName;
            } else {
                tmp += ":" + port + serviceContextPath + "/" + serviceName;
            }
            // This trailing / is needed for REST to work. The http Location is resolved against
            // this base URI and produce completely dirrefent results if this is not present.
            return new EndpointReference(tmp + "/");
        } catch (SocketException e) {
            throw AxisFault.makeFault(e);
        }
    }
}
